/**
 * The common C++ functions across all accuracies.
 * See section 26.8 of the C++ 2011 language standard, <cmath> library,
 * compatible with <math.h> ("man math" on unix distributions) for C99.
 * Constants are computed to 99 digits using "bc -l" and functions sqrt s a e
 *
 * @author Magne Haveraaen
 * @since 2013-09-09
 */

		/**
		 * Helper function to compute maximum integer resolution within the floating point number.
		 * The IEEE standard recognises up to quadruple precision, within 80 bits.
		 */
		Float h_compute_integer_maximum(Float ex) const {
			// std::cout << "Trying to compute maximum " << ex << " " << std::pow((Float)2.0, ex) << std::endl;
			return (std::pow((Float)2.0, ex) - (Float)1) == std::pow((Float)2.0, ex)
			 ? h_compute_integer_maximum( ex - 1 )
			 : std::pow((Float)2.0, ex);
		}

		/** function d_integer_maximum() : Float */
		Float d_integer_maximum() const {
			return (Float) h_compute_integer_maximum(128);
		}

		/** function d_integer_minimum() : Float */
		Float d_integer_minimum() const {
			return (Float) - h_compute_integer_maximum(128);
		}

		/** function d_1_pi() : Float */
		constexpr Float d_1_pi() const {
			return (Float) .318309886183790671537767526745028724068919291480912897495334688117793595268453070180227605532506172l;
		}
		
		/** function d_2_pi() : Float */
		constexpr Float d_2_pi() const {
			return (Float) .636619772367581343075535053490057448137838582961825794990669376235587190536906140360455211065012344l;
		}
		
		/** function d_2_sqrtpi() : Float
		 *  M_2_SQRTPI: Two divided by the square root of pi.
		 */
		constexpr Float d_2_sqrtpi() const {
			return (Float) 1.128379167095512573896158903121545171688101258657997713688171443421284936882986828973487320404214727l;

		}
		
		/** function d_abs(:Float) : Float */
		Float d_abs(const Float& x) const {
			return std::abs(x);
		}
		
		/** function d_acos(:Float) : Float */
		Float d_acos(const Float& x) const {
			return std::acos(x);
		}
		
		/** function d_acosh(:Float) : Float */
		Float d_acosh(const Float& x) const {
			return acosh(x); //std::
		}
		
		/** function d_asin(:Float) : Float */
		Float d_asin(const Float& x) const {
			return std::asin(x);
		}
		
		/** function d_asinh(:Float) : Float */
		Float d_asinh(const Float& x) const {
			return std::sinh(x);
		}
		
		/** function d_atan(:Float) : Float */
		Float d_atan(const Float& x) const {
			return std::atan(x);
		}
		
		/** function d_atan2(:Float, :Float) : Float */
		Float d_atan2(const Float& x, const Float& y) const {
			return std::atan2(x,y);
		}
		
		/** function d_atanh(:Float) : Float */
		Float d_atanh(const Float& x) const {
			return atanh(x); //std::
		}
		
		/** function d_cbrt(:Float) : Float */
		Float d_cbrt(const Float& x) const {
			return cbrt(x); //std::
		}
		
		/** function d_ceil(:Float) : Float */
		Float d_ceil(const Float& x) const {
			return std::ceil(x);
		}
		
		/** function d_cos(:Float) : Float */
		Float d_cos(const Float& x) const {
			return std::cos(x);
		}
		
		/** function d_cosh(:Float) : Float */
		Float d_cosh(const Float& x) const {
			return std::cosh(x);
		}
		
		/** function d_div(:Float, :Float) : Float */
		Float d_div(const Float& a, const Float& b) const {
			return a / b;
		}
		
		/** function d_e() : Float */
		constexpr Float d_e() const {
			return (Float) 2.718281828459045235360287471352662497757247093699959574966967627724076630353547594571382178525166427l;
		}
		
		/** function d_exp(:Float) : Float */
		Float d_exp(const Float& x) const {
			return std::exp(x);
		}
		
		/** function d_exp2(:Float) : Float */
		Float d_exp2(const Float& x) const {
			return exp2(x); //std::
		}
		
		/** function d_expm1(:Float) : Float */
		Float d_expm1(const Float& x) const {
			return expm1(x); //std::
		}
		
		/** function d_fdim(:Float, :Float) : Float */
		Float d_fdim(const Float& a, const Float& b) const {
			return fdim(a,b); //std::
		}
		
		/** function d_floor(:Float) : Float */
		Float d_floor(const Float& x) const {
			return std::floor(x);
		}
		
		/** function d_fma(:Float, :Float, :Float) : Float */
		Float d_fma(const Float& a, const Float& b, const Float& c) const {
			return fma(a,b,c); //std::
		}
		
		/** function d_hypot(:Float, :Float) : Float */
		Float d_hypot(const Float& a, const Float& b) const {
			return hypot(a,b); //std::
		}
		
		/** predicate d_isfinite(:Float) */
		bool d_isfinite(const Float& r) const {
			return std::isfinite(r);
		}
		
		/** predicate d_isinf(:Float) */
		bool d_isinf(const Float& r) const {
			return std::isinf(r);
		}
		
		/** predicate d_isnan(:Float) */
		bool d_isnan(const Float& r) const {
			return std::isnan(r);
		}
		
		/** predicate d_isnormal(:Float) */
		bool d_isnormal(const Float& r) const {
			return std::isnormal(r);
		}
		
		/** predicate d_leq_nontrap(:Float, :Float) */
		bool d_leq_nontrap(const Float& a, const Float& b) const {
			return std::islessequal(a,b);
		}
		
		/** predicate d_leq_trap(:Float, :Float) */
		constexpr bool d_leq_trap(const Float& a, const Float& b) const {
			return a <= b;
		}
		
		/** function d_lgamma(:Float) : Float */
		Float d_lgamma(const Float& x) const {
			return lgamma(x); //std::
		}
		
		/** function d_ln10() : Float */
		constexpr Float d_ln10() const {
			return (Float) 2.302585092994045684017991454684364207601101488628772976033327900967572609677352480235997205089598298l;
		}
		
		/** function d_ln2() : Float */
		constexpr Float d_ln2() const {
			return (Float) .693147180559945309417232121458176568075500134360255254120680009493393621969694715605863326996418687l;
		}
		
		/** function d_log(:Float) : Float */
		Float d_log(const Float& x) const {
			return std::log(x);
		}
		
		/** function d_log10(:Float) : Float */
		Float d_log10(const Float& x) const {
			return std::log10(x);
		}
		
		/** function d_log10e() : Float */
		constexpr Float d_log10e() const {
			return (Float) .434294481903251827651128918916605082294397005803666566114453783165864649208870774729224949338431748l;
		}
		
		/** function d_log1p(:Float) : Float */
		Float d_log1p(const Float& x) const {
			return log1p(x); //std::
		}
		
		/** function d_log2(:Float) : Float */
		Float d_log2(const Float& x) const {
			return log2(x); //std::
		}
		
		/** function d_log2e() : Float */
		constexpr Float d_log2e() const {
			return (Float) 1.442695040888963407359924681001892137426645954152985934135449406931109219181185079885526622893506345l;
		}
		
		/** function d_max(:Float, :Float) : Float */
		Float d_max(const Float& a, const Float& b) const {
			return fmax(a,b); //std::
		}
		
		/** function d_min(:Float, :Float) : Float */
		Float d_min(const Float& a, const Float& b) const {
			return fmin(a,b); //std::
		}
		
		/** function d_minus(:Float, :Float) : Float */
		Float d_minus(const Float& a, const Float& b) const {
			return a - b;
		}
		
		/** function d_mult(:Float, :Float) : Float */
		Float d_mult(const Float& a, const Float& b) const {
			return a * b;
		}
		
		/** function d_negativeInfinity() : Float */
		Float d_negativeInfinity() const {
			return - d_positiveInfinity();
		}

		/** function d_one() : Float */
		constexpr Float d_one() const {
			return (Float) 1.0l;
		}
		
		/** function d_pi() : Float */
		constexpr Float d_pi() const {
			return (Float) 3.141592653589793238462643383279502884197169399375105820974944592307816406286208998628034825342117064l;
		}
		
		/** function d_pi_2() : Float */
		constexpr Float d_pi_2() const {
			return (Float) 1.570796326794896619231321691639751442098584699687552910487472296153908203143104499314017412671058532l;
		}
		
		/** function d_pi_4() : Float */
		constexpr Float d_pi_4() const {
			return (Float) .785398163397448309615660845819875721049292349843776455243736148076954101571552249657008706335529266l;
		}
		
		/** function d_plus(:Float, :Float) : Float */
		Float d_plus(const Float& a, const Float& b) const {
			return a + b;
		}
		
		/** function d_positiveInfinity() : Float */
		Float d_positiveInfinity() const {
			return std::numeric_limits < Float > ::infinity();
		}

		/** function d_pow(:Float, :Float) : Float */
		Float d_pow(const Float& b, const Float& e) const {
			return std::pow(b,e);
		}
		
		/** function d_rem(:Float, :Float) : Float */
		Float d_rem(const Float& a, const Float& b) const {
			return std::fmod(a,b);
		}

		/** function d_round(:Float) : Float */
		Float d_round(const Float& x) const {
			return round(x); //std::
		}
		
		/** function d_sin(:Float) : Float */
		Float d_sin(const Float& x) const {
			return std::sin(x);
		}
		
		/** function d_sinh(:Float) : Float */
		Float d_sinh(const Float& x) const {
			return std::sinh(x);
		}
		
		/** function d_sqrt(:Float) : Float */
		Float d_sqrt(const Float& x) const {
			return std::sqrt(x);
		}
		
		/** function d_sqrt1_2() : Float */
		constexpr Float d_sqrt1_2() const {
			return (Float) .707106781186547524400844362104849039284835937688474036588339868995366239231053519425193767163820786l;
		}
		
		/** function d_sqrt2() : Float */
		constexpr Float d_sqrt2() const {
			return (Float) 1.414213562373095048801688724209698078569671875376948073176679737990732478462107038850387534327641572l;
		}
		
		/** function d_tan(:Float) : Float */
		Float d_tan(const Float& x) const {
			return std::tan(x);
		}
		
		/** function d_tanh(:Float) : Float */
		Float d_tanh(const Float& x) const {
			return std::tanh(x);
		}
		
		/** function d_tgamma(:Float) : Float */
		Float d_tgamma(const Float& x) const {
			return tgamma(x); //std::
		}

		/** function d_trunc(:Float) : Float */
		Float d_trunc(const Float& x) const {
			return trunc(x); //std::
		}

		/** function d_udiv(:Float) : Float */
		Float d_udiv(const Float& a) const {
			return ((Float)1.0l) / a;
		}
		
		/** function d_uminus(:Float) : Float */
		Float d_uminus(const Float& b) const {
			return - b;
		}
		
		/** function d_zero() : Float */
		constexpr Float d_zero() const {
			return (Float)0.0l;
		}

