(comment "

This project definition is for the console variant of Magnolia. It
assumes that input (i.e. '.mg' source files) is static, and provides
no way to propagate changes. It outputs compilation feedback onto the
console.

This project hosts Magnolia unit tests. To run them, invoke the shell
command 'lein test'. It is also possible to use a Clojure REPL to play
with the compiler interactively. To launch a REPL on the console, use
the command 'lein repl'. To launch a REPL under Emacs you may simply
use M-x clojure-jack-in, provided you have a bleeding edge setup of
Emacs with clojure-mode.

It is assumed that the Magnolia code base Java code is built
separately, so that classfiles are under 'bin', and dependencies under
'lib'. (The Eclipse-based project will take care of this.)

This project file requires Java and Leiningen. If you don't have the
latter, you can install it as follows:

  wget https://raw.github.com/technomancy/leiningen/preview/bin/lein
  mkdir -p ~/bin
  mv -i lein ~/bin/
  chmod 755 ~/bin/lein
  export PATH=$HOME/bin:$PATH

To actually use Magnolia, you'll need to provide some specifics about
your build environment. This is possible to do by setting the
following environment variables.

  MAGNOLIA_ECLIPSE_HOME - specifies (for Makefile) where to look for
  the required external Java libraries; this should point to the root
  directory of an Eclipse installation that has the Rascal Eclipse
  plugin installed - e.g. $HOME/eclipse

  MAGNOLIA_JAVA_CLASSPATH - may optionally be used to override the
  'extra-classpath-dirs' setting (for Leiningen), which is useful if
  you for instance want to use a development build of Rascal rather
  than the one installed as an Eclipse plugin - e.g.
  'bin:lib/*:'$MAGNOLIA_ECLIPSE_HOME'/workspace/rascal/src:ext-lib/*'
  - any '_DEFAULT_' component is replaced by the default
  MAGNOLIA_JAVA_CLASSPATH - e.g.
  $MAGNOLIA_ECLIPSE_HOME/workspace/rascal/src:_DEFAULT_

  MAGNOLIA_MODULE_PATH - in PATH env var style manner specifies (for
  the Magnolia compiler) where to look for '.mg' files - e.g.
  magnolia-tests/src

  MAGNOLIA_RASCAL_MODULE_PATH - optional search path for additional
  '.rsc' files, which allows for making use of Magnolia compiler
  functionality in other Rascal projects - e.g.
  /tmp/rascal-tests/src:/tmp/rascal-tools/src

After you have set the above env vars, execute 'make update-ext-lib'
to arrange for the required external libraries to be copied to the
'ext-lib' directory. To check that Leiningen ends up with the correct
idea of your CLASSPATH, run 'lein classpath'.

")

(defn dir-path [x]
  (.getAbsolutePath (java.io.File. x)))

(defn jar-path [x]
  (str (dir-path x) "/*"))

(defn get-extra-classpath-dirs []
  (let [ev (System/getenv "MAGNOLIA_JAVA_CLASSPATH")
        de [(dir-path "bin")
            (jar-path "lib")
            (jar-path "ext-lib")]]
    (if-not ev
      de
      (let [x (seq (.split #":" ev))
            y (mapcat #(if (= % "_DEFAULT_") de [(dir-path %)]) x)]
        y))))

(defproject org.magnolialang "0.2-SNAPSHOT"
  :description "Magnolia language"
  :url "http://magnolia-lang.org/"
  :compile-path "clj-bin"
  :library-path "clj-lib"
  ;; This may be deprecated, but have yet to figure out how to do the
  ;; same with :repl-init and/or :repl-options.
  :repl-init-script "src/org/magnolialang/testutil/repl_init.clj"
  ;;:repl-init org.magnolialang.testutil.core
  ;; :repl-options [:init
  ;;                (use 'org.magnolialang.testutil.core)
  ;;                (init-javadoc)
  ;;                ]
  :aot [org.magnolialang.testutil.bankers-deque
        org.magnolialang.testutil.core
        org.magnolialang.testutil.pdb-values
        org.magnolialang.testutil.pdb-values-pprint
        ;;xxx org.magnolialang.testutil.rascal-shell
        ]
  :jvm-opts [
             "-ea"
             ;; Use the same memory settings as for Eclipse.
             "-Xss32m" "-Xms40m" "-Xmx1G" "-XX:MaxPermSize=512M"
             ;; Time-consuming, but shows what classes loaded from
             ;; where.
             ;;"-verbose:class"
             ;; To see memory consumption info.
             ;;"-verbose:gc"
             ]
  :extra-classpath-dirs ~(get-extra-classpath-dirs)
  :dependencies [[org.clojure/clojure "1.4.0"]
                 [org.clojure/math.numeric-tower "0.0.1"]]
  :plugins [[lein-swank "1.4.4"]])
