/**************************************************************************
 * Copyright (c) 2010-2014 Anya Helene Bagge
 * Copyright (c) 2010-2014 Tero Hasu
 * Copyright (c) 2010-2014 University of Bergen
 *
 * All rights reserved. This program and the accompanying materials
 * are made available under the terms of the Eclipse Public License v1.0
 * which accompanies this distribution, and is available at
 * http://www.eclipse.org/legal/epl-v10.html
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version. See http://www.gnu.org/licenses/
 *
 *
 * See the file COPYRIGHT for more information.
 *
 * Contributors:
 * * Anya Helene Bagge
 * * Tero Hasu
 *
 *************************************************************************/
package org.magnolialang.eclipse;

import java.io.File;
import java.io.IOException;
import java.net.URI;
import java.net.URISyntaxException;
import java.net.URL;
import java.util.HashMap;
import java.util.Map;

import org.eclipse.core.resources.IFile;
import org.eclipse.core.resources.IProject;
import org.eclipse.core.resources.ResourcesPlugin;
import org.eclipse.core.runtime.FileLocator;
import org.eclipse.core.runtime.IPath;
import org.eclipse.core.runtime.Path;
import org.eclipse.core.runtime.Platform;
import org.eclipse.core.runtime.URIUtil;
import org.eclipse.core.runtime.jobs.Job;
import io.usethesource.impulse.model.ISourceProject;
import io.usethesource.impulse.preferences.PreferencesService;
import io.usethesource.impulse.runtime.PluginBase;
import org.eclipse.jdt.annotation.NonNullByDefault;
import org.eclipse.jdt.annotation.Nullable;
import org.eclipse.jface.resource.ImageDescriptor;
import org.eclipse.jface.resource.ImageRegistry;
import org.magnolialang.eclipse.commands.ToggleHandler;
import org.magnolialang.eclipse.commands.langdev.DeveloperMode;
import org.magnolialang.eclipse.icons.StyleBase;
import org.magnolialang.eclipse.preferences.Pref;
import org.magnolialang.infra.EclipseInfra;
import org.magnolialang.infra.Infra;
import org.magnolialang.magnolia.Magnolia;
import org.magnolialang.uri.ProjectURIResolver;
import org.nuthatchery.pica.errors.ImplementationError;
import org.nuthatchery.pica.resources.ILanguage;
import org.nuthatchery.pica.resources.IWorkspaceManager;
import org.nuthatchery.pica.resources.IWorkspaceManager.ThreadingPreference;
import org.nuthatchery.pica.resources.LanguageRegistry;
import org.osgi.framework.Bundle;
import org.osgi.framework.BundleContext;
import org.osgi.framework.BundleException;
import org.rascalmpl.uri.BadURIException;
import org.rascalmpl.uri.URIResolverRegistry;
import org.rascalmpl.uri.UnsupportedSchemeException;

@NonNullByDefault
public class MagnoliaPlugin extends PluginBase {

	public static final String MAGNOLIA_PROJECT_NAME = "magnolia-eclipse";
	public static final String PLUGIN_ID = "magnolia_eclipse";
	public static final String LANGUAGE_ID = "Magnolia";

	/**
	 * The unique instance of this plugin class.
	 */
	@Nullable
	private static MagnoliaPlugin sPlugin;
	public static final Bundle MAGNOLIA_BUNDLE = Platform.getBundle(PLUGIN_ID);
	public static final Object JOB_FAMILY_MAGNOLIA = new Object();
	public static final String MAGNOLIA_CONSOLE_NAME = "Magnolia Console";

	private final Map<String, StyleBase> styles = new HashMap<String, StyleBase>();

	private final boolean isInSecondLevel;


	public MagnoliaPlugin() {
		super();
		sPlugin = this;

		EclipseInfra.setInfra();

		refreshPaths();
		//URIResolverRegistry uriResolverRegistry = Infra.getResolverRegistry();

		// project scheme
		// TODO: load config file from rascal-eclipse?
		//final ProjectURIResolver resolver = new ProjectURIResolver();
		//uriResolverRegistry.registerInput(resolver);
		//uriResolverRegistry.registerOutput(resolver);

		styles.put(Magnolia.MAGNOLIA, new StyleBase(Magnolia.MAGNOLIA));
		styles.put("MetaXa", new StyleBase(Magnolia.MAGNOLIA));

		if(getFileOrNull(MAGNOLIA_BUNDLE, "bin") != null) {
			isInSecondLevel = true;
		}
		else {
			isInSecondLevel = false;
		}
		System.err.println("Running in second level: " + isInSecondLevel);

		Magnolia magnolia = Magnolia.getInstance();
		LanguageRegistry.registerLanguage(magnolia);
		magnolia.getParser().initialize();
		IWorkspaceManager workspaceManager = Infra.getWorkspaceManager();
		String threadingPref = MagnoliaPlugin.getInstance().getPreferenceStore().getString(Pref.MAIN_MULTITHREADING.name());
		assert threadingPref != null;
		workspaceManager.setThreadingPreference(ThreadingPreference.getPref(threadingPref));
		magnolia.getCompiler().refresh();
		ToggleHandler.setState(ToggleHandler.COMMAND_AUTO_REFRESH, false);
	}


	@Override
	public String getID() {
		return PLUGIN_ID;
	}


	@Override
	public String getLanguageID() {
		return LANGUAGE_ID;
	}


	@Nullable
	public StyleBase getStyleBase(String language) {
		return styles.get(language);
	}


	public boolean isInSecondLevel() {
		return isInSecondLevel;
	}


	@Override
	public void logException(@Nullable String msg, @Nullable Throwable t) {
		super.logException(msg == null ? "unknown error" : msg, t);
	}


	@Override
	public void stop(@Nullable BundleContext context) throws Exception {
		try {
			Infra.getWorkspaceManager().stop();
			Job.getJobManager().cancel(JOB_FAMILY_MAGNOLIA);
		}
		finally {
			super.stop(context);
		}
	}


	@Override
	@NonNullByDefault(false)
	protected void initializeImageRegistry(final ImageRegistry reg) {
		assert reg != null;
		for(String s : MagnoliaResources.MAGNOLIA_ICONS) {
			ImageDescriptor desc = MagnoliaResources.getImageDescriptor(s);
			if(desc != null) {
				reg.put(s, desc);
			}
		}
	}


	public static URI constructProjectURI(IProject project, IPath path) {
		return constructProjectURI(project.getName(), path);
	}


	public static URI constructProjectURI(ISourceProject project, IPath path) {
		return constructProjectURI(project.getRawProject().getName(), path);
	}


	public static URI constructProjectURI(String project, IPath path) {
		try {
			// making sure that spaces in 'path' are properly escaped
			path = path.makeAbsolute();
			return new URI("project", project, path.toString(), null, null);
		}
		catch(URISyntaxException usex) {
			throw new BadURIException(usex);
		}
	}


	public static String getFile(final Bundle bundle, final String path) {
		URI uri = getFileURI(bundle, path);
		if(uri.getScheme().equals("file")) {
			return uri.getPath();
		}
		else {
			throw new ImplementationError("getFile: Unsupported scheme: " + uri);
		}
	}


	public static String getFile(final String path) {
		return getFile(MAGNOLIA_BUNDLE, path);
	}


	/**
	 * @param uri
	 *            The URI of the desired file
	 * @return An IFile representing the URI
	 */
	@Nullable
	public static IFile getFileHandle(URI uri) {
//		IPath path = null;
//		try {
		// TODO
		throw new UnsupportedOperationException();
//			path = new Path(new File(Infra.getResolverRegistry().getResourceURI(uri)).getAbsolutePath());
//		}
//		catch(UnsupportedSchemeException e) {
//			getInstance().logException(e.getMessage(), e);
//			e.printStackTrace();
//			return null;
//		}
//		catch(IOException e) {
//			getInstance().logException(e.getMessage(), e);
//			e.printStackTrace();
//			return null;
//		}
//		return ResourcesPlugin.getWorkspace().getRoot().getFileForLocation(path);

	}


	@Nullable
	public static String getFileOrNull(final Bundle bundle, final String path) {
		try {
			URI uri = getFileURI(bundle, path);
			if(uri.getScheme().equals("file")) {
				return uri.getPath();
			}
			else {
				throw new ImplementationError("getFile: Unsupported scheme: " + uri);
			}
		}
		catch(ImplementationError e) {
			return null;
		}
	}


	@SuppressWarnings("null")
	public static URI getFileURI(final Bundle bundle, final String path) {
		try {
			return URIUtil.toURI(FileLocator.toFileURL(bundle.getEntry(path)));
		}
		catch(Exception ex) {
			try {

				return URIUtil.toURI(FileLocator.toFileURL(bundle.getResource(path)));
			}
			catch(Exception ex2) {
				URL file = FileLocator.find(bundle, new Path(path), null);

				if(file == null) {
					throw new ImplementationError("Can't find: " + bundle + ":/" + path, ex2);
				}
				else {
					try {
						file = FileLocator.toFileURL(file);
					}
					catch(IOException e) {
						// ignore
					}
					try {
						return URIUtil.toURI(file);
					}
					catch(URISyntaxException e) {
						throw new ImplementationError("getFileURI: URISyntaxException while resolving " + bundle + " / " + path, e);
					}
				}
			}
		}
	}


	public static synchronized MagnoliaPlugin getInstance() {

		final MagnoliaPlugin p = sPlugin;
		if(p != null) {
			return p;
		}
		else {
			throw new ImplementationError("Magnolia plugin not started yet");
		}

	}


	public static synchronized MagnoliaPlugin getOrStartInstance() throws BundleException {
		if(sPlugin == null) {
			MAGNOLIA_BUNDLE.start();
		}
		return getInstance();
	}


	public static String getPref(final String key) {
		final PreferencesService prefs = getInstance().getPreferencesService();
		String pref;
		if(prefs.isDefined("useDefault" + key) && prefs.getBooleanPreference("useDefault" + key)) {
			pref = prefs.getStringPreference("default", key);
		}
		else {
			pref = prefs.getStringPreference(key);
		}

		return pref;
	}


	public static void refresh() {
		for(StyleBase s : getInstance().styles.values()) {
			s.refresh();
		}
		refreshPaths();
		MagnoliaResources.refresh();
	}


	public static void refreshPaths() {
		String path = "";

		if(DeveloperMode.enabled()) {
			IFile file = MagnoliaFileLocator.findInProject(MAGNOLIA_PROJECT_NAME, "src");
			if(file != null) {
				path = path + file.getLocation().toString() + ":";
			}
		}

		try {
			path = path + MagnoliaPlugin.getFile(MagnoliaPlugin.MAGNOLIA_BUNDLE, "src") + ":";
		}
		catch(ImplementationError e) {
			// no worries, src/ is not available in binary builds.
		}
		path = path + MagnoliaPlugin.getFile(MagnoliaPlugin.MAGNOLIA_BUNDLE, "");
		System.setProperty("rascal.path", "");
	}
}
