/**************************************************************************
 * Copyright (c) 2012-2014 Anya Helene Bagge
 * Copyright (c) 2012-2014 University of Bergen
 * 
 * All rights reserved. This program and the accompanying materials
 * are made available under the terms of the Eclipse Public License v1.0
 * which accompanies this distribution, and is available at
 * http://www.eclipse.org/legal/epl-v10.html
 * 
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version. See http://www.gnu.org/licenses/
 * 
 * 
 * See the file COPYRIGHT for more information.
 * 
 * Contributors:
 * * Anya Helene Bagge
 * 
 *************************************************************************/
package org.magnolialang.eclipse;

import java.net.URL;
import java.util.ArrayList;
import java.util.Arrays;
import java.util.Collection;
import java.util.Collections;
import java.util.HashMap;
import java.util.List;
import java.util.Map;

import org.eclipse.core.runtime.FileLocator;
import org.eclipse.core.runtime.IPath;
import org.eclipse.core.runtime.Path;
import org.eclipse.jdt.annotation.Nullable;
import org.eclipse.jface.resource.ImageDescriptor;
import org.eclipse.jface.resource.ImageRegistry;
import org.eclipse.swt.graphics.Image;
import org.eclipse.ui.ISharedImages;
import org.eclipse.ui.PlatformUI;

public final class MagnoliaResources {

	public static final String MAGNOLIA_FILE = "magnolia_file_source";
	public static final String MAGNOLIA_FOLDER = "folder+magnolia";
	public static final String MAGNOLIA_PROJECT = "project+magnolia";
	/**
	 * Icons that should be preloaded
	 */
	public static final Collection<String> MAGNOLIA_ICONS = Collections.unmodifiableCollection(Arrays.asList(MAGNOLIA_FILE));

	public static final String MAGNOLIA_DEFAULT_OUTLINE_ITEM = "default";

	public static final String SUFFIX_OBJECT = "_obj";
	public static final String SUFFIX_OVERLAY = "_ovr";
	public static final IPath ICONS_PATH = new Path("icons"); //$NON-NLS-1$("icons");
	/**
	 * Relative to the plugin bundle
	 */
	private static final IPath[] ICONS_SEARCH_PATH = { ICONS_PATH, ICONS_PATH.append("obj16"), ICONS_PATH.append("ovr16") };
	private static final String[] ICONS_EXTENSIONS = { ".png", ".gif", ".PNG", ".GIF" };

	private static final Map<String, ImageDescriptor> imageDescriptors = new HashMap<String, ImageDescriptor>();
	private static final List<String> registeredImages = new ArrayList<String>();
	static {
		refresh();
	}


	private MagnoliaResources() {

	}


	/**
	 * Return an Image for the given image ID.
	 * 
	 * This method makes use of the plugin's ImageRegistry, so the resulting
	 * Image should not be disposed of (it is owned and managed by the
	 * registry).
	 * 
	 * If the image isn't already in the registry, it will be loaded from disk,
	 * in the same manner as with getImageDescriptor(): The plugin's icon search
	 * path is search, with the file name extensions '.png' and '.gif'.
	 * 
	 * @param imageId
	 * @return An Image object, managed by the image registry, or null if the
	 *         image wasn't found.
	 */
	public static Image getImage(String imageId) {
		MagnoliaPlugin plugin = MagnoliaPlugin.getInstance();
		Image img = plugin.getImageRegistry().get(imageId);
		if(img == null) {
			getImageDescriptor(imageId, true);
			img = plugin.getImageRegistry().get(imageId);
		}
		return img;
	}


	/**
	 * Return an image descriptor for the given image ID. The plugin's icon
	 * search path is searched, and if the file name does not contain a dot,
	 * '.png' and '.gif' are tried.
	 * 
	 * Use createResource()/destroyResource() on the returned descriptor to
	 * ensure maximum sharing of images.
	 * 
	 * @param imageId
	 *            Identifier for the image, without file name extension and
	 *            relative to the plugin icon directory
	 * @return An ImageDescriptor, or null.
	 */
	@Nullable
	public static ImageDescriptor getImageDescriptor(String imageId) {
		return getImageDescriptor(imageId, false);
	}


	public static void putImageDescriptor(String imageId, ImageDescriptor desc) {
		imageDescriptors.put(imageId, desc);
	}


	/**
	 * Reinitialize all resources, forcing icons to be reloaded from disk.
	 */
	public static void refresh() {
		imageDescriptors.clear();
		for(String s : registeredImages) {
			MagnoliaPlugin.getInstance().getImageRegistry().remove(s);
		}
		ISharedImages images = PlatformUI.getWorkbench().getSharedImages();
		imageDescriptors.put("warning_ovr", images.getImageDescriptor(ISharedImages.IMG_DEC_FIELD_WARNING));
		imageDescriptors.put("error_ovr", images.getImageDescriptor(ISharedImages.IMG_DEC_FIELD_ERROR));
		imageDescriptors.put("file_obj", images.getImageDescriptor(ISharedImages.IMG_OBJ_FILE));
		imageDescriptors.put("folder_obj", images.getImageDescriptor(ISharedImages.IMG_OBJ_FOLDER));
		imageDescriptors.put("default_obj", images.getImageDescriptor(ISharedImages.IMG_OBJ_ELEMENT));
		imageDescriptors.put("project_obj", images.getImageDescriptor(org.eclipse.ui.ide.IDE.SharedImages.IMG_OBJ_PROJECT));
	}


	@Nullable
	private static ImageDescriptor getImageDescriptor(String imageId, boolean useImageRegistry) {
		MagnoliaPlugin plugin = MagnoliaPlugin.getInstance();
		ImageRegistry registry = plugin.getImageRegistry();

		// first, see if the ImageRegistry has the image
		ImageDescriptor desc = registry.getDescriptor(imageId);
		if(desc != null) {
			return desc;
		}
		// then, see if we've already found the image
		desc = imageDescriptors.get(imageId);
		if(desc != null) {
			if(useImageRegistry) {
				plugin.getImageRegistry().put(imageId, desc);
				registeredImages.add(imageId);
			}
			return desc;
		}

		// otherwise, do a path search
		for(IPath path : ICONS_SEARCH_PATH) {
			for(String ext : ICONS_EXTENSIONS) {
				IPath p = path.append(imageId + ext);
				URL url = FileLocator.find(MagnoliaPlugin.MAGNOLIA_BUNDLE, p, null);
				if(url != null) {
					desc = ImageDescriptor.createFromURL(url);
					if(desc != null) {
						if(useImageRegistry) {
							plugin.getImageRegistry().put(imageId, desc);
							registeredImages.add(imageId);
						}
						else {
							imageDescriptors.put(imageId, desc);
						}
						return desc;
					}

				}
			}
		}
		return null;
	}
}
