/**************************************************************************
 * Copyright (c) 2007 IBM Corporation.
 * Copyright (c) 2010-2014 Anya Helene Bagge
 * Copyright (c) 2010-2014 University of Bergen
 *
 * All rights reserved. This program and the accompanying materials
 * are made available under the terms of the Eclipse Public License v1.0
 * which accompanies this distribution, and is available at
 * http://www.eclipse.org/legal/epl-v10.html
 *
 *
 * See the file COPYRIGHT for more information.
 *
 * Contributors:
 * * Anya Helene Bagge
 * + Robert Fuhrer
 *
 *************************************************************************/

package org.magnolialang.eclipse.builders;

import java.io.ByteArrayInputStream;
import java.io.IOException;
import java.io.InputStream;
import java.nio.charset.Charset;
import java.util.ArrayList;
import java.util.Collection;
import java.util.HashSet;
import java.util.List;
import java.util.Map;
import java.util.Set;
import java.util.concurrent.Callable;
import java.util.concurrent.ExecutionException;
import java.util.concurrent.Future;

import org.eclipse.core.resources.IContainer;
import org.eclipse.core.resources.IFile;
import org.eclipse.core.resources.IFolder;
import org.eclipse.core.resources.IProject;
import org.eclipse.core.resources.IResource;
import org.eclipse.core.resources.IResourceDelta;
import org.eclipse.core.resources.IResourceVisitor;
import org.eclipse.core.resources.IWorkspaceRoot;
import org.eclipse.core.resources.IncrementalProjectBuilder;
import org.eclipse.core.runtime.CoreException;
import org.eclipse.core.runtime.IPath;
import org.eclipse.core.runtime.IProgressMonitor;
import org.eclipse.core.runtime.IStatus;
import org.eclipse.core.runtime.OperationCanceledException;
import org.eclipse.core.runtime.Path;
import org.eclipse.core.runtime.Status;
import org.eclipse.core.runtime.SubMonitor;
import org.rascalmpl.value.IConstructor;
import org.rascalmpl.value.IString;
import org.eclipse.jdt.annotation.NonNullByDefault;
import org.eclipse.jdt.annotation.Nullable;
import org.magnolialang.eclipse.MagnoliaFileLocator;
import org.magnolialang.eclipse.MagnoliaPlugin;
import org.magnolialang.eclipse.preferences.Pref;
import org.magnolialang.errors.RascalErrors;
import org.magnolialang.infra.Infra;
import org.magnolialang.magnolia.Magnolia;
import org.magnolialang.magnolia.MagnoliaFacts;
import org.magnolialang.magnolia.backend.Instantiator;
import org.magnolialang.magnolia.resources.IMagnoliaModule;
import org.magnolialang.magnolia.resources.IMagnoliaPackage;
import org.nuthatchery.pica.eclipse.EclipsePicaInfra;
import org.nuthatchery.pica.errors.ErrorMarkers;
import org.nuthatchery.pica.errors.Severity;
import org.nuthatchery.pica.rascal.ICompiler;
import org.nuthatchery.pica.resources.ILanguage;
import org.nuthatchery.pica.resources.IProjectManager;
import org.nuthatchery.pica.resources.LanguageRegistry;
import org.nuthatchery.pica.resources.eclipse.EclipseWorkspaceManager;
import org.nuthatchery.pica.resources.managed.IManagedCodeUnit;
import org.nuthatchery.pica.resources.managed.IManagedResource;
import org.nuthatchery.pica.resources.marks.MarkBuilder;
import org.nuthatchery.pica.tasks.ITaskMonitor;
import org.nuthatchery.pica.tasks.TaskId;
import org.nuthatchery.pica.tasks.eclipse.EclipseTaskMonitor;
import org.nuthatchery.pica.terms.TermFactory;
import org.nuthatchery.pica.util.NullnessHelper;
import org.nuthatchery.pica.util.Pair;
import org.nuthatchery.pica.util.concurrent.FuturesQueue;
import org.nuthatchery.pica.util.depgraph.IDepGraph;
import org.nuthatchery.pica.util.depgraph.ITopologicalWorkQueue;
import org.rascalmpl.eclipse.nature.RascalMonitor;
import org.rascalmpl.eclipse.nature.WarningsToMarkers;
import org.rascalmpl.debug.IRascalMonitor;
import org.rascalmpl.interpreter.control_exceptions.Throw;
import org.rascalmpl.parser.gtd.io.InputConverter;

/**
 * Builder for Magnolia projects.
 *
 * Has a concept of a source folder and an output folder. All Magnolia modules
 * in the source folder
 * will be build, with code generated to the output folder. Any other files in
 * the source folder are
 * copied directly to the output folder.
 *
 * The builder will never use files in the output folder as input. By default,
 * the clean process will
 * avoid actually deleting files until it sees which files are overwritten by
 * the build process.
 *
 * The output code can be built automatically by running another builder on the
 * output folder after the Magnolia builder. Source files in other languages are
 * copied automatically to the output folder.
 *
 * @author anya
 *
 */
@NonNullByDefault
public class MagnoliaBuilder extends IncrementalProjectBuilder {
	public static final String PREF_PRELUDE_FILE = "org/magnolialang/magnolia/backend/MagnoliaPrelude.cxx";
	public static final String BUILDER_ID = MagnoliaPlugin.PLUGIN_ID + ".magnolia.builder";


	static void checkCancel(@Nullable IProgressMonitor monitor) {
		if(monitor != null && monitor.isCanceled()) {
			throw new OperationCanceledException();
		}
	}

	@Nullable
	private String prelude = null;


	private final ILanguage magnoliaLanguage = NullnessHelper.checkNonNull(LanguageRegistry.getLanguage("Magnolia"));


	/**
	 * Generate code for a module and store the results in the file system.
	 *
	 * @param manager
	 *            The project manager.
	 * @param tm
	 *            A progress monitor.
	 * @param pkg
	 *            The name of the module that should be build
	 * @param toClean
	 *            List of files to clean
	 * @return Updated list of files to clean (excluding newly generated files)
	 * @throws CoreException
	 */
	public List<IResource> buildOutputFiles(IProjectManager manager, ITaskMonitor tm, IMagnoliaPackage pkg, List<IResource> toClean) throws CoreException {
		FuturesQueue<Pair<String, IPath>> queue = new FuturesQueue<>(manager.getResourceManager().getThreadPool());
		prepareOutputFiles(manager, tm, pkg, queue);
		doOutput(toClean, queue, true);
		return toClean;
	}

		if(kind == FULL_BUILD || kind == CLEAN_BUILD) { // FULL_BUILD implies CLEAN_BUILD

	/**
	 * Copy all interesting files and directories from the src folder to the
	 * output folder.
	 * Excludes any files that are managed by Magnolia and will be run through
	 * the Magnolia compiler.
	 *
	 * @param monitor
	 *            Progress monitor
	 * @param toClean
	 *            List of files to clean
	 * @return Updated list of files to clean (exculding any copied files)
	 * @throws CoreException
	 *             In case of file system trouble
	 */
	public List<IResource> copyFiles(@Nullable final IProgressMonitor monitor, final List<IResource> toClean) throws CoreException {
		final IProject project = getProject();
		final IProjectManager manager = Infra.getResourceManager(project.getName());
		final IPath outPath = EclipsePicaInfra.toIPath(manager.getOutFolder());
		final IPath srcPath = EclipsePicaInfra.toIPath(manager.getSrcFolder());
		final IFolder outFolder = project.getWorkspace().getRoot().getFolder(outPath);
		if(!outFolder.exists()) {
			outFolder.create(IResource.DERIVED, true, monitor);
		}
		outFolder.refreshLocal(IResource.DEPTH_INFINITE, monitor);
		IWorkspaceRoot root = project.getWorkspace().getRoot();
		final IResource srcFolder = root.findMember(srcPath);
		srcFolder.accept(new IResourceVisitor() {
			@Override
			public boolean visit(@Nullable IResource resource) throws CoreException {
				assert resource != null;
				checkCancel(monitor);
				//if(isInterrupted())
				//	return false;
				// System.err.println(srcPath);
				IPath relPath = resource.getFullPath().makeRelativeTo(srcPath);
				if(getPref(Pref.BUILD_DEBUG)) {
					System.err.println("Visiting " + resource.getFullPath());
				}
				if(resource.getFullPath().equals(outFolder.getFullPath())) {
					if(getPref(Pref.BUILD_DEBUG)) {
						System.err.println("  ... is the output folder!");
					}
					return false;
				}
				else if(resource.isHidden() || resource.isTeamPrivateMember() || resource.getName().startsWith(".")) {
					if(getPref(Pref.BUILD_DEBUG)) {
						System.err.println("  ... is hidden!");
					}
					return false;
				}
				else if(resource.getType() == IResource.FOLDER) {
					IFolder folder = outFolder.getFolder(relPath);
					if(folder.exists()) {
						if(folder.getType() == IResource.FOLDER) {
							if(getPref(Pref.BUILD_DEBUG)) {
								System.err.println("  ... already exists: " + folder.getFullPath());
							}
							toClean.remove(folder);
							return true;
						}
						else if(folder.isDerived()) {
							folder.delete(getPref(Pref.BUILD_CLEAN_OUT_OF_SYNC), monitor);
							folder.create(IResource.DERIVED, true, monitor);
							if(getPref(Pref.BUILD_DEBUG)) {
								System.err.println("  ... replaced: " + folder.getFullPath());
							}
							toClean.remove(folder);
							return true;
						}
						else {
							throw new CoreException(new Status(IStatus.ERROR, MagnoliaPlugin.PLUGIN_ID, "Path already exists, and is not a folder: " + folder.getFullPath()));
						}
					}

			toClean = findCleanFiles(monitor, toClean);

					IManagedCodeUnit mres = manager.findCodeUnit(MagnoliaPlugin.constructProjectURI(resource.getProject(), resource.getProjectRelativePath()));
					if(mres != null) {
						if(getPref(Pref.BUILD_DEBUG)) {
							System.err.println("  ... is managed by " + mres.getLanguage());
						}
						return false;
					}
					else if(file.exists() && file.getType() == IResource.FILE) {
						// System.err.println(file.getLocalTimeStamp() + "\n" +
// pkg.getLocalTimeStamp());
						if(file.getLocalTimeStamp() >= resource.getLocalTimeStamp()) {
							if(getPref(Pref.BUILD_DEBUG)) {
								System.err.println("  ... exists and up-to-date: " + file.getFullPath());
							}
							toClean.remove(file);
							return false;
						}
						else if(file.isDerived()) {
							file.delete(getPref(Pref.BUILD_CLEAN_OUT_OF_SYNC), monitor);
							resource.copy(file.getFullPath(), IResource.DERIVED, monitor);
							if(getPref(Pref.BUILD_DEBUG)) {
								System.err.println("  ... exists and updated: " + file.getFullPath());
							}
							toClean.remove(file);
							return false;
						}
						else {
							return false;
							// else
							// throw new CoreException(new Status(Status.ERROR,
// MagnoliaPlugin.PLUGIN_ID, "Path already exists: " + file.getFullPath()));
						}
					}
					else {
						resource.copy(file.getFullPath(), IResource.DERIVED, monitor);
						if(getPref(Pref.BUILD_DEBUG)) {
							System.err.println("  ... copied: " + file.getFullPath());
						}
						toClean.remove(file);
						return false;
					}

		try {
			if(kind == FULL_BUILD || kind == AUTO_BUILD || kind == INCREMENTAL_BUILD) {
				toClean = copyFiles(monitor, toClean);
				if(!getPref(Pref.BUILD_COPY_ONLY)) {
					toClean = build(monitor, toClean);
				}
			}
		}
		catch(OperationCanceledException e) {
			rememberLastBuiltState();
		}

		cleanFiles(toClean, monitor);

		return new IProject[0];
	}


	/**
	 * Magnolia Builder. Typecheck all modules in the project, and generate code
	 * for them.
	 *
	 * @param monitor
	 *            A progress monitor
	 * @param toClean
	 *            List of files that should be cleaned later on
	 * @return Updated list of files to clean (excluding any recently built
	 *         files)
	 * @throws CoreException
	 *             In case of any file system errors
	 */
	private List<IResource> build(@Nullable IProgressMonitor monitor, List<IResource> toClean) throws CoreException {
		final int COMPILE_SINGLE_MODULE_WORK = 100;
		IProject project = getProject();
		IResourceDelta delta = getDelta(project);
		Collection<IMagnoliaPackage> packages = null;
		IProjectManager manager = Infra.getResourceManager(project.getName());

		ITaskMonitor tm = new EclipseTaskMonitor(SubMonitor.convert(monitor));
		tm.begin(new TaskId(getClass().getCanonicalName(), "Building project " + project.getName(), project.getName()));
		manager.processChanges(tm);
		long t0 = System.currentTimeMillis();
		try {
			tm.subTask("Obtaining dependency graph for " + project.getName());
			IDepGraph<IManagedCodeUnit> depGraph = manager.getPackageDependencyGraph(tm);
			if(delta != null) {
				tm.subTask("Checking resource delta for " + project.getName());
				Pair<Set<IManagedResource>, Set<IPath>> pair = ((EclipseWorkspaceManager) Infra.getWorkspaceManager()).getResourcesForDelta(delta);
				Set<IManagedResource> changed = pair.getFirst();
				Set<IPath> removed = pair.getSecond();
				cleanOutputFor(removed);
				packages = new HashSet<>();
				tm.setWorkTodo(changed.size() * 2);
				for(IManagedResource r : changed) {
					if(r instanceof IManagedCodeUnit) {
						IManagedCodeUnit cu = (IManagedCodeUnit) r;
						if(cu != null && cu.getLanguage().equals(Magnolia.getInstance())) {
							tm.subTask("Adding dependencies for " + cu.getName());
							packages.add((IMagnoliaPackage) cu);
							for(IManagedCodeUnit u : depGraph.getTransitiveDependents(cu)) {
								if(u instanceof IMagnoliaPackage) {
									packages.add((IMagnoliaPackage) u);
								}
							}
						}
					}
				}
			}
			else {
				packages = (Collection<IMagnoliaPackage>) manager.allPackages(magnoliaLanguage);
			}
			System.out.println("Building " + project.getName());
			System.out.println("  The following " + packages.size() + " will be built:");
			for(IManagedCodeUnit pkg : packages) {
				System.out.println("    " + pkg.getName());
			}

			tm.setWorkTodo(packages.size() * COMPILE_SINGLE_MODULE_WORK);
			FuturesQueue<Pair<String, IPath>> outputQueue = new FuturesQueue<>(manager.getResourceManager().getThreadPool());
			ITopologicalWorkQueue<IManagedCodeUnit> queue = depGraph.topologicalWorkQueue();
			IManagedCodeUnit cu = null;
			cu = queue.waitForNext();
			while(cu != null) {
				try {
					if(packages.contains(cu) && cu instanceof IMagnoliaPackage) {
						IMagnoliaPackage p = (IMagnoliaPackage) cu;
						checkCancel(monitor);

						try {
							System.err.println("COMPILING: " + p);
							tm.subTask("Compiling " + p);
							((EclipseWorkspaceManager) Infra.getWorkspaceManager()).dumpThreadingStats();
							//rm.startJob("Compiling " + p, COMPILE_SINGLE_MODULE_WORK);
							p.getTypechecked(tm);
							manager.commitMarks();
							prepareOutputFiles(manager, tm, p, outputQueue);

						}
						catch(Throwable t) {
							MagnoliaPlugin.getInstance().logException("INTERNAL COMPILER ERROR (while compiling " + p + ")", t);
							if(t.getStackTrace() != null) {
								t.printStackTrace();
							}
							else {
								System.err.println(t);
							}
						}
					}
				}
				finally {
					queue.done(cu);
				}
				cu = queue.next();
				doOutput(toClean, outputQueue, false);
				if(cu == null) {// no work available, do some output instead

					cu = queue.waitForNext();// wait if necessary
				}
				if(isInterrupted()) {
					System.err.println("BUILD INTERRUPTED!" + Thread.currentThread().getId());
					//if(delta != null) {
					//	rememberLastBuiltState();
					//	break;
					//}
				}
			}

			// flush the output queue
			toClean = doOutput(toClean, outputQueue, true);
		}
		finally {
			manager.commitMarks();
			System.err.println("Magnolia builder for project " + project.getName() + " finished: " + (System.currentTimeMillis() - t0) + "ms");
			tm.done();
		}
		return toClean;
	}


	/**
	 * Generate code for a module and store the results in the file system.
	 *
	 * @param manager
	 *            The project manager.
	 * @param rm
	 *            A progress monitor.
	 * @param pkg
	 *            The name of the module that should be build
	 * @param toClean
	 *            List of files to clean
	 * @return Updated list of files to clean (excluding newly generated files)
	 * @throws CoreException
	 */
	public List<IResource> buildOutputFiles(IProjectManager manager, IRascalMonitor rm, IMagnoliaPackage pkg, List<IResource> toClean) throws CoreException {
		FuturesQueue<Pair<String, IPath>> queue = new FuturesQueue<>(manager.getResourceManager().getThreadPool());
		prepareOutputFiles(manager, rm, pkg, queue);
		doOutput(toClean, queue, true);
		return toClean;
	}


	@Override
	protected void clean(@Nullable IProgressMonitor monitor) throws CoreException {
		clean(new RascalMonitor(monitor, new WarningsToMarkers()));
		List<IResource> toClean = findCleanFiles(monitor, new ArrayList<IResource>());
		cleanFiles(toClean, monitor);
	}


	/**
	 * Clean internal Magnolia resources.
	 *
	 * @param monitor
	 *            A progress monitor.
	 */
	private void clean(IRascalMonitor monitor) {
		IProject project = getProject();
		IProjectManager manager = Infra.getResourceManager(project.getName());
		manager.refresh();
		prelude = null;
	}


	/**
	 * Remove all files in the toClean list.
	 *
	 * @param toClean
	 *            Files to be removed
	 * @param monitor
	 * @throws CoreException
	 */
	private void cleanFiles(List<IResource> toClean, @Nullable IProgressMonitor monitor) throws CoreException {
		while(!toClean.isEmpty()) {
			checkCancel(monitor);
			if(isInterrupted()) {
				break;
			}
			IResource res = toClean.remove(toClean.size() - 1);
			if(getPref(Pref.BUILD_DEBUG)) {
				System.err.println("Removing " + res);
			}
			if(res != null) {
				res.delete(getPref(Pref.BUILD_CLEAN_OUT_OF_SYNC), monitor);
			}
		}
	}


	private void cleanOutputFor(Set<IPath> removed) {
		// TODO Auto-generated method stub

	}


	protected void clearMarkersOn(final Collection<IFile> files) {
		for(final IFile file : files) {
			clearMarkersOn(file);
		}
	}


	/**
	 * Clears all problem markers (all markers whose type derives from
	 * IMarker.PROBLEM) from the given file. A utility method for the use of
	 * derived builder classes.
	 */
	protected void clearMarkersOn(final IFile file) {
		try {
			// SMS 28 Mar 2007
			// Clear the markers for this builder only (and clear all of them)
			// (may be a simpler way to do this, given a more complex set up of
			// marker types)
			// file.deleteMarkers(IMarker.PROBLEM, true,
			// IResource.DEPTH_INFINITE);
			file.deleteMarkers(ErrorMarkers.TYPE_COMPILATION_ERROR, true, IResource.DEPTH_INFINITE);
		}
		catch(final CoreException e) { // NOPMD by anya on 1/5/12 5:30 AM
		}
	}


	/**
	 * Copy all interesting files and directories from the src folder to the
	 * output folder.
	 * Excludes any files that are managed by Magnolia and will be run through
	 * the Magnolia compiler.
	 *
	 * @param monitor
	 *            Progress monitor
	 * @param toClean
	 *            List of files to clean
	 * @return Updated list of files to clean (exculding any copied files)
	 * @throws CoreException
	 *             In case of file system trouble
	 */
	public List<IResource> copyFiles(@Nullable final IProgressMonitor monitor, final List<IResource> toClean) throws CoreException {
		final IProject project = getProject();
		final IProjectManager manager = Infra.getResourceManager(project.getName());
		final IPath outPath = manager.getOutFolder();
		final IPath srcPath = manager.getSrcFolder();
		final IFolder outFolder = project.getWorkspace().getRoot().getFolder(outPath);
		if(!outFolder.exists()) {
			outFolder.create(IResource.DERIVED, true, monitor);
		}
		outFolder.refreshLocal(IResource.DEPTH_INFINITE, monitor);
		IWorkspaceRoot root = project.getWorkspace().getRoot();
		final IResource srcFolder = root.findMember(manager.getSrcFolder());
		srcFolder.accept(new IResourceVisitor() {
			@Override
			public boolean visit(@Nullable IResource resource) throws CoreException {
				assert resource != null;
				checkCancel(monitor);
				//if(isInterrupted())
				//	return false;
				// System.err.println(srcPath);
				IPath relPath = resource.getFullPath().makeRelativeTo(srcPath);
				if(getPref(Pref.BUILD_DEBUG)) {
					System.err.println("Visiting " + resource.getFullPath());
				}
				if(resource.getFullPath().equals(outFolder.getFullPath())) {
					if(getPref(Pref.BUILD_DEBUG)) {
						System.err.println("  ... is the output folder!");
					}
					return false;
				}
				else if(resource.isHidden() || resource.isTeamPrivateMember() || resource.getName().startsWith(".")) {
					if(getPref(Pref.BUILD_DEBUG)) {
						System.err.println("  ... is hidden!");
					}
					return false;
				}
				else if(resource.getType() == IResource.FOLDER) {
					IFolder folder = outFolder.getFolder(relPath);
					if(folder.exists()) {
						if(folder.getType() == IResource.FOLDER) {
							if(getPref(Pref.BUILD_DEBUG)) {
								System.err.println("  ... already exists: " + folder.getFullPath());
							}
							toClean.remove(folder);
							return true;
						}
						else if(folder.isDerived()) {
							folder.delete(getPref(Pref.BUILD_CLEAN_OUT_OF_SYNC), monitor);
							folder.create(IResource.DERIVED, true, monitor);
							if(getPref(Pref.BUILD_DEBUG)) {
								System.err.println("  ... replaced: " + folder.getFullPath());
							}
							toClean.remove(folder);
							return true;
						}
						else {
							throw new CoreException(new Status(IStatus.ERROR, MagnoliaPlugin.PLUGIN_ID, "Path already exists, and is not a folder: " + folder.getFullPath()));
						}
					}

					folder.create(IResource.DERIVED, true, monitor);
					if(getPref(Pref.BUILD_DEBUG)) {
						System.err.println("  ... created: " + folder.getFullPath());
					}
					toClean.remove(folder);
					return true;
				}
				else if(resource.getType() == IResource.FILE) {
					IFile file = outFolder.getFile(relPath);

					IManagedCodeUnit mres = manager.findPackage(MagnoliaPlugin.constructProjectURI(resource.getProject(), resource.getProjectRelativePath()));
					if(mres != null) {
						if(getPref(Pref.BUILD_DEBUG)) {
							System.err.println("  ... is managed by " + mres.getLanguage());
						}
						return false;
					}
					else if(file.exists() && file.getType() == IResource.FILE) {
						// System.err.println(file.getLocalTimeStamp() + "\n" +
// pkg.getLocalTimeStamp());
						if(file.getLocalTimeStamp() >= resource.getLocalTimeStamp()) {
							if(getPref(Pref.BUILD_DEBUG)) {
								System.err.println("  ... exists and up-to-date: " + file.getFullPath());
							}
							toClean.remove(file);
							return false;
						}
						else if(file.isDerived()) {
							file.delete(getPref(Pref.BUILD_CLEAN_OUT_OF_SYNC), monitor);
							resource.copy(file.getFullPath(), IResource.DERIVED, monitor);
							if(getPref(Pref.BUILD_DEBUG)) {
								System.err.println("  ... exists and updated: " + file.getFullPath());
							}
							toClean.remove(file);
							return false;
						}
						else {
							return false;
							// else
							// throw new CoreException(new Status(Status.ERROR,
// MagnoliaPlugin.PLUGIN_ID, "Path already exists: " + file.getFullPath()));
						}
					}
					else {
						resource.copy(file.getFullPath(), IResource.DERIVED, monitor);
						if(getPref(Pref.BUILD_DEBUG)) {
							System.err.println("  ... copied: " + file.getFullPath());
						}
						toClean.remove(file);
						return false;
					}

				}
				else {
					if(getPref(Pref.BUILD_DEBUG)) {
						System.err.println("  Not a file or folder: " + resource.getFullPath());
					}
					IResource res = outFolder.findMember(relPath);
					if(res != null) {
						toClean.remove(res);
					}

					return true;
				}
			}

		});
		return toClean;
	}


	/**
	 * Create a file. All necessary parent directories will be created as well.
	 *
	 * @param code
	 *            The new file's contents
	 * @param path
	 *            Path to the new file
	 * @param toClean
	 *            list of files to clean
	 * @return updated list of files to clean (excluding the new file and its
	 *         parents)
	 * @throws CoreException
	 *             if anything went wrong
	 */
	private List<IResource> createFile(String code, IPath path, List<IResource> toClean) throws CoreException {
		IPath dir = path.removeLastSegments(1);
		IContainer parent = EclipseWorkspaceManager.mkdir(dir, IResource.DERIVED);
		IFile file = parent.getFile(new Path(path.lastSegment()));
		if(file.exists()) {
			file.setContents(new ByteArrayInputStream(code.getBytes()), IResource.FORCE, null);//getPref(Pref.BUILD_CLEAN_OUT_OF_SYNC) ? IResource.FORCE : IResource.NONE, null);
		}
		else {
			file.create(new ByteArrayInputStream(code.getBytes()), IResource.DERIVED, null);
		}
		toClean.remove(file);
		return toClean;
	}


	@SuppressWarnings("null")
	private List<IResource> doOutput(List<IResource> toClean, FuturesQueue<Pair<String, IPath>> outputQueue, boolean block) throws CoreException {
		Future<Pair<String, IPath>> outputFuture = block ? outputQueue.take() : outputQueue.poll();
		while(outputFuture != null) {
			Pair<String, IPath> output;
			try {
				output = outputFuture.get();
				if(output != null) {
					System.err.println("doOutput [" + Thread.currentThread().getId() + "]: writing file " + output.getSecond());
					toClean = createFile(output.getFirst(), output.getSecond(), toClean);
				}
				else {
					System.err.println("doOutput: null result from future ");
				}
			}
			catch(InterruptedException e) {
				e.printStackTrace();
			}
			catch(ExecutionException e) {
				Throwable t = e.getCause();
				if(t instanceof Throw) {
					System.err.println(((Throw) t).getTrace());
					t = RascalErrors.decodeRascalError((Throw) t);
				}
				t.printStackTrace();
			}
			outputFuture = block ? outputQueue.take() : outputQueue.poll();
		}
		return toClean;
	}


	/**
	 * Find all files and directories eligible for cleaning.
	 *
	 * @param monitor
	 * @param toClean
	 *            list to which resources will be added
	 * @return a list of resources to clean
	 * @throws CoreException
	 */
	private List<IResource> findCleanFiles(@Nullable final IProgressMonitor monitor, final List<IResource> toClean) throws CoreException {
		final IProject project = getProject();
		final IProjectManager manager = Infra.getResourceManager(project.getName());
		final IPath outPath = EclipsePicaInfra.toIPath(manager.getOutFolder());
		final IPath srcPath = EclipsePicaInfra.toIPath(manager.getSrcFolder());
		final IWorkspaceRoot root = project.getWorkspace().getRoot();
		final IFolder outFolder = root.getFolder(outPath);
		final boolean onlyDerived = getPref(Pref.BUILD_CLEAN_ONLY_DERIVED);
		final boolean retainHierarchy = getPref(Pref.BUILD_CLEAN_RETAIN_HIERARCHY);
		if(outPath.isPrefixOf(srcPath)) {
			throw new CoreException(new Status(IStatus.ERROR, MagnoliaPlugin.PLUGIN_ID, "Cannot clean: src folder is located within out folder!"));
		}
		outFolder.refreshLocal(IResource.DEPTH_INFINITE, monitor);
		if(outFolder.exists()) {
			outFolder.accept(new IResourceVisitor() {
				@Override
				public boolean visit(@Nullable IResource resource) throws CoreException {
					assert resource != null;
					MagnoliaBuilder.checkCancel(monitor);
					if(!onlyDerived || resource.isDerived()) {
						IFolder folder = root.getFolder(srcPath.append(resource.getFullPath().makeRelativeTo(outPath)));
						if(retainHierarchy && folder.exists()) {
							;
						}
						else {
							if(getPref(Pref.BUILD_DEBUG)) {
								System.err.println("To clean " + resource.getFullPath());
							}
							toClean.add(resource);
						}
					}
					return true;
				}

			});
		}
		return toClean;
	}


	protected boolean getPref(Pref pref) {
		return MagnoliaPlugin.getInstance().getPreferenceStore().getBoolean(pref.name());
	}


	private void loadPrelude() {
		try {
			InputStream inputStream = MagnoliaFileLocator.openStream(new Path(PREF_PRELUDE_FILE));
			if(inputStream != null) {
				char[] cs = InputConverter.toChar(inputStream, Charset.forName("UTF-8"));
				inputStream.close();
				prelude = String.valueOf(cs);
			}
		}
		catch(IOException e) {
			prelude = "";
			e.printStackTrace();
		}
	}


	/**
	 * Generate implementation code and store the result in the file system.
	 *
	 * @param manager
	 *            The project manager
	 * @param compiler
	 *            The compiler
	 * @param tm
	 *            A monitor
	 * @param module
	 *            Name of the implementation to build
	 * @param toClean
	 *            Files to clean
	 * @return Updated list of files to clean (excluding the newly generated
	 *         file(s))
	 * @throws CoreException
	 *             if any file system operation went wrong
	 */
	private void prepareOutput(final IProjectManager manager, final ITaskMonitor tm, final IMagnoliaModule module, FuturesQueue<Pair<String, IPath>> outputQueue) throws CoreException {
		final String outname = module.getName().replaceAll("\\.", "/") + ".cxx";
		if(module.hasModifier("NoBuild")) {
			System.err.println("MagnoliaBuilder: not building @nobuild " + module);
			return;
		}
		else if(module.hasModifier("Broken") && !getPref(Pref.BUILD_BROKEN)) {
			System.err.println("MagnoliaBuilder: not building @broken " + module);
			return;
		}
		System.err.println("MagnoliaBuilder: building " + module);
		IConstructor typechecked = module.getTypechecked(tm);
		IConstructor instantiated = Instantiator.instantiate(typechecked);
		final IConstructor defInfo = module.getDefInfo(tm).set("body", instantiated);
		final ICompiler compiler = Magnolia.getInstance().getCompiler();

		Callable<Pair<String, IPath>> callable = new Callable<Pair<String, IPath>>() {
			@Override
			@Nullable
			public Pair<String, IPath> call() {
//				final IConstructor defInfo = module.getDefInfo(rm).set("body", module.getTypechecked(rm));
//				final ICompiler compiler = module.getLanguage().getCompiler();
				IConstructor plainified = (IConstructor) compiler.call(tm, "getPlainified", defInfo);
				IConstructor myDefInfo = defInfo.set("body", plainified.get("val"));
				IConstructor mutified = (IConstructor) compiler.call(tm, "getMutified", myDefInfo);
				myDefInfo = myDefInfo.set("body", mutified.get("val"));
				IConstructor cxxResult = (IConstructor) compiler.call(tm, "cxxCode", myDefInfo, TermFactory.vf.string(outname), TermFactory.vf.bool(getPref(Pref.BUILD_ERRONEOUS)));
				MarkBuilder buildInfoMark = new MarkBuilder().region(module.getNameRegion()).severity(Severity.INFO).source(getClass().getName() + "#backend").context(module.getURI());
				try {
					if(cxxResult.has("val")) {
						String code = ((IString) cxxResult.get("val")).getValue();
						if(prelude == null) {
							loadPrelude();
						}
						if(code.contains("\n#error")) {
							System.err.print("prepareOutput: generated cxxCode for " + module + " contains errors");
							if(getPref(Pref.BUILD_ERRONEOUS)) {
								System.err.println(" -- continuing anyway");
								buildInfoMark = buildInfoMark.message("C++ code output in '" + outname + "' (contains errors)").severity(Severity.WARNING);
							}
							else {
								System.err.println(" -- suppressing output");
								buildInfoMark = buildInfoMark.message("No C++ code generated, " + module + " contains errors").severity(Severity.WARNING);
								return null;
							}
						}
						else {
							buildInfoMark = buildInfoMark.message("C++ code output in '" + outname + "'");
						}

						code = prelude + "\n" + code;
						IPath path = EclipsePicaInfra.toIPath(manager.getOutFolder()).append(outname);
						return new Pair<>(code, path);
					}
					else {
						System.err.println("prepareOutput: no result from cxxCode for '" + module + "'");
						buildInfoMark = buildInfoMark.message("No C++ code generated for " + module);
						return null;
					}
				}
				finally {
					IMagnoliaPackage pkg = module.getParent();
					pkg.clearMarks(getClass().getName() + "#backend", module);
					pkg.addMarks(plainified, getClass().getName() + "#backend", module);
					pkg.addMarks(mutified, getClass().getName() + "#backend", module);
					pkg.addMarks(cxxResult, getClass().getName() + "#backend", module);
					manager.addMark(buildInfoMark.done());
				}
			}
		};

		outputQueue.submit(callable);
	}


	private void prepareOutputFiles(IProjectManager manager, ITaskMonitor tm, IMagnoliaPackage pkg, FuturesQueue<Pair<String, IPath>> outputQueue) throws CoreException {
		for(IMagnoliaModule module : pkg.getChildren(tm)) {
			if(module.getKind(tm).equals(MagnoliaFacts.KIND_PROGRAM) || module.getKind(tm).equals(MagnoliaFacts.KIND_SATISFACTION)) {
				prepareOutput(manager, tm, module, outputQueue);
			}
			else {
				module.getTypechecked(tm);
				System.err.println("MagnoliaBuilder: not building " + module.getKind(tm) + " " + module);
			}
		}
	}


	@SuppressWarnings("rawtypes")
	@Override
	protected IProject[] build(final int kind, final @Nullable Map args, final @Nullable IProgressMonitor monitor) throws CoreException {
		System.err.println("Build started for project " + getProject().getName());
		List<IResource> toClean = new ArrayList<IResource>();

		if(kind == FULL_BUILD || kind == CLEAN_BUILD) {// FULL_BUILD implies CLEAN_BUILD

			clean(new RascalMonitor(monitor, new WarningsToMarkers()));

			toClean = findCleanFiles(monitor, toClean);

			if(!getPref(Pref.BUILD_VIRTUAL_CLEAN)) {
				cleanFiles(toClean, monitor);
				toClean.clear();
			}
		}

		try {
			if(kind == FULL_BUILD || kind == AUTO_BUILD || kind == INCREMENTAL_BUILD) {
				toClean = copyFiles(monitor, toClean);
				if(!getPref(Pref.BUILD_COPY_ONLY)) {
					toClean = build(monitor, toClean);
				}
			}
		}
		catch(OperationCanceledException e) {
			rememberLastBuiltState();
		}

		cleanFiles(toClean, monitor);

		return new IProject[0];
	}


	@Override
	protected void clean(@Nullable IProgressMonitor monitor) throws CoreException {
		clean(new RascalMonitor(monitor, new WarningsToMarkers()));
		List<IResource> toClean = findCleanFiles(monitor, new ArrayList<IResource>());
		cleanFiles(toClean, monitor);
	}


	protected void clearMarkersOn(final Collection<IFile> files) {
		for(final IFile file : files) {
			clearMarkersOn(file);
		}
	}


	/**
	 * Clears all problem markers (all markers whose type derives from
	 * IMarker.PROBLEM) from the given file. A utility method for the use of
	 * derived builder classes.
	 */
	protected void clearMarkersOn(final IFile file) {
		try {
			// SMS 28 Mar 2007
			// Clear the markers for this builder only (and clear all of them)
			// (may be a simpler way to do this, given a more complex set up of
			// marker types)
			// file.deleteMarkers(IMarker.PROBLEM, true,
			// IResource.DEPTH_INFINITE);
			file.deleteMarkers(ErrorMarkers.TYPE_COMPILATION_ERROR, true, IResource.DEPTH_INFINITE);
		}
		catch(final CoreException e) {// NOPMD by anya on 1/5/12 5:30 AM
		}
	}


	protected boolean getPref(Pref pref) {
		return MagnoliaPlugin.getInstance().getPreferenceStore().getBoolean(pref.name());
	}


	static void checkCancel(@Nullable IProgressMonitor monitor) {
		if(monitor != null && monitor.isCanceled()) {
			throw new OperationCanceledException();
		}
	}

}
