/**************************************************************************
 * Copyright (c) 2011-2014 Anya Helene Bagge
 * Copyright (c) 2011-2014 University of Bergen
 *
 * All rights reserved. This program and the accompanying materials
 * are made available under the terms of the Eclipse Public License v1.0
 * which accompanies this distribution, and is available at
 * http://www.eclipse.org/legal/epl-v10.html
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version. See http://www.gnu.org/licenses/
 *
 *
 * See the file COPYRIGHT for more information.
 *
 * Contributors:
 * * Andreas Hjortland
 * * Anya Helene Bagge
 *
 *************************************************************************/
package org.magnolialang.eclipse.commands;

import java.io.InputStreamReader;
import java.io.PrintWriter;
import java.io.Reader;

import org.eclipse.core.commands.ExecutionException;
import io.usethesource.impulse.editor.UniversalEditor;
import io.usethesource.impulse.model.ISourceProject;
import io.usethesource.impulse.parser.IParseController;
import org.rascalmpl.value.IConstructor;
import org.rascalmpl.value.IValue;
import org.eclipse.jdt.annotation.NonNullByDefault;
import org.eclipse.jdt.annotation.Nullable;
import org.eclipse.swt.SWT;
import org.eclipse.swt.widgets.Display;
import org.eclipse.ui.IEditorPart;
import org.eclipse.ui.IWorkbenchPage;
import org.eclipse.ui.PartInitException;
import org.eclipse.ui.PlatformUI;
import org.eclipse.ui.console.ConsolePlugin;
import org.eclipse.ui.console.IConsole;
import org.eclipse.ui.console.IConsoleConstants;
import org.eclipse.ui.console.IConsoleManager;
import org.eclipse.ui.console.IConsoleView;
import org.eclipse.ui.console.IOConsole;
import org.eclipse.ui.console.IOConsoleInputStream;
import org.eclipse.ui.console.IOConsoleOutputStream;
import org.magnolialang.eclipse.MagnoliaPlugin;
import org.magnolialang.infra.Infra;
import org.magnolialang.magnolia.Magnolia;
import org.magnolialang.magnolia.MagnoliaCompiler;
import org.magnolialang.magnolia.resources.IMagnoliaPackage;
import org.nuthatchery.pica.resources.ILanguage;
import org.nuthatchery.pica.resources.IProjectManager;
import org.nuthatchery.pica.resources.LanguageRegistry;
import org.nuthatchery.pica.resources.managed.IManagedCodeUnit;

@NonNullByDefault
public final class CommandUtil {
	private CommandUtil() {
	}


	public static void commitMarks(IParseController parseCtrl) {
		try {
			getModuleManager(parseCtrl).commitMarks();
		}
		catch(ExecutionException e) {
			e.printStackTrace();
		}
	}


	public static MagnoliaCompiler getCompiler(IParseController parseCtrl) throws ExecutionException {
		return (MagnoliaCompiler) Magnolia.getInstance().getCompiler();
	}


	public static IOConsole getConsole() {
		ConsolePlugin plugin = ConsolePlugin.getDefault();
		IConsoleManager conMan = plugin.getConsoleManager();
		IConsole[] existing = conMan.getConsoles();
		for(IConsole element : existing) {
			if(MagnoliaPlugin.MAGNOLIA_CONSOLE_NAME.equals(element.getName())) {
				return (IOConsole) element;
			}
		}
		//no console found, so create a new one
		IOConsole myConsole = new IOConsole(MagnoliaPlugin.MAGNOLIA_CONSOLE_NAME, null);
		conMan.addConsoles(new IConsole[] { myConsole });

		return myConsole;
	}


	public static PrintWriter getConsoleErrorWriter() {
		IOConsole console = getConsole();
		IOConsoleOutputStream stream = console.newOutputStream();
		stream.setColor(Display.getDefault().getSystemColor(SWT.COLOR_RED));

		return new PrintWriter(stream);
	}


	public static Reader getConsoleInputReader() {
		IOConsole console = getConsole();
		IOConsoleInputStream stream = console.getInputStream();
		stream.setColor(Display.getDefault().getSystemColor(SWT.COLOR_GREEN));

		return new InputStreamReader(stream);
	}


	public static PrintWriter getConsoleOutputWriter() {
		IOConsole console = getConsole();
		IOConsoleOutputStream stream = console.newOutputStream();
		stream.setColor(Display.getDefault().getSystemColor(SWT.COLOR_DARK_MAGENTA));

		return new PrintWriter(stream);
	}


	public static ILanguage getLanguage(IParseController parseCtrl) throws ExecutionException {
		ILanguage language = LanguageRegistry.getLanguage(parseCtrl.getLanguage().getName());

		if(language == null) {
			throw new ExecutionException("Language not found for this editor");
		}

		return language;
	}


	public static IConstructor getModuleId(IParseController parseCtrl) throws ExecutionException {
		IConstructor id = getResource(parseCtrl).getId();

		return id;
	}


	public static IProjectManager getModuleManager(IParseController parseCtrl) throws ExecutionException {
		ISourceProject srcProject = parseCtrl.getProject();

		if(srcProject == null) {
			throw new ExecutionException("Module manager not found for file " + parseCtrl.getPath());
		}
		return Infra.getResourceManager(srcProject.getRawProject().getName());
	}


	public static IParseController getParseController(IEditorPart editorPart) {
		assert editorPart instanceof UniversalEditor;
		return ((UniversalEditor) editorPart).getParseController();
	}


	public static IMagnoliaPackage getResource(IParseController parseCtrl) throws ExecutionException {
		IManagedCodeUnit res = getModuleManager(parseCtrl).findCodeUnit(MagnoliaPlugin.constructProjectURI(parseCtrl.getProject(), parseCtrl.getPath()));

		if(res == null) {
			throw new ExecutionException("No managed pkg found for " + parseCtrl.getPath());
		}

		if(!(res instanceof IMagnoliaPackage)) {
			throw new ExecutionException("Not a Magnolia package " + parseCtrl.getPath());
		}
		return (IMagnoliaPackage) res;
	}


	@Nullable
	public static IValue getVal(IValue val) {
		if(val instanceof IConstructor) {
			return ((IConstructor) val).get("val");
		}
		else {
			return null;
		}
	}


	public static void showConsole() {
		IConsole myConsole = CommandUtil.getConsole();
		IWorkbenchPage page = PlatformUI.getWorkbench().getActiveWorkbenchWindow().getActivePage();
		String id = IConsoleConstants.ID_CONSOLE_VIEW;
		IConsoleView view;
		try {
			view = (IConsoleView) page.showView(id);
			view.display(myConsole);
		}
		catch(PartInitException e2) {
			// ignored
		}
	}

}
