package org.magnolialang.eclipse.commands;

import java.io.ByteArrayInputStream;
import java.net.URISyntaxException;
import java.util.ArrayList;
import java.util.List;

import org.eclipse.core.resources.IContainer;
import org.eclipse.core.resources.IFile;
import org.eclipse.core.resources.IFolder;
import org.eclipse.core.resources.IProject;
import org.eclipse.core.resources.IResource;
import org.eclipse.core.resources.IWorkspaceRoot;
import org.eclipse.core.runtime.CoreException;
import org.eclipse.core.runtime.IPath;
import org.eclipse.core.runtime.IStatus;
import org.eclipse.core.runtime.Path;
import org.eclipse.core.runtime.Status;
import org.eclipse.jface.dialogs.ErrorDialog;
import org.eclipse.jface.dialogs.MessageDialog;
import org.eclipse.swt.widgets.Display;
import org.magnolialang.eclipse.MagnoliaPlugin;
import org.magnolialang.eclipse.editor.Editor;
import org.nuthatchery.pica.errors.ImplementationError;
import org.nuthatchery.pica.resources.IProjectManager;
import org.rascalmpl.uri.URIUtil;

class GeneratedFileSaver implements Runnable {
	final List<GeneratedFile> files = new ArrayList<GeneratedFile>();;

	boolean running = false;
	String title;


	public GeneratedFileSaver(String title) {
		this.title = title;
	}


	/**
	 * Save a file and show it in an editor.
	 *
	 * The saver should be started with
	 * <code>Display.getDefault().asyncExec(saver);</code>.
	 *
	 * The file will then asynchroniously be picked up by the file saver and
	 * saved to
	 * disk + shown in a new editor. The user will be asked for confirmation
	 * before overwriting the file.
	 *
	 * @param name
	 *            Filename, relative to the folder indicated by 'where'
	 * @param code
	 *            File contents
	 * @param manager
	 *            Pica project manager
	 * @param project
	 *            Eclipse project
	 * @param where
	 *            Where to save the file
	 * @param Flags
	 *            to be applied if the file is created (e.g.,
	 *            {@link IResource#DERIVED})
	 */
	public synchronized void addFile(String name, String code, IProjectManager manager, IProject project, Root where, int flags) {
		files.add(new GeneratedFile(name, code, manager, project, where, flags));
	}


	@Override
	public void run() {
		GeneratedFile genFile = null;
		synchronized(this) {
			if(running) {
				return;
			}

			genFile = getNextFile();
		}
		try {
			while(genFile != null) {
				IPath path;
				switch(genFile.where) {
				case PROJECTROOT:
					path = genFile.project.getFullPath().append(genFile.fileName);
					break;
				case SOURCEFOLDER:
					path = new Path(genFile.manager.getSrcFolder().toString()).append(genFile.fileName);
					break;
				case OUTPUTFOLDER:
					path = new Path(genFile.manager.getOutFolder().toString()).append(genFile.fileName);
					break;
				case WORKSPACEROOT:
					path = new Path(genFile.fileName);
					break;
				default:
					throw new ImplementationError("This can't happen");
				}
				IWorkspaceRoot root = genFile.project.getWorkspace().getRoot();
				final String dialogTitle = "Generating " + title;
				try {
					IFile file = root.getFile(path);
					IContainer parent = file.getParent();

					if(!parent.exists()) {
						boolean confirm = MessageDialog.openConfirm(Display.getDefault().getActiveShell(), dialogTitle, "Parent directory " + parent.getName() + " does not exist. Create?");
						if(confirm) {
							createDir(genFile.project, parent.getFullPath());
						}
						else {
							file = null;
						}
					}

					if(file != null) {
						if(file.exists()) {
							boolean confirm = MessageDialog.openConfirm(Display.getDefault().getActiveShell(), dialogTitle, "File " + path.toString() + " already exists. Overwrite?");
							if(confirm) {
								file.setContents(new ByteArrayInputStream(genFile.code.getBytes()), IResource.KEEP_HISTORY, null);
								Editor.edit(path);
							}
							else {
								try {
									Editor.edit(genFile.code, URIUtil.create("mggen", genFile.project.getName(), file.getFullPath().removeFirstSegments(1).makeAbsolute().toString(), null, null));
								}
								catch(URISyntaxException e) {
									// TODO Auto-generated catch block
									e.printStackTrace();
								}
							}
						}
						else {
							file.create(new ByteArrayInputStream(genFile.code.getBytes()), genFile.flags, null);
							Editor.edit(path);
						}
					}
				}
				catch(CoreException e) {
					ErrorDialog.openError(Display.getCurrent().getActiveShell(), dialogTitle, "Failed to create " + path, e.getStatus());
				}
				genFile = getNextFile();
			}
		}
		finally {
			synchronized(this) {
				running = false;
			}
		}

	}


	private void createDir(IProject project, IPath dir) throws CoreException {
		IContainer parent = project.getWorkspace().getRoot();
		for(String s : dir.segments()) {
			IResource member = parent.findMember(s);
			if(member == null) {
				parent = parent.getFolder(new Path(s));
				((IFolder) parent).create(IResource.NONE, true, null);
			}
			else if(member instanceof IContainer) {
				parent = (IContainer) member;
			}
			else {
				throw new CoreException(new Status(IStatus.ERROR, MagnoliaPlugin.PLUGIN_ID, "Path already exists, and is not a folder: " + member.getFullPath()));
			}
		}
	}


	private synchronized GeneratedFile getNextFile() {
		if(files.isEmpty()) {
			running = false;
			return null;
		}
		else {
			running = true;
			return files.remove(0);
		}
	}


	static class GeneratedFile {
		final String fileName;
		final String code;
		final IProjectManager manager;
		final IProject project;
		final Root where;
		final int flags;


		public GeneratedFile(String fileName, String code, IProjectManager manager, IProject project, Root where, int flags) {
			super();
			this.fileName = fileName;
			this.code = code;
			this.manager = manager;
			this.project = project;
			this.where = where;
			this.flags = flags;
		}

	}


	enum Root {
		WORKSPACEROOT, PROJECTROOT, SOURCEFOLDER, OUTPUTFOLDER
	}

}
