/**************************************************************************
 * Copyright (c) 2012-2014 Anya Helene Bagge
 * Copyright (c) 2012-2014 University of Bergen
 *
 * All rights reserved. This program and the accompanying materials
 * are made available under the terms of the Eclipse Public License v1.0
 * which accompanies this distribution, and is available at
 * http://www.eclipse.org/legal/epl-v10.html
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version. See http://www.gnu.org/licenses/
 *
 *
 * See the file COPYRIGHT for more information.
 *
 * Contributors:
 * * Anya Helene Bagge
 *
 *************************************************************************/
package org.magnolialang.eclipse.commands;

import java.util.ArrayList;
import java.util.Collection;
import java.util.Collections;
import java.util.Comparator;
import java.util.List;
import java.util.Set;

import org.eclipse.core.commands.AbstractHandler;
import org.eclipse.core.commands.ExecutionEvent;
import org.eclipse.core.commands.ExecutionException;
import org.eclipse.core.runtime.IProgressMonitor;
import org.eclipse.core.runtime.IStatus;
import org.eclipse.core.runtime.Status;
import org.eclipse.core.runtime.SubMonitor;
import org.eclipse.core.runtime.jobs.Job;
import io.usethesource.impulse.editor.UniversalEditor;
import io.usethesource.impulse.parser.IParseController;
import org.rascalmpl.value.IConstructor;
import org.rascalmpl.value.IValue;
import org.eclipse.ui.IEditorPart;
import org.eclipse.ui.handlers.HandlerUtil;
import org.magnolialang.eclipse.MagnoliaPlugin;
import org.magnolialang.eclipse.editor.Editor;
import org.magnolialang.eclipse.views.ProjectGraph;
import org.magnolialang.errors.CompilationErrorException;
import org.magnolialang.magnolia.resources.IMagnoliaCodeUnit;
import org.magnolialang.magnolia.resources.IMagnoliaPackage;
import org.nuthatchery.pica.resources.IProjectManager;
import org.nuthatchery.pica.resources.managed.IManagedCodeUnit;
import org.nuthatchery.pica.resources.managed.IManagedResource;
import org.nuthatchery.pica.tasks.ITaskMonitor;
import org.nuthatchery.pica.tasks.TaskId;
import org.nuthatchery.pica.tasks.eclipse.EclipseTaskMonitor;
import org.nuthatchery.pica.terms.TermAdapter;
import org.nuthatchery.pica.util.depgraph.IDepGraph;
import org.nuthatchery.pica.util.depgraph.UnsyncedDepGraph;
import org.nuthatchery.pica.util.tests.DepGraphAxioms;
import org.rascalmpl.eclipse.nature.RascalMonitor;
import org.rascalmpl.eclipse.nature.WarningsToMarkers;
import org.rascalmpl.debug.IRascalMonitor;
import org.rascalmpl.interpreter.control_exceptions.Throw;

public class ProjectOverview extends AbstractHandler {
	public static final String CMD_NAME = "Project Overview";


	@Override
	public Object execute(final ExecutionEvent event) throws ExecutionException {
		final IEditorPart editorPart = HandlerUtil.getActiveEditor(event);
		if(editorPart instanceof UniversalEditor) {
			final IParseController parseCtrl = CommandUtil.getParseController(editorPart);
			final IConstructor symbol = CommandUtil.getModuleId(parseCtrl);
			final IProjectManager mng = CommandUtil.getModuleManager(parseCtrl);
			final IMagnoliaPackage pkg = CommandUtil.getResource(parseCtrl);
			Job job = new Job(CMD_NAME) {
				@Override
				public boolean belongsTo(Object obj) {
					return obj == MagnoliaPlugin.JOB_FAMILY_MAGNOLIA;
				}


				@Override
				protected IStatus run(IProgressMonitor monitor) {
					ITaskMonitor tm = new EclipseTaskMonitor(SubMonitor.convert(monitor));
					tm.begin(new TaskId(getClass().getCanonicalName(), CMD_NAME, pkg.getResource()));
					long t0 = System.currentTimeMillis();
					RefreshHandler.autoRefresh(event);
					long t1 = System.currentTimeMillis();
					try {

						IConstructor info = pkg.getDefInfo(tm);
						Collection<? extends IMagnoliaCodeUnit> contents = pkg.getChildren(tm);
						System.out.println(CMD_NAME + " for " + TermAdapter.yieldTerm(symbol, false));
						StringBuilder result = new StringBuilder();
						result.append("INFO FOR CURRENT FILE\n\n");
						result.append(TermAdapter.yieldTerm(info, false));
						result.append("\n\n\n");
						result.append("INFO FOR CURRENT FILE CONTENTS\n\n");
						for(IMagnoliaCodeUnit c : contents) {
							result.append("\t");
							result.append(c);
							result.append("\n");
						}

						result.append("\n\n\n");
						result.append("PACKAGES\n");
						result.append("--------\n\n");
						List<IMagnoliaPackage> pkgs = new ArrayList<IMagnoliaPackage>();
						for(IManagedResource p : mng.getChildren(tm)) {
							if(p instanceof IMagnoliaPackage) {
								pkgs.add((IMagnoliaPackage) p);
							}
						}
						Collections.sort(pkgs, new Comparator<IMagnoliaPackage>() {
							@Override
							public int compare(IMagnoliaPackage arg0, IMagnoliaPackage arg1) {
								return arg0.getName().compareTo(arg1.getName());
							}
						});
						for(IMagnoliaPackage res : pkgs) {
							result.append("\tpackage ");
							result.append(res.getName());
							result.append("\n");
							result.append("\t\tDependencies: ");
							for(IManagedCodeUnit dep : res.getDepends(tm)) {
								result.append(" ");
								result.append(dep.getName());
							}
							result.append("\n");
							result.append("\t\tDependencies*: ");
							for(IManagedCodeUnit dep : res.getTransitiveDepends(tm)) {
								result.append(" ");
								result.append(dep.getName());
							}
							result.append("\n");
							result.append("\t\tContents:\n");

							for(IMagnoliaCodeUnit c : res.getChildren(tm)) {
								Collection<? extends IManagedCodeUnit> deps = c.getDepends(tm);
								result.append("\t\t\t");
								result.append(c.getName());
								result.append(": ");
								result.append("(");
								result.append(String.valueOf(deps.size()));
								result.append(") ");
								for(IManagedCodeUnit d : deps) {
									result.append(d.getName());
									result.append(" ");
								}
								result.append("\n");

								deps = c.getTransitiveDepends(tm);
								result.append("\t\t\t");
								result.append(c.getName());
								result.append("*: ");
								result.append("(");
								result.append(String.valueOf(deps.size()));
								result.append(") ");
								for(IManagedCodeUnit d : deps) {
									result.append(d.getName());
									result.append(" ");
								}
								result.append("\n");

							}
						}

						result.append("\n\n\n");
						result.append("DEPENDENCY GRAPH\n");
						result.append("----------------\n\n");
						IDepGraph<IManagedCodeUnit> depGraph = mng.getPackageDependencyGraph(tm);

						for(IMagnoliaPackage res : pkgs) {
							result.append("\tpackage ");
							result.append(res.getName());
							result.append("\n");
							result.append("\t\tDependencies: ");
							for(IManagedCodeUnit dep : depGraph.getDepends(res)) {
								result.append(" ");
								result.append(dep.getName());
							}
							Set<IManagedCodeUnit> d1 = depGraph.getDepends(res);
							Collection<? extends IManagedCodeUnit> d2 = res.getDepends(tm);
							if(!UnsyncedDepGraph.isSetEquals(d1, d2)) {
								System.err.println("************** dependency graph differs for " + res);
							}
							result.append("\n");
							result.append("\t\tDependencies*: ");
							for(IManagedCodeUnit dep : depGraph.getTransitiveDepends(res)) {
								result.append(" ");
								result.append(dep.getName());
							}
							d1 = depGraph.getTransitiveDepends(res);
							d2 = res.getTransitiveDepends(tm);
							if(!UnsyncedDepGraph.isSetEquals(d1, d2)) {
								System.err.println("************** transitive dependency graph differs for " + res);
							}
							result.append("\n");
						}

						result.append("\n\n\n");
						result.append("TOPOLOGICAL SORT\n");
						result.append("----------------\n\n");

						DepGraphAxioms.runTests(depGraph);
						for(IManagedCodeUnit res : depGraph.topological()) {
							result.append("\t");
							result.append(res.getName());
							result.append("\n");
							result.append("\t\tDependencies: ");
							for(IManagedCodeUnit dep : depGraph.getDepends(res)) {
								result.append(" ");
								result.append(dep.getName());
							}
							result.append("\n");
						}

						result.append("\n\n\n");
						result.append("FILE RESOURCES\n");
						result.append("--------------\n\n");
						List<IManagedResource> children = new ArrayList<>(mng.getChildren(tm));
						Collections.sort(children, new Comparator<IManagedResource>() {
							@Override
							public int compare(IManagedResource arg0, IManagedResource arg1) {
								return arg0.getURI().toString().compareTo(arg1.getURI().toString());
							}
						});
						for(IManagedResource res : children) {
							result.append("\t");
							result.append(res.getURI());
							result.append("\n");
						}
						Editor.edit(result.toString(), "ProjectOverview");
					}
					catch(final CompilationErrorException ce) {
						ce.printStackTrace();
					}
					catch(final Throw t) {
						if(t.getStackTrace() != null) {
							t.printStackTrace();
						}
						else {
							System.err.println(t);
						}
					}
					catch(final Throwable e1) {
						e1.printStackTrace();
					}
					finally {
						CommandUtil.commitMarks(parseCtrl);
						tm.done();
						monitor.done();
					}
					long t2 = System.currentTimeMillis();
					System.err.printf(CMD_NAME + ": %dms refresh + %dms work = %dms total%n", t1 - t0, t2 - t1, t2 - t0);
					return Status.OK_STATUS;
				}

			};
//			TODO: ProjectGraph projectGraph = new ProjectGraph(mng);
//			TODO: projectGraph.display();
			// job.schedule();
		}
		return null;
	}
}
