/**************************************************************************
 * Copyright (c) 2010-2013 Anya Helene Bagge
 * Copyright (c) 2010-2013 University of Bergen
 * 
 * All rights reserved. This program and the accompanying materials
 * are made available under the terms of the Eclipse Public License v1.0
 * which accompanies this distribution, and is available at
 * http://www.eclipse.org/legal/epl-v10.html
 * 
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version. See http://www.gnu.org/licenses/
 * 
 * 
 * See the file COPYRIGHT for more information.
 * 
 * Contributors:
 * * Anya Helene Bagge
 * 
 *************************************************************************/
package org.magnolialang.eclipse.commands;

import java.util.HashMap;
import java.util.Map;

import org.eclipse.core.commands.AbstractHandler;
import org.eclipse.core.commands.ExecutionEvent;
import org.eclipse.core.commands.ExecutionException;
import org.eclipse.core.resources.IProject;
import org.eclipse.core.runtime.IProgressMonitor;
import org.eclipse.core.runtime.IStatus;
import org.eclipse.core.runtime.OperationCanceledException;
import org.eclipse.core.runtime.Status;
import org.eclipse.core.runtime.jobs.Job;
import io.usethesource.impulse.editor.UniversalEditor;
import org.eclipse.ui.IEditorPart;
import org.eclipse.ui.handlers.HandlerUtil;
import org.magnolialang.eclipse.MagnoliaPlugin;
import org.magnolialang.infra.Infra;
import org.magnolialang.magnolia.Magnolia;
import org.magnolialang.magnolia.lib.MagnoliaCompileLib;
import org.nuthatchery.pica.rascal.RascalParserLoader;
import org.nuthatchery.pica.resources.IProjectManager;
import org.rascalmpl.eclipse.nature.RascalMonitor;
import org.rascalmpl.eclipse.nature.WarningsToMarkers;
import org.rascalmpl.debug.IRascalMonitor;

public class RefreshHandler extends AbstractHandler {
	private static final Map<IProject, RefreshJob> jobs = new HashMap<IProject, RefreshJob>();
	public static final String COMMAND_REFRESH = "magnolia_eclipse.commands.Refresh";
	public static final String COMMAND_RELOAD = "magnolia_eclipse.commands.Reload";


	@Override
	public Object execute(ExecutionEvent event) throws ExecutionException {
		IEditorPart editorPart = HandlerUtil.getActiveEditor(event);
		if(editorPart instanceof UniversalEditor) {
			IProject project = ((UniversalEditor) editorPart).getParseController().getProject().getRawProject();
			RefreshJob job = getRefreshJob(project);
			job.setReload(event.getCommand().getId().equals(COMMAND_RELOAD));
			job.schedule();

		}
		return null;
	}


	public static void autoRefresh(ExecutionEvent event) {
		if(ToggleHandler.enabled(ToggleHandler.COMMAND_AUTO_REFRESH) || ToggleHandler.enabled(ToggleHandler.COMMAND_AUTO_RELOAD)) {
			IEditorPart editorPart = HandlerUtil.getActiveEditor(event);
			if(editorPart instanceof UniversalEditor) {
				IProject project = ((UniversalEditor) editorPart).getParseController().getProject().getRawProject();
				RefreshJob job = getRefreshJob(project);
				job.setReload(ToggleHandler.enabled(ToggleHandler.COMMAND_AUTO_RELOAD));
				if(job.getState() == Job.NONE) {
					job.schedule();
				}
				try {
					job.join();
				}
				catch(InterruptedException e) {
				}
			}
		}
	}


	private static synchronized RefreshJob getRefreshJob(IProject proj) {
		RefreshJob job = jobs.get(proj);
		if(job == null) {
			job = new RefreshJob("Refresh Magnolia Plugin", proj);
			jobs.put(proj, job);
		}
		return job;
	}


	private static final class RefreshJob extends Job {
		private final IProject project;
		private boolean reload = false;


		private RefreshJob(String name, IProject project) {
			super(name);
			this.project = project;
		}


		@Override
		public boolean belongsTo(Object obj) {
			return obj == MagnoliaPlugin.JOB_FAMILY_MAGNOLIA;
		}


		public void setReload(boolean value) {
			reload = value;
		}


		@Override
		protected IStatus run(IProgressMonitor monitor) {
			IRascalMonitor rm = new RascalMonitor(monitor, new WarningsToMarkers());

			rm.startJob("Reloading magnolia-eclipse", reload ? 100 : 40);

			rm.event("Refresh plugin", 5);
			if(rm.isCanceled()) {
				throw new OperationCanceledException();
			}
			MagnoliaPlugin.refresh();

			rm.event("Refresh manager", 5);
			if(rm.isCanceled()) {
				throw new OperationCanceledException();
			}
			IProjectManager manager = Infra.getResourceManager(project.getName());
			manager.refresh();

			rm.event("Refresh parser", 25);
			if(rm.isCanceled()) {
				throw new OperationCanceledException();
			}
			RascalParserLoader.refresh();

			if(reload) {
				rm.event("Reload compiler", 60);
				if(rm.isCanceled()) {
					throw new OperationCanceledException();
				}
				Magnolia.getInstance().getCompiler().refresh();
			}

			MagnoliaCompileLib.refresh();
			rm.endJob(true);

			monitor.done();
			return Status.OK_STATUS;
		}
	}

}
