/**************************************************************************
 * Copyright (c) 2013-2014 Anya Helene Bagge
 * Copyright (c) 2013-2014 University of Bergen
 *
 * All rights reserved. This program and the accompanying materials
 * are made available under the terms of the Eclipse Public License v1.0
 * which accompanies this distribution, and is available at
 * http://www.eclipse.org/legal/epl-v10.html
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version. See http://www.gnu.org/licenses/
 *
 *
 * See the file COPYRIGHT for more information.
 *
 * Contributors:
 * * Anya Helene Bagge
 *
 *************************************************************************/
package org.magnolialang.eclipse.commands;

import java.io.BufferedInputStream;
import java.io.BufferedOutputStream;
import java.io.BufferedWriter;
import java.io.InputStreamReader;
import java.io.OutputStreamWriter;
import java.nio.charset.Charset;
import java.nio.file.Files;
import java.nio.file.Path;
import java.nio.file.Paths;
import java.util.Collection;

import org.eclipse.core.commands.AbstractHandler;
import org.eclipse.core.commands.ExecutionEvent;
import org.eclipse.core.commands.ExecutionException;
import org.eclipse.core.runtime.CoreException;
import org.eclipse.core.runtime.IProgressMonitor;
import org.eclipse.core.runtime.IStatus;
import org.eclipse.core.runtime.Status;
import org.eclipse.core.runtime.SubMonitor;
import org.eclipse.core.runtime.jobs.Job;
import io.usethesource.impulse.editor.UniversalEditor;
import io.usethesource.impulse.parser.IParseController;
import org.rascalmpl.value.IConstructor;
import org.rascalmpl.value.IList;
import org.rascalmpl.value.ITuple;
import org.rascalmpl.value.IValue;
import org.eclipse.jface.dialogs.MessageDialog;
import org.eclipse.ui.IEditorPart;
import org.eclipse.ui.handlers.HandlerUtil;
import org.magnolialang.eclipse.MagnoliaPlugin;
import org.magnolialang.eclipse.editor.Editor;
import org.magnolialang.eclipse.preferences.Pref;
import org.magnolialang.magnolia.resources.IMagnoliaCodeUnit;
import org.magnolialang.magnolia.resources.IMagnoliaPackage;
import org.nuthatchery.pica.rascal.ICompiler;
import org.nuthatchery.pica.tasks.ITaskMonitor;
import org.nuthatchery.pica.tasks.TaskId;
import org.nuthatchery.pica.tasks.eclipse.EclipseTaskMonitor;
import org.nuthatchery.pica.terms.TermAdapter;
import org.nuthatchery.pica.terms.TermLoader;
import org.rascalmpl.eclipse.nature.RascalMonitor;
import org.rascalmpl.eclipse.nature.WarningsToMarkers;
import org.rascalmpl.debug.IRascalMonitor;
import org.rascalmpl.interpreter.control_exceptions.Throw;

public class RunSolver extends AbstractHandler {
	@Override
	public Object execute(final ExecutionEvent event) throws ExecutionException {
		final IEditorPart editorPart = HandlerUtil.getActiveEditor(event);
		if(editorPart instanceof UniversalEditor) {
			System.out.println("----------------------------------------------------------------------------------------------------------------------------------------");
			final IParseController parseCtrl = CommandUtil.getParseController(editorPart);
			final ICompiler compiler = CommandUtil.getCompiler(parseCtrl);
			final IConstructor symbol = CommandUtil.getModuleId(parseCtrl);
			final IMagnoliaPackage pkg = CommandUtil.getResource(parseCtrl);
			final String CMD_NAME = "Run Solver";
			try {
				final String solverPath = MagnoliaPlugin.getInstance().getPreferenceStore().getString(Pref.MAIN_OPTIMIZER_PATH.name());
				final String z3Path = MagnoliaPlugin.getInstance().getPreferenceStore().getString(Pref.MAIN_Z3_PATH.name());

				if(solverPath.equals("") || !Files.isExecutable(Paths.get(solverPath))) {
					String msg = "MagnoliaOptimizer executable not available. Please check Magnolia → Builder preferences.";
					if(solverPath.equals("")) {
						msg += "\n\n(path to executable not set)";
					}
					else {
						msg += "\n\n(path to executable: '" + solverPath + "')";
					}
					MessageDialog.openError(null, "Error Running Solver", msg);
					throw new CoreException(new Status(Status.ERROR, MagnoliaPlugin.PLUGIN_ID, msg));
				}
				if(solverPath.equals("") || z3Path.equals("") || !Files.isExecutable(Paths.get(solverPath)) || !Files.isExecutable(Paths.get(z3Path))) {
					String msg = "Z3 solver executable not available. Please check Magnolia → Builder preferences.";
					if(z3Path.equals("")) {
						msg += "\n\n(Path to executable: not set)";
					}
					else {
						msg += "\n\n(Path to executable: '" + z3Path + "')";
					}
					MessageDialog.openError(null, "Error Running Solver", msg);
					throw new CoreException(new Status(Status.ERROR, MagnoliaPlugin.PLUGIN_ID, msg));
				}

				Job job = new Job(CMD_NAME) {
					@Override
					public boolean belongsTo(Object obj) {
						return obj == MagnoliaPlugin.JOB_FAMILY_MAGNOLIA;
					}


					@Override
					protected IStatus run(IProgressMonitor monitor) {
						ITaskMonitor tm = new EclipseTaskMonitor(SubMonitor.convert(monitor));
						tm.begin(new TaskId(getClass().getCanonicalName(), CMD_NAME, pkg.getResource()));
						RefreshHandler.autoRefresh(event);

						try {
							Collection<? extends IMagnoliaCodeUnit> children = pkg.getChildren(tm);

							for(IMagnoliaCodeUnit c : children) {
								StringBuilder b = new StringBuilder();
								System.out.println("Run solver: " + c.getKind(tm) + " " + c.getName());
								if(c.getKind(tm).equals("satisfaction") || c.getKind(tm).equals("program")) {
									IConstructor defInfo = c.getDefInfo(tm);
									defInfo = defInfo.set("body", c.getTypechecked(tm));
									IConstructor result = (IConstructor) compiler.call("collectAxioms", defInfo.get("body"));
									if(result.has("val")) {
										IList defs = (IList) ((ITuple) result.get("val")).get(0);
										IList axioms = (IList) ((ITuple) result.get("val")).get(1);
										b.append("ToTransform(\n");
										b.append("[");
										boolean first = true;
										for(IValue v : defs) {
											if(!first) {
												b.append(",");
											}
											if(v instanceof IConstructor) {
												IConstructor cons = (IConstructor) v;
												if(cons.getName().equals("Define")) {
													b.append("Define(");
													b.append(TermAdapter.yieldTerm(cons.get(0), false));
													b.append(", ");
													b.append(TermAdapter.yieldTerm(cons.get(2), false));
													b.append(")");
												}
											}
											else {
												b.append(TermAdapter.yieldTerm(v, false));
											}
											b.append("\n");
											first = false;
										}
										b.append("],\n");
										b.append("[");
										first = true;
										for(IValue v : axioms) {
											if(!first) {
												b.append(",");
											}
											b.append(TermAdapter.yieldTerm(v, false));
											b.append("\n");
											first = false;
										}
										b.append("])\n");
									}
									System.out.println("Displaying axioms for " + TermAdapter.yieldTerm(symbol, false));
									// pretty printed AST
									final Path tmpIn = Files.createTempFile("solverIn", ".mgt");
									boolean solverOk = true;
									try {
										final Path tmpOut = Files.createTempFile("solverOut", ".mgt");
										try {
											BufferedWriter writer = Files.newBufferedWriter(tmpIn, Charset.forName("UTF-8"));
											writer.append(b);
											writer.close();
											System.err.println("Solver input written to " + tmpIn);

											String[] cmdArray = { solverPath, z3Path, tmpIn.toString() };

											Process process = Runtime.getRuntime().exec(cmdArray);
											try (InputStreamReader reader = new InputStreamReader(new BufferedInputStream(process.getInputStream()))) {
												try (OutputStreamWriter cmd = new OutputStreamWriter(new BufferedOutputStream(process.getOutputStream()))) {
													cmd.write("Halt()\n");
													cmd.flush();
													System.err.println(">>> Halt()");
													char[] cbuf = new char[1];
													int r;
													StringBuilder solverOut = new StringBuilder();
													while((r = reader.read(cbuf)) >= 0) {
														solverOut.append(cbuf[0]);
													}
													solverOut.append("\n");

													int exit = process.waitFor();
													if(exit == 0) {
														String output = solverOut.toString();
														if(output.startsWith("\"")) {
															solverOk = false;
														}
														try {
															TermLoader loader = new TermLoader(output);
															IConstructor term = loader.parseTerm();
															System.out.println(TermAdapter.yieldTerm(term));
														}
														catch(TermLoader.ParseError e) {
															System.err.println(e.getMessage());
														}
														Editor.edit(output, /*parseCtrl.getPath().removeFileExtension().toString()*/c.getName() + ".solverOut");
													}
													else {
														solverOk = false;
														System.err.println("Solver exited with return code " + exit);
													}
												}
											}
										}
										finally {
											Files.deleteIfExists(tmpOut);
										}
									}
									finally {
										if(solverOk) {
											Files.deleteIfExists(tmpIn);
										}
									}
								}

							}
						}
						catch(Throw t) {
							System.err.println(t);
						}
						catch(Throwable t) { // NOPMD by anya on 1/5/12 3:52 AM
							t.printStackTrace();
						}

						CommandUtil.commitMarks(parseCtrl);
						// compiler.visualize();
						System.out.println("Done!!");
						tm.done();
						monitor.done();
						return Status.OK_STATUS;
					}

				};
				job.schedule();
			}
			catch(final Throwable e1) { // NOPMD by anya on 1/5/12 3:52 AM
				e1.printStackTrace();
			}
		}
		return null;
	}
}
