/**************************************************************************
 * Copyright (c) 2014 Anya Helene Bagge
 * Copyright (c) 2014 University of Bergen
 *
 * All rights reserved. This program and the accompanying materials
 * are made available under the terms of the Eclipse Public License v1.0
 * which accompanies this distribution, and is available at
 * http://www.eclipse.org/legal/epl-v10.html
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version. See http://www.gnu.org/licenses/
 *
 *
 * See the file COPYRIGHT for more information.
 *
 * Contributors:
 * * Anya Helene Bagge
 *
 *************************************************************************/
package org.magnolialang.eclipse.commands.tests;

import java.io.IOException;
import java.io.PrintWriter;
import java.io.Reader;
import java.util.ArrayList;
import java.util.Collection;
import java.util.List;
import java.util.Random;
import java.util.concurrent.BlockingQueue;
import java.util.concurrent.LinkedBlockingQueue;
import java.util.concurrent.TimeUnit;

import org.eclipse.core.commands.AbstractHandler;
import org.eclipse.core.commands.ExecutionEvent;
import org.eclipse.core.commands.ExecutionException;
import org.eclipse.core.runtime.IProgressMonitor;
import org.eclipse.core.runtime.IStatus;
import org.eclipse.core.runtime.Status;
import org.eclipse.core.runtime.jobs.Job;
import io.usethesource.impulse.editor.UniversalEditor;
import io.usethesource.impulse.parser.IParseController;
import org.rascalmpl.value.IConstructor;
import org.rascalmpl.value.ISet;
import org.rascalmpl.value.ISourceLocation;
import org.eclipse.ui.IEditorPart;
import org.eclipse.ui.handlers.HandlerUtil;
import org.magnolialang.eclipse.MagnoliaPlugin;
import org.magnolialang.eclipse.commands.CommandUtil;
import org.magnolialang.eclipse.commands.RefreshHandler;
import org.magnolialang.errors.CompilationErrorException;
import org.magnolialang.magnolia.Magnolia;
import org.magnolialang.magnolia.resources.IMagnoliaModule;
import org.magnolialang.magnolia.resources.IMagnoliaPackage;
import org.magnolialang.magnolia.resources.MagnoliaPackage;
import org.nuthatchery.pica.resources.ILanguage;
import org.nuthatchery.pica.resources.IProjectManager;
import org.nuthatchery.pica.resources.handles.IResourceHandle;
import org.nuthatchery.pica.resources.managed.IManagedCodeUnit;
import org.nuthatchery.pica.resources.managed.IManagedResource;
import org.nuthatchery.pica.resources.regions.IOffsetLength;
import org.nuthatchery.pica.tasks.ITaskMonitor;
import org.nuthatchery.pica.tasks.NullTaskMonitor;
import org.nuthatchery.pica.util.ISignature;
import org.nuthatchery.pica.util.Pair;
import org.rascalmpl.eclipse.nature.RascalMonitor;
import org.rascalmpl.eclipse.nature.WarningsToMarkers;
import org.rascalmpl.debug.IRascalMonitor;
import org.rascalmpl.interpreter.NullRascalMonitor;

public class ConcurrencyTest extends AbstractHandler {
	@Override
	public Object execute(final ExecutionEvent event) throws ExecutionException {
		final IEditorPart editorPart = HandlerUtil.getActiveEditor(event);

		if(editorPart instanceof UniversalEditor) {
			final IParseController parseCtrl = CommandUtil.getParseController(editorPart);
			final PrintWriter errorWriter = CommandUtil.getConsoleErrorWriter();
			final PrintWriter outputWriter = CommandUtil.getConsoleOutputWriter();
			final Reader inputReader = CommandUtil.getConsoleInputReader();
			CommandUtil.showConsole();

			final IProjectManager projectManager = CommandUtil.getModuleManager(parseCtrl);

			Job job = new Job("Concurrency Test") {
				@Override
				public boolean belongsTo(Object obj) {
					return obj == MagnoliaPlugin.JOB_FAMILY_MAGNOLIA;
				}


				@Override
				protected IStatus run(IProgressMonitor monitor) {
					IRascalMonitor rm = new RascalMonitor(monitor, new WarningsToMarkers());
					rm.startJob("Concurrency Test");
					RefreshHandler.autoRefresh(event);
					BlockingQueue<TestRunnable> queue = new LinkedBlockingQueue<>();
					int n = 16;
					TestRunnable[] tests = new TestRunnable[n];
					Thread[] threads = new Thread[n];
					int[] count = new int[n];

					for(int i = 0; i < n; i++) {
						count[i] = 0;
						tests[i] = new TestRunnable(projectManager, queue, i);
						threads[i] = new Thread(null, tests[i], "ConcurrencyTest-" + i);
						threads[i].start();
					}

					try {
						int j = 0;
						while(!rm.isCanceled()) {
							for(int i = 0; i < n; i++) {
								count[i]++;
							}

							TestRunnable poll = queue.poll(10, TimeUnit.SECONDS);
							while(poll != null) {
								count[poll.id] = 0;
								poll = queue.poll();
							}

							System.err.flush();
							System.out.print("Status after " + j + " iterations: ");
							for(int i = 0; i < n; i++) {
								System.out.printf("%5d ", count[i]);
							}
							System.out.println();
							System.out.flush();
							j++;
						}
						for(Thread t : threads) {
							t.interrupt();
						}
						for(Thread t : threads) {
							t.join();
						}
					}
					catch(final CompilationErrorException ce) {
						ce.printStackTrace();
						outputWriter.flush();
						ce.printStackTrace(errorWriter);
					}
					catch(final Throwable e1) { // NOPMD by anya on 1/5/12 3:52 AM

						e1.printStackTrace();
						outputWriter.flush();
						e1.printStackTrace(errorWriter);
						// throw new ImplementationError("Flattening failed:",
						// e1);
					}
					finally {
						CommandUtil.commitMarks(parseCtrl);
						errorWriter.close();
						outputWriter.close();
						try {
							inputReader.close();
						}
						catch(IOException e) {
							// ignore
						}

						rm.endJob(true);
						monitor.done();
					}
					return Status.OK_STATUS;
				}
			};
			job.schedule();
		}
		return null;
	}


	public static class TestRunnable implements Runnable {
		protected final List<IMagnoliaPackage> pkgs = new ArrayList<IMagnoliaPackage>();
		protected Random random = new Random();
		protected Collection<IMagnoliaModule> children;
		protected int id;
		protected BlockingQueue<TestRunnable> queue;


		public TestRunnable(IProjectManager manager, BlockingQueue<TestRunnable> queue, int i) {
			this.id = i;
			this.queue = queue;
			for(IManagedCodeUnit pkg : manager.allPackages(Magnolia.getInstance())) {
				pkgs.add((IMagnoliaPackage) pkg);
			}
		}


		@SuppressWarnings("unused")
		public void randomCall() {
			IMagnoliaPackage pkg = pkgs.get(random.nextInt(pkgs.size()));
			ITaskMonitor rm = new NullTaskMonitor();

			try {
				switch(random.nextInt(25)) {
				case 0:
					pkg.containsOffset(random.nextInt(Integer.MAX_VALUE));
					break;
				case 1:
					IConstructor ast = pkg.getAST(rm);
					break;
				case 2:
					children = pkg.getChildren(rm);
					break;
				case 3:
					IResourceHandle containingFile = pkg.getResource();
					break;
				case 4:
					IConstructor defInfo = pkg.getDefInfo(rm);
					break;
				case 5:
					Collection<? extends IManagedCodeUnit> depends = pkg.getDepends(rm);
					break;
				case 6:
					ISignature fullSignature = pkg.getFullSignature(rm);
					break;
				case 7:
					IConstructor pkgId = pkg.getId();
					break;
				case 8:
					String kind = pkg.getKind(rm);
					break;
				case 9:
					ILanguage language = pkg.getLanguage();
					break;
				case 10:
					break;
				case 11:
					IConstructor packageEnv = pkg.getPackageEnv(rm);
					break;
				case 12:
					ISignature sourceSignature = pkg.getSourceSignature(rm);
					break;
				case 13:
					Collection<? extends IManagedCodeUnit> transitiveDepends = pkg.getTransitiveDepends(rm);
					break;
				case 14:
					IConstructor typechecked = pkg.getTypechecked(rm);
					break;
				case 15:
					pkg.onDependencyChanged();
					break;
				case 16:
					pkg.onResourceChanged();
					break;
				case 17:
					Collection<Pair<IOffsetLength, Object>> xRefs = pkg.getXRefs(rm);
					break;
				case 18:
					((MagnoliaPackage) pkg).clearFacts();
					break;
				default:
					randomChildCall();
				}
			}
			catch(Throwable t) {
				t.printStackTrace();
			}
		}


		@Override
		public void run() {
			int i = 0;
			while(!Thread.interrupted()) {
				randomCall();
				if(i++ > 10) {
					System.out.println("SYNC!");
					i = 0;
					queue.add(this);
				}
			}
		}


		@SuppressWarnings("unused")
		private void randomChildCall() {
			if(children != null) {
				IMagnoliaModule mod = (IMagnoliaModule) children.toArray()[random.nextInt(children.size())];
				ITaskMonitor rm = new NullTaskMonitor();

				try {
					switch(random.nextInt(11)) {
					case 0:
						IConstructor ast = mod.getAST(rm);
						break;
					case 1:
						IConstructor baseName = mod.getBaseName();
						break;
					case 2:
						IConstructor body = mod.getBody();
						break;
					case 3:
						Collection<? extends IManagedResource> children2 = mod.getChildren(rm);
						break;
					case 4:
						IConstructor defClause = mod.getDefClause();
						break;
					case 5:
						IConstructor defInfo = mod.getDefInfo(rm);
						break;
					case 6:
						Collection<? extends IManagedCodeUnit> depends = mod.getDepends(rm);
						break;
					case 7:
						ISet externals = mod.getExternals(rm);
						break;
					case 8:
						Collection<? extends IManagedCodeUnit> transitiveDepends = mod.getTransitiveDepends(rm);
						break;
					case 9:
						IConstructor typechecked = mod.getTypechecked(rm);
						break;
					case 10:
						Collection<Pair<IOffsetLength, Object>> xRefs = mod.getXRefs(rm);
						break;
					}
				}
				catch(Throwable t) {
					t.printStackTrace();
				}
			}

		}
	}
}
