/**************************************************************************
 * Copyright (c) 2012-2014 Anya Helene Bagge
 * Copyright (c) 2012-2014 University of Bergen
 *
 * All rights reserved. This program and the accompanying materials
 * are made available under the terms of the Eclipse Public License v1.0
 * which accompanies this distribution, and is available at
 * http://www.eclipse.org/legal/epl-v10.html
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version. See http://www.gnu.org/licenses/
 *
 *
 * See the file COPYRIGHT for more information.
 *
 * Contributors:
 * * Anya Helene Bagge
 *
 *************************************************************************/
package org.magnolialang.eclipse.editor;

import java.util.List;

import org.rascalmpl.value.IString;
import io.usethesource.impulse.utils.AnnotationUtils;
import org.eclipse.jdt.annotation.Nullable;
import org.eclipse.jface.text.BadLocationException;
import org.eclipse.jface.text.IRegion;
import org.eclipse.jface.text.ITextHover;
import org.eclipse.jface.text.ITextHoverExtension2;
import org.eclipse.jface.text.ITextViewer;
import org.eclipse.jface.text.Region;
import org.eclipse.jface.text.source.Annotation;
import org.eclipse.jface.text.source.ISourceViewer;
import org.magnolialang.eclipse.MagnoliaPlugin;
import org.magnolialang.errors.RascalErrors;
import org.magnolialang.infra.Infra;
import org.magnolialang.magnolia.Magnolia;
import org.magnolialang.magnolia.services.IMgXRefInfo;
import org.nuthatchery.pica.rascal.RascalUtil;
import org.nuthatchery.pica.resources.IProjectManager;
import org.nuthatchery.pica.resources.IXRefInfo;
import org.nuthatchery.pica.resources.managed.IManagedCodeUnit;
import org.nuthatchery.pica.tasks.NullTaskMonitor;
import org.nuthatchery.pica.terms.TermFactory;
import org.rascalmpl.interpreter.NullRascalMonitor;
import org.rascalmpl.interpreter.control_exceptions.Throw;

import edu.umd.cs.findbugs.annotations.SuppressFBWarnings;

public class HoverHelper implements ITextHover, ITextHoverExtension2 {

	private final MagnoliaParseController ctrl;


	public HoverHelper(MagnoliaParseController ctrl) {
		this.ctrl = ctrl;
	}


	@Override
	@Nullable
	public String getHoverInfo(ITextViewer textViewer, IRegion hoverRegion) {
		IProjectManager manager = Infra.getResourceManager(ctrl.getProject().getName());
		manager.processChanges(new NullTaskMonitor());
		IManagedCodeUnit pkg = manager.findCodeUnit(ctrl.getURI());
		try {
			if(pkg == null) {
				return null;
			}
			int offset = hoverRegion.getOffset();

			/*
			IConstructor ast = pkg.getTypechecked(null);
			if(ast != null) {
				IConstructor locate = TermLocator.locateName(ast, TermFactory.vf.sourceLocation(pkg.getURI(), offset, 0));

				if(locate != null) {
					System.out.println("getHoverInfo: " + TermAdapter.yieldTerm(locate, false));
					return TermAdapter.yieldTerm(locate, false);
				}
			}
			 */

			// commit first, otherwise we won't see the marks
			manager.commitMarks();
			@SuppressFBWarnings("BC_UNCONFIRMED_CAST")
			List<Annotation> annos = AnnotationUtils.getAnnotationsForOffset((ISourceViewer) textViewer, offset);
			String annoString = null;
			String xrefResult = null;
			if(annos != null && annos.size() > 0) {
				annoString = AnnotationUtils.formatAnnotationList(annos);
			}

			try {
				int length = hoverRegion.getLength();
				String s = textViewer.getDocument().get(offset, length > 0 ? length : 1);
				if(s.equals("\t") || s.equals(" ")) {
					return null;
				}
				else {
					// Collection<Pair<ISourceLocation, IConstructor>> xRefs = XRefHelper.getXRefs(pkg, offset, new NullRascalMonitor());
					IMgXRefInfo info = (IMgXRefInfo) pkg.getXRefs(RascalUtil.toCodeRegion(TermFactory.vf.sourceLocation(ctrl.getSourceLocation(), offset, length)), new NullTaskMonitor());

					if(info != null) {
						xrefResult = formatOne(info);
						/*
						if(xRefs.size() == 1)
						xrefResult = formatOne(xRefs.iterator().next().getSecond());
						else
						xrefResult = formatMany(xRefs);
						 */
						//System.err.println("getHoverInfo: " + result);
					}
				}
			}
			catch(BadLocationException e) {
			}

			String result = null;
			if(annoString != null && xrefResult != null) {
				result = annoString + "<br/><p>" + xrefResult + "</p>";
			}
			else if(annoString != null) {
				result = annoString;
			}
			else if(xrefResult != null) {
				result = xrefResult;
			}

			if(result != null) {
				if(result.contains("\t")) {
					System.err.println("Tab in hover help string!");
					result.replaceAll("\t", "    ");
				}
			}
			return result;
		}
		catch(Exception e) {
			if(e instanceof Throw) {
				e = RascalErrors.decodeRascalError((Throw) e);
			}
			System.err.println(e);
			MagnoliaPlugin.getInstance().logException("Errør in hover helper", e);
			return null;
		}
		finally {
			manager.commitMarks();
		}
	}


	@Override
	public Object getHoverInfo2(ITextViewer textViewer, IRegion hoverRegion) {
		return getHoverInfo(textViewer, hoverRegion);
	}


	@Override
	public IRegion getHoverRegion(ITextViewer textViewer, int offset) {
		return new Region(offset, 0);
	}


/*
	private String formatMany(Collection<Pair<ISourceLocation, IConstructor>> xRefs) {
		StringBuilder buf = new StringBuilder(1024);
		//buf.append("<ul>");
		for(Pair<ISourceLocation, IConstructor> xRef : xRefs) {
			String s = formatOne(xRef.getSecond());
			if(s != null) {
				//buf.append("<li>");
				buf.append(s);
				//buf.append("</li>");
			}
		}
		//buf.append("</ul>");
		return buf.toString();
	}
 */

	private String formatOne(IMgXRefInfo info) {
		if(info.hasDefInfo()) {
			return ((IString) Magnolia.getInstance().getCompiler().call(new NullTaskMonitor(), "formatXRef", info.getDefInfo(), info.getBindings())).getValue();
		}
		else {
			return ((IString) Magnolia.getInstance().getCompiler().call(new NullTaskMonitor(), "formatXRef", info.getId(), info.getBindings())).getValue();
		}
	}

}
