/**************************************************************************
 * Copyright (c) 2009-2011 CWI
 * Copyright (c) 2011-2014 Anya Helene Bagge
 * Copyright (c) 2011-2014 University of Bergen
 * 
 * All rights reserved. This program and the accompanying materials
 * are made available under the terms of the Eclipse Public License v1.0
 * which accompanies this distribution, and is available at
 * http://www.eclipse.org/legal/epl-v10.html
 * 
 * 
 * See the file COPYRIGHT for more information.
 * 
 * Contributors:
 * * Andreas Hjortland
 * * Anya Helene Bagge
 * + Arnold Lankamp
 * + Jurgen J. Vinju
 * 
 *************************************************************************/
package org.magnolialang.eclipse.editor;

import java.util.HashMap;
import java.util.HashSet;
import java.util.Map;
import java.util.Set;

import org.eclipse.core.resources.IFile;
import org.eclipse.core.resources.IResource;
import org.eclipse.core.resources.ResourcesPlugin;
import org.eclipse.core.runtime.IPath;
import io.usethesource.impulse.editor.ModelTreeNode;
import io.usethesource.impulse.language.ILanguageService;
import org.rascalmpl.value.IConstructor;
import org.rascalmpl.value.IInteger;
import org.rascalmpl.value.IString;
import org.rascalmpl.value.IValue;
import io.usethesource.impulse.services.ILabelProvider;
import org.eclipse.jdt.annotation.Nullable;
import org.eclipse.jface.resource.ImageDescriptor;
import org.eclipse.jface.viewers.IColorProvider;
import org.eclipse.jface.viewers.ILabelProviderListener;
import org.eclipse.swt.graphics.Color;
import org.eclipse.swt.graphics.Image;
import org.eclipse.swt.widgets.Display;
import org.eclipse.ui.PlatformUI;
import org.magnolialang.eclipse.MagnoliaPlugin;
import org.magnolialang.eclipse.MagnoliaResources;
import org.magnolialang.eclipse.icons.StyleBase;
import org.magnolialang.magnolia.Magnolia;

public class LabelProvider implements ILabelProvider, IColorProvider, ILanguageService {
	private final Set<ILabelProviderListener> fListeners = new HashSet<ILabelProviderListener>();
	private final Map<ImageDescriptor, Image> images = new HashMap<ImageDescriptor, Image>();
	private final Map<Integer, Color> colors = new HashMap<Integer, Color>();
	private final StyleBase style = MagnoliaPlugin.getInstance().getStyleBase(Magnolia.MAGNOLIA);


	@Override
	public void addListener(ILabelProviderListener listener) {
		fListeners.add(listener);
	}


	@Override
	public void dispose() {
		for(ImageDescriptor desc : images.keySet()) {
			desc.destroyResource(images.get(desc));
		}
		for(Color c : colors.values()) {
			c.dispose();
		}
		fListeners.clear();
		images.clear();
		colors.clear();
	}


	@Override
	public Color getBackground(Object element) {
		return getColor(element, "bgcolor");
	}


	@Override
	@Nullable
	public Color getForeground(Object element) {
		return getColor(element, "fgcolor");
	}


	@Override
	public Image getImage(Object element) {
		String iconName;
		if((element instanceof ModelTreeNode)) {
			element = ((ModelTreeNode) element).getASTNode();
		}
		if(element instanceof IPath) {
			IResource resource = ResourcesPlugin.getWorkspace().getRoot().findMember((IPath) element);
			if(resource != null) {
				element = resource;
			}
		}
		if(element instanceof IResource) {
			switch(((IResource) element).getType()) {
			case IResource.FILE:
				iconName = MagnoliaResources.MAGNOLIA_FILE;
				break;
			case IResource.PROJECT:
				iconName = MagnoliaResources.MAGNOLIA_PROJECT;
				break;
			case IResource.FOLDER:
				iconName = MagnoliaResources.MAGNOLIA_FOLDER;
				break;
			default:
				iconName = MagnoliaResources.MAGNOLIA_FILE;
				break;
			}
			if(!((IResource) element).exists()) {
				iconName = iconName + "+virtual";
			}
		}
		else if(element instanceof IPath) {
			iconName = MagnoliaResources.MAGNOLIA_FILE + "+virtual";
		}
		else {
			IValue value = getAnno(element, "icon");
			if(value instanceof IString) {
				iconName = ((IString) value).getValue();
			}
			else {
				iconName = MagnoliaResources.MAGNOLIA_DEFAULT_OUTLINE_ITEM;
			}
		}
		if(element instanceof TreeNode) {
			TreeNode node = (TreeNode) element;
			iconName = node.getType();
			// System.out.println("Element is of type: " + node.getType());
		}
		ImageDescriptor descriptor = style.getIcon(iconName, MagnoliaResources.SUFFIX_OBJECT);

		if(descriptor == null) {
			MagnoliaPlugin.getInstance().logException("ImageDescriptor not found for icon '" + iconName + "'", null);
			return null;
		}
		if(images.containsKey(descriptor)) {
			return images.get(descriptor);
		}

		Display disp = style.getDevice();

		Image img = (Image) descriptor.createResource(disp);
		images.put(descriptor, img);
		return img;
	}


	@Override
	@Nullable
	public String getText(Object element) {
		if(element instanceof ModelTreeNode) {
			ModelTreeNode node = (ModelTreeNode) element;
			if(node.getASTNode() instanceof TreeNode) {
				return ((TreeNode) node.getASTNode()).getName();
			}
		}
		// Everything in this method after this comment is debugging code. 
		// This should not be reachable unless we have set debug = true in 
		// TreeModelBuilder.
		if(element instanceof ModelTreeNode) {
			ModelTreeNode n = (ModelTreeNode) element;
			if(n.getASTNode() instanceof IConstructor) {
				IConstructor c = (IConstructor) n.getASTNode();
				if(c.getName().equals("char")) {
					StringBuilder s = new StringBuilder();
					for(IValue i : c.getChildren()) {
						if(i instanceof IInteger) {
							s.append((char) ((IInteger) i).intValue());
						}
					}
					return s.toString();
				}

				return truncate(c.getName() + " - " + c.getType().getName() + " - " + c.getClass());
			}
		}
		if(element instanceof IFile) {
			return truncate(((IFile) element).getName());
		}
		IValue anno = getAnno(element, "label");
		if(anno instanceof IString) {
			return truncate(((IString) anno).getValue());
		}
		if(element != null) {
			return truncate(element.toString());
		}
		return null;
	}


	@Override
	public boolean isLabelProperty(Object element, String property) {
		return false;
	}


	@Override
	public void removeListener(ILabelProviderListener listener) {
		fListeners.remove(listener);
	}


	@Nullable
	private IValue getAnno(Object element, String annotationName) {
		if(element instanceof ModelTreeNode) {
			element = ((ModelTreeNode) element).getASTNode();
		}
		if(element instanceof IConstructor) {
			return ((IConstructor) element).asAnnotatable().getAnnotation(annotationName);
		}

		return null;
	}


	@Nullable
	private Color getColor(Object element, String annoName) {
		if(element instanceof IFile) {
			return null;
		}
		else {
			IValue value = getAnno(element, annoName);
			if(value instanceof IInteger) {
				int color = ((IInteger) value).intValue();
				if(colors.containsKey(color)) {
					return colors.get(color);
				}
				else {
					int b = color & 0xff;
					int g = color >> 8 & 0xff;
					int r = color >> 16 & 0xff;
					Color c = new Color(getDevice(), r, g, b);
					colors.put(color, c);
					return c;
				}
			}
			else if(value instanceof IString) {
				StyleBase base = MagnoliaPlugin.getInstance().getStyleBase("Magnolia");
				if(base != null) {
					return base.getColor(((IString) value).getValue());
				}
				else {
					return null;
				}
			}
			else {
				return null;
			}
		}
	}


	private Display getDevice() {
		Display disp = Display.getCurrent();
		if(disp == null) {
			disp = PlatformUI.getWorkbench().getDisplay();
		}
		return disp;
	}


	private String truncate(String s) {
		return s.substring(0, Math.min(120, s.length()));
	}
}
