/**************************************************************************
 * Copyright (c) 2010-2014 Anya Helene Bagge
 * Copyright (c) 2010-2014 University of Bergen
 *
 * All rights reserved. This program and the accompanying materials
 * are made available under the terms of the Eclipse Public License v1.0
 * which accompanies this distribution, and is available at
 * http://www.eclipse.org/legal/epl-v10.html
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version. See http://www.gnu.org/licenses/
 *
 *
 * See the file COPYRIGHT for more information.
 *
 * Contributors:
 * * Anya Helene Bagge
 *
 *************************************************************************/
package org.magnolialang.eclipse.editor;

import java.net.URI;
import java.net.URISyntaxException;
import java.util.Iterator;

import org.eclipse.core.filesystem.EFS;
import org.eclipse.core.runtime.CoreException;
import org.eclipse.core.runtime.IPath;
import org.eclipse.core.runtime.IProgressMonitor;

import io.usethesource.impulse.language.Language;
import io.usethesource.impulse.language.LanguageRegistry;
import io.usethesource.impulse.model.ISourceProject;
import io.usethesource.impulse.parser.IMessageHandler;

import org.rascalmpl.value.ISourceLocation;
import org.rascalmpl.value.exceptions.FactTypeUseException;

import io.usethesource.impulse.services.IAnnotationTypeInfo;
import io.usethesource.impulse.services.ILanguageSyntaxProperties;

import org.eclipse.jdt.annotation.NonNullByDefault;
import org.eclipse.jdt.annotation.Nullable;
import org.eclipse.jface.text.IDocument;
import org.eclipse.jface.text.IRegion;
import org.magnolialang.eclipse.MagnoliaPlugin;
import org.magnolialang.load.ModuleParser;
import org.magnolialang.magnolia.Magnolia;
import org.nuthatchery.pica.errors.ImplementationError;
import org.nuthatchery.pica.resources.ILanguage;
import org.nuthatchery.pica.terms.TermFactory;
import org.rascalmpl.parser.gtd.exception.ParseError;
import org.rascalmpl.parser.gtd.exception.UndeclaredNonTerminalException;
import org.rascalmpl.uri.BadURIException;
import org.rascalmpl.values.uptr.ITree;

public class MagnoliaParseController implements IMagnoliaParseController {

	private ModuleParser parser;

	private IMessageHandler handler;
	private ISourceProject project;

	private ITree parseTree;
	private IPath path;

	private String languageName;

	private Language impLanguage;

	private MagnoliaEditor editor;

	private URI uri;

	private ISourceLocation loc;

	private IDocument document;

	private boolean initialized = false;

	private String lastInput = null;


	void dispose() {
		initialized = false;
	}


	@Override
	@Nullable
	public IAnnotationTypeInfo getAnnotationTypeInfo() {
		return null;
	}


	@Override
	@Nullable
	public ITree getCurrentAst() {
		return parseTree;
	}


	@Override
	@Nullable
	public IDocument getDocument() {
		return document;
	}


	@Override
	public MagnoliaEditor getEditor() {
		return editor;
	}


	@Override
	public Language getLanguage() {
		if(!isInitialized()) {
			throw new ImplementationError("Not initialized");
		}
		return impLanguage;
	}


	@Override
	public IPath getPath() {
		if(!isInitialized()) {
			throw new ImplementationError("Not initialized");
		}
		return path;
	}


	@Override
	public ISourceProject getProject() {
		if(!isInitialized()) {
			throw new ImplementationError("Not initialized");
		}
		return project;
	}


	public ISourceLocation getSourceLocation() {
		if(!isInitialized()) {
			throw new ImplementationError("Not initialized");
		}
		return loc;
	}


	@Override
	public NodeLocator getSourcePositionLocator() {
		return new NodeLocator();
	}


	@Override
	public ILanguageSyntaxProperties getSyntaxProperties() {
		return new MagnoliaSyntaxProperties();
	}


	@Override
	@Nullable
	public Iterator<Object> getTokenIterator(final @Nullable IRegion region) {
		if(parseTree == null) {
			return null;
		}
		else {
			return new TokenIterator(parseTree);
		}
	}


	/*
	 * (non-Javadoc)
	 *
	 * @see org.magnolialang.eclipse.editor.IMagnoliaParseController#getURI()
	 */
	@Override
	public URI getURI() {
		if(!isInitialized()) {
			throw new ImplementationError("Not initialized");
		}
		return uri;
	}


	@Override
	@NonNullByDefault(false)
	public void initialize(final IPath filePath, final ISourceProject project, final IMessageHandler handler) {
		assert filePath != null;
		assert handler != null;

		this.handler = handler;
		this.project = project;
		this.path = filePath;

		initLanguage(LanguageRegistry.findLanguage(filePath, null));

		if(project == null) {
			URI fileURI, genURI;
			try {
				genURI = new URI("mggen://" + path.toString());
				fileURI = new URI("file://" + path.toString());
			}
			catch(URISyntaxException e) {
				throw new BadURIException(e);
			}

			try {
				if(EFS.getStore(genURI).fetchInfo().exists()) {
					uri = genURI;
				}
				else {
					uri = fileURI;
				}

			}
			catch(CoreException e) {
				uri = fileURI;
			}
		}
		else {
			uri = MagnoliaPlugin.constructProjectURI(project, path);
		}
		loc = TermFactory.vf.sourceLocation(uri);
		initialized = true;
	}


	private void initLanguage(Language lang) {
		this.impLanguage = lang;
		this.languageName = impLanguage.getName();

		ILanguage l = org.nuthatchery.pica.resources.LanguageRegistry.getLanguage(languageName);
		if(l == null) {
			throw new ImplementationError("Language '" + languageName + "' not registered in Magnolia Language Registry");
		}

		this.parser = Magnolia.getInstance().getParser();
	}


	/*
	 * (non-Javadoc)
	 *
	 * @see
	 * org.magnolialang.eclipse.editor.IMagnoliaParseController#isInitialized()
	 */
	private boolean isInitialized() {
		return initialized;
	}


	@Override
	@NonNullByDefault(false)
	public Object parse(IDocument document, IProgressMonitor monitor) {
		assert document != null;
		Object o = parse(document.get(), monitor);
		this.document = document;
		return o;
	}


	@Override
	@NonNullByDefault(false)
	public Object parse(final String input, final IProgressMonitor monitor) {
		assert monitor != null;
		assert input != null;
		if(input.equals(lastInput)) {
			return parseTree;
		}
		parseTree = null;
		document = null;
		if(!isInitialized()) {
			throw new ImplementationError("Not initialized");
		}
		handler.clearMessages();
		monitor.beginTask("parsing " + languageName, 1);
		// System.out.println("MagnoliaParseController: parsing: " + path);
		try {
			lastInput = input;
			parseTree = parser.parseModule(uri, input);
			ParseTreeChecker.checkParseTree(parseTree, handler);
		}
		catch(final FactTypeUseException e) {
			MagnoliaPlugin.getInstance().logException("parsing " + languageName + " failed", e);
		}
		catch(final ParseError e) {
			String msg = e.getMessage();
			if(msg == null) {
				msg = "unknown";
			}
			if(e.getOffset() >= 0) {
				handler.handleSimpleMessage("Editor: " + msg, e.getOffset(), e.getOffset() + e.getLength(), e.getBeginColumn(), e.getEndColumn(), e.getBeginLine(), e.getEndLine());
			}
			else {
				handler.handleSimpleMessage("Editor: " + msg, 0, 0, 0, 0, 1, 1);
			}
		}
		catch(final UndeclaredNonTerminalException e) {
			e.printStackTrace();
		}
		catch(Throwable t) {
			t.printStackTrace();
		}
		finally {
			monitor.worked(1);
			monitor.done();
		}

		return parseTree;
	}


	void setEditor(final MagnoliaEditor magnoliaEditor) {
		editor = magnoliaEditor;
		if(impLanguage == null) {
			initLanguage(magnoliaEditor.getLanguage());
		}
	}

}
