/**************************************************************************
 * Copyright (c) 2010-2014 Anya Helene Bagge
 * Copyright (c) 2010-2014 University of Bergen
 *
 * All rights reserved. This program and the accompanying materials
 * are made available under the terms of the Eclipse Public License v1.0
 * which accompanies this distribution, and is available at
 * http://www.eclipse.org/legal/epl-v10.html
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version. See http://www.gnu.org/licenses/
 *
 *
 * See the file COPYRIGHT for more information.
 *
 * Contributors:
 * + Andreas Hjortland
 * * Anya Helene Bagge
 *
 *************************************************************************/
package org.magnolialang.eclipse.editor;

import org.eclipse.core.runtime.IPath;
import io.usethesource.impulse.editor.ModelTreeNode;
import io.usethesource.impulse.parser.ISourcePositionLocator;
import org.rascalmpl.value.IConstructor;
import org.rascalmpl.value.ISourceLocation;
import org.rascalmpl.value.IValue;
import org.eclipse.jdt.annotation.Nullable;
import org.nuthatchery.pica.terms.INodePredicate;
import org.rascalmpl.values.uptr.ITree;
import org.rascalmpl.values.uptr.TreeAdapter;
import org.rascalmpl.values.uptr.RascalValueFactory;

public class NodeLocator implements ISourcePositionLocator {
	public enum LocatorStrategy {
		INNERMOST, OUTERMOST
	}

	public static final LocatorStrategy INNERMOST = LocatorStrategy.INNERMOST;


	public static final LocatorStrategy OUTERMOST = LocatorStrategy.OUTERMOST;;


	@Nullable
	public ITree findNode(final ITree tree, final int offset, final INodePredicate pred, final LocatorStrategy strategy) {
		final ISourceLocation l = TreeAdapter.getLocation(tree);

		if(l == null) {
			throw new IllegalArgumentException("locate assumes position information on the tree");
		}

		if(TreeAdapter.isLexical(tree)) {
			if(l.getOffset() <= offset && offset < l.getOffset() + l.getLength() && pred.match(tree)) {
				return tree;
			}

			// we're at a leaf, but not in position (so zoom out)
			return null;
		}

		if(TreeAdapter.isAmb(tree)) {
			if(pred.match(tree)) {
				return tree;
			}

			return null;
		}

		if(TreeAdapter.isAppl(tree)) {
			// final IList children = tree.

			for(final IValue child : TreeAdapter.getArgs(tree)) {
				final ISourceLocation childLoc = TreeAdapter.getLocation((ITree) child);

				if(childLoc == null) {
					continue;
				}

				if(childLoc.getOffset() <= offset && offset < childLoc.getOffset() + childLoc.getLength()) {
					final ITree result = findNode((ITree) child, offset, pred, strategy);

					if(result != null) {
						return result;
					}
				}
			}

			if(l.getOffset() <= offset && l.getOffset() + l.getLength() >= offset) {
				if(pred.match(tree)) {
					return tree;
				}

				// in scope, but no annotation, so zoom out
				return null;
			}
		}

		return null;
	}


	@Override
	@Nullable
	public Object findNode(final Object ast, final int offset) {
		if(ast instanceof IConstructor) {
			final IConstructor cons = (IConstructor) ast;
			if(cons.getType().equivalent(RascalValueFactory.Tree)) {
				return TreeAdapter.locateLexical((ITree) ast, offset);
			}
		}
		else if(ast instanceof ModelTreeNode) {
			return findNode(((ModelTreeNode) ast).getASTNode(), offset);
		}
		return null;
	}


	@Override
	@Nullable
	public Object findNode(final Object ast, final int startOffset, final int endOffset) {
		if(ast instanceof IConstructor) {
			final IConstructor cons = (IConstructor) ast;
			if(cons.getType().equivalent(RascalValueFactory.Tree)) {
				return TreeAdapter.locateLexical((ITree) ast, startOffset);
			}
		}
		else if(ast instanceof ModelTreeNode) {
			return findNode(((ModelTreeNode) ast).getASTNode(), startOffset);
		}

		return null;
	}


	@Override
	public int getEndOffset(final Object node) {
		return getStartOffset(node) + getLength(node) - 1;
	}


	@Override
	public int getLength(final Object node) {
		if(node instanceof Token) {
			return ((Token) node).getLength();
		}
		ISourceLocation loc = getLocation(node);
		return loc == null ? 0 : loc.getLength();
	}


	private ISourceLocation getLocation(final Object node) {
		if(node instanceof TreeNode) {
			return ((TreeNode) node).getLocation();
		}
		else if(node instanceof IConstructor) {
			return TreeAdapter.getLocation((ITree) node);
		}
		else if(node instanceof ModelTreeNode) {
			return getLocation(((ModelTreeNode) node).getASTNode());
		}
		/*
		 * else if (node instanceof Group<?>) { Group<?> group = (Group<?>)
		 * node; Iterator<?> i = group.iterator(); if (i.hasNext()) { return
		 * getLocation(i.next()); } return group.getLocation(); }
		 */
		return null;
//		throw new ImplementationError("Unknown node type " + node.getClass());
	}


	@Override
	@Nullable
	public IPath getPath(final Object node) {
		return null;
	}


	@Override
	public int getStartOffset(final Object node) {
		if(node instanceof Token) {
			return ((Token) node).getOffset();
		}

		return getLocation(node) == null ? 0 : getLocation(node).getOffset();
	}
}
