/**************************************************************************
 * Copyright (c) 2013 Anya Helene Bagge
 * Copyright (c) 2013 University of Bergen
 *
 * All rights reserved. This program and the accompanying materials
 * are made available under the terms of the Eclipse Public License v1.0
 * which accompanies this distribution, and is available at
 * http://www.eclipse.org/legal/epl-v10.html
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version. See http://www.gnu.org/licenses/
 *
 *
 * See the file COPYRIGHT for more information.
 *
 * Contributors:
 * * Anya Helene Bagge
 *
 *************************************************************************/
package org.magnolialang.eclipse.editor;

import java.util.Arrays;

import org.eclipse.core.runtime.IProgressMonitor;
import org.eclipse.core.runtime.IStatus;
import org.eclipse.core.runtime.Status;
import org.eclipse.core.runtime.SubMonitor;
import org.eclipse.core.runtime.jobs.Job;
import io.usethesource.impulse.parser.IMessageHandler;
import org.rascalmpl.value.IConstructor;
import org.rascalmpl.value.IList;
import org.rascalmpl.value.INode;
import org.rascalmpl.value.ISet;
import org.rascalmpl.value.ISourceLocation;
import org.rascalmpl.value.IString;
import org.rascalmpl.value.IValue;
import org.magnolialang.infra.Infra;
import org.nuthatchery.pica.tasks.ITaskMonitor;
import org.nuthatchery.pica.tasks.eclipse.EclipseTaskMonitor;
import org.rascalmpl.eclipse.nature.RascalMonitor;
import org.rascalmpl.eclipse.nature.WarningsToMarkers;
import org.rascalmpl.values.uptr.RascalValueFactory;
import org.rascalmpl.values.uptr.ITree;
import org.rascalmpl.values.uptr.ProductionAdapter;
import org.rascalmpl.values.uptr.TreeAdapter;
import org.rascalmpl.values.uptr.visitors.TreeVisitor;

public class ParseTreeChecker {
	static class DoctorJob extends Job {
		protected volatile ITree tree;


		public DoctorJob() {
			super("Dr. Ambiguity");
		}


		@Override
		protected IStatus run(IProgressMonitor monitor) {
			ITaskMonitor tm = new EclipseTaskMonitor(SubMonitor.convert(monitor));

			IList msgs = (IList) Infra.get().getEvaluatorPool("Dr. Ambiguity", Arrays.asList("Ambiguity")).call(tm, "diagnose", tree);
			for(IValue msg : msgs) {
				System.out.println("  " + msg);
			}
			tree = null;
			return Status.OK_STATUS;
		}

	}


	static final DoctorJob drJob = new DoctorJob();;


	public static void checkParseTree(ITree tree, final IMessageHandler handler) {
		tree.accept(new TreeVisitor<RuntimeException>() {
			boolean doctorHasBeenCalled = false;


			@Override
			public ITree visitTreeAmb(ITree arg) {
				ISourceLocation loc = TreeAdapter.getLocation(arg);
				if(loc != null) {
					handler.handleSimpleMessage("Ambiguity detected (check log for Dr. Ambiguity's diagnosis)", loc.getOffset(), loc.getOffset() + loc.getLength(), loc.getBeginColumn(), loc.getEndColumn(), loc.getBeginLine(), loc.getEndLine());
				}
				// only use doctor once per tree
				if(!doctorHasBeenCalled) {
					// share a common job, to avoid excessive doctor calls for multiple trees
					drJob.tree = arg;
					drJob.schedule();
					doctorHasBeenCalled = true;
				}
				TreeAdapter.getAlternatives(arg).iterator().next().accept(this);
				return null;
			}


			@Override
			public ITree visitTreeAppl(ITree arg) {
				ISet attrs = ProductionAdapter.getAttributes(TreeAdapter.getProduction(arg));
				for(IValue attr : attrs) {
					if(attr.getType().isAbstractData() && ((IConstructor) attr).getConstructorType() == RascalValueFactory.Attr_Tag) {
						final IValue value = ((IConstructor) attr).get("tag");
						if(value.getType().isNode()) {
							INode node = (INode) value;
							if(node.getName().equals("deprecated")) {
								String msg = "";
								if(node.arity() == 1) {
									msg = ": " + ((IString) node.get(0)).getValue();
								}
								ISourceLocation loc = TreeAdapter.getLocation(arg);
								handler.handleSimpleMessage("Deprecated syntax" + msg, loc.getOffset(), loc.getOffset() + loc.getLength(), loc.getBeginColumn(), loc.getEndColumn(), loc.getBeginLine(), loc.getEndLine());
							}
						}
					}
				}

				for(IValue child : TreeAdapter.getArgs(arg)) {
					child.accept(this);
				}
				return null;
			}


			@Override
			public ITree visitTreeChar(ITree arg) {
				return null;
			}


			@Override
			public ITree visitTreeCycle(ITree arg) {
				return null;
			}
		});
	}
}
