/**************************************************************************
 * Copyright (c) 2009-2011 CWI
 * Copyright (c) 2011-2014 Anya Helene Bagge
 * Copyright (c) 2011-2014 University of Bergen
 *
 * All rights reserved. This program and the accompanying materials
 * are made available under the terms of the Eclipse Public License v1.0
 * which accompanies this distribution, and is available at
 * http://www.eclipse.org/legal/epl-v10.html
 *
 *
 * See the file COPYRIGHT for more information.
 *
 * Contributors:
 * * Andreas Hjortland
 * * Anya Helene Bagge
 * + Jurgen J. Vinju
 *
 *************************************************************************/
package org.magnolialang.eclipse.editor;

import org.rascalmpl.value.IBool;
import org.rascalmpl.value.IConstructor;
import org.rascalmpl.value.IDateTime;
import org.rascalmpl.value.IExternalValue;
import org.rascalmpl.value.IInteger;
import org.rascalmpl.value.IList;
import org.rascalmpl.value.IMap;
import org.rascalmpl.value.INode;
import org.rascalmpl.value.IRational;
import org.rascalmpl.value.IReal;
import org.rascalmpl.value.ISet;
import org.rascalmpl.value.ISourceLocation;
import org.rascalmpl.value.IString;
import org.rascalmpl.value.ITuple;
import org.rascalmpl.value.IValue;
import org.rascalmpl.value.visitors.IValueVisitor;
import io.usethesource.impulse.services.base.TreeModelBuilderBase;
import org.magnolialang.eclipse.MagnoliaPlugin;
import org.rascalmpl.values.uptr.ITree;
import org.rascalmpl.values.uptr.TreeAdapter;

public class TreeModelBuilder extends TreeModelBuilderBase {
	protected final boolean debug = false;


	public TreeModelBuilder() {
		super();
	}


	@Override
	protected void visitTree(Object root) {
		try {
			if(root instanceof INode) {
				INode node = (INode) root;
				for(IValue child : node) {
					/*
					 * The only method which does any proper work in
					 * this anonymous class is the visitConstructor
					 * method.
					 *
					 * The rest of the methods will just return
					 * null unless we have the debug flag enabled.
					 *
					 * If the debug flag is enabled, they will just push
					 * and pop the item which goes in, such that we get
					 * the complete AST in the outline view
					 *
					 * This is only usable if there is some vertices
					 * which you need to add to the outline view.
					 *
					 */
					child.accept(new IValueVisitor<Object, RuntimeException>() {
						@Override
						public Object visitBoolean(IBool o) {
							if(debug) {
								createSubItem(o);
							}
							return null;
						}


						@Override
						public Object visitConstructor(IConstructor o) {
							if(o instanceof ITree) {
								ITree t = (ITree) o;
								if(TreeAdapter.isAppl(t)) {
									String label = TreeNode.getLabel(t);

									boolean shouldPop = false;
									if(label.equals("PackageHead")) {
										pushSubItem(new TreeNode(t));
										shouldPop = true;
									}
									else if(label.equals("TopDefineSugar")) {
										pushSubItem(new TreeNode((ITree) TreeAdapter.getASTArgs(t).get(1)));
										shouldPop = true;
									}
									else if(label.equals("DefineSugar")) {
										pushSubItem(new TreeNode((ITree) TreeAdapter.getASTArgs(t).get(1)));
										shouldPop = true;
									}

									for(IValue child : o) {
										child.accept(this);
									}

									if(shouldPop) {
										popSubItem();
									}
								}
							}
							return null;
						}


						@Override
						public Object visitDateTime(IDateTime o) {
							if(debug) {
								return createSubItem(o);
							}
							return null;
						}


						@Override
						public Object visitExternal(IExternalValue o) {
							if(debug) {
								return createSubItem(o);
							}
							return null;
						}


						@Override
						public Object visitInteger(IInteger o) {
							if(debug) {
								return createSubItem(o);
							}
							return null;
						}


						@Override
						public Object visitList(IList o) {
							if(debug) {
								pushSubItem(o);
							}
							for(IValue elem : o) {
								elem.accept(this);
							}
							if(debug) {
								popSubItem();
							}
							return null;
						}


						@Override
						public Object visitListRelation(IList o) {
							if(debug) {
								pushSubItem(o);
							}
							for(IValue tuple : o) {
								tuple.accept(this);
							}
							if(debug) {
								popSubItem();
							}
							return null;
						}


						@Override
						public Object visitMap(IMap o) {
							if(debug) {
								pushSubItem(o);
							}
							for(IValue key : o) {
								o.get(key).accept(this);
							}
							if(debug) {
								popSubItem();
							}
							return null;
						}


						@Override
						public Object visitNode(INode o) {
							if(debug) {
								pushSubItem(o);
							}
							for(IValue child : o) {
								child.accept(this);
							}
							if(debug) {
								popSubItem();
							}
							return null;
						}


						@Override
						public Object visitRational(IRational o) {
							if(debug) {
								return createSubItem(o);
							}
							return null;
						}


						@Override
						public Object visitReal(IReal o) {
							if(debug) {
								return createSubItem(o);
							}
							return null;
						}


						@Override
						public Object visitRelation(ISet o) {
							if(debug) {
								pushSubItem(o);
							}
							for(IValue tuple : o) {
								tuple.accept(this);
							}
							if(debug) {
								popSubItem();
							}
							return null;
						}


						@Override
						public Object visitSet(ISet o) {
							if(debug) {
								pushSubItem(o);
							}
							for(IValue tuple : o) {
								tuple.accept(this);
							}
							if(debug) {
								popSubItem();
							}
							return null;
						}


						@Override
						public Object visitSourceLocation(ISourceLocation o) {
							if(debug) {
								return createSubItem(o);
							}
							return null;
						}


						@Override
						public Object visitString(IString o) {
							if(debug) {
								return createSubItem(o);
							}
							return null;
						}


						@Override
						public Object visitTuple(ITuple o) {
							if(debug) {
								pushSubItem(o);
							}
							for(IValue field : o) {
								field.accept(this);
							}
							if(debug) {
								popSubItem();
							}
							return null;
						}
					});
				}
			}
		}
		catch(Throwable e) {
			MagnoliaPlugin.getInstance().logException("outliner failed: " + e.getMessage(), e);
			return;
		}
	}
}
