/**************************************************************************
 * Copyright (c) 2013-2014 Anya Helene Bagge
 * Copyright (c) 2013-2014 University of Bergen
 *
 * All rights reserved. This program and the accompanying materials
 * are made available under the terms of the Eclipse Public License v1.0
 * which accompanies this distribution, and is available at
 * http://www.eclipse.org/legal/epl-v10.html
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version. See http://www.gnu.org/licenses/
 *
 *
 * See the file COPYRIGHT for more information.
 *
 * Contributors:
 * * Andreas Hjortland
 * * Anya Helene Bagge
 *
 *************************************************************************/
package org.magnolialang.eclipse.editor;

import org.rascalmpl.value.IConstructor;
import org.rascalmpl.value.IList;
import org.rascalmpl.value.ISourceLocation;
import org.rascalmpl.value.IString;
import org.rascalmpl.value.IValue;
import org.rascalmpl.values.uptr.ITree;
import org.rascalmpl.values.uptr.ProductionAdapter;
import org.rascalmpl.values.uptr.RascalValueFactory;
import org.rascalmpl.values.uptr.TreeAdapter;

/**
 * This class will parse an IConstructor and extract the information we need
 * to populate the Outline.
 *
 * @author Andreas Hjortland
 *
 */
public class TreeNode {
	private String name;

	private ISourceLocation location;

	private String type;


	/**
	 * Constructor for the TreeNode object. Will take a {@link IConstructor},
	 * try to parse it and set the properties of this {@link TreeNode} object.
	 *
	 * @param constructor
	 *            The {@link IConstructor} to parse.
	 * @param shortName
	 *            If this boolean is true, we will only use the name or
	 *            identifier in our outline view. If it is false, we will use
	 *            the string in the source location of the constructor
	 */
	public TreeNode(ITree constructor) {
		if(getLabel(constructor).equals("PackageHead")) {
			ITree tmp = (ITree) TreeAdapter.getASTArgs(constructor).get(0);
			constructor = tmp;
			location = getLocation(constructor);
			name = TreeAdapter.yield(constructor);
			type = "package";
		}
		else {
			location = getLocation(constructor);
			String str = TreeAdapter.yield(constructor);
			type = str.split(" ")[0];
			name = str.substring(type.length()).replaceAll("\\s+", " ").trim();
		}
	}


	/**
	 * This is the source location of the element the vertex in the outline
	 * graph represents
	 *
	 * @return {@link ISourceLocation} of the element.
	 */
	public ISourceLocation getLocation() {
		return location;
	}


	/**
	 * This is the name of the element. This is what will be shown in the
	 * outline view.
	 *
	 * @return Name of the vertex.
	 */
	public String getName() {
		return name;
	}


	/**
	 * This is the type of the element. This should be used to figure out which
	 * image to use to represent it.
	 *
	 * @return String representing the element type.
	 */
	public String getType() {
		return type;
	}


	@Override
	public String toString() {
		return String.format("Name: %s%nType: %s%nLocation: %s", name, type, location);
	}


	/**
	 * Retrieve the label for a {@link IConstructor}. This will traverse the
	 * children of the {@link IConstructor}, and look for a Production vertex,
	 * and then look for the label vertex among the cild vertices of the
	 * Produciton vertex.
	 *
	 * @param constructor
	 *            The vertex we will search for a label
	 * @return A string containing the label of the {@link IConstructor}.
	 */
	public static String getLabel(ITree appl) {
		IConstructor prod = TreeAdapter.getProduction(appl);
		String string = ProductionAdapter.getConstructorName(prod);
		return string == null ? "" : string;
	}


	/**
	 * This helper method will try to get the location of the Name of the
	 * {@link IConstructor} such that we will only highlight the name, and not
	 * the whole signature of the {@link IConstructor}
	 *
	 * @param c
	 *            The {@link IConstructor} which we will find the name for
	 * @return A {@link ISourceLocation} for the name of the
	 *         {@link IConstructor}. If no such location is found, it will
	 *         return the ISourceLocation of the {@link IConstructor} itself
	 */
	private static ISourceLocation getLocation(ITree c) {
		String str = getSort(c);
		if(str.contains("Name")) {
			return (ISourceLocation) c.asAnnotatable().getAnnotation(RascalValueFactory.Location);
		}

		for(IValue v : c) {
			if(v instanceof IList) {
				for(IValue v2 : (IList) v) {
					str = getSort((IConstructor) v2);
					if(str.contains("Name")) {
						return (ISourceLocation) v2.asAnnotatable().getAnnotation(RascalValueFactory.Location);
					}
				}
			}
		}
		System.err.println("Did not find location for " + getLabel(c));
		return (ISourceLocation) c.asAnnotatable().getAnnotation(RascalValueFactory.Location);
	}


	/**
	 * A helper method to get the sort of a IConstructor. This method is used in
	 * getLocation to figure out what kind of node we are looking for.
	 *
	 * @param constructor
	 *            The {@link IConstructor} to search for a sort in.
	 * @return A string representing the sort, if it exists, if not, the empty
	 *         string.
	 */
	private static String getSort(IConstructor constructor) {
		String str = getString(query(query(query(constructor, "prod"), "label"), "sort"));
		if(str.equals("")) {
			return getString(query(query(constructor, "prod"), "sort"));
		}
		return str;
	}


	/**
	 * This method will go through all children of a {@link IConstructor} until
	 * it finds a String
	 *
	 * @param constructor
	 *            The {@link IConstructor} to search through.
	 * @return The first string that is found among the children of the input
	 */
	private static String getString(IConstructor constructor) {
		if(constructor == null) {
			return "";
		}
		for(IValue v : constructor) {
			if(v instanceof IString) {
				return ((IString) v).getValue();
			}
		}
		return "";
	}


	/**
	 * This method will query all the children of an {@link IConstructor} based
	 * on their name
	 *
	 * @param root
	 *            The parent {@link IConstructor} which we will search from.
	 * @param name
	 *            Name of the {@link IConstructor} we are looking for
	 * @return The first {@link IConstructor} matching the input name.
	 */
	private static IConstructor query(IConstructor root, String name) {
		if(root == null) {
			return null;
		}
		for(IValue val : root) {
			if(val instanceof IConstructor) {
				IConstructor con = (IConstructor) val;
				if(con.getName().equals(name)) {
					return con;
				}
			}
		}
		return null;
	}
}
