/**************************************************************************
 * Copyright (c) 2011-2014 Anya Helene Bagge
 * Copyright (c) 2011-2014 University of Bergen
 * 
 * All rights reserved. This program and the accompanying materials
 * are made available under the terms of the Eclipse Public License v1.0
 * which accompanies this distribution, and is available at
 * http://www.eclipse.org/legal/epl-v10.html
 * 
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version. See http://www.gnu.org/licenses/
 * 
 * 
 * See the file COPYRIGHT for more information.
 * 
 * Contributors:
 * * Anya Helene Bagge
 * 
 *************************************************************************/
package org.magnolialang.eclipse.filesystem;

import java.io.ByteArrayInputStream;
import java.io.InputStream;
import java.net.URI;
import java.net.URISyntaxException;
import java.util.HashMap;
import java.util.Map;

import org.eclipse.core.filesystem.EFS;
import org.eclipse.core.filesystem.IFileInfo;
import org.eclipse.core.filesystem.IFileStore;
import org.eclipse.core.filesystem.provider.FileInfo;
import org.eclipse.core.filesystem.provider.FileStore;
import org.eclipse.core.filesystem.provider.FileSystem;
import org.eclipse.core.runtime.CoreException;
import org.eclipse.core.runtime.IPath;
import org.eclipse.core.runtime.IProgressMonitor;
import org.eclipse.core.runtime.IStatus;
import org.eclipse.core.runtime.Path;
import org.eclipse.core.runtime.Status;
import org.magnolialang.eclipse.MagnoliaPlugin;
import org.eclipse.jdt.annotation.Nullable;
import org.nuthatchery.pica.errors.ImplementationError;

public class MgGenFileSystem extends FileSystem {
	private final Map<String, MgGenFileStore> files = new HashMap<String, MgGenFileStore>();


	public void addFile(URI uri, String contents) {
		String path = uri.getPath();
		if(files.containsKey(path)) {
			files.get(path).update(contents);
		}
		else {
			files.put(path, new MgGenFileStore(new Path(path), contents));
		}
	}


	@Override
	public IFileStore getStore(URI uri) {
		String path = uri.getPath();
		if(files.containsKey(path)) {
			return files.get(path);
		}
		else {
			return EFS.getNullFileSystem().getStore(uri);
		}
	}


	public void removeFile(URI uri) {
		MgGenFileStore store = files.remove(uri.getPath());
		if(store != null) {
			store.info.setExists(false);
			store.value = null; // NOPMD by anya on 1/5/12 3:04 AM
		}

	}


	class MgGenFileStore extends FileStore {
		private final IPath path;
		private String value;
		private final FileInfo info;


		public MgGenFileStore(IPath path, String value) {
			super();
			this.path = path;
			this.value = value;
			this.info = new FileInfo(path.toString());
			info.setDirectory(false);
			info.setExists(true);
			info.setLastModified(System.currentTimeMillis());
			info.setLength(value.length());
		}


		@Override
		public String[] childNames(int options, IProgressMonitor monitor) throws CoreException {
			return new String[0];
		}


		@Override
		public IFileInfo fetchInfo(int options, IProgressMonitor monitor) throws CoreException {
			return info;
		}


		@Override
		public IFileStore getChild(String name) {
			return getStore(path.append(name));
		}


		@Override
		public String getName() {
			return path.toString();
		}


		@Override
		@Nullable
		public IFileStore getParent() {
			if(path.isRoot()) {
				return null;
			}
			else {
				return getStore(path.removeLastSegments(1));
			}
		}


		@Override
		public InputStream openInputStream(int options, IProgressMonitor monitor) throws CoreException {
			if(value == null) {
				throw new CoreException(new Status(IStatus.ERROR, MagnoliaPlugin.PLUGIN_ID, "Generated file does not exist: " + path));
			}

			return new ByteArrayInputStream(value.getBytes());
		}


		@Override
		public URI toURI() {
			try {
				return new URI("mggen://" + path.toString());
			}
			catch(URISyntaxException e) {
				throw new ImplementationError("Unable to create URI for path " + path.toString(), e);
			}
		}


		private void update(String value) {
			this.value = value;
			info.setLastModified(System.currentTimeMillis());
		}

	}

}
