/**************************************************************************
 * Copyright (c) 2011-2014 Anya Helene Bagge
 * Copyright (c) 2011-2014 University of Bergen
 * 
 * All rights reserved. This program and the accompanying materials
 * are made available under the terms of the Eclipse Public License v1.0
 * which accompanies this distribution, and is available at
 * http://www.eclipse.org/legal/epl-v10.html
 * 
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version. See http://www.gnu.org/licenses/
 * 
 * 
 * See the file COPYRIGHT for more information.
 * 
 * Contributors:
 * * Anya Helene Bagge
 * 
 *************************************************************************/
package org.magnolialang.eclipse.icons;

import java.io.BufferedReader;
import java.io.IOException;
import java.io.InputStreamReader;
import java.util.HashMap;
import java.util.Map;

import org.eclipse.core.runtime.FileLocator;
import org.eclipse.core.runtime.IPath;
import org.eclipse.jface.resource.ImageDescriptor;
import org.eclipse.jface.viewers.DecorationOverlayIcon;
import org.eclipse.jface.viewers.IDecoration;
import org.eclipse.swt.graphics.Color;
import org.eclipse.swt.graphics.Image;
import org.eclipse.swt.widgets.Display;
import org.eclipse.ui.PlatformUI;
import org.magnolialang.eclipse.MagnoliaPlugin;
import org.magnolialang.eclipse.MagnoliaResources;
import org.eclipse.jdt.annotation.Nullable;

public class StyleBase {
	public final Map<String, Color> colors = new HashMap<String, Color>();
	public final Map<String, String> letters = new HashMap<String, String>();
	private final String prefix;
	private final Map<String, ImageDescriptor> icons = new HashMap<String, ImageDescriptor>();
	private final Map<String, Integer> positions = new HashMap<String, Integer>();


	public StyleBase(String prefix) {
		if(prefix == null) {
			throw new IllegalArgumentException("Prefix should not be null");
		}

		this.prefix = prefix;
	}


	/**
	 * Return the color associated with the given kind of entity
	 */
	@Nullable
	public Color getColor(String kind) {
		if(colors.isEmpty()) {
			loadTables();
		}
		return colors.get(kind);
	}


	public Display getDevice() {
		Display disp = Display.getCurrent();
		if(disp == null) {
			disp = PlatformUI.getWorkbench().getDisplay();
		}
		return disp;
	}


	/**
	 * Obtain an ImageDescriptor for an icon, with optional overlays.
	 * 
	 * @param key
	 *            An icon id, compose of a base icon name, followed by an
	 *            optional list of overlays (all separated by '+'); e.g.,
	 *            "class+error"
	 * @param suffix
	 *            Icon suffix, e.g., "_obj"
	 * @return An ImageDescriptor, or null if no such icon was found
	 */
	@Nullable
	public ImageDescriptor getIcon(String key, String suffix) {
		if(icons.containsKey(key)) {
			return icons.get(key);
		}
		if(suffix == null) {
			suffix = "";
		}

		String[] desc = key.split("\\+");
		String baseIcon = desc[0] + suffix;

		if(desc.length > 1) {
			Image base = MagnoliaResources.getImage(baseIcon);
			if(base == null) {
				return null;
			}

			@Nullable
			ImageDescriptor[] compImages = new ImageDescriptor[5];
			int pos = 0;
			for(int i = 1; i < desc.length; i++) {
				String imageName = desc[i];
				if(imageName.equals("")) {
					pos++;
				}
				else {
					// if the icon name is associated with a particular
					// position, use that, otherwise go clockwise from
					// top left corner.
					int p = getPosition(imageName);
					imageName = imageName + "_ovr";
					ImageDescriptor img = MagnoliaResources.getImageDescriptor(imageName);
					if(p < 0 || p >= compImages.length) {
						while(compImages[pos] != null && pos < compImages.length) {
							pos++;
						}
						if(pos < compImages.length) {
							compImages[pos++] = img;
						}
					}
					else {
						compImages[p] = img;
					}
				}
			}

			DecorationOverlayIcon icon = new DecorationOverlayIcon(base, compImages);
			MagnoliaResources.putImageDescriptor(key, icon);
			return icon;
		}
		else {
			return MagnoliaResources.getImageDescriptor(baseIcon);
		}
	}


	/**
	 * Return the letter associated with the given kind of entity
	 */
	@Nullable
	public String getLetter(String kind) {
		if(letters.isEmpty()) {
			loadTables();
		}
		return letters.get(kind);
	}


	/**
	 * Return the position of the icon for use as an overlay. Positions are
	 * numbered clockwise, with the top-left corner being 0, and the bottom-left
	 * corner being 3. The special 'underlay' position is 5.
	 * 
	 * Returns -1 if there is no position associated with the icon.
	 */
	public int getPosition(String iconId) {
		if(positions.isEmpty()) {
			loadTables();
		}
		if(positions.containsKey(iconId)) {
			return positions.get(iconId);
		}
		else {
			return -1;
		}
	}


	public void refresh() {
		positions.clear();
		colors.clear();
		letters.clear();
		icons.clear();
	}


	private Color cacheColor(String string, Color color) {
		colors.put(string, color);
		return color;
	}


	private int decodeHex(String string) {
		return Integer.parseInt(string, 16);
	}


	private int decodePosition(String posString) throws SyntaxErrorException {
		if(posString.equals("top-left")) {
			return IDecoration.TOP_LEFT;
		}
		else if(posString.equals("top-right")) {
			return IDecoration.TOP_RIGHT;
		}
		else if(posString.equals("bottom-right")) {
			return IDecoration.BOTTOM_RIGHT;
		}
		else if(posString.equals("bottom-left")) {
			return IDecoration.BOTTOM_LEFT;
		}
		else if(posString.equals("underlay")) {
			return IDecoration.UNDERLAY;
		}
		else {
			throw new SyntaxErrorException("Unknown icon overlay position '" + posString + "'");
		}
	}


	private void loadTables() {
		IPath fileName = MagnoliaResources.ICONS_PATH.append(prefix + ".style");
		try {
			BufferedReader reader = new BufferedReader(new InputStreamReader(FileLocator.openStream(MagnoliaPlugin.getInstance().getBundle(), fileName, false)));
			try {
				String line = reader.readLine();
				int lineNum = 1;
				while(line != null) {
					line = line.trim();
					if(!line.startsWith("#")) {
						try {
							String[] strings = line.split(":", 6);
							if(strings.length < 3) {
								throw new SyntaxErrorException("Ill-formed line (too few fields)");
							}
							String iconType = strings[0];
							if(iconType.equals("object")) {
								if(strings.length != 5) {
									throw new SyntaxErrorException("Ill-formed line (should have 5 fields)");
								}
								String key = strings[1];
								letters.put(key, strings[3]);
								colors.put(key, makeColor(strings[4], "#000"));
							}
							else if(iconType.equals("overlay")) {
								if(strings.length != 6) {
									throw new SyntaxErrorException("Ill-formed line (should have 6 fields)");
								}
								String key = strings[1];
								positions.put(key, decodePosition(strings[3]));
								letters.put(key, strings[4]);
								colors.put(key, makeColor(strings[5], "#000"));
							}
						}
						catch(SyntaxErrorException e) { // NOPMD by anya on 1/5/12 5:33 AM
							MagnoliaPlugin.getInstance().logException("Error in style file " + fileName + ":" + lineNum + ": " + (e.getMessage() == null ? "unknown error" : e.getMessage()), null);
						}
					}
					line = reader.readLine();
					lineNum++;
				}
			}
			finally {
				reader.close();
			}
		}
		catch(IOException e) {
			// TODO Auto-generated catch block
			e.printStackTrace();
		}
	}


	private Color makeColor(String string, String dflt) {
		if(colors.containsKey(string)) {
			return colors.get(string);
		}

		if(string.length() == 0) {
			string = dflt;
		}

		if(string.startsWith("#")) {
			string = string.substring(1, string.length());
		}
		if(string.length() == 6) {
			int r = decodeHex(string.substring(0, 2));
			int g = decodeHex(string.substring(2, 4));
			int b = decodeHex(string.substring(4, 6));
			return cacheColor(string, new Color(getDevice(), r, g, b));
		}
		else if(string.length() == 3) {
			int r = decodeHex(string.substring(0, 1) + string.substring(0, 1));
			int g = decodeHex(string.substring(1, 2) + string.substring(1, 2));
			int b = decodeHex(string.substring(2, 3) + string.substring(2, 3));
			return cacheColor(string, new Color(getDevice(), r, g, b));
		}
		else if(string.length() == 2) {
			int r = decodeHex(string.substring(0, 2));
			return cacheColor(string, new Color(getDevice(), r, r, r));
		}
		else {
			throw new IllegalArgumentException("Illegal color specification: '" + string + "'");
		}
	}


	private static class SyntaxErrorException extends Exception {
		private static final long serialVersionUID = 7257627873775658472L;


		SyntaxErrorException(String message) {
			super(message);
		}
	}
}
