/**************************************************************************
 * Copyright (c) 2010-2013 Anya Helene Bagge
 * Copyright (c) 2010-2013 University of Bergen
 * 
 * All rights reserved. This program and the accompanying materials
 * are made available under the terms of the Eclipse Public License v1.0
 * which accompanies this distribution, and is available at
 * http://www.eclipse.org/legal/epl-v10.html
 * 
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version. See http://www.gnu.org/licenses/
 * 
 * 
 * See the file COPYRIGHT for more information.
 * 
 * Contributors:
 * * Anya Helene Bagge
 * 
 *************************************************************************/
package org.magnolialang.eclipse.preferences;

import org.eclipse.jface.preference.BooleanFieldEditor;
import org.eclipse.jface.preference.FieldEditorPreferencePage;
import org.eclipse.jface.preference.FileFieldEditor;
import org.eclipse.jface.preference.RadioGroupFieldEditor;
import org.eclipse.jface.util.IPropertyChangeListener;
import org.eclipse.jface.util.PropertyChangeEvent;
import org.eclipse.ui.IWorkbench;
import org.eclipse.ui.IWorkbenchPreferencePage;
import org.magnolialang.eclipse.MagnoliaPlugin;
import org.magnolialang.infra.Infra;
import org.nuthatchery.pica.resources.IWorkspaceManager.ThreadingPreference;

/**
 * This class represents a preference page that
 * is contributed to the Preferences dialog. By
 * subclassing <samp>FieldEditorPreferencePage</samp>, we
 * can use the field support built into JFace that allows
 * us to create a page that is small and knows how to
 * save, restore and apply itself.
 * <p>
 * This page is used to modify preferences only. They are stored in the
 * preference store that belongs to the main plug-in class. That way,
 * preferences can be accessed directly via the preference store.
 */

public class MagnoliaPreferencePage extends FieldEditorPreferencePage implements IWorkbenchPreferencePage {

	public MagnoliaPreferencePage() {
		super(GRID);
		setPreferenceStore(MagnoliaPlugin.getInstance().getPreferenceStore());
		setDescription("General Magnolia Preferences:");
	}


	/**
	 * Creates the field editors. Field editors are abstractions of
	 * the common GUI blocks needed to manipulate various types
	 * of preferences. Each field editor knows how to save and
	 * restore itself.
	 */
	@Override
	public void createFieldEditors() {
		for(Pref p : Pref.values()) {
			if(p.getPage().equals("MAIN")) {
				if(p.isString() && p.name().endsWith("_PATH")) {
					FileFieldEditor editor = new FileFieldEditor(p.name(), p.getLabel(), true, getFieldEditorParent());
					if(!p.getTooltip().equals("")) {
						editor.getLabelControl(getFieldEditorParent()).setToolTipText(p.getTooltip());
					}
					addField(editor);
				}
				else if(p.isBoolean()) {
					BooleanFieldEditor editor = new BooleanFieldEditor(p.name(), p.getLabel(), getFieldEditorParent());
					if(!p.getTooltip().equals("")) {
						editor.getDescriptionControl(getFieldEditorParent()).setToolTipText(p.getTooltip());
					}
					addField(editor);
				}
				else if(p == Pref.MAIN_MULTITHREADING) {
					RadioGroupFieldEditor editor = new RadioGroupFieldEditor(p.name(), p.getLabel(), 4, //
					new String[][] { { "&Aggressive", "AGGRESSIVE" },//
					{ "&Normal", "NORMAL" },//
					{ "&Moderate", "MODERATE" },//
					{ "N&one", "NONE" },//
					}, getFieldEditorParent());
					if(!p.getTooltip().equals("")) {
						editor.getLabelControl(getFieldEditorParent()).setToolTipText(p.getTooltip());
					}
					addField(editor);
				}
			}
		}

		getPreferenceStore().addPropertyChangeListener(new IPropertyChangeListener() {

			@Override
			public void propertyChange(PropertyChangeEvent event) {
				if(event.getProperty().equals(Pref.MAIN_MULTITHREADING.name())) {
					Object value = event.getNewValue();
					if(value instanceof String) {
						Infra.getWorkspaceManager().setThreadingPreference(ThreadingPreference.getPref((String) value));
					}
				}
			}
		});

		// org.eclipse.jface.preference.

		/*
		addField(new DirectoryFieldEditor(PreferenceConstants.P_PATH, "&Directory preference:", getFieldEditorParent()));
		addField(new BooleanFieldEditor(PreferenceConstants.P_BOOLEAN, "&An example of a boolean preference", getFieldEditorParent()));

		addField(new RadioGroupFieldEditor(PreferenceConstants.P_CHOICE, "An example of a multiple-choice preference", 1, new String[][] { { "&Choice 1", "choice1" }, { "C&hoice 2", "choice2" } }, getFieldEditorParent()));
		addField(new StringFieldEditor(PreferenceConstants.P_STRING, "A &text preference:", getFieldEditorParent()));
		*/
	}


	/* (non-Javadoc)
	 * @see org.eclipse.ui.IWorkbenchPreferencePage#init(org.eclipse.ui.IWorkbench)
	 */
	@Override
	public void init(IWorkbench workbench) {
	}

}
