/**************************************************************************
 * Copyright (c) 2013-2014 Anya Helene Bagge
 * Copyright (c) 2013-2014 University of Bergen
 * 
 * All rights reserved. This program and the accompanying materials
 * are made available under the terms of the Eclipse Public License v1.0
 * which accompanies this distribution, and is available at
 * http://www.eclipse.org/legal/epl-v10.html
 * 
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version. See http://www.gnu.org/licenses/
 * 
 * 
 * See the file COPYRIGHT for more information.
 * 
 * Contributors:
 * * Anya Helene Bagge
 * 
 *************************************************************************/
package org.magnolialang.eclipse.preferences;

import org.eclipse.jdt.annotation.NonNullByDefault;
import org.eclipse.jdt.annotation.Nullable;
import org.nuthatchery.pica.errors.ImplementationError;

/**
 * Constant definitions for plug-in preferences
 */
@NonNullByDefault
public enum Pref {
	MAIN_OPTIMIZER_PATH("Path to optimizer executable", "[Experimental]", ""), //
	MAIN_Z3_PATH("Path to Z3 executable", "[Experimental]", ""), //
	MAIN_MULTITHREADING("Multithreading", "Controls how many threads and Rascal evaluators are created. Aggressive threading uses a lot of Java heap memory. Aggressive: 2 threads/core; normal: 1.4 threads/core; moderate: 0.8 threads/core; none: avoid multithreading if possible.", "normal"), //
	BUILD_COPY_ONLY("Don't build, just &copy files", "If true, build will only copy files, not build any source files.", false), //
	BUILD_ERRONEOUS("Also build &erroneous files", "If true, build will create stubs for erroneous operations, and proceed with build.", false), //
	BUILD_BROKEN("Also build @&broken (disabled due to bugs in code or compiler) modules", "If true, build will attempt to build modules marked as @broken", false), //
	BUILD_CLEAN_RETAIN_HIERARCHY("&Keep directories when cleaning if they exist in src/", "Whether to keep directories in cxx/ that have a counterpart in src/ when cleaning", true), //
	BUILD_CLEAN_ONLY_DERIVED("Clean &derived files only", "Only clean files marked in Eclipse as 'derived'. Enable this if you have other tools producing resources in the output folder that you don't want Magnolia to clean.", false), //
	BUILD_VIRTUAL_CLEAN("Clean only files that won't be &overwritten by build", "Don't remove files or directories that will be overwritten anyway by an immediately following build. This is usually a good thing.", true), //
	BUILD_CLEAN_OUT_OF_SYNC("Clean files even when workspace is out of &sync", "Whether files that are out-of-sync with the workspace should be deleted. If false, an exception is thrown when a 'clean' operation encounters out-of-sync resources.", true), //
	BUILD_DEBUG("Debu&g build decisions", "If true, print a debug trace of decisions made while copying files", false), //
	;

	private final String label;
	private final String tooltip;
	@Nullable
	private final String defString;
	private final boolean defBoolean;
	private final String page;


	Pref(String label, String tooltip, boolean defValue) {
		this.label = label;
		this.tooltip = tooltip;
		this.defString = null;
		this.defBoolean = defValue;
		this.page = name().substring(0, name().indexOf('_'));
	}


	Pref(String label, String tooltip, String defValue) {
		this.label = label;
		this.tooltip = tooltip;
		this.defString = defValue;
		this.defBoolean = false;
		this.page = name().substring(0, name().indexOf('_'));
	}


	public boolean getBooleanValue() {
		if(defString != null) {
			throw new ImplementationError("Not a boolean preference");
		}
		return defBoolean;
	}


	public String getLabel() {
		return label;
	}


	public String getPage() {
		return page;
	}


	public String getStringValue() {
		if(defString != null) {
			return defString;
		}
		throw new ImplementationError("Not a string preference");
	}


	public String getTooltip() {
		return tooltip;
	}


	public boolean isBoolean() {
		return defString == null;
	}


	public boolean isString() {
		return defString != null;
	}
}
