/**************************************************************************
 * Copyright (c) 2012-2014 Anya Helene Bagge
 * Copyright (c) 2012-2014 University of Bergen
 *
 * All rights reserved. This program and the accompanying materials
 * are made available under the terms of the Eclipse Public License v1.0
 * which accompanies this distribution, and is available at
 * http://www.eclipse.org/legal/epl-v10.html
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version. See http://www.gnu.org/licenses/
 *
 *
 * See the file COPYRIGHT for more information.
 *
 * Contributors:
 * * Anya Helene Bagge
 *
 *************************************************************************/
package org.magnolialang.eclipse.tests;

import static org.junit.Assert.*;

import java.io.ByteArrayInputStream;
import java.io.UnsupportedEncodingException;
import java.util.ArrayList;
import java.util.Collections;
import java.util.Comparator;
import java.util.HashMap;
import java.util.List;
import java.util.Set;

import org.eclipse.core.resources.IFile;
import org.eclipse.core.resources.IFolder;
import org.eclipse.core.resources.IProject;
import org.eclipse.core.resources.WorkspaceJob;
import org.eclipse.core.runtime.CoreException;
import org.eclipse.core.runtime.IProgressMonitor;
import org.eclipse.core.runtime.IStatus;
import org.eclipse.core.runtime.NullProgressMonitor;
import org.eclipse.core.runtime.Status;
import org.eclipse.core.runtime.SubMonitor;
import org.eclipse.core.runtime.jobs.Job;
import org.junit.After;
import org.junit.Before;
import org.junit.Test;
import org.magnolialang.eclipse.MagnoliaPlugin;
import org.magnolialang.infra.Infra;
import org.magnolialang.magnolia.Magnolia;
import org.nuthatchery.pica.resources.IProjectManager;
import org.nuthatchery.pica.resources.eclipse.EclipseProjectManager;
import org.nuthatchery.pica.resources.managed.IManagedCodeUnit;
import org.nuthatchery.pica.tasks.ITaskMonitor;
import org.nuthatchery.pica.tasks.eclipse.EclipseTaskMonitor;
import org.nuthatchery.pica.util.depgraph.IDepGraph;
import org.nuthatchery.pica.util.depgraph.IWritableDepGraph;
import org.nuthatchery.pica.util.tests.DepGraphGenerator;
import org.rascalmpl.eclipse.nature.RascalMonitor;
import org.rascalmpl.eclipse.nature.WarningsToMarkers;

import edu.umd.cs.findbugs.annotations.SuppressFBWarnings;

public class DepGraphTest {
	private final NullProgressMonitor monitor = new NullProgressMonitor();

	private IProject project;

	private IFolder src;

	private static final String PROJECT_NAME = "DepGraphTestProject";


	@Before
	public void setUp() throws Exception {
		MagnoliaPlugin.getOrStartInstance();
		project = ProjectUtil.createProject(PROJECT_NAME, true);
		src = project.getFolder("src");
		src.create(true, true, monitor);
	}


	@After
	public void tearDown() throws Exception {
		project.delete(true, monitor);
	}


	@Test
	@SuppressFBWarnings("SBSC_USE_STRINGBUFFER_CONCATENATION")
	public final void test() throws UnsupportedEncodingException, CoreException, InterruptedException {
		int N = 5000, OPS = 20000;
		long t0;

		Magnolia.getInstance().getCompiler().ensureInit();
		List<IFile> files = new ArrayList<IFile>();
		for(int i = 0; i < N; i++) {
			String name = "";
			int j = i;
			do {
				name += String.valueOf((char) ('A' + j % 26));
				j /= 26;
			}
			while(j != 0);
			files.add(src.getFile(name + ".mg"));
		}
		System.err.print("Generating graph...");
		t0 = System.currentTimeMillis();
		final IWritableDepGraph<IFile> graph = DepGraphGenerator.genDepGraph(files, OPS, true);
		System.err.println("done (" + (System.currentTimeMillis() - t0) + "ms)");

		System.err.print("Creating files...");
		Job job = new WorkspaceJob("Creating files") {
			@Override
			public IStatus runInWorkspace(IProgressMonitor monitor) throws CoreException {
				monitor.beginTask("Creating files", graph.getElements().size());
				for(IFile f : graph.topological()) {
					monitor.subTask("Creating " + f);
					Set<IFile> depends = graph.getDepends(f);
					StringBuilder code = new StringBuilder();
					code.append("package ");
					code.append(pkgNameOf(f));
					code.append(" imports ");
					String comma = "";
					for(IFile d : depends) {
						code.append(comma);
						code.append(pkgNameOf(d));
						comma = ", ";
					}
					code.append(";\n");
					//System.err.print(code.toString());
					try {
						f.create(new ByteArrayInputStream(code.toString().getBytes("UTF-8")), true, monitor);
					}
					catch(UnsupportedEncodingException e) {
						e.printStackTrace();
					}
					monitor.worked(1);
				}
				monitor.done();
				return Status.OK_STATUS;
			}
		};
		job.schedule();
		job.join();
		System.err.println("done");

		IProjectManager manager = Infra.getResourceManager(project.getName());
		assert manager != null;

		ITaskMonitor tm = new EclipseTaskMonitor(SubMonitor.convert(monitor));
		System.err.println("Ready to process changes");
		// Thread.sleep(5000);
		System.err.print("Processing changes...");
		t0 = System.currentTimeMillis();
		manager.processChanges(tm);
		System.err.println("done (" + (System.currentTimeMillis() - t0) + "ms)");

		System.err.print("Getting graph...");
		t0 = System.currentTimeMillis();
		IDepGraph<IManagedCodeUnit> depGraph = ((EclipseProjectManager) manager).getPackageDependencyGraph(tm);
		System.err.println("done (" + (System.currentTimeMillis() - t0) + "ms)");

		System.err.println("STARTING COMPARISON");
		assertSameGraph(graph, depGraph);
		System.err.println("OK!");
	}


	public static void assertSameGraph(IDepGraph<IFile> fileGraph, IDepGraph<IManagedCodeUnit> depGraph) {
		List<IFile> fileElements = new ArrayList<IFile>(fileGraph.getElements());
		List<IManagedCodeUnit> pkgElements = new ArrayList<IManagedCodeUnit>(depGraph.getElements());
		assertEquals(fileElements.size(), pkgElements.size());

		Collections.sort(fileElements, new Comparator<IFile>() {
			@Override
			public int compare(IFile arg0, IFile arg1) {
				return arg0.getName().compareTo(arg1.getName());
			}
		});
		Collections.sort(pkgElements, new Comparator<IManagedCodeUnit>() {
			@Override
			public int compare(IManagedCodeUnit o1, IManagedCodeUnit o2) {
				return o1.getName().compareTo(o2.getName());
			}
		});

		HashMap<IFile, IManagedCodeUnit> map = new HashMap<>();

		for(int i = 0; i < fileElements.size(); i++) {
			assertSamePackage(fileElements.get(i), pkgElements.get(i));
			map.put(fileElements.get(i), pkgElements.get(i));
		}

		for(IFile f : fileElements) {
			Set<IFile> fileDepends = fileGraph.getDepends(f);
			Set<IManagedCodeUnit> pkgDepends = depGraph.getDepends(map.get(f));
			System.err.print(f.getName() + " ");
			assertEquals(fileDepends.size(), pkgDepends.size());
			for(IFile g : fileDepends) {
				assertTrue(pkgDepends.contains(map.get(g)));
			}
		}
		System.err.println();
	}


	public static void assertSamePackage(IFile file, IManagedCodeUnit iManagedCodeUnit) {
		String name = file.getName().substring(0, file.getName().lastIndexOf('.'));
		assertEquals(name, iManagedCodeUnit.getName());
	}


	public static boolean isSamePackage(IFile file, IManagedCodeUnit pkg) {
		return pkgNameOf(file).equals(pkg.getName());
	}


	public static String pkgNameOf(IFile file) {
		return file.getName().substring(0, file.getName().lastIndexOf('.'));
	}
}
