/**************************************************************************
 * Copyright (c) 2012-2014 Anya Helene Bagge
 * Copyright (c) 2012-2014 Tero Hasu
 * Copyright (c) 2012-2014 University of Bergen
 *
 * All rights reserved. This program and the accompanying materials
 * are made available under the terms of the Eclipse Public License v1.0
 * which accompanies this distribution, and is available at
 * http://www.eclipse.org/legal/epl-v10.html
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version. See http://www.gnu.org/licenses/
 *
 *
 * See the file COPYRIGHT for more information.
 *
 * Contributors:
 * * Anya Helene Bagge
 * * Tero Hasu
 *
 *************************************************************************/
package org.magnolialang.infra;

import java.io.File;
import java.io.FileNotFoundException;
import java.io.IOException;
import java.io.InputStream;
import java.nio.charset.Charset;
import java.util.List;

import org.rascalmpl.value.IConstructor;
import org.rascalmpl.value.ISourceLocation;
import org.rascalmpl.value.IValue;
import org.rascalmpl.value.IValueFactory;
import org.rascalmpl.value.io.BinaryValueReader;
import org.rascalmpl.value.io.BinaryValueWriter;
import org.rascalmpl.value.type.TypeStore;
import org.eclipse.jdt.annotation.Nullable;
import org.magnolialang.magnolia.Magnolia;
import org.nuthatchery.pica.IPica;
import org.nuthatchery.pica.errors.ParserLoadError;
import org.nuthatchery.pica.errors.ParserNotFoundError;
import org.nuthatchery.pica.rascal.ICompiler;
import org.nuthatchery.pica.rascal.IEvaluatorPool;
import org.nuthatchery.pica.rascal.RascalParserLoader;
import org.nuthatchery.pica.resources.IWorkspaceManager;
import org.nuthatchery.pica.util.NullnessHelper;
import org.rascalmpl.parser.gtd.IGTD;
import org.rascalmpl.parser.gtd.io.InputConverter;
import org.rascalmpl.values.uptr.ITree;

public abstract class AbstractInfra implements IInfra {
	@Nullable
	private RascalParserLoader parserLoader = null;
	protected final IPica picaInfra;


	public AbstractInfra(org.nuthatchery.pica.IPica picaInfra) {
		this.picaInfra = picaInfra;
	}


	@Override
	public void clearParser(String moduleName) {
		RascalParserLoader.clearParser(moduleName);
	}


	@Override
	public synchronized ICompiler getCompiler() {
		return Magnolia.getInstance().getCompiler();
	}


	@Override
	public IEvaluatorPool getEvaluatorPool(String name, List<String> imports) {
		return getEvaluatorFactory().getEvaluatorPool(name, imports);
	}


	@Override
	public IGTD<IConstructor, ITree, ISourceLocation> getParser(String moduleName) throws ParserLoadError, ParserNotFoundError {
		RascalParserLoader loader = parserLoader;
		if(loader == null) {
			loader = parserLoader = new RascalParserLoader(NullnessHelper.assertNonNull(getClass().getClassLoader()), true);
		}
		return loader.getParser(moduleName);

	}


	@Override
	public IWorkspaceManager getWorkspaceManager() {
		return picaInfra.getWorkspaceManager();
	}


	@Override
	public IValue loadData(String fileName, IValueFactory vf, TypeStore ts) throws IOException {
		File file = getDataFile(fileName);
		if(file != null) {
			return NullnessHelper.assertNonNull(BinaryValueReader.readValueFromFile(vf, ts, file));
		}
		else {
			throw new FileNotFoundException(fileName);
		}
	}


	@Override
	public String readFile(String path) throws IOException {
		InputStream inputStream = openStream(path);
		char[] cs;
		try {
			cs = InputConverter.toChar(inputStream, Charset.forName("UTF-8"));
		}
		finally {
			inputStream.close();
		}
		return NullnessHelper.assertNonNull(String.valueOf(cs));
	}


	@Override
	public void refresh() {
		getEvaluatorFactory().refresh();
	}


	@Override
	public boolean removeDataFile(String fileName) {
		File file = getDataFile(fileName);
		if(file != null) {
			return file.delete();
		}
		else {
			return false;
		}
	}


	@Override
	public void saveData(String fileName, IValue value, TypeStore typeStore) throws IOException {
		File file = getDataFile(fileName);
		if(file != null) {
			BinaryValueWriter.writeValueToFile(value, file, typeStore);
		}
		else {
			throw new FileNotFoundException(fileName);
		}

	}
}
