/**************************************************************************
 * Copyright (c) 2012-2014 Anya Helene Bagge
 * Copyright (c) 2012-2014 Tero Hasu
 * Copyright (c) 2012-2014 University of Bergen
 * 
 * All rights reserved. This program and the accompanying materials
 * are made available under the terms of the Eclipse Public License v1.0
 * which accompanies this distribution, and is available at
 * http://www.eclipse.org/legal/epl-v10.html
 * 
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version. See http://www.gnu.org/licenses/
 * 
 * 
 * See the file COPYRIGHT for more information.
 * 
 * Contributors:
 * * Anya Helene Bagge
 * * Tero Hasu
 * 
 *************************************************************************/
package org.magnolialang.infra;

import java.io.File;
import java.io.FileInputStream;
import java.io.IOException;
import java.io.InputStream;
import java.util.Collections;
import java.util.List;

import org.rascalmpl.value.IConstructor;
import org.eclipse.jdt.annotation.Nullable;
import org.nuthatchery.pica.errors.Severity;
import org.nuthatchery.pica.rascal.EvaluatorFactory;
import org.nuthatchery.pica.rascal.IEvaluatorFactory;
import org.nuthatchery.pica.resources.ILanguage;
import org.nuthatchery.pica.resources.IWorkspaceManager;
import org.rascalmpl.debug.IRascalMonitor;
import org.rascalmpl.interpreter.NullRascalMonitor;

/**
 * For use in a static setting, where the input files do not change.
 * Any user feedback is output to the console.
 */
public final class ConsoleInfra extends AbstractInfra {
	private static List<String> rascalSearchPath = Collections.emptyList();

	private static List<String> magnoliaSearchPath = Collections.emptyList();
	private final IRascalMonitor rm = new NullRascalMonitor();

	protected static final String parserPackageName = "org.rascalmpl.java.parser.object";

	private final static MagnoliaWorkspaceConfig config = new MagnoliaWorkspaceConfig();
	private final static IEvaluatorFactory evaluatorFactory = new EvaluatorFactory(config);


	public ConsoleInfra() {
		super(new org.nuthatchery.pica.ConsolePicaInfra(config));
	}


	@Override
	public boolean areModuleFactsPreloaded() {
		return false;
	}


	@Override
	@Nullable
	public File getDataFile(String fileName) {
		String home = System.getenv("HOME");
		if(home != null) {
			return new File(home + File.pathSeparator + ".magnolia" + File.pathSeparator + fileName);
		}
		else {
			return null;
		}
	}


	@Override
	public IEvaluatorFactory getEvaluatorFactory() {
		return evaluatorFactory;
	}


	public IRascalMonitor getMonitor() {
		return rm;
	}


	@Override
	public IWorkspaceManager getWorkspaceManager() {
		throw new UnsupportedOperationException();
	}


	@Override
	public void logException(@Nullable String msg, @Nullable Throwable t) {
		System.err.println("error: " + (t != null ? t.toString() : "no exception") + " (" + (msg != null ? msg : "no details") + ")");
		if(t != null) {
			t.printStackTrace();
		}
	}


	@Override
	public void logMessage(String msg) {
		logMessage(msg, Severity.DEFAULT);
	}


	@Override
	public void logMessage(String msg, Severity severity) {
		System.err.println(severity.toString() + ": " + msg);
	}


	@Override
	public InputStream openStream(String path) throws IOException {
		return new FileInputStream(path);
	}


	/** Returns null if not found from the search path. */
	@Nullable
	public static File findModuleFile(IConstructor nameAst, ILanguage lang) {
		String modName = lang.getNameString(nameAst);
		//System.err.println("---modName is " + modName);
		for(String loadPath : magnoliaSearchPath) {
			//System.err.println("---checking load path " + loadPath);
			String relPath = lang.getFileName(modName);
			//System.err.println("---relPath is " + relPath);
			File file = new File(loadPath, relPath);
			//System.err.println("---looking for file " + file);
			if(file.exists()) {
				//System.err.println("produce ModuleResource " + file + " for " + modName);
				return file;
			}
		}
		return null;
	}


	public static ConsoleInfra getInfra() {
		return (ConsoleInfra) Infra.get();
	}


	public static List<String> getMagnoliaSearchPath() {
		return magnoliaSearchPath;
	}


	public static List<String> getRascalSearchPath() {
		return rascalSearchPath;
	}


	/**
	 * This always sets a fresh instance, even if already set.
	 * This is to help get the system in a known state, e.g.
	 * for repeatable testing (cf. test fixture).
	 * (A caveat is that there is a lot of mutable static final state in the
	 * rest of the system which is not easily reset.)
	 */
	public static void setInfra() {
		Infra.set(new ConsoleInfra());
	}


	public static void setMagnoliaSearchPath(List<String> magnoliaSearchPath) {
		ConsoleInfra.magnoliaSearchPath = magnoliaSearchPath;
	}


	/**
	 * @param rascalSearchPath
	 *            list of absolute paths from which to look for Rascal modules
	 */
	public static void setRascalSearchPath(List<String> rascalSearchPath) {
		ConsoleInfra.rascalSearchPath = rascalSearchPath;
	}

}
