/**************************************************************************
 * Copyright (c) 2012-2014 Anya Helene Bagge
 * Copyright (c) 2012-2014 Tero Hasu
 * Copyright (c) 2012-2014 University of Bergen
 *
 * All rights reserved. This program and the accompanying materials
 * are made available under the terms of the Eclipse Public License v1.0
 * which accompanies this distribution, and is available at
 * http://www.eclipse.org/legal/epl-v10.html
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version. See http://www.gnu.org/licenses/
 *
 *
 * See the file COPYRIGHT for more information.
 *
 * Contributors:
 * * Anya Helene Bagge
 * * Tero Hasu
 *
 *************************************************************************/
package org.magnolialang.infra;

import java.io.File;
import java.io.IOException;
import java.io.InputStream;
import java.util.List;

import org.rascalmpl.value.IConstructor;
import org.rascalmpl.value.ISourceLocation;
import org.rascalmpl.value.IValue;
import org.rascalmpl.value.IValueFactory;
import org.rascalmpl.value.io.BinaryValueReader;
import org.rascalmpl.value.type.TypeStore;
import org.eclipse.jdt.annotation.Nullable;
import org.nuthatchery.pica.errors.ParserLoadError;
import org.nuthatchery.pica.errors.ParserNotFoundError;
import org.nuthatchery.pica.errors.Severity;
import org.nuthatchery.pica.rascal.ICompiler;
import org.nuthatchery.pica.rascal.IEvaluatorFactory;
import org.nuthatchery.pica.rascal.IEvaluatorPool;
import org.nuthatchery.pica.resources.IWorkspaceManager;
import org.rascalmpl.parser.gtd.IGTD;
import org.rascalmpl.values.uptr.ITree;

/**
 * This interface abstracts away from underlying platform/infrastructure
 * differences.
 * It is a mixed bag of all sorts of platform dependent functionality.
 * (This is necessary as Java does not support #ifdefs, which we might use
 * instead in C code. We also cannot use runtime checks, as we cannot expect all
 * of the imported classes to be available.)
 *
 */
public interface IInfra {

	/**
	 * Clear stored parser and associated files. Will force a regeneration of
	 * parser on next call to getParser().
	 *
	 * This is needed if the Rascal runtime has changed and the generated parser
	 * is no longer compatible.
	 *
	 * @param moduleName
	 *            Rascal module name for the grammar
	 */
	public void clearParser(String moduleName);


	/**
	 * Return a parser for the given grammar. The generated parser should
	 * be somewhere in the class path.
	 *
	 *
	 * @param moduleName
	 *            Rascal module name for the grammar
	 * @return A new parser instance
	 * @throws ParserLoadError
	 *             if the parser could not be loaded
	 * @throws ParserNotFoundError
	 *             if the parser could not be found
	 */
	public IGTD<IConstructor, ITree, ISourceLocation> getParser(String moduleName) throws ParserLoadError, ParserNotFoundError;


	/**
	 * Log an exception
	 *
	 * @param msg
	 *            A log message
	 * @param t
	 *            The exception
	 */
	public void logException(@Nullable String msg, @Nullable Throwable t);


	/**
	 * Write a message to the log
	 *
	 * @param msg
	 *            A log message
	 */
	public void logMessage(String msg);


	/**
	 * Write a message to the log
	 *
	 * @param msg
	 *            A log message
	 * @param severity
	 *            The severity
	 */
	public void logMessage(String msg, Severity severity);


	/**
	 * @return true if ModuleResource facts are preloaded, and false if produced
	 *         on demand
	 */
	boolean areModuleFactsPreloaded();


	/**
	 * @return A Magnolia compiler
	 * @deprecated Use Magnolia.getInstance().getCompiler()
	 */
	@Deprecated
	ICompiler getCompiler();


	/**
	 * Get a data file from the platform metadata area (the workspace .metadata
	 * directory in Eclipse, or a .-directory in the user's home directory on a
	 * stand-alone system).
	 *
	 * The metadata area can be used to store data that should be persistent
	 * across sessions, which is not associated with a particular project or
	 * file.
	 *
	 *
	 * @param fileName
	 *            The relative filename
	 * @return A File handle
	 */
	@Nullable
	File getDataFile(String fileName);


	/**
	 * The newEvaluator() method is hidden away behind the IEvaluatorFactory
	 * method to encourage the use of evaluator pools instead.
	 *
	 * @return An evaluator factory
	 */
	IEvaluatorFactory getEvaluatorFactory();


	/**
	 * Get an evaluator pool, with the given list of modules imported.
	 *
	 * This may return a previously generated pool. If you really need a new
	 * pool, use getEvaluatorFactory().makeEvaluatorPool().
	 *
	 * @param name
	 *            Name of the pool, e.g. "Magnolia loader"
	 * @param imports
	 *            A list of Rascal modules to import
	 * @return The new pool
	 */
	IEvaluatorPool getEvaluatorPool(String name, List<String> imports);


	/**
	 * Get the workspace manager.
	 *
	 * The workspace manager provides an entry point to all projects and files
	 * in the workspace.
	 *
	 * @return The workspace manager
	 */
	IWorkspaceManager getWorkspaceManager();


	/**
	 * Load data from the internal plugin metadata storage. 'fileName' should be
	 * a relative path.
	 *
	 * @see {@link #getDataFile(String)}
	 * @param fileName
	 *            Relative path of the of a PDB binary file (PBF format)
	 * @param vf
	 *            A value factory for producing PDB values
	 * @param ts
	 *            The type store, using in constructing values
	 * @return The value stored in the file
	 * @see {@link BinaryValueReader#readValueFromFile}
	 */
	IValue loadData(String fileName, IValueFactory vf, TypeStore ts) throws IOException;


	/**
	 * Open a file in the from the plugin bundle.
	 *
	 * This is for files / data bundled with the installation, see
	 * {@link #getDataFile(String)} for
	 * accessing/storing data across sessions.
	 *
	 * @param path
	 *            The relative file path (with no src/ or bin/ prefix)
	 * @return An input stream
	 * @throws IOException
	 */
	InputStream openStream(String path) throws IOException;


	/**
	 * Open a file in the from the plugin bundle and return it as a string.
	 *
	 * This is for files / data bundled with the installation, see
	 * {@link #getDataFile(String)} for
	 * accessing/storing data across sessions.
	 *
	 * @param path
	 *            The relative file path (with no src/ or bin/ prefix)
	 * @return A string containing the entire file
	 * @throws IOException
	 */
	String readFile(String path) throws IOException;


	/**
	 * Reload code and clear cached data
	 */
	void refresh();


	/**
	 * Remove file from the internal plugin storage. 'fileName' should be a
	 * relative path.
	 *
	 * @return True if the file was successfully deleted
	 */
	boolean removeDataFile(String fileName);


	/**
	 * Save data to the internal plugin storage. 'fileName' should be a relative
	 * path.
	 */
	void saveData(String fileName, IValue value, TypeStore ts) throws IOException;
}
