/**************************************************************************
 * Copyright (c) 2012-2014 Anya Helene Bagge
 * Copyright (c) 2012-2014 Tero Hasu
 * Copyright (c) 2012-2014 University of Bergen
 *
 * All rights reserved. This program and the accompanying materials
 * are made available under the terms of the Eclipse Public License v1.0
 * which accompanies this distribution, and is available at
 * http://www.eclipse.org/legal/epl-v10.html
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version. See http://www.gnu.org/licenses/
 *
 *
 * See the file COPYRIGHT for more information.
 *
 * Contributors:
 * * Anya Helene Bagge
 * * Tero Hasu
 *
 *************************************************************************/
package org.magnolialang.infra;

import java.io.File;
import java.io.IOException;
import java.util.List;

import org.rascalmpl.value.IValue;
import org.rascalmpl.value.IValueFactory;
import org.rascalmpl.value.io.BinaryValueReader;
import org.rascalmpl.value.type.TypeStore;
import org.eclipse.jdt.annotation.Nullable;
import org.nuthatchery.pica.errors.ImplementationError;
import org.nuthatchery.pica.rascal.IEvaluatorFactory;
import org.nuthatchery.pica.resources.IProjectManager;
import org.nuthatchery.pica.resources.IWorkspaceManager;
import org.rascalmpl.uri.URIResolverRegistry;

public final class Infra {
	@Nullable
	private static IInfra platform;
	public static final URIResolverRegistry URI_RESOLVER_REGISTRY = URIResolverRegistry.getInstance();


	@Nullable
	public static String findFileInPath(final String path, final List<String> searchPath) {
		String file = null;
		for(final String dir : searchPath) {
			file = dir + "/" + path;
			if(new File(file).canRead()) {
				break;
			}
		}
		return file;
	}


	public static IInfra get() {
		IInfra p = platform;
		if(p == null) {
			throw new ImplementationError("Magnolia not initialised");
		}

		return p;
	}


	/**
	 * Get a data file from the platform metadata area (the workspace .metadata
	 * directory in Eclipse, or a .-directory in the user's home directory on a
	 * stand-alone system).
	 *
	 * The metadata area can be used to store data that should be persistent
	 * across sessions, which is not associated with a particular project or
	 * file.
	 *
	 *
	 * @param fileName
	 *            The relative filename
	 * @return A File handle
	 * @see {@link IInfra#getDataFile()}
	 */
	@Nullable
	public static File getDataFile(String fileName) {
		return get().getDataFile(fileName);
	}


	/**
	 * The newEvaluator() method is hidden away behind the IEvaluatorFactory
	 * method to encourage the use of evaluator pools instead.
	 *
	 * @return An evaluator factory
	 * @see {@link IInfra#getEvaluatorFactory()}
	 */
	public static IEvaluatorFactory getEvaluatorFactory() {
		return get().getEvaluatorFactory();
	}


	public static URIResolverRegistry getResolverRegistry() {
		return URI_RESOLVER_REGISTRY;
	}


	/**
	 * Get the manager for a project
	 *
	 * @param project
	 *            The project name (in Eclipse) or directory name
	 * @return A project resource manager
	 * @see {@link IWorkspaceManager#getManager(String)}
	 */
	public static IProjectManager getResourceManager(String project) {
		return get().getWorkspaceManager().getManager(project);
	}


	/**
	 * Get the workspace manager.
	 *
	 * The workspace manager provides an entry point to all projects and files
	 * in the workspace.
	 *
	 * @return The workspace manager
	 * @see {@link IInfra#getWorkspaceManager()}
	 */
	public static IWorkspaceManager getWorkspaceManager() {
		return get().getWorkspaceManager();
	}


	/**
	 * Load data from the internal plugin metadata storage. 'fileName' should be
	 * a relative path.
	 *
	 * @see {@link #getDataFile(String)}
	 * @param fileName
	 *            Relative path of the of a PDB binary file (PBF format)
	 * @param vf
	 *            A value factory for producing PDB values
	 * @param ts
	 *            The type store, using in constructing values
	 * @return The value stored in the file
	 * @see {@link IInfra#loadData(String, IValueFactory, TypeStore)}
	 *      {@link BinaryValueReader#readValueFromFile}
	 */
	public static IValue loadData(String fileName, IValueFactory vf, TypeStore ts) throws IOException {
		return get().loadData(fileName, vf, ts);
	}


	public static void removeDataFile(String fileName) throws IOException {
		get().removeDataFile(fileName);
	}


	public static void saveData(String fileName, IValue value, TypeStore ts) throws IOException {
		get().saveData(fileName, value, ts);
	}


	public static void set(IInfra platform) {
		Infra.platform = platform;
	}
}
