/**************************************************************************
 * Copyright (c) 2010-2013 Anya Helene Bagge
 * Copyright (c) 2010-2013 University of Bergen
 *
 * All rights reserved. This program and the accompanying materials
 * are made available under the terms of the Eclipse Public License v1.0
 * which accompanies this distribution, and is available at
 * http://www.eclipse.org/legal/epl-v10.html
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version. See http://www.gnu.org/licenses/
 *
 *
 * See the file COPYRIGHT for more information.
 *
 * Contributors:
 * * Anya Helene Bagge
 *
 *************************************************************************/
package org.magnolialang.load;

import java.io.IOException;
import java.net.URI;
import java.net.URISyntaxException;
import java.util.ArrayList;
import java.util.Collection;
import java.util.HashMap;
import java.util.List;
import java.util.Map;

import org.eclipse.core.runtime.IPath;
import org.eclipse.core.runtime.Path;
import org.rascalmpl.value.ISourceLocation;
import io.usethesource.impulse.utils.Pair;
import org.magnolialang.infra.Infra;
import org.nuthatchery.pica.terms.TermFactory;

public class MagnoliaBuildPath implements IBuildPath {
	private final List<URI> srcPath;
	private final Map<URI, URI> pathMap;
	private URI mainBinPath;
	private final org.rascalmpl.uri.URIResolverRegistry resolverRegistry;


	public MagnoliaBuildPath(final URI mainBinPath) {
		this.mainBinPath = mainBinPath;
		srcPath = new ArrayList<URI>();
		pathMap = new HashMap<URI, URI>();
		resolverRegistry = Infra.getResolverRegistry();
	}


	@Override
	public void addPath(URI src) {
		src = addSlash(src);
		if(!srcPath.contains(src)) {
			srcPath.add(src);
		}
	}


	@Override
	public void addPath(URI src, URI bin) {
		src = addSlash(src);
		bin = addSlash(bin);
		if(!srcPath.contains(src)) {
			srcPath.add(src);
		}
		if(bin != null) {
			pathMap.put(src, bin);
		}
	}


	@Override
	public URI find(final String moduleName, final Collection<String> extensions) {
		final String[] components = moduleName.split("\\.");
		for(URI path : srcPath) {
			URI binPath = getBinPath(path);
			for(int i = 0; i < components.length; i++) {
				for(String extension : extensions) {
					final URI srcURI = path.resolve(components[i] + "." + extension);
					final URI binURI = binPath.resolve(components[i] + "." + extension + "c");
					final ISourceLocation srcLoc = TermFactory.vf.sourceLocation(srcURI);
					final ISourceLocation binLoc = TermFactory.vf.sourceLocation(binURI);
					long srcMod = 0;
					long binMod = 0;
					try {
						srcMod = resolverRegistry.lastModified(srcLoc);
						binMod = resolverRegistry.lastModified(binLoc);
					}
					catch(IOException e1) {// NOPMD by anya on 1/5/12 5:24 AM
					}

					if(srcMod > 0 || binMod > 0 || resolverRegistry.exists(srcLoc) || resolverRegistry.exists(binLoc)) {
						final StringBuilder fragment = new StringBuilder(moduleName.length());
						for(int j = i + 1; j < components.length; j++) {
							fragment.append(components[j]);
							if(j < components.length - 1) {
								fragment.append('.');
							}
						}

						try {
							if(binMod > srcMod) {
								return fragment.length() == 0 ? binURI : binURI.resolve(new URI(null, null, fragment.toString()));
							}
							else if(srcMod > 0 || resolverRegistry.exists(srcLoc)) {
								return fragment.length() == 0 ? srcURI : srcURI.resolve(new URI(null, null, fragment.toString()));
							}
							else if(binMod > 0 || resolverRegistry.exists(binLoc)) {
								return fragment.length() == 0 ? binURI : binURI.resolve(new URI(null, null, fragment.toString()));
							}
						}
						catch(final URISyntaxException e) {
							e.printStackTrace();
						}
					}
				}
				path = path.resolve(components[i] + "/");
				binPath = binPath.resolve(components[i] + "/");

			}
		}
		return null;
	}


	@Override
	public URI getBinPath(final URI src) {
		if(pathMap.containsKey(src)) {
			return pathMap.get(src);
		}
		else if(mainBinPath == null) {
			return src;
		}
		else {
			return mainBinPath;
		}
	}


	@Override
	public URI getMainBinPath() {
		return mainBinPath;
	}


	@Override
	public Pair<URI, String> getModuleName(URI uri) {
		uri = uri.normalize();
		for(URI path : srcPath) {
			path = path.normalize();
			if(path.getScheme().equals(uri.getScheme()) && (path.getHost() == null && uri.getHost() == null || path.getHost().equalsIgnoreCase(uri.getHost())) && (path.getAuthority() == null && uri.getAuthority() == null || path.getAuthority().equalsIgnoreCase(uri.getAuthority()))) {
				final IPath p1 = new Path(path.getPath());
				IPath p2 = new Path(uri.getPath());
				if(p1.isPrefixOf(p2)) {
					p2 = p2.makeRelativeTo(p1);
					final String moduleName = p2.removeFileExtension().toString().replace('/', '.');
					return new Pair<URI, String>(path, moduleName);
				}
			}
		}
		return new Pair<URI, String>(null, null);
	}


	@Override
	public IBuildPath newPath() {
		return new MagnoliaBuildPath(mainBinPath);
	}


	@Override
	public IBuildPath newPath(final URI mainBinPath) {
		return new MagnoliaBuildPath(mainBinPath);
	}


	@Override
	public void setMainBinPath(final URI bin) {
		mainBinPath = bin;
	}


	private URI addSlash(final URI uri) {
		if(uri.getPath().endsWith("/")) {
			return uri;
		}
		else {
			try {
				return new URI(uri.getScheme(), uri.getSchemeSpecificPart() + "/", uri.getFragment());
			}
			catch(final URISyntaxException e) {
				return uri;
			}
		}
	}

}
