/**************************************************************************
 * Copyright (c) 2010-2014 Anya Helene Bagge
 * Copyright (c) 2010-2014 Tero Hasu
 * Copyright (c) 2010-2014 University of Bergen
 *
 * All rights reserved. This program and the accompanying materials
 * are made available under the terms of the Eclipse Public License v1.0
 * which accompanies this distribution, and is available at
 * http://www.eclipse.org/legal/epl-v10.html
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version. See http://www.gnu.org/licenses/
 *
 *
 * See the file COPYRIGHT for more information.
 *
 * Contributors:
 * * Anya Helene Bagge
 * * Tero Hasu
 *
 *************************************************************************/
package org.magnolialang.magnolia;

import static org.nuthatchery.pica.terms.TermAdapter.*;

import java.util.Arrays;
import java.util.Collection;

import org.rascalmpl.value.IConstructor;
import org.rascalmpl.value.IListWriter;
import org.rascalmpl.value.IValue;
import org.eclipse.jdt.annotation.NonNullByDefault;
import org.eclipse.jdt.annotation.Nullable;
import org.magnolialang.load.ModuleParser;
import org.nuthatchery.pica.rascal.ICompiler;
import org.nuthatchery.pica.resources.ILanguage;
import org.nuthatchery.pica.terms.TermAdapter;
import org.nuthatchery.pica.terms.TermFactory;

import edu.umd.cs.findbugs.annotations.SuppressFBWarnings;

@NonNullByDefault
public class Magnolia implements ILanguage {
	public static final String MAGNOLIA = "Magnolia";
	public static final String MAGNOLIA_EXTENSION = "mg";
	public static final Collection<String> MAGNOLIA_EXTENSIONS = Arrays.asList(MAGNOLIA_EXTENSION);
	private static final Magnolia instance = new Magnolia();

	private final MagnoliaCompiler compiler = new MagnoliaCompiler();


	@Override
	public boolean equals(@Nullable Object other) {
		return this == other;
	}


	public synchronized ICompiler getCompiler() {
		return compiler;
	}


	@Override
	public Collection<String> getExtensions() {
		return MAGNOLIA_EXTENSIONS;
	}


	@Override
	public String getFileName(String modName) {
		return modName.replace(".", "/") + "." + MAGNOLIA_EXTENSION;
	}


	@Override
	public String getId() {
		return MAGNOLIA;
	}


	@Override
	public String getModuleName(String fileName) {
		return fileName.substring(0, fileName.lastIndexOf('.')).replaceFirst("^/", "").replace("/", ".");
	}


	@Override
	public String getName() {
		return MAGNOLIA;
	}


	@Override
	public IConstructor getNameAST(String name) {
		String[] split = name.split("\\.");
		if(split.length == 1) {
			return TermFactory.cons("Name", TermFactory.leaf(split[0]));
		}
		else {
			IListWriter list = TermFactory.vf.listWriter(TermFactory.Type_AST);
			for(String s : split) {
				list.append(TermFactory.cons("Name", TermFactory.leaf(s)));
			}
			return TermFactory.cons("QName", TermFactory.seq(list.done()));
		}
	}


	@Override
	public String getNameString(Object nameAST) {
		return yieldName((IValue) nameAST);
	}


	public ModuleParser getParser() {
		return ModuleParser.getParser(this, "org::magnolialang::magnolia::grammar::Magnolia");
	}


	@Override
	public String getPreferredExtension() {
		return MAGNOLIA_EXTENSION;
	}


	@Override
	public String getStoreExtension() {
		return "mgf";
	}


	@Override
	public boolean hasExtension(String ext) {
		if(ext != null && ext.startsWith(".")) {
			return hasExtension(ext.substring(1));
		}
		else {
			return MAGNOLIA_EXTENSION.equals(ext);
		}
	}


	@Override
	public int hashCode() {
		return System.identityHashCode(this);
	}


	public static IConstructor basename(IConstructor name) {
		return (IConstructor) getInstance().getCompiler().call("basename", name);
	}


	public static synchronized Magnolia getInstance() {
		return instance;
	}


	@SuppressFBWarnings("BC_UNCONFIRMED_CAST")
	public static boolean isSymbol(IValue symbol) {
		if(isCons(symbol, "Name", 1)) {
			return true;
		}
		else if(isCons(symbol, "QName", 2)) {
			return isSymbol(TermAdapter.getArg((IConstructor) symbol, 0));
		}
		else {
			return false;
		}
	}


	public static IConstructor qualify(IConstructor modName, IConstructor name) {
		return (IConstructor) getInstance().getCompiler().call("qualify", modName, name);
	}


	public static IConstructor unqualify(IConstructor name) {
		return (IConstructor) getInstance().getCompiler().call("unqualify", name);
	}


	public static IConstructor qualifierOf(IConstructor name) {
		return (IConstructor) getInstance().getCompiler().call("qualifierOf", name);
	}


	@SuppressFBWarnings("BC_UNCONFIRMED_CAST")
	public static String yieldName(IValue symbol) {
		if(isCons(symbol, "Name", 1)) {
			return yield(symbol);
		}
		else if(isCons(symbol, "QName", 1)) {
			StringBuilder str = new StringBuilder(256);
			String dot = "";
			for(IConstructor n : TermAdapter.getChildren(TermAdapter.getArg((IConstructor) symbol, 0))) {
				str.append(dot);
				str.append(yieldName(n));
				dot = ".";
			}
			return str.toString();
		}
		else if(isCons(symbol, "VarName", 2)) {
			return yieldName(getArg((IConstructor) symbol, 0)) + ":" + yieldName(getArg((IConstructor) symbol, 1));
		}
		else if(isCons(symbol, "TopName", 2)) {
			return yieldName(getArg((IConstructor) symbol, 0));
		}
		else {
			return yield(symbol);
			// throw new ImplementationError("Not a name: "
			// + yieldTerm(symbol, true));
		}
	}
}
