/**************************************************************************
 * Copyright (c) 2010-2014 Anya Helene Bagge
 * Copyright (c) 2010-2014 University of Bergen
 *
 * All rights reserved. This program and the accompanying materials
 * are made available under the terms of the Eclipse Public License v1.0
 * which accompanies this distribution, and is available at
 * http://www.eclipse.org/legal/epl-v10.html
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version. See http://www.gnu.org/licenses/
 *
 *
 * See the file COPYRIGHT for more information.
 *
 * Contributors:
 * * Anya Helene Bagge
 *
 *************************************************************************/
package org.magnolialang.magnolia;

import java.util.Arrays;
import java.util.HashMap;
import java.util.List;
import java.util.Map;

import org.rascalmpl.value.IValue;
import org.eclipse.jdt.annotation.NonNullByDefault;
import org.eclipse.jdt.annotation.Nullable;
import org.magnolialang.infra.Infra;
import org.nuthatchery.pica.rascal.ICompiler;
import org.nuthatchery.pica.rascal.IEvaluatorFactory;
import org.nuthatchery.pica.rascal.IEvaluatorPool;
import org.nuthatchery.pica.resources.IWorkspaceManager.ThreadingPreference;
import org.nuthatchery.pica.tasks.ITaskMonitor;
import org.nuthatchery.pica.tasks.NullTaskMonitor;
import org.nuthatchery.pica.tasks.rascal.RascalTaskMonitor;
import org.rascalmpl.debug.IRascalMonitor;
import org.rascalmpl.interpreter.NullRascalMonitor;

/**
 * Interface to the Magnolia compiler.
 *
 * @author anya
 *
 */
@NonNullByDefault
public class MagnoliaCompiler implements ICompiler {
	public static final String TEMPLATE_MAIN_ID = "org.magnolialang.magnolia.templates.MagicMain";

	public static final String TEMPLATE_MAIN = "org/magnolialang/magnolia/templates/MagicMain.cxx";
	@SuppressWarnings("null")
	private final List<String> xrefImports = Arrays.asList("org::magnolialang::magnolia::services::XRefs");
	@SuppressWarnings("null")
	private final List<String> inlinerImports = Arrays.asList("org::magnolialang::magnolia::lib::Inliner", "org::magnolialang::magnolia::lib::Renaming");
	@SuppressWarnings("null")
	private final List<String> backendImports = Arrays.asList("org::magnolialang::magnolia::backend::Plainifier", "org::magnolialang::magnolia::backend::Mutifier", "org::magnolialang::magnolia::backend::CxxBackend", "org::magnolialang::magnolia::backend::CxxSkeleton", "org::magnolialang::magnolia::lib::CollectAxioms", "org::magnolialang::magnolia::analysis::Optimiser");
	@SuppressWarnings("null")
	private final List<String> frontendImports = Arrays.asList("org::nuthatchery::pica::terms::Terms", "org::magnolialang::magnolia::typechecker::Flattener", "org::magnolialang::magnolia::typechecker::TypeChecker", "org::magnolialang::magnolia::lib::Environment", "org::magnolialang::magnolia::lib::Names", "org::magnolialang::magnolia::lib::MagnoliaUtil", "org::magnolialang::magnolia::lib::MagnoliaCompileLib");
	@SuppressWarnings("null")
	private final List<String> backendCommands = Arrays.asList("getPlainified", "getMutified", "cxxCode", "generateCxxSkeleton", "collectAxioms", "getOptimised");
	@SuppressWarnings("null")
	private final List<String> inlinerCommands = Arrays.asList("simpleRename", "inlineByPoint", "treeEditDiff");
	@Nullable
	private IEvaluatorPool xrefs;
	@Nullable
	private IEvaluatorPool backend;
	@Nullable
	private IEvaluatorPool frontend;
	@Nullable
	private IEvaluatorPool inliner;
	private final Map<String, IEvaluatorPool> cmdSwitch = new HashMap<String, IEvaluatorPool>();


	MagnoliaCompiler() {
	}


	@Override
	public IValue call(ITaskMonitor tm, String funName, IValue... args) {
		IEvaluatorPool pool;
		synchronized(this) {
			IEvaluatorPool fe = frontend;
			if(fe == null) {
				throw new IllegalStateException("Not initialized");
			}

			pool = cmdSwitch.get(funName);
			if(pool == null && funName.endsWith("_impl")) {
				pool = cmdSwitch.get(funName.substring(0, funName.lastIndexOf("_")));
			}

			if(pool == null) {
				pool = fe;
			}
		}

		return pool.call(tm, funName, args);
	}


	@Override
	public IValue call(String funName, IValue... args) {
		return call(new NullTaskMonitor(), funName, args);
	}


	@Override
	public synchronized void ensureInit() {
		//loader.ensureInit();

		if(xrefs != null) {
			xrefs.ensureInit();
		}
		if(backend != null) {
			backend.ensureInit();
		}
		if(frontend != null) {
			frontend.ensureInit();
		}
		//	if(inliner != null) {
		//		inliner.ensureInit();
		//	}

	}


	@Override
	public String getLanguage() {
		return Magnolia.MAGNOLIA;
	}


	@Override
	// @SuppressWarnings("null")
	public synchronized void refresh() {
		IEvaluatorFactory factory = Infra.get().getEvaluatorFactory();
//		IEvaluatorPool pool1 = factory.getEvaluatorPool("Installation Check (Rascal standard library)", Arrays.asList("IO"), 1);
//		IEvaluatorPool pool2 = factory.getEvaluatorPool("Installation Check (Magnolia compiler)", Arrays.asList("org::magnolialang::magnolia::Errors"), 1);
//		IEvaluatorPool pool3 = factory.getEvaluatorPool("Installation Check (Pica library)", Arrays.asList("org::nuthatchery::pica::terms::Terms"), 1);
//		pool1.ensureInit();
//		pool2.ensureInit();
//		pool3.ensureInit();
		System.err.println("Compile: Loading Backend");
		int nEvals = numEvaluators(1.25f);
		if(backend != null) {
			backend = factory.refresh(backend, nEvals);
		}
		else {
			backend = factory.getEvaluatorPool("Magnolia backend", backendImports, nEvals);
		}
		System.err.println("Compile: Loading Frontend");
		if(frontend != null) {
			frontend = factory.refresh(frontend, nEvals);
		}
		else {
			frontend = factory.getEvaluatorPool("Magnolia frontend", frontendImports, nEvals);
		}

		nEvals = numEvaluators(.1f);
		System.err.println("Compile: Loading Xrefs");
		if(xrefs != null) {
			xrefs = factory.refresh(xrefs, nEvals);
		}
		else {
			xrefs = factory.getEvaluatorPool("Magnolia xref formatter", xrefImports, nEvals);
		}

//		System.err.println("Compile: Loading Inliner");
//		if(inliner != null) {
//			inliner = factory.refresh(inliner, nEvals);
//		}
//		else {
//			inliner = factory.getEvaluatorPool("Magnolia inliner", inlinerImports, nEvals);
//		}

		System.err.println("Compile: done");
		cmdSwitch.clear();

		for(String s : backendCommands) {
			cmdSwitch.put(s, backend);
		}
//		for(String s : inlinerCommands) {
//			cmdSwitch.put(s, inliner);
//		}
		cmdSwitch.put("formatXRef", xrefs);
		//loader.reload();
	}


	private int numEvaluators(float n) {
		ThreadingPreference pref = Infra.getWorkspaceManager().getThreadingPreference();
		int nprocs = Runtime.getRuntime().availableProcessors();
		float factor;
		if(nprocs < 2) {
			factor = 0.5f;
		}
		else if(nprocs <= 4) {
			factor = 1.0f;
		}
		else {
			factor = 1.5f;
		}
		float nevals = n * factor * pref.threadFactor;

		System.err.printf("numEvaluators: %f ~= %d%n", nevals, Math.max(1, Math.round(nevals)));
		return Math.max(1, Math.round(nevals));

	}
}
