/**************************************************************************
 * Copyright (c) 2014 Anya Helene Bagge
 * Copyright (c) 2014 University of Bergen
 *
 * All rights reserved. This program and the accompanying materials
 * are made available under the terms of the Eclipse Public License v1.0
 * which accompanies this distribution, and is available at
 * http://www.eclipse.org/legal/epl-v10.html
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version. See http://www.gnu.org/licenses/
 *
 *
 * See the file COPYRIGHT for more information.
 *
 * Contributors:
 * * Anya Helene Bagge
 *
 *************************************************************************/
package org.magnolialang.magnolia.evaluator;

import static nuthatch.rascal.pattern.StaticValuesPatternFactory.var;
import static org.magnolialang.magnolia.grammar.MagnoliaPatterns.VarDecl;

import java.io.InputStreamReader;
import java.io.PrintWriter;
import java.io.Reader;
import java.lang.reflect.InvocationTargetException;
import java.lang.reflect.Method;
import java.net.URLClassLoader;
import java.util.HashMap;
import java.util.Map;
import java.util.Set;

import nuthatch.pattern.Environment;
import nuthatch.pattern.EnvironmentFactory;
import nuthatch.rascal.adapter.ValuesCursor;

import org.rascalmpl.value.IConstructor;
import org.rascalmpl.value.IListWriter;
import org.rascalmpl.value.IMap;
import org.rascalmpl.value.ISourceLocation;
import org.rascalmpl.value.IValue;
import org.eclipse.jdt.annotation.Nullable;
import org.magnolialang.magnolia.Magnolia;
import org.magnolialang.magnolia.evaluator.callables.BuiltinFunction;
import org.magnolialang.magnolia.evaluator.callables.Callable;
import org.magnolialang.magnolia.evaluator.callables.ExternalFunction;
import org.magnolialang.magnolia.evaluator.callables.ExternalProcedure;
import org.magnolialang.magnolia.evaluator.callables.Function;
import org.magnolialang.magnolia.evaluator.callables.Procedure;
import org.magnolialang.magnolia.evaluator.controls.GuardFailed;
import org.magnolialang.magnolia.evaluator.controls.Returned;
import org.magnolialang.magnolia.evaluator.errors.AssertFailed;
import org.magnolialang.magnolia.evaluator.errors.EvaluationError;
import org.magnolialang.magnolia.evaluator.types.BuiltinType;
import org.magnolialang.magnolia.evaluator.types.ExternalModule;
import org.magnolialang.magnolia.evaluator.types.ExternalType;
import org.magnolialang.magnolia.evaluator.types.OpaqueType;
import org.magnolialang.magnolia.evaluator.types.StructType;
import org.magnolialang.magnolia.evaluator.types.Type;
import org.magnolialang.magnolia.evaluator.values.BooleanValue;
import org.magnolialang.magnolia.evaluator.values.ICopyable;
import org.magnolialang.magnolia.evaluator.values.IValueBuilder;
import org.magnolialang.magnolia.evaluator.values.StructValue;
import org.magnolialang.magnolia.evaluator.values.Value;
import org.magnolialang.magnolia.evaluator.values.VarValue;
import org.magnolialang.util.StackTraceUtil;
import org.nuthatchery.pica.terms.TermAdapter;
import org.nuthatchery.pica.terms.TermFactory;
import org.nuthatchery.pica.util.NullnessHelper;
import org.nuthatchery.pica.util.depgraph.IMultiMap;
import org.nuthatchery.pica.util.depgraph.MultiHashMap;

/**
 * Evaluator for Magnolia programs.
 *
 * <p>
 *
 * The evaluator should be initialised with a fully composed Magnolia module
 * (i.e., a program or a satisfaction). Missing, incomplete or erroneous code
 * will cause EvaluationExceptions at runtime.
 *
 *
 */
public class AlgoEvaluator {

	private final IMap toplevelEnv;
	private final IMultiMap<String, Callable> nameEnv = new MultiHashMap<>();
	private final PatternDispatcher dispatcher = new PatternDispatcher();
	private final ClassLoader classLoader;
	private final Map<String, ExternalModule> externals = new HashMap<>();
	private final PrintWriter stdout;

	private final PrintWriter stderr;

	private final Reader stdin;
	private String name;
	private final EvaluationContext context;


	/**
	 * Construct an evaluator for a Magnolia module.
	 *
	 * @param declList
	 *            A module, in the form of a DeclBody list of declarations
	 * @param classLoader
	 *            For loading external Java code (the evaluator's class loader
	 *            is chosen if null)
	 * @param out
	 *            Output stream (System.out if null)
	 * @param err
	 *            Error stream (System.err if null)
	 * @param inputReader
	 *            Input stream (System.in if null)
	 */

	public AlgoEvaluator(String name, IConstructor declList, @Nullable URLClassLoader classLoader, @Nullable PrintWriter out, @Nullable PrintWriter err, @Nullable Reader inputReader) {
		this.name = name;
		this.context = new EvaluationContext();
		this.classLoader = classLoader != null ? classLoader : NullnessHelper.assertNonNull(getClass().getClassLoader());
		this.stdout = out != null ? out : new PrintWriter(System.out);
		this.stderr = err != null ? err : new PrintWriter(System.err);
		this.stdin = inputReader != null ? inputReader : new InputStreamReader(System.in);

		if(!TermAdapter.isCons(declList, "DeclList", 1) && !TermAdapter.isCons(declList, "DeclBody", 1)) {
			throw new EvaluationError("Not a declaration list");
		}
		IMap env = TermFactory.vf.mapWriter().done();
		assert env != null;
		for(IConstructor decl : TermAdapter.getChildren(TermAdapter.getArg(declList, 0))) {
			assert decl != null;
			if(TermAdapter.isCons(decl, "Define", 3)) {
				IConstructor clause = TermAdapter.getArg(decl, 0);
				Value value;

				if(TermAdapter.isCons(clause, "TypeClause", 1)) {
					IConstructor body = TermAdapter.getArg(decl, 2);
					if(TermAdapter.isCons(body, "Body", 1)) {
						body = TermAdapter.getArg(body, 0);
					}
					if(TermAdapter.isCons(body, "Struct", 1)) {
						value = new StructType(decl);
					}
					else if(TermAdapter.isCons(body, "ExternalType", 4)) {
						value = new ExternalType(getExternalModule(TermAdapter.getArg(body, 0), TermAdapter.getArg(body, 1), TermAdapter.getArg(body, 2)), decl);
					}
					else {
						value = new OpaqueType(decl);
					}

				}
				else if(TermAdapter.isCons(clause, "FunClause", 3)) {
					value = new Function(decl, env);
				}
				else if(TermAdapter.isCons(clause, "ProcClause", 2)) {
					value = new Procedure(decl, env);
				}
				else {
					throw new EvaluationError("Unknown declaration kind " + TermAdapter.getName(clause));
				}

				if(value instanceof Callable) {
					String n = Magnolia.yieldName(Magnolia.basename(Magnolia.unqualify(value.getId())));
					nameEnv.put(n, (Callable) value);
				}
				env = env.put(value.getId(), value);
				assert env != null;
			}
		}
		toplevelEnv = env;
	}


	/**
	 * Apply a function.
	 *
	 * The first function with the given name that matches the argument list
	 * will be called. Currently types are not considered in overload
	 * resolution, only the number of arguments.
	 *
	 * @param name
	 *            Name of the function
	 * @param args
	 *            Argument list
	 * @return Return value
	 */
	public Object apply(String name, Object... args) {
		Set<Callable> cands = nameEnv.get(name);
		if(cands != null) {
			for(Callable c : cands) {
				if(c.isFunction() && c.numParams() == args.length) {
					IListWriter argList = TermFactory.vf.listWriter();
					for(int i = 0; i < args.length; i++) {
						IConstructor mode = c.getParamMode(i);
						if(isOut(mode) && args[i] instanceof VarValue) {
							argList.append((VarValue<?>) args[i]);
						}
						else {
							argList.append(new VarValue<Object>(args[i], mode));
						}
					}
					EvaluationContext ctx = context.enterOperation(c);
					return c.call(ctx, toplevelEnv, TermFactory.seq(NullnessHelper.assertNonNull(argList.done())));
				}
			}
		}
		throw new EvaluationError("InternalOperation not found: " + name);
	}


	/**
	 * Call a procedure.
	 *
	 * The first procedure with the given name that matches the argument list
	 * will be called. Currently types are not considered in overload
	 * resolution, only the number of arguments.
	 *
	 * Arguments in upd or out positions should be VarValues; a temporary
	 * variable will be supplied if necessary.
	 *
	 * @param name
	 *            Name of the procedure
	 * @param args
	 *            Argument list
	 * @see {@link VarValue}
	 */
	public void call(String name, Object... args) {
		Set<Callable> cands = nameEnv.get(name);
		for(Callable c : cands) {
			if(c.isProcedure() && c.numParams() == args.length) {
				IListWriter argList = TermFactory.vf.listWriter();
				for(int i = 0; i < args.length; i++) {
					IConstructor mode = c.getParamMode(i);
					if(isOut(mode) && args[i] instanceof VarValue) {
						argList.append((VarValue<?>) args[i]);
					}
					else {
						argList.append(new VarValue<Object>(args[i], mode));
					}
				}
				EvaluationContext ctx = context.enterOperation(c);
				c.call(ctx, toplevelEnv, TermFactory.seq(NullnessHelper.assertNonNull(argList.done())));
				return;
			}
		}
		throw new EvaluationError("InternalOperation not found: " + name);
	}


	/**
	 * Evaluate an expression.
	 *
	 * @param expr
	 *            The expression.
	 * @param env
	 *            The current environment.
	 * @param ctx
	 *            The current evaluation context.
	 * @return The expression's value
	 */
	public Object evalExpr(IConstructor expr, IMap env, IEvaluationContext ctx) {
		return dispatch("evalExpr", expr, env, ctx);
	}


	/**
	 * Evaluate a function body.
	 *
	 * @param expr
	 *            The expression.
	 * @param env
	 *            The current environment.
	 * @param ctx
	 *            The current evaluation context.
	 * @return The expression's value
	 */
	public Object evalExprDefBody(IConstructor body, IMap env, IEvaluationContext ctx) {
		return dispatch("evalExprDefBody", body, env, ctx);
	}


	/**
	 * Evaluate a function expression.
	 *
	 * @param fun
	 *            The function expression.
	 * @param env
	 *            The current environment.
	 * @param ctx
	 *            The current evaluation context.
	 * @return A callable, suitable for application
	 */
	public Callable evalFun(IConstructor fun, IMap env, EvaluationContext ctx) {
		return (Callable) dispatch("evalFun", fun, env, ctx);
	}


	/**
	 * Evaluate a literal expression.
	 *
	 * @param literal
	 *            The expression.
	 * @param env
	 *            The current environment.
	 * @param ctx
	 *            The current evaluation context.
	 * @return The expression's value
	 */
	public Object evalIntegerLiteral(IConstructor literal, IMap env, EvaluationContext ctx) {
		return dispatch("evalIntegerLiteral", literal, env, ctx);
	}


	/**
	 * Evaluate a procedure expression.
	 *
	 * @param proc
	 *            The procedure expression.
	 * @param env
	 *            The current environment.
	 * @param ctx
	 *            The current evaluation context.
	 * @return A callable, suitable for calling
	 */
	public Callable evalProc(IConstructor proc, IMap env, EvaluationContext ctx) {
		return (Callable) dispatch("evalProc", proc, env, ctx);
	}


	/**
	 * Evaluate a statement.
	 *
	 * @param stat
	 *            The statement.
	 * @param env
	 *            The current environment.
	 * @param ctx
	 *            The current evaluation context.
	 * @return {@link Value#UNDEFINED}
	 */
	public Object evalStat(IConstructor stat, IMap env, EvaluationContext ctx) {
		return dispatch("evalStat", stat, env, ctx);
	}


	/**
	 * Evaluate statement block.
	 *
	 * @param stat
	 *            The statement block.
	 * @param env
	 *            The current environment.
	 * @param ctx
	 *            The current evaluation context.
	 * @return {@link Value#UNDEFINED}
	 */
	public Object evalStatDefBody(IConstructor body, IMap env, IEvaluationContext ctx) {
		return dispatch("evalStatDefBody", body, env, ctx);
	}


	/**
	 * Evaluate a variable identifier, yielding the variable itself.
	 *
	 * @param arg
	 *            Name of the variable.
	 * @param env
	 *            The current environment.
	 * @param ctx
	 *            The current evaluation context.
	 * @return The variable
	 * @throws EvaluationError
	 *             if the identifier is undefined or does not refer to a
	 *             variable
	 */
	public VarValue<?> evalVar(IConstructor arg, IMap env, IEvaluationContext ctx) {
		if(TermAdapter.isCons(arg, "Var", 1)) { // Var(VarName(name, type))
			Value value = (Value) env.get(TermAdapter.getArg(arg, 0));
			if(value instanceof VarValue) {
				return (VarValue<?>) value;
			}
			else {
				throw new EvaluationError("Undefined variable: " + arg);
			}
		}
		else {
			throw new EvaluationError("Expected a variable: " + arg);
		}

	}


	/**
	 * Find an external Java module.
	 *
	 * @param langId
	 *            Language identifier.
	 * @param moduleId
	 *            Module identifier, corresponding to the Java class name.
	 * @param onPart
	 *            Parameter list
	 * @return The module; freshly created if necessary.
	 * @throws EvaluationError
	 *             if Java class loading fails.
	 */
	public ExternalModule getExternalModule(IConstructor langId, IConstructor moduleId, IConstructor onPart) {
		String className = Magnolia.yieldName(moduleId);
		ExternalModule m = externals.get(className);

		if(m == null) {
			m = new ExternalModule(classLoader, className, onPart);
			externals.put(className, m);
		}
		return m;
	}


	/**
	 * @return Name of this evaluator's Magnolia module.
	 */
	public String getName() {
		return name;
	}


	/**
	 * @return The standard error stream
	 */
	public PrintWriter getStderr() {
		return stderr;
	}


	/**
	 * @return The standard input stream
	 */
	public Reader getStdin() {
		return stdin;
	}


	/**
	 * @return The standard output stream
	 */
	public PrintWriter getStdout() {
		return stdout;
	}


	/**
	 * @return This module's top-level environment, containing all operation and
	 *         type declarations.
	 */
	public IMap getTopEnv() {
		return toplevelEnv;
	}


	/**
	 * Do a pattern dispatch on a constructor.
	 *
	 * @param baseName
	 *            Name prefix for the method to call (e.g. EvalExpr)
	 * @param cons
	 *            The constructor.
	 * @param env
	 *            The current environment.
	 * @param ctx
	 *            The currrent evaluation context.
	 * @return The result of calling
	 *         PatternDispatcher.&lt;baseName&gt;_&lt;consName(cons)&gt;(&lt;
	 *         consArgs(cons)&gt;)
	 */
	private Object dispatch(String baseName, IConstructor cons, IMap env, IEvaluationContext ctx) {
		String consName = TermAdapter.getName(cons);

		String method = baseName + "_" + consName;

		Class<?>[] paramTypes = new Class<?>[TermAdapter.arity(cons) + 2];
		Object[] paramValues = new Object[TermAdapter.arity(cons) + 2];

		int i = 0;
		for(IValue arg : TermAdapter.getChildren(cons)) {
			paramTypes[i] = IConstructor.class;
			paramValues[i] = arg;
			i++;
		}

		ctx = ctx.enterLocation(TermAdapter.locOf(cons));

		paramTypes[i] = IMap.class;
		paramValues[i++] = env;
		paramTypes[i] = EvaluationContext.class;
		paramValues[i++] = ctx;
		try {
			Method meth = PatternDispatcher.class.getMethod(method, paramTypes);
			System.err.println("Calling " + meth);
			// System.err.println("  With: " + Arrays.toString(paramValues));

			Object result = meth.invoke(dispatcher, paramValues);
			if(result == null) {
				result = Value.UNDEFINED;
			}
			return result;
		}
		catch(ClassCastException | NoSuchMethodException | SecurityException | IllegalAccessException | IllegalArgumentException e) {
			// e.printStackTrace();
			throw new EvaluationError("Failed to find eval method", e);
		}
		catch(InvocationTargetException e) {
			Throwable targetException = e.getTargetException();
			if(targetException instanceof Returned) {
				throw (Returned) targetException;
			}
			ISourceLocation loc = ctx.getCurrentLocation();

			String path = loc.getPath();
			int line = -1;
			if(loc.hasLineColumn()) {
				line = loc.getBeginLine();
			}
			targetException = StackTraceUtil.patch(targetException, name, ctx.getCurrentOperation().getName(), path, line);

			if(targetException instanceof RuntimeException) {
				throw (RuntimeException) targetException;
			}
			else {
				throw new EvaluationError("Exception caught while evaluating: ", e);
			}
		}
	}


	/**
	 * @param mode
	 *            A parameter mode
	 * @return True if mode represents an inflow mode (obs or upd)
	 */
	public static boolean isIn(IConstructor mode) {
		String name = TermAdapter.getName(mode);
		switch(name) {
		case "Obs":
			return true;
		case "Out":
			return false;
		case "Upd":
			return true;
		default:
			throw new EvaluationError("Unknown parameter mode: " + name);
		}
	}


	/**
	 * @param mode
	 *            A parameter mode
	 * @return True if mode represents an outflow mode (out or upd)
	 */
	public static boolean isOut(IConstructor mode) {
		String name = TermAdapter.getName(mode);
		switch(name) {
		case "Obs":
			return false;
		case "Out":
			return true;
		case "Upd":
			return true;
		default:
			throw new EvaluationError("Unknown parameter mode: " + name);
		}
	}


	/**
	 * Look up a type name in an environment
	 *
	 * @param name
	 *            The name.
	 * @param env
	 *            The environment.
	 * @return The type, or {@link BuiltinType#UNKNOWN_TYPE}.
	 */
	public static Type lookupType(IConstructor name, IMap env) {
		IValue value = env.get(name);
		if(value instanceof Type) {
			return (Type) value;
		}
		else {
			return BuiltinType.UNKNOWN_TYPE;
		}
	}


	class EvaluationContext implements IEvaluationContext {
		@Nullable
		final ISourceLocation location;
		@Nullable
		final Callable operation;
		@Nullable
		final EvaluationContext next;


		EvaluationContext() {
			location = null;
			operation = null;
			next = null;
		}


		EvaluationContext(Callable op, EvaluationContext next) {
			this.location = next.location;
			this.operation = op;
			this.next = next;
		}


		EvaluationContext(ISourceLocation loc, EvaluationContext next) {
			this.location = loc;
			this.operation = next.operation;
			this.next = next;
		}


		@Override
		public EvaluationContext enterLocation(ISourceLocation loc) {
			return new EvaluationContext(loc, this);
		}


		@Override
		public EvaluationContext enterOperation(Callable op) {
			return new EvaluationContext(op, this);
		}


		@Override
		public ISourceLocation getCurrentLocation() {
			ISourceLocation loc = location;
			if(loc == null) {
				throw new EvaluationError("Bottom of stack");
			}
			else {
				return loc;
			}
		}


		@SuppressWarnings("null")
		@Override
		public String getCurrentLocationString() {
			ISourceLocation loc = location;
			if(loc == null) {
				throw new EvaluationError("Bottom of stack");
			}
			String s = loc.getURI().toString();
			if(loc.hasLineColumn()) {
				s = s + ":" + loc.getBeginColumn();
			}
			return s;
		}


		@Override
		public Callable getCurrentOperation() {
			Callable op = operation;
			if(op == null) {
				throw new EvaluationError("Bottom of stack");
			}
			return op;
		}


		@Override
		public AlgoEvaluator getEvaluator() {
			return AlgoEvaluator.this;
		}


		@Override
		public String getModuleName() {
			return AlgoEvaluator.this.getName();
		}


		@Override
		public PrintWriter getStderr() {
			return AlgoEvaluator.this.getStderr();
		}


		@Override
		public Reader getStdin() {
			return AlgoEvaluator.this.getStdin();
		}


		@Override
		public PrintWriter getStdout() {
			return AlgoEvaluator.this.getStdout();
		}
	}


	@SuppressWarnings("unused")
	class PatternDispatcher {
		public Object evalExpr_Apply(IConstructor fun, IConstructor argList, IMap env, EvaluationContext context) {
			Callable funValue = evalFun(fun, env, context);

			context = context.enterOperation(funValue);
			return funValue.call(context, env, argList);
		}


		public Object evalExpr_Block(IConstructor statList, IMap env, EvaluationContext context) {
			try {
				evalStatList(statList, env, context);
			}
			catch(Returned e) {
				return e.getValue();
			}
			throw new EvaluationError("No return value provided");
		}


		public Object evalExpr_BlockExpr(IConstructor statList, IConstructor retType, IMap env, EvaluationContext context) {
			try {
				evalStatList(statList, env, context);
			}
			catch(Returned e) {
				return e.getValue();
			}
			throw new EvaluationError("No return value provided");
		}


		public Object evalExpr_DotOp(IConstructor expr, IConstructor fieldName, IConstructor type, IMap env, EvaluationContext context) {
			Object value = evalExpr(expr, env, context);
			if(value instanceof StructValue) {
				return ((StructValue) value).get(fieldName);
			}
			else {
				throw new EvaluationError("Value has no field " + Magnolia.yieldName(fieldName) + ": " + value);
			}
		}


		public Object evalExpr_Equal(IConstructor lhs, IConstructor op, IConstructor rhs, IMap env, EvaluationContext context) {
			Object lhsValue = evalExpr(lhs, env, context);
			Object rhsValue = evalExpr(rhs, env, context);
			String opName = TermAdapter.getString(op);

			boolean equals;
			if(lhsValue instanceof IValue && rhsValue instanceof IValue) {
				equals = ((IValue) lhsValue).isEqual((IValue) rhsValue);
			}
			else {
				equals = lhsValue.equals(rhsValue);
			}
			if(opName.equals("==")) {
				return BooleanValue.create(equals);
			}
			else {
				return BooleanValue.create(!equals);
			}
		}


		public Object evalExpr_Expect(IConstructor arg0, IConstructor arg1, IMap env, EvaluationContext context) {
			return evalExpr(arg0, env, context);
		}


		public Object evalExpr_Guard(IConstructor arg0, IConstructor arg1, IMap env, EvaluationContext context) {
			throw new UnsupportedOperationException();
		}


		public Object evalExpr_IfThenElseExpr(IConstructor cond, IConstructor thenPart, IConstructor elsePart, IMap env, EvaluationContext context) {
			if(Value.valueOf(evalExpr(cond, env, context)).isTrue()) {
				return evalExpr(thenPart, env, context);
			}
			else {
				return evalExpr(elsePart, env, context);
			}
		}


		public Object evalExpr_ListCons(IConstructor arg0, IMap env, EvaluationContext context) {
			throw new UnsupportedOperationException();
		}


		public Object evalExpr_Literal(IConstructor arg0, IConstructor arg1, IMap env, EvaluationContext context) {
			throw new UnsupportedOperationException();
		}


		public Object evalExpr_Literal(IConstructor arg0, IMap env, EvaluationContext context) {
			throw new UnsupportedOperationException();
		}


		public Object evalExpr_Recover(IConstructor arg0, IConstructor arg1, IMap env, EvaluationContext context) {
			try {
				return evalExpr(arg0, env, context);
			}
			catch(GuardFailed e) {
				return evalExpr(arg1, env, context);
			}
		}


		public Object evalExpr_Struct(IConstructor type, IConstructor inits, IMap env, EvaluationContext context) {
			Value value = (Value) env.get(type);
			if(value instanceof StructType) {
				StructType typeValue = (StructType) value;
				IValueBuilder builder = typeValue.getBuilder();
				for(IConstructor init : TermAdapter.getChildren(inits)) {
					assert init != null;
					IConstructor name = TermAdapter.getArg(init, 0);
					IConstructor expr = TermAdapter.getArg(init, 1);
					builder.put(name, evalExpr(expr, env, context));
				}
				return builder.getValue();
			}

			throw new EvaluationError("Trying to construct non-struct");
		}


		public Object evalExpr_Tuple(IConstructor arg0, IMap env, EvaluationContext context) {
			throw new UnsupportedOperationException();
		}


		public Object evalExpr_TypeExpr(IConstructor arg0, IMap env, EvaluationContext context) {
			throw new UnsupportedOperationException();
		}


		public Object evalExpr_Undefined(IMap env, EvaluationContext context) {
			return Value.UNDEFINED;
		}


		public Object evalExpr_Var(IConstructor name, IMap env, EvaluationContext context) {
			Value varValue = (Value) env.get(name);
			if(varValue instanceof VarValue) {
				Object val = ((VarValue<?>) varValue).get();
				if(val != null) {
					return val;
				}
				else {
					return Value.UNDEFINED;
				}
			}
			else {
				throw new EvaluationError("Undefined variable " + Magnolia.yieldName(name));
			}
		}


		public Object evalExprDefBody_Body(IConstructor expr, IMap env, EvaluationContext context) {
			return evalExpr(expr, env, context);
		}


		public Object evalExprDefBody_EmptyBody(IMap env, EvaluationContext context) {
			return Value.UNDEFINED;
		}


		public Object evalFun_Builtin(IConstructor qName, IConstructor idString, IMap env, EvaluationContext context) {
			return BuiltinFunction.valueOf(idString);
		}


		public Object evalFun_ExternalFun(IConstructor langId, IConstructor moduleId, IConstructor onPart, IConstructor funName, IConstructor paramList, IConstructor retType, IMap env, EvaluationContext context) {
			return new ExternalFunction(getExternalModule(langId, moduleId, onPart), funName, paramList, retType, env);
		}


		public Object evalFun_Fun(IConstructor name, IMap env, EvaluationContext context) {
			IValue val = env.get(name);
			if(val != null) {
				return val;
			}
			else {
				throw new EvaluationError("Undefined operation: " + Magnolia.yieldName(name));
			}
		}


		public Object evalFun_FunOf(IConstructor arg0, IConstructor arg1, IMap env, EvaluationContext context) {
			throw new UnsupportedOperationException();
		}


		public Object evalIntegerLiteral_DecimalIntegerLiteral(IConstructor decimal, IMap env, EvaluationContext context) {
			throw new UnsupportedOperationException();
		}


		public Object evalIntegerLiteral_HexIntegerLiteral(IConstructor hex, IMap env, EvaluationContext context) {
			throw new UnsupportedOperationException();
		}


		public Object evalIntegerLiteral_OctalIntegerLiteral(IConstructor octal, IMap env, EvaluationContext context) {
			throw new UnsupportedOperationException();
		}


		public Object evalProc_ExternalProc(IConstructor langId, IConstructor moduleId, IConstructor onPart, IConstructor procName, IConstructor paramList, IMap env, EvaluationContext context) {
			return new ExternalProcedure(getExternalModule(langId, moduleId, onPart), procName, paramList, env);
		}


		public Object evalProc_Proc(IConstructor name, IMap env, EvaluationContext context) {
			IValue val = env.get(name);
			if(val != null) {
				return val;
			}
			else {
				throw new EvaluationError("Undefined operation: " + Magnolia.yieldName(name));
			}
		}


		public Object evalProc_ProcOf(IConstructor arg0, IMap env, EvaluationContext context) {
			throw new UnsupportedOperationException();
		}


		public Object evalStat_Assert(IConstructor expr, IConstructor clauses, IMap env, EvaluationContext context) {
			Value value = Value.valueOf(evalExpr(expr, env, context));
			if(value.isFalse()) {
				throw new AssertFailed();
			}
			return Value.UNDEFINED;
		}


		@SuppressWarnings("unchecked")
		public Object evalStat_Assign(IConstructor name, IConstructor expr, IMap env, EvaluationContext context) {
			Object result = evalExpr(expr, env, context);
			if(result instanceof VarValue) {
				result = ((VarValue<?>) result).get();
				if(result == null) {
					result = Value.UNDEFINED;
				}
				else if(result instanceof ICopyable) {
					result = ((ICopyable<?>) result).copy();
				}
			}

			if(TermAdapter.isCons(name, "Var", 1)) {// name is Var(VarName(...,...))
				IValue var = env.get(TermAdapter.getArg(name, 0));
				if(var instanceof VarValue) {
					((VarValue<Object>) var).set(result);
				}
				else {
					throw new EvaluationError("Variable not bound: " + TermAdapter.yieldTerm(name, false));
				}
			}
			else if(TermAdapter.isCons(name, "DotOp", 3)) { // DotOp(expr, name, type)
				Object value = evalExpr(TermAdapter.getArg(name, 0), env, context);
				IConstructor fieldName = TermAdapter.getArg(name, 1);
				if(value instanceof StructValue) {
					((StructValue) value).put(fieldName, result);
				}
				else {
					throw new EvaluationError("Value has no field " + Magnolia.yieldName(fieldName) + ": " + value);
				}

			}

			return Value.UNDEFINED;
		}


		public Object evalStat_Assume(IConstructor expr, IConstructor clauses, IMap env, EvaluationContext context) {
			Value value = Value.valueOf(evalExpr(expr, env, context));
			if(value.isFalse()) {
				throw new AssertFailed();
			}
			return Value.UNDEFINED;
		}


		public Object evalStat_Block(IConstructor statList, IMap env, EvaluationContext context) {
			return evalStatList(statList, env, context);
		}


		public Object evalStat_Break(IMap env, EvaluationContext context) {
			throw new UnsupportedOperationException();
		}


		public Object evalStat_Call(IConstructor proc, IConstructor argList, IMap env, EvaluationContext context) {
			Callable procValue = evalProc(proc, env, context);

			context = context.enterOperation(procValue);
			return procValue.call(context, env, argList);
		}


		public Object evalStat_Expect(IConstructor stat, IConstructor error, IMap env, EvaluationContext context) {
			return evalStat(stat, env, context);
		}


		public Object evalStat_For(IConstructor arg0, IConstructor arg1, IConstructor arg2, IMap env, EvaluationContext context) {
			throw new UnsupportedOperationException();
		}


		public Object evalStat_Guard(IConstructor stat, IConstructor guard, IMap env, EvaluationContext context) {
			Value guardVal = Value.valueOf(evalExpr(guard, env, context));
			if(guardVal.isTrue()) {
				return evalStatList(stat, env, context);
			}
			else {
				throw new GuardFailed();
			}
		}


		public Object evalStat_If(IConstructor cond, IConstructor thenPart, IConstructor elsePart, IMap env, EvaluationContext context) {
			Value condVal = Value.valueOf(evalExpr(cond, env, context));
			if(condVal.isTrue()) {
				return evalStatList(thenPart, env, context);
			}
			else {
				return evalStatList(thenPart, env, context);
			}

		}


		@SuppressWarnings("null")
		public Object evalStat_Let(IConstructor decl, IConstructor body, IMap env, EvaluationContext context) {
			Environment<ValuesCursor> matchEnv = EnvironmentFactory.env();
			if(VarDecl(var("mode"), var("name"), var("type"), var("expr")).match(new ValuesCursor(decl), matchEnv)) {
				VarValue<Object> varValue = new VarValue<>();
				IConstructor varName = TermFactory.cons("VarName", matchEnv.get("name").getData(), matchEnv.get("type").getData());
				env = env.put(varName, varValue);
				varValue.set(evalExpr((IConstructor) matchEnv.get("expr").getData(), env, context));
				varValue.setMode(matchEnv.get("mode").getData());

				evalStatList(body, env, context);
			}
			else {
				throw new EvaluationError("Not a variable declaration " + decl);
			}
			return Value.UNDEFINED;
		}


		public Object evalStat_Nop(IMap env, EvaluationContext context) {
			return Value.UNDEFINED;
		}


		public Object evalStat_Return(IConstructor expr, IMap env, EvaluationContext context) {
			throw new Returned(evalExpr(expr, env, context));
		}


		public Object evalStat_Return(IMap env, EvaluationContext context) {
			throw new Returned();
		}


		public Object evalStat_While(IConstructor arg0, IConstructor arg1, IMap env, EvaluationContext context) {
			throw new UnsupportedOperationException();
		}


		public Object evalStat_Yield(IConstructor arg0, IMap env, EvaluationContext context) {
			throw new UnsupportedOperationException();
		}


		public Object evalStatDefBody_Body(IConstructor body, IMap env, EvaluationContext context) {
			return evalStat(body, env, context);
		}


		public Object evalStatDefBody_EmptyBody(IMap env, EvaluationContext context) {
			return Value.UNDEFINED;
		}


		public Object evalStatList(IConstructor statList, IMap env, EvaluationContext context) {
			for(IConstructor s : TermAdapter.getChildren(statList)) {
				assert s != null;
				evalStat(s, env, context);
			}
			return Value.UNDEFINED;
		}


		public Object evalStringLiteral_StringLiteral(IConstructor string, IMap env, EvaluationContext context) {
			throw new UnsupportedOperationException();
		}

	}
}
