/**************************************************************************
 * Copyright (c) 2014 Anya Helene Bagge
 * Copyright (c) 2014 University of Bergen
 * 
 * All rights reserved. This program and the accompanying materials
 * are made available under the terms of the Eclipse Public License v1.0
 * which accompanies this distribution, and is available at
 * http://www.eclipse.org/legal/epl-v10.html
 * 
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version. See http://www.gnu.org/licenses/
 * 
 * 
 * See the file COPYRIGHT for more information.
 * 
 * Contributors:
 * * Anya Helene Bagge
 * 
 *************************************************************************/
package org.magnolialang.magnolia.evaluator.callables;

import org.rascalmpl.value.IConstructor;
import org.rascalmpl.value.IMap;
import org.magnolialang.magnolia.evaluator.IEvaluationContext;
import org.magnolialang.magnolia.evaluator.errors.EvaluationError;
import org.magnolialang.magnolia.evaluator.errors.WrongNumberOfArguments;
import org.magnolialang.magnolia.evaluator.types.BuiltinType;
import org.magnolialang.magnolia.evaluator.types.Type;
import org.magnolialang.magnolia.evaluator.values.BooleanValue;
import org.magnolialang.magnolia.evaluator.values.Value;
import org.nuthatchery.pica.terms.TermAdapter;
import org.nuthatchery.pica.terms.TermFactory;

public abstract class BuiltinFunction extends Callable {
	protected static final IConstructor A = TermFactory.cons("Name", TermFactory.leaf("a"));
	protected static final IConstructor B = TermFactory.cons("Name", TermFactory.leaf("b"));
	protected final int numParams;

	protected static final IConstructor OBS = TermFactory.cons("Obs");


	protected BuiltinFunction(int numParams) {
		if(numParams < 0 || numParams > 2) {
			throw new IllegalArgumentException();
		}
		this.numParams = numParams;
	}


	@Override
	@SuppressWarnings("null")
	public Value call(IEvaluationContext ctx, IMap env, IConstructor argList) {
		if(numParams() != TermAdapter.arity(argList)) {
			throw new WrongNumberOfArguments(this, TermAdapter.arity(argList));
		}

		BooleanValue[] args = new BooleanValue[numParams];
		int i = 0;
		for(IConstructor arg : TermAdapter.getChildren(argList)) {
			Object val = ctx.getEvaluator().evalExpr(arg, env, ctx);
			if(val instanceof BooleanValue) {
				args[i] = (BooleanValue) val;
			}
			else {
				throw new EvaluationError("Expected predicate argument");
			}
		}

		if(numParams == 0) {
			return call0();
		}
		else if(numParams == 1) {
			return call1(args[0]);
		}
		else if(numParams == 2) {
			return call2(args[0], args[1]);
		}
		else {
			throw new EvaluationError("Internal error: wrong number of arguments");
		}
	}


	@Override
	public String getKind() {
		return "function";
	}


	@Override
	public IConstructor getParamMode(int i) {
		if(i < 0 || i >= numParams) {
			throw new IllegalArgumentException();
		}
		return OBS;
	}


	@Override
	public IConstructor getParamName(int i) {
		if(i < 0 || i >= numParams) {
			throw new IllegalArgumentException();
		}
		if(i == 0) {
			return A;
		}
		else {
			return B;
		}
	}


	@Override
	public Type getParamType(int i) {
		if(i < 0 || i >= numParams) {
			throw new IllegalArgumentException();
		}
		return BuiltinType.BOOLEAN_TYPE;
	}


	@Override
	public boolean isFunction() {
		return true;
	}


	@Override
	public int numParams() {
		return numParams;
	}


	/**
	 * @return Result of the operation
	 */
	protected Value call0() {
		throw new UnsupportedOperationException();
	}


	/**
	 * @param a
	 *            The first argument
	 * @return Result of the operation applied to a
	 */
	protected Value call1(BooleanValue a) {
		throw new UnsupportedOperationException();
	}


	/**
	 * @param a
	 *            The first argument
	 * @param b
	 *            The second argument
	 * @return Result of the operation applied to a and b
	 */
	protected Value call2(BooleanValue a, BooleanValue b) {
		throw new UnsupportedOperationException();
	}


	public static Value valueOf(IConstructor idString) {
		String op = TermAdapter.getString(idString);
		switch(op) {
		case "TRUE()":
			return new TrueValue();
		case "FALSE()":
			return new FalseValue();
		case "!_(\\predicate)":
			return new NotValue();
		case "_=>_(\\predicate,\\predicate)":
			return new ImpliesValue();
		case "_<=>_(\\predicate,\\predicate)":
			return new IffValue();
		case "_&&_(\\predicate,\\predicate)":
			return new AndValue();
		case "_||_(\\predicate,\\predicate)":
			return new OrValue();
		default:
			throw new EvaluationError("Unknown builtin " + op);
		} // TODO Auto-generated method stub
	}


	private static class AndValue extends BuiltinFunction {
		protected AndValue() {
			super(2);
		}


		@Override
		protected Value call2(BooleanValue a, BooleanValue b) {
			return a.and(b);
		}

	}


	private static class FalseValue extends BuiltinFunction {
		protected FalseValue() {
			super(0);
		}


		@Override
		protected Value call0() {
			return BooleanValue.FALSE;
		}
	}


	private static class IffValue extends BuiltinFunction {
		protected IffValue() {
			super(2);
		}


		@Override
		protected Value call2(BooleanValue a, BooleanValue b) {
			return a.iff(b);
		}
	}


	private static class ImpliesValue extends BuiltinFunction {
		protected ImpliesValue() {
			super(2);
		}


		@Override
		protected Value call2(BooleanValue a, BooleanValue b) {
			return a.implies(b);
		}
	}


	private static class NotValue extends BuiltinFunction {
		protected NotValue() {
			super(1);
		}


		@Override
		protected Value call1(BooleanValue a) {
			return a.not();
		}
	}


	private static class OrValue extends BuiltinFunction {
		protected OrValue() {
			super(2);
		}


		@Override
		protected Value call2(BooleanValue a, BooleanValue b) {
			return a.or(b);
		}
	}


	private static class TrueValue extends BuiltinFunction {
		protected TrueValue() {
			super(0);
		}


		@Override
		protected Value call0() {
			return BooleanValue.TRUE;
		}
	}

}
