/**************************************************************************
 * Copyright (c) 2014 Anya Helene Bagge
 * Copyright (c) 2014 University of Bergen
 * 
 * All rights reserved. This program and the accompanying materials
 * are made available under the terms of the Eclipse Public License v1.0
 * which accompanies this distribution, and is available at
 * http://www.eclipse.org/legal/epl-v10.html
 * 
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version. See http://www.gnu.org/licenses/
 * 
 * 
 * See the file COPYRIGHT for more information.
 * 
 * Contributors:
 * * Anya Helene Bagge
 * 
 *************************************************************************/
package org.magnolialang.magnolia.evaluator.types;

import java.lang.reflect.Constructor;
import java.lang.reflect.Field;
import java.lang.reflect.InvocationTargetException;
import java.lang.reflect.Method;
import java.util.Arrays;
import java.util.List;

import org.rascalmpl.value.IConstructor;
import org.magnolialang.magnolia.evaluator.AlgoEvaluator;
import org.magnolialang.magnolia.evaluator.IEvaluationContext;
import org.magnolialang.magnolia.evaluator.errors.EvaluationError;
import org.magnolialang.magnolia.evaluator.values.Value;
import org.magnolialang.magnolia.evaluator.values.VarValue;

public class ExternalModule extends Value {

	private final Object module;
	private Class<?> clazz;
	private String moduleName;


	@SuppressWarnings("null")
	public ExternalModule(ClassLoader classLoader, String className, IConstructor onPart) {
		try {
			this.moduleName = className;
			clazz = classLoader.loadClass(className);
			Constructor<?> constructor = clazz.getConstructor();
			module = constructor.newInstance();
		}
		catch(ClassNotFoundException | NoSuchMethodException | SecurityException | InstantiationException | IllegalAccessException | IllegalArgumentException | InvocationTargetException e) {
			e.printStackTrace();
			throw new EvaluationError("Unable to instantiate external module: " + e.getMessage(), e);
		}

	}


	@Override
	public String getKind() {
		return "module";
	}


	/**
	 * Find a method with the given name and parameter list.
	 * 
	 * The method is looked up in the external Java class according to the
	 * following scheme: First we try to find a method which takes an extra
	 * AlgoEvaluator last argument, if not found we attempt an exact match.
	 * 
	 * @param opName
	 *            Fully-qualified operation name
	 * @param paramTypes
	 *            Formal parameter list (wrapped in ParamList)
	 * @return A handle to the method, ready for calling
	 */
	@SuppressWarnings("null")
	public MethodHandle getMethod(String methodName, List<Type> paramTypes, List<IConstructor> paramModes) {

		Class<?>[] paramTypes1 = new Class[paramTypes.size()];
		Class<?>[] paramTypes2 = new Class[paramTypes.size() + 1];

		int i = 0;
		for(; i < paramTypes.size(); i++) {
			Class<?> clz;
			if(AlgoEvaluator.isOut(paramModes.get(i))) {
				clz = VarValue.class;
			}
			else {
				clz = paramTypes.get(i).getValueClass();
			}
			paramTypes1[i] = clz;
			paramTypes2[i] = clz;

		}

		paramTypes2[i] = IEvaluationContext.class;
		try {
			return new MethodHandle(module, clazz.getMethod(methodName, paramTypes2), true);
		}
		catch(NoSuchMethodException | SecurityException e) {
			try {
				return new MethodHandle(module, clazz.getMethod(methodName, paramTypes1), false);
			}
			catch(NoSuchMethodException | SecurityException e1) {
				throw new EvaluationError("Unable to find external operation '" + methodName + "' in module '" + moduleName + ": " + e.getMessage(), e);
			}

		}
	}


	public Class<?> getValueClass(String className) {
		try {
			Field field = clazz.getField(className);
			Object object = field.get(module);
			if(object instanceof Class) {
				return (Class<?>) object;
			}
			else {
				throw new EvaluationError("Finding external type: field '" + moduleName + "." + className + "' is not a class object");
			}
		}
		catch(NoSuchFieldException | SecurityException | IllegalArgumentException | IllegalAccessException e) {
			throw new EvaluationError("Unable to find external type '" + className + "' in module '" + moduleName + ": " + e.getMessage(), e);
		}
	}


	/**
	 * A MethodHandle keeps track of a method, the object it should be called
	 * on, and its calling conventions.
	 * 
	 */
	public static class MethodHandle {
		private final Method method;
		private final boolean needsEvalutorArg;
		private final Object module;


		public MethodHandle(Object module, Method method, boolean needsEvalutorArg) {
			this.module = module;
			this.method = method;
			this.needsEvalutorArg = needsEvalutorArg;
		}


		/**
		 * Call the method.
		 * 
		 * @param eval
		 *            The evaluator (to be passed to the method if it needs it)
		 * @param args
		 *            The argument list
		 * @return Return value of the method
		 * @throws IllegalAccessException
		 * @throws IllegalArgumentException
		 * @throws InvocationTargetException
		 */
		public Object invoke(IEvaluationContext ctx, Object... args) throws IllegalAccessException, IllegalArgumentException, InvocationTargetException {
			if(needsEvalutorArg) {
				args = Arrays.copyOf(args, args.length + 1);
				args[args.length - 1] = ctx;
			}
			return method.invoke(module, args);
		}
	}
}
