/**************************************************************************
 * Copyright (c) 2012-2014 Anya Helene Bagge
 * Copyright (c) 2012-2014 University of Bergen
 *
 * All rights reserved. This program and the accompanying materials
 * are made available under the terms of the Eclipse Public License v1.0
 * which accompanies this distribution, and is available at
 * http://www.eclipse.org/legal/epl-v10.html
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version. See http://www.gnu.org/licenses/
 *
 *
 * See the file COPYRIGHT for more information.
 *
 * Contributors:
 * * Anya Helene Bagge
 *
 *************************************************************************/
package org.magnolialang.magnolia.frontend;

import static nuthatch.rascal.pattern.StaticValuesPatternFactory.*;
import static org.magnolialang.magnolia.AbstractSyntax.*;
import static org.magnolialang.magnolia.grammar.MagnoliaPatterns.*;
import static org.nuthatchery.pica.terms.TermFactory.vf;

import java.util.ArrayList;
import java.util.HashMap;
import java.util.List;

import nuthatch.library.ActionFactory;
import nuthatch.library.FactoryFactory;
import nuthatch.library.MatchAction;
import nuthatch.library.MatchBuilder;
import nuthatch.library.Walk;
import nuthatch.pattern.Environment;
import nuthatch.pattern.EnvironmentFactory;
import nuthatch.pattern.Pattern;
import nuthatch.pattern.PatternBuildAction;
import nuthatch.rascal.adapter.ValuesCursor;
import nuthatch.rascal.adapter.ValuesWalker;
import nuthatch.rascal.pattern.ValuesBuildContext;
import nuthatch.tree.TreeCursor;

import org.rascalmpl.value.IConstructor;
import org.rascalmpl.value.ISet;
import org.rascalmpl.value.ISetWriter;
import org.rascalmpl.value.IValue;
import org.rascalmpl.value.type.Type;
import org.eclipse.jdt.annotation.NonNull;
import org.eclipse.jdt.annotation.NonNullByDefault;
import org.magnolialang.magnolia.Magnolia;
import org.magnolialang.magnolia.MagnoliaFacts;
import org.magnolialang.magnolia.resources.IMagnoliaPackage;
import org.magnolialang.magnolia.resources.internal.DefinitionData;
import org.nuthatchery.pica.errors.ImplementationError;
import org.nuthatchery.pica.errors.Severity;
import org.nuthatchery.pica.resources.IProjectManager;
import org.nuthatchery.pica.tasks.ITaskMonitor;
import org.nuthatchery.pica.terms.TermAdapter;
import org.nuthatchery.pica.terms.TermFactory;
import org.nuthatchery.pica.util.Pair;
import org.rascalmpl.debug.IRascalMonitor;


public class MagnoliaLoader {
	protected final static String MARK_SOURCE = MagnoliaLoader.class.getName();
	protected final static String MARK_SOURCE_GET_PKG_INFO = MARK_SOURCE + "#getPkgInfo";
	protected final static String MARK_SOURCE_GET_PKG_CONTENTS = MARK_SOURCE + "#getPkgContents";
	static long tOld = 0;
	static long tNew = 0;
	static int numTries = 0;
	static int numFailed = 0;
	private static final ValuesBuildContext context = new ValuesBuildContext(vf, TermFactory.ts);
	private static final Walk<ValuesWalker> desugarWalk = desugarWalk();
	private static boolean COMPARE_DESUGAR_WITH_RASCAL = false;


	public static IConstructor desugar(@NonNull ITaskMonitor tm, IConstructor tree) {
		long t = 0, oldTime = 0, newTime = 0;
		IConstructor oldTree = null;
/*		if(true) {
			t = System.currentTimeMillis();
			oldTree = (IConstructor) Magnolia.getInstance().getCompiler().call(rm, "desugarTree", tree);
			oldTime = System.currentTimeMillis() - t;
			tOld += oldTime;
			numTries++;
			System.err.printf("  Desugar Summary: %d ok. Old: %dms\n", MagnoliaLoader.numTries, MagnoliaLoader.tOld);
			return oldTree;
		}
 */
		if(COMPARE_DESUGAR_WITH_RASCAL) {
			t = System.currentTimeMillis();
			oldTree = (IConstructor) Magnolia.getInstance().getCompiler().call(tm, "desugarTree", tree);
			oldTime = System.currentTimeMillis() - t;
			MagnoliaLoader.tOld += oldTime;
		}

		t = System.currentTimeMillis();
		ValuesWalker walker = new ValuesWalker(tree, desugarWalk);
		walker.start();
		IConstructor result = (IConstructor) walker.getData();
		newTime = System.currentTimeMillis() - t;
		MagnoliaLoader.tNew += newTime;
		MagnoliaLoader.numTries++;

		if(oldTree != null) {
			System.err.println("Desugar: old: " + oldTime + "ms, new: " + newTime + "ms");
			if(result.isEqual(oldTree)) {
				if(result.toString().equals(oldTree.toString())) {

					System.err.println("*************** Desugared trees are equal!");
				}
				else {
					Pair<IValue, Pair<HashMap<String, IValue>, HashMap<String, IValue>>> diff = TermAdapter.oneDiffWithAnnos(result, oldTree);
					System.err.println("--------------- Desugared trees are NOT STRICTLY EQUAL!");
					if(diff != null) {
						System.err.println("  Diff: Term: " + diff.getFirst());
						Pair<HashMap<String, IValue>, HashMap<String, IValue>> maps = diff.getSecond();
						System.err.println("        A: " + maps.getFirst());
						System.err.println("        B: " + maps.getSecond());
					}
					MagnoliaLoader.numFailed++;

				}

			}
			else {
				Pair<IValue, IValue> diff = TermAdapter.oneDiff(result, oldTree);
				System.err.println("!!!!!!!!!!!!!!! Desugared trees are NOT EQUAL!");
				if(diff != null) {
					System.err.println("  Diff: A: " + TermAdapter.yieldTerm(diff.getFirst(), false));
					System.err.println("        B: " + TermAdapter.yieldTerm(diff.getSecond(), false));
				}
				MagnoliaLoader.numFailed++;
			}
			System.err.printf("  Desugar Summary: %d/%d ok. Old: %dms New: %dms%n", MagnoliaLoader.numTries - MagnoliaLoader.numFailed, MagnoliaLoader.numTries, MagnoliaLoader.tOld, MagnoliaLoader.tNew);
		}

		return result;
	}


	public static Pair<List<DefinitionData>, ISet> getPkgContents(IMagnoliaPackage pkg, IProjectManager manager, ITaskMonitor tm) {
		ISetWriter xrefsWriter = vf.setWriter();
		List<DefinitionData> contents = new ArrayList<DefinitionData>();
		IConstructor pkgInfo = pkg.getDefInfo(tm);

		manager.clearMarks(MARK_SOURCE_GET_PKG_CONTENTS, pkg.getURI());
		Walk<ValuesWalker> walk = makeDefInfoWalk(pkg, xrefsWriter, contents);
		for(IConstructor decl : TermAdapter.getChildren((IConstructor) pkgInfo.get("body"))) {
			ValuesWalker walker = new ValuesWalker(decl, walk);
			walker.start();

		}
		return new Pair<>(contents, xrefsWriter.done());
	}


	public static Pair<IConstructor, ISet> getPkgInfo(IMagnoliaPackage pkg, IProjectManager manager, ITaskMonitor tm) {
		IConstructor tree = pkg.getAST(tm);
		if(TermAdapter.isCons(tree, "MagnoliaTree")) {
			IConstructor head = TermAdapter.getArg(tree, 0);
			IConstructor decls = TermAdapter.getArg(tree, 1);
			ISetWriter xrefs = MagnoliaFacts.xrefsWriter();
			IConstructor moduleName;
			IConstructor clauses;
			manager.clearMarks(MARK_SOURCE_GET_PKG_INFO, pkg.getURI());
			if(TermAdapter.isCons(head, "PackageHead")) {
				moduleName = TermAdapter.getArg(head, 0);
				clauses = TermAdapter.getArg(head, 1);
			}
			else if(TermAdapter.isCons(head, "ModuleHead")) {
				moduleName = TermAdapter.getArg(head, 0);
				clauses = TermAdapter.getArg(head, 1);
				manager.addMark("'module' deprecated', use 'package' instead", TermAdapter.codeRegionOf(moduleName), Severity.WARNING, MARK_SOURCE_GET_PKG_INFO, pkg.getURI());
			}
			else {
				throw new ImplementationError("Not a valid Magnolia tree");
			}

			if(moduleName.asAnnotatable().hasAnnotation("loc")) {
				xrefs.insert(vf.tuple(moduleName.asAnnotatable().getAnnotation("loc"), vf.constructor(MagnoliaFacts.Cons_Ref, pkg.getId())));
			}
			if(TermAdapter.isCons(moduleName, "Unknown")) {
				moduleName = TermAdapter.preserveAnnos(pkg.getId(), moduleName);
			}
			else if(!pkg.getId().isEqual(moduleName)) {
				manager.addMark("Package name mismatch, should be '" + pkg.getName() + "' not '" + Magnolia.getInstance().getNameString(moduleName) + "'", TermAdapter.codeRegionOf(moduleName), Severity.WARNING, MARK_SOURCE_GET_PKG_INFO, pkg.getURI());
				moduleName = TermAdapter.preserveAnnos(pkg.getId(), moduleName);
			}

			ISetWriter imports = vf.setWriter();
			if(TermAdapter.isSeq(clauses)) {
				for(IConstructor clause : TermAdapter.getChildren(clauses)) {
					if(TermAdapter.isCons(clause, "Imports")) {
						for(IConstructor imp : TermAdapter.getChildren(TermAdapter.getArg(clause, 0))) {
							if(TermAdapter.isCons(imp, "ImportAll")) {
								imports.insert(TermAdapter.getArg(imp, 0));
							}
						}
					}
				}
			}

			IConstructor defInfo = MagnoliaFacts.defInfo(vf);
			defInfo = defInfo.set("kind", vf.string(MagnoliaFacts.KIND_PACKAGE));
			defInfo = defInfo.set("body", decls);
			defInfo = defInfo.set("depends", imports.done());
			defInfo = defInfo.set("qName", moduleName);
			defInfo = defInfo.set("decl", moduleName);
			defInfo = defInfo.set("defName", moduleName);
			defInfo = defInfo.set("locInfo", vf.constructor(MagnoliaFacts.Cons_LocInfo1, TermAdapter.getLocation(moduleName)));
			defInfo = defInfo.set("parent", TermFactory.cons("QName", TermFactory.seq()));
			return new Pair<IConstructor, ISet>(defInfo, xrefs.done());
		}
		else {
			throw new ImplementationError("Not a Magnolia tree");
		}
	}


	@NonNullByDefault(false)
	private static Pattern<IValue, Type> addAxiomReturn(Pattern<IValue, Type> pat) {
		return new PatternBuildAction<IValue, Type, ValuesCursor>(pat) {

			@Override
			public TreeCursor<IValue, Type> build(ValuesCursor tree) {
				Environment<ValuesCursor> env = EnvironmentFactory.env();
				if(var("body", Body(Block(seq(listvar("ss"))))).match(tree, env)) {
					return preserveAnnos(Body(Block(seq(listvar("ss"), Return(Apply(Fun(Name(leaf("TRUE"))), seq()))))), var("body")).build(context, env);
				}
				else {
					return tree;
				}
			}
		};
	}


	private static void addRewrite(MatchBuilder<IValue, Type, ValuesCursor, ValuesWalker> builder, Pattern<IValue, Type> pat, Pattern<IValue, Type> replace) {
		builder.add(pat, preserveTreeAnnos(replace));
	}


	@SuppressWarnings("null")
	private static Walk<ValuesWalker> desugarWalk() {
		ActionFactory<IValue, Type, ValuesCursor, ValuesWalker> af = FactoryFactory.getActionFactory(IValue.class, Type.class, ValuesCursor.class, ValuesWalker.class);

		MatchBuilder<IValue, Type, ValuesCursor, ValuesWalker> act = af.matchBuilder(context);

		Pattern<IValue, Type> ds = var("ds");
		Pattern<IValue, Type> a = var("a");
		Pattern<IValue, Type> op = var("op");
		Pattern<IValue, Type> b = var("b");
		Pattern<IValue, Type> n = var("n");
		Pattern<IValue, Type> c = var("c");
		Pattern<IValue, Type> t = var("t");
		Pattern<IValue, Type> e = var("e");
		Pattern<IValue, Type> ns = listvar("ns");
		Pattern<IValue, Type> ss = listvar("ss");
		Pattern<IValue, Type> as = listvar("as");
		Pattern<IValue, Type> bs = listvar("bs");
		Pattern<IValue, Type> cs = listvar("cs");
		addRewrite(act, Undefined(_), Undefined());
		addRewrite(act, MagnoliaTreeSugar(ds), MagnoliaTree(PackageHead(preserveTreeAnnos(Unknown()), seq()), ds));
		addRewrite(act, PackageHeadSugar(a), PackageHead(preserveTreeAnnos(Unknown()), a));
		addRewrite(act, TopAppSugar(op, a), TopApp(op, preserveAnnos(DeclList(seq(a)), a)));
		addRewrite(act, TopBinOpSugar(a, op, b), DeclList(seq(a, TopMap(preserveAnnos(Name(op), op), b))));
		addRewrite(act, TopBinOpAsUnarySugar(op, a), TopMap(preserveAnnos(Name(op), op), a));
		addRewrite(act, RequireSugar(a), TopMap(Modifier(Require()), preserveAnnos(SignatureOf(a), a)));
		addRewrite(act, ExtendSugar(a), TopMap(Modifier(Extend()), a));
		addRewrite(act, UseSugar(a), TopMap(Modifier(Use()), a));
		addRewrite(act, ProtectSugar(a), TopMap(Modifier(Protect()), preserveAnnos(SignatureOf(a), a)));
		addRewrite(act, QName(seq(n)), n);
		addRewrite(act, QNameSugar(seq(ns), n), QName(seq(ns, n)));
		addRewrite(act, IfSugar(c, t), If(c, t, seq()));
		addRewrite(act, VarDeclSugar(n, t, e), VarDecl(Upd(), n, t, e));
		addRewrite(act, VarDeclSugarNoType(n, e), VarDecl(Upd(), n, Unknown(), e));
		addRewrite(act, VarDeclSugarNoExpr(n, t), VarDecl(Upd(), n, t, Undefined()));
		addRewrite(act, ValDeclSugar(n, t, e), VarDecl(Obs(), n, t, e));
		addRewrite(act, ValDeclSugarNoType(n, e), VarDecl(Obs(), n, Unknown(), e));
		addRewrite(act, BlockExprSugar1(seq(ss)), BlockExpr(seq(ss), Unknown()));
		addRewrite(act, BlockExprSugar2(seq(ss), e), BlockExpr(seq(ss, Return(e)), Unknown()));
		addRewrite(act, DotOp(e, n), DotOp(e, n, Unknown()));
		addRewrite(act, Assign(IndexSugar(a, seq(bs)), e), Call(Proc(preserveTreeAnnos(Name(leaf("_[_]=_")))), seq(a, bs, e)));
		addRewrite(act, IndexSugar(a, seq(bs)), Apply(Fun(preserveTreeAnnos(Name(leaf("_[_]")))), seq(a, bs)));
		addRewrite(act, BinOpSugar(a, op, b), Apply(Fun(preserveAnnos(Name(makeBinOp(op)), op)), seq(a, b)));
		addRewrite(act, PreOpSugar(op, b), Apply(Fun(preserveAnnos(Name(makePreOp(op)), op)), seq(b)));
		addRewrite(act, InSugar(a, op, b), Apply(Fun(preserveAnnos(Name(leaf("_in_")), op)), seq(a, b)));
		addRewrite(act, NotInSugar(a, op, b), Apply(Fun(preserveAnnos(Name(leaf("_notin_")), op)), seq(a, b)));
		addRewrite(act, ParamSugar(t), Param(preserveAnnos(Obs(), t), Unknown(), t));
		addRewrite(act, ParamSugar(n, t), Param(preserveAnnos(Obs(), n), n, t));
		addRewrite(act, seqElement(as, LetSugar(a), bs), seq(as, Let(a, seq(bs))));

		Pattern<IValue, Type> body = var("body");
		Pattern<IValue, Type> mods = listvar("mods");
		Pattern<IValue, Type> attrs = listvar("attrs");
		addRewrite(act, DefineSugar(seq(mods), var("cl", PredClauseSugar(n, a)), seq(attrs), body), Define(preserveAnnos(FunClause(n, a, SpecialType(Predicate())), var("cl")), seq(Predicate(), mods, attrs), body));
		addRewrite(act, DefineSugar(seq(mods), var("cl", AxiomClauseSugar(n, a)), seq(attrs), body), Define(preserveAnnos(ProcClause(n, a), var("cl")), seq(Axiom(), mods, attrs), body));
		addRewrite(act, DefineSugar(seq(mods), var("cl", TestClauseSugar(n, a)), seq(attrs), body), Define(preserveAnnos(ProcClause(n, a), var("cl")), seq(Test(), mods, attrs), body));
		addRewrite(act, DefineSugar(seq(mods), var("cl", not(or(PredClauseSugar(_, _), AxiomClauseSugar(_, _), TestClauseSugar(_, _)))), seq(attrs), body), Define(var("cl"), seq(mods, attrs), body));
		addRewrite(act, TopDefineSugar(seq(mods), var("cl"), seq(attrs), body), TopDefine(var("cl"), seq(mods, attrs), body));
		addRewrite(act, AlertClauseSugar(n), AlertClause(n, seq()));
		addRewrite(act, AlertPreSugar(e), Alert(Name(preserveAnnos(leaf("PreconditionFailed"), e)), preserveTreeAnnos(AlertPreNegSugar(e))));
		addRewrite(act, AlertPreNegSugar(e), AlertPre(Apply(Fun(preserveTreeAnnos(Name(leaf("!_")))), seq(e))));
		addRewrite(act, AlertPostNegSugar(e), AlertPost(Apply(Fun(preserveTreeAnnos(Name(leaf("!_")))), seq(e))));
		addRewrite(act, Assert(e, seq(cs)), Assert(e, seq(cs, Src(unparse(e)))));
		addRewrite(act, Assume(e, seq(cs)), Assume(e, seq(cs, Src(unparse(e)))));
		addRewrite(act, Expect(a, e), Expect(unquote(a), e));

		return af.walk(af.down(af.match(var("_tree_"), act.oneEnv())));
	}


	@NonNullByDefault(false)
	private static Pattern<IValue, Type> makeBinOp(Pattern<IValue, Type> pat) {
		return new PatternBuildAction<IValue, Type, ValuesCursor>(pat) {

			@Override
			public TreeCursor<IValue, Type> build(ValuesCursor tree) {
				IValue data = tree.getData();
				if(TermAdapter.isLeaf(data)) {
					String string = TermAdapter.getString((IConstructor) data);
					return new ValuesCursor(TermFactory.leaf("_" + string + "_"));

				}
				else {
					throw new IllegalArgumentException();
				}
			}

		};
	}


	@NonNullByDefault(false)
	private static MatchAction<IValue, Type, ValuesCursor, ValuesWalker> makeDefInfo(final IMagnoliaPackage pkg, final List<DefinitionData> result, final ISetWriter xrefsWriter) {
		return new MatchAction<IValue, Type, ValuesCursor, ValuesWalker>() {

			@Override
			public void init(ValuesWalker walker) {

			}


			@Override
			public int step(ValuesWalker walker, Environment<ValuesCursor> env) {
				IConstructor def = (IConstructor) walker.getData();

				DefinitionData defData = new DefinitionData(def, pkg, DefinitionData.OPTION_QUALIFY);

				xrefsWriter.insert(vf.tuple(defData.getNameLoc(), vf.constructor(MagnoliaFacts.Cons_Ref, defData.getId())));
				result.add(defData);
				return PARENT;
			}
		};
	}


	private static Walk<ValuesWalker> makeDefInfoWalk(IMagnoliaPackage pkg, ISetWriter xrefsWriter, List<DefinitionData> contents) {
		ActionFactory<IValue, Type, ValuesCursor, ValuesWalker> af = FactoryFactory.getActionFactory(IValue.class, Type.class, ValuesCursor.class, ValuesWalker.class);
		MatchBuilder<IValue, Type, ValuesCursor, ValuesWalker> act = af.matchBuilder(context);

		act.add(TopDefine(var("clause"), var("attrs"), var("body")), makeDefInfo(pkg, contents, xrefsWriter));
		act.add(Define(var("clause"), var("attrs"), var("body")), makeDefInfo(pkg, contents, xrefsWriter));

		return af.walk(act.one());
	}


	@NonNullByDefault(false)
	private static Pattern<IValue, Type> makePreOp(Pattern<IValue, Type> pat) {
		return new PatternBuildAction<IValue, Type, ValuesCursor>(pat) {

			@Override
			public TreeCursor<IValue, Type> build(ValuesCursor tree) {
				IValue data = tree.getData();
				if(TermAdapter.isLeaf(data)) {
					String string = TermAdapter.getString((IConstructor) data);
					return new ValuesCursor(TermFactory.leaf(string + "_"));

				}
				else {
					throw new IllegalArgumentException();
				}
			}

		};
	}


	private static Pattern<IValue, Type> preserveTreeAnnos(Pattern<IValue, Type> pat) {
		return preserveAnnos(pat, var("_tree_"));
	}


	@NonNullByDefault(false)
	private static Pattern<IValue, Type> unparse(Pattern<IValue, Type> pat) {
		return new PatternBuildAction<IValue, Type, ValuesCursor>(pat) {

			@Override
			public TreeCursor<IValue, Type> build(ValuesCursor tree) {
				return new ValuesCursor(TermFactory.leaf(TermAdapter.yield(tree.getData())));
			}

		};
	}


	@NonNullByDefault(false)
	private static Pattern<IValue, Type> unquote(Pattern<IValue, Type> a) {
		return new PatternBuildAction<IValue, Type, ValuesCursor>(a) {

			@Override
			public TreeCursor<IValue, Type> build(ValuesCursor tree) {
				IValue data = tree.getData();
				if(TermAdapter.isLeaf(data)) {
					String string = TermAdapter.getString((IConstructor) data);
					string = string.replaceAll("^\"|\"$", "");
					string = string.replace("\\n", "\n").replace("\\r", "\r").replace("\\\"", "\"");
					string = string.replace("\\\\", "\\").replaceAll("\\\\(.)", "$1");
					return new ValuesCursor(TermFactory.leaf(string).asAnnotatable().setAnnotations(data.asAnnotatable().getAnnotations()));

				}
				else {
					throw new IllegalArgumentException();
				}
			}

		};
	}
}
