/**************************************************************************
 * Copyright (c) 2010-2014 Anya Helene Bagge
 * Copyright (c) 2010-2014 University of Bergen
 * 
 * All rights reserved. This program and the accompanying materials
 * are made available under the terms of the Eclipse Public License v1.0
 * which accompanies this distribution, and is available at
 * http://www.eclipse.org/legal/epl-v10.html
 * 
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version. See http://www.gnu.org/licenses/
 * 
 * 
 * See the file COPYRIGHT for more information.
 * 
 * Contributors:
 * * Anya Helene Bagge
 * 
 *************************************************************************/
module org::magnolialang::magnolia::lib::MagnoliaCompileLib
import String;
import IO;
import Set;
import List;
import Map;
import org::magnolialang::magnolia::lib::Environment;
import org::magnolialang::magnolia::MagnoliaFacts;
import org::nuthatchery::pica::terms::Terms;
import org::magnolialang::magnolia::AbstractSyntax;
import org::magnolialang::compiler::Util;
import org::magnolialang::magnolia::lib::Names;
import org::magnolialang::magnolia::lib::CxxAPI;
import Node;

data AST
	= RenameList(AST arg0)  // a list of renaming operations
	| TopName(AST arg0, AST arg1)
	;
data RuntimeException = LookupError(AST astName)
	| InternalError(str msg, AST ast)
	;

public loc unknownLoc = |unknown:///|;

map[str,str] severityMsg = (
	"warning" : "Warning",
	"error" : "Error",
	"internal" : "Internal Error",
	"info" : "Info");

public Severity defaultSeverity = Error();

public str trunc(AST tree) {
	s = "<tree>";
	if(size(s) > 80)
		return substring(s, 0, 80);
	else
		return s;
}


public AST preserveAnnos(AST tree, AST orig) {
	treeAnnos = getAnnotations(tree);
	origAnnos = getAnnotations(orig);
	if(treeAnnos == origAnnos)
		return tree;
	for(a <- origAnnos) {
		//if(a notin treeAnnos) {
			if(a == "concrete") {
				if(consOf(tree) == consOf(orig))
					treeAnnos[a] = origAnnos[a];
			}
			else
				treeAnnos[a] = origAnnos[a];
		//}
	}
	return setAnnotations(tree, treeAnnos);
}

public AST setLocRec(AST tree, loc location) {
	return top-down visit(tree) {
		case AST x => (x@\loc?) ? x : x[@\loc=location]
	}
}

public loc fromOrLocOf(AST tree) {
	if((tree@from)?)
		return tree@from;
	else
		return locOf(tree);
}

public loc locOf(AST tree) {
	return tree@\loc ? estimateLocOf(tree);
}

public loc pickLoc(list[loc] ls) {
	for(l <- ls) {
		if(l != unknownLoc)
			return l;
	}
	
	return unknownLoc;
} 

private loc estimateLocOf(AST tree) {
	if(QName(seq([*ns, n])) := tree) {
		return locOf(n);
	}
	else {
		top-down visit(tree) {
			case n:Name(_): {
				if((n@\loc)?) {
					return n@\loc;
				}
				else
					fail;
			}
			case s:leaf(_): {
				if((s@\loc)?) {
					return s@\loc;
				}
				else
					fail;
			}
			case QName(seq([*ns, n])): {
				return locOf(n);
			}
			case TopMap(f,_): {
				return estimateLocOf(f);
			}
			case Modifier(m): {
				if((m@\loc)?) {
					return m@\loc;
				}
				else
					fail;
			}
		}

		println("Unable to estimate location of <tree>");

		return unknownLoc;
	}
}

public &T locOnly(&T tree) {
  return visit(tree) { case m: node n => delAnnotations(m) };
}

@doc{Return true if the 'contained' location is textually within the 'container'.
The URIs must match exactly, and the contained area must not extend outside
the container. If the 'container' has no offset/length information, it is
assumed to extend to the entire file. Line/column information is ignored.}
public bool isIn(loc contained, loc container) {
	return sameFile(contained, container)
		&&  ((contained.offset >= container.offset 
						&& (contained.offset + contained.length) <= (container.offset + container.length))
			|| !(container.offset?));	
}

public bool sameFile(loc loc1, loc loc2) {
	return loc1 != unknownLoc && loc2 != unknownLoc
		&& loc1.uri == loc2.uri;
}

public str locInfo(loc location, loc base) {
	str filename = "";
	if(!sameFile(location, base))
		filename = location.path + " ";
	if(location.begin.line == location.end.line)
		return "<filename> line <location.begin.line>:<location.begin.column>-<location.end.column>";
	else
		return "<filename> line <location.begin.line>:<location.begin.column>-<location.end.line>:<location.end.column>";
}

@doc{
Find the location where error messages should be displayed for this declaration.
}
public loc dispLocOf(AST decl) {
	if(Define(dcl,_,_) := decl) {
		orgs = getOrigins(decl);
		if(size(orgs) > 0) {
			return orgs[0].location;
		}
		else {
			return locOf(dcl);
		}
	}
	else {
		return locOf(decl);
	}
}


public str kindOf(AST tree, str dflt) {
	switch(tree) {
		case Define(d,_,_):
			return kindOf(d, dflt);
		case TopDefine(d,_,_):
			return kindOf(d, dflt);
		case TypeClause(_):
			return "type";
		case FunClause(n, _, _):
			return "function";
		case ProcClause(n, _):
			return "procedure";
		case TypeName(n):
			return "type";
		case AlertName(n):
			return "alert";
		case AlertClause(_,_):
			return "alert";
		case FunName(_, _ , _):
			return "function";
		case ProcName(_, _):
			return "procedure";
		case VarName(_,_):
			return "var";
		case TopName(_,leaf(k)):
			return k;
		case ConceptClause(_):
			return "concept";
		case SatClause(_):
			return "satisfaction";
		case ImplClause(_):
			return "implementation";
		case RenClause(_):
			return "renaming";
		case ProgramClause(_):
			return "program";
		case Instance(n, _):
			return kindOf(n, dflt);
		default:
			return dflt;
	}
}

public bool isOperationKind(str kind) {
	return kind in ["function", "procedure", "axiom", "test"];
}

public bool isOperation(AST tree) {
	switch(tree) {
		case Define(d,_,_):
			return isOperation(d);
		case Instance(n, _):
			return isOperation(n);
		case FunClause(n, _, _):
			return true;
		case ProcClause(n, _):
			return true;
		case FunName(_, _ , _):
			return true;
		case ProcName(_, _):
			return true;
	}
	return false;
}

public bool isFunctionDecl(AST tree) {
	switch(tree) {
		case Define(d,_,_):
			return isFunctionDecl(d);
		case Instance(n, _):
			return isFunctionDecl(n);
		case FunClause(n, _, _):
			return true;
		case FunName(_, _ , _):
			return true;
	}
	return false;
}

public bool isProcedureDecl(AST tree) {
	switch(tree) {
		case Define(d,_,_):
			return isProcedureDecl(d);
		case Instance(n, _):
			return isProcedureDecl(n);
		case ProcClause(n, _):
			return true;
		case ProcName(_, _):
			return true;
	}
	return false;
}

public bool isTypeDecl(AST tree) {
	switch(tree) {
		case Define(d,_,_):
			return isTypeDecl(d);
		case Instance(n, _):
			return isTypeDecl(n);
		case TypeClause(n):
			return true;
		case TypeName(_):
			return true;
		case SpecialType(_):
			return true;
	}
	return false;
}

public bool isAlertDecl(AST tree) {
	switch(tree) {
		case Define(d,_,_):
			return isAlertDecl(d);
		case Instance(n, _):
			return isAlertDecl(n);
		case AlertClause(_,_):
			return true;
		case AlertName(_):
			return true;
	}
	return false;
}

public list[AST] paramsToArgs(AST paramList) {
	if(ParamList(seq(pList)) := paramList) {
		return paramsToArgs(pList);
	}
	else
		throw InternalError("MagnoliaCompileLib::Unknown parameter list constructor <consOf(paramList)>", paramList);
}

public list[AST] paramsToArgs(list[AST] pList) {
	for(i <- index(pList)) {
		if(Param(m, n, t) := pList[i])
			pList[i] = preserveAnnos(Var(preserveAnnos(VarName(n, t), n)), pList[i]);
		else
			throw InternalError("MagnoliaCompileLib::Unknown parameter constructor <consOf(pList[i])>", pList[i]);
	}
	return pList;
}

public bool isConcrete(AST decl) {
	if(Define(_,_,_) := decl)
		return hasModifier(decl, Impl());
	else if(Nop() := decl)
		return false;
	else
		throw InternalError("MagnoliaCompileLib::isConcrete: not a definition: <consOf(decl)>", decl);
}

public bool isInstance(AST decl) {
	if(Define(_,_,_) := decl)
		return hasModifier(decl, Api());
	else if(Instance(_,_) := decl)
		return true;
	else
		return false;
}

public bool isInstance(DefInfo info) {
	return Api() in info.attrs;
}

public default AST instantiate(AST tree, Env env) {
	return tree;
}

public AST instantiate(Instance(idName, seq(bs)), Env env) {
	DefInfo info = lookupDef(idName, env);
	
	map[AST,AST] renaming = ();
	for(Binding(x, y) <- bs) {
		renaming[x] =  y;
	}
	
	body = rename(info.body, renaming);

	return body;
}

public AST instantiateName(Instance(n, seq(bs))) {
	map[AST,AST] renaming = ();
	for(Binding(x, y) <- bs) {
		renaming[x] =  y;
	}
	
	return rename(n, renaming);
} 

public AST instantiateName(AST n) {
	return n;
}

public AST addModifier(decls:DeclBody(seq(ds)), AST  modifier) {
	decls.arg0.args = [addModifier(d, modifier) | d <- ds];
	return decls;
}

public default AST addModifier(AST decl, AST modifier) {
	if(Define(def,seq(attrs),body) := decl)
		return preserveAnnos(Define(def,seq(toList(toSet(attrs) + modifier)),body), decl);
	else if(Nop() := decl)
		return decl;
	else
		throw InternalError("MagnoliaCompileLib::addModifier: not a definition: <consOf(decl)>", decl);
}

public AST changeModifiers(decls:DeclBody(seq(ds)), set[AST] remove, set[AST] add) {
	decls.arg0.args = [changeModifiers(d, remove, add) | d <- ds];
	return decls;
}

public default AST changeModifiers(AST decl, set[AST] remove, set[AST] add) {
	if(Define(def,seq(attrs),body) := decl)
		return preserveAnnos(Define(def,seq(toList((toSet(attrs) - remove) + add)),body), decl);
	else if(Nop() := decl)
		return decl;
	else
		throw InternalError("MagnoliaCompileLib::addModifier: not a definition: <consOf(decl)>", decl);
}

public AST setModifier(decls:DeclBody(seq(ds)), AST modifier) {
	decls.arg0.args = [setModifier(d, modifier) | d <- ds];
	return decls;
}

public AST setModifier(AST decl, AST modifier) {
	set[AST] computeModifiers(set[AST] mods) {
		switch(modifier) {
			case Require():
				if(size(mods & {Axiom(), Protect(), External(), Assume()}) == 0)
					return (mods - {Extend(), Use(), External()}) + {Require()};
				else
					return mods;
			case Assume():
				if(size(mods & {Axiom(), Protect(), External()}) == 0)
					return (mods - {Extend(), Use(), External(), Require(), ImplyMod()}) + {Assume()};
				else
					return mods;
			case ImplyMod():
				if(size(mods & {Protect(), External(), Assume()}) == 0)
					return (mods - {Extend(), Use(), External()}) + {ImplyMod()};
				else
					return mods;
			case Extend():
				if(size(mods & {Use(), Require(), Protect(), Assume()}) == 0)
					return mods + {Extend()};
				else
					return mods;
			case Use():
				if(size(mods & {Require(), Protect(), Assume()}) == 0)
					return (mods - {Extend()}) + {Use()};
				else
					return mods;
			case External():
				if(size(mods & {Require(), Protect()}) == 0)
					return (mods - {Assume(), Extend()}) + {External()};
				else
					return mods;
			case Protect():
				if(Axiom() notin mods)
					return (mods - {Extend(), Use(), External(), Require(), Assume(), ImplyMod()}) + {Protect()};
				else
					return mods;
			case Api():
				return (mods - {Impl()}) + {Api()};
			case Impl():
				return (mods - {Api()}) + {Impl()};
			case Retain():
				return (mods - {Remove()}) + {Retain()};
			case Remove():
				return (mods - {Retain()}) + {Remove()};
			default:
				throw InternalError("MagnoliaCompileLib::setModifier: modifier should be one of require, extend, use, external, protect, api, impl: <consOf(modifier)>", modifier);
		}
	}
	
	if(Define(def,seq(attrs),body) := decl) {
		return preserveAnnos(Define(def,seq(toList(computeModifiers(toSet(attrs)))),body), decl);
	}
	else if(Nop() := decl) {
		return decl;
	}
	else {
		throw InternalError("MagnoliaCompileLib::addModifier: not a definition: <consOf(decl)>", decl);
	}
}

public bool hasModifier(AST decl, AST modifier) {
	if(Define(def,seq(attrs),body) := decl)
		return modifier in attrs;
	else if(Nop() := decl)
		return false;
	else
		throw InternalError("MagnoliaCompileLib::hasModifier: not a definition: <consOf(decl)>", decl);
}

@doc{True if expr has one or more of the modifiers.}
public bool hasModifier(AST decl, set[AST] modifiers) {
	if(Define(def,seq(attrs),body) := decl)
		return size(toSet(attrs) & modifiers) > 0 ;
	else if(Nop() := decl)
		return false;
	else
		throw InternalError("MagnoliaCompileLib::hasModifier: not a definition: <consOf(decl)>", decl);
}

public AST addOrigin(decls:DeclBody(seq(ds)), str name, loc origin) {
	decls.arg0.args = [addOrigin(d, name, origin) | d <- ds];
	return decls;
}

public default AST addOrigin(AST decl, str name, loc origin) {
	if(Define(def,seq(attrs),body) := decl) {
		for(i <- index(attrs)) {
			if(Origins(os) := attrs[i]) {
				attrs[i] = Origins([Origin(name, origin), *os]);
				decl.arg1.args = attrs;
				return decl; 
			}
		}
		decl.arg1.args = [Origins([Origin(name, origin)]), *attrs];
		return decl;
	}
	else if(Nop() := decl)
		return Nop();
	else
		throw InternalError("MagnoliaCompileLib::addOrigin: not a definition: <consOf(decl)>", decl);
}

public list[Origin] getOrigins(AST decl) {
	if(Define(def,seq(attrs),body) := decl) {
		if([*_, Origins(os), *_] := attrs) {
			return os;
		}
		else {
			return [];
		}
	}
	else if(Nop() := decl)
		return [];
	else
		throw InternalError("MagnoliaCompileLib::getOrigins: not a definition: <consOf(decl)>", decl);
}

public tuple[str,loc] getLatestOrigin(AST decl, loc fallback) {
	orgs = getOrigins(decl);
	if([Origin(name,org), *_] := orgs) {
		return <name, org>;
	}
	else {
		return <"", fallback>;
	}
}

public list[AST] getDeclAttrs(AST decl) {
	if(Define(def,seq(attrs),body) := decl)
		return attrs;
	else if(Nop() := decl)
		return [];
	else
		throw InternalError("MagnoliaCompileLib::getDeclAttrs: not a definition: <consOf(decl)>", decl);
}

public AST addDeclAttr(AST decl, AST attr) {
	if(Define(def,s:seq(attrs),body) := decl)
		return decl[arg1=preserveAnnos(seq([attr, *attrs]), s)];
	else if(Nop() := decl)
		return decl;
	else
		throw InternalError("MagnoliaCompileLib::addDeclAttr: not a definition: <consOf(decl)>", decl);
}
public AST getDefBody(AST decl) {
	if(Define(def,seq(attrs),body) := decl)
		return body;
	else if(Nop() := decl)
		return EmptyBody();
	else
		throw InternalError("MagnoliaCompileLib::getDefBody: not a definition: <consOf(decl)>", decl);
}

public bool hasBody(AST decl) {
	return getDefBody(decl) != EmptyBody();
}

public AST getDeclarator(AST decl) {
	if(Define(def,seq(attrs),body) := decl)
		return def;
	else if(Nop() := decl)
		return Nop();
	else
		throw InternalError("MagnoliaCompileLib::getDeclarator: not a definition: <consOf(decl)>", decl);
}

public bool isFlat(AST tree) {
	return DeclBody(_) := tree;
}

public bool isTopLevel(DefInfo info) {
	return info.kind in {"concept", "satisfaction", "implementation", "renaming", "program"}; 
}

public AST trackOrigin(AST tree, AST from) {
	l = locOf(from);
	if(l != unknownLoc)
		return tree[@from=l];
	else
		return tree;
}

public bool isType(Instance(tree, _)) {
	return isType(tree);
}

public default bool isType(AST tree) {
	return TypeName(_) := tree || Unknown() := tree || SpecialType(_) := tree;
}

public list[AST] paramsOf(Define(dtor, _,_ )) {
	return paramsOf(dtor);
}

public list[AST] paramsOf(FunClause(_,ParamList(seq(ps)),_)) {
	return ps;
}

public list[AST] paramsOf(ProcClause(_,ParamList(seq(ps)))) {
	return ps;
}

public AST returnOf(Define(dtor, _,_ )) {
	return returnOf(dtor);
}

public AST returnOf(FunClause(_,_,rt)) {
	return rt;
}

public AST inParams(pl:ParamList(seq(ps))) {
	return pl[arg0 = seq(inParams(ps))];
}
public AST outParams(pl:ParamList(seq(ps))) {
	return pl[arg0 = seq(outParams(ps))];
}
public AST filterParams(pl:ParamList(seq(ps)), set[AST] modes) {
	return pl[arg0 = seq(filterParams(ps, modes))];
}

public list[AST] inParams(list[AST] ps) {
	return filterParams(ps, {Obs(), Upd(), Del()});
}

public list[AST] outParams(list[AST] ps) {
	return filterParams(ps, {Out(), Upd(), Giv()});
}


public list[AST] filterParams(list[AST] ps, set[AST] modes) {
	list[AST] result = [];
	for(p:Param(m, n, t) <- ps) {
		if(m in modes)
			result += p;
	}
	return result;
}

public bool isValidLanguage(AST language, Env env) {
	if(isCxxLanguage(language)) {
		language = checkCxxLanguage(language, env);
		return true;
	}
	else if(isJavaLanguage(language)) {
		return true;
	}
	
	return false;
}

public list[AST] sortDefs(list[AST] defs) {
	return sort(defs, bool(AST a, AST b) {
			if(<Define(dcl1, _, _), Define(dcl2, _, _)> := <a, b>) {
				if(isOperation(dcl1) && isOperation(dcl2)) {
					if(hasModifier(a, Axiom()) == hasModifier(b, Axiom()))
						return yieldTerm(dcl1) < yieldTerm(dcl2);
					else
						return !hasModifier(a, Axiom());
				}
				else if(!isOperation(dcl1) && !isOperation(dcl2))
					return yieldTerm(dcl1) < yieldTerm(dcl2);
				else					
					return !isOperation(dcl1);
			}
			else
				return yieldTerm(a) < yieldTerm(b);
	});
}

public AST getExprType(AST e) {
	try {
		return _getExprType(e);
	}
	catch ex: {
		termPrintln("getExprType: failed: ", e);
		throw ex;
	}
}

AST _getExprType(Var(v))                  { return _getExprType(v); }
AST _getExprType(VarName(v, t))           { return t; }
AST _getExprType(TopName(n, k))           { return SpecialType(DeclList()); }
AST _getExprType(Apply(f, _))             { return _getExprType(f); }
AST _getExprType(Fun(f))                  { return _getExprType(f); }
AST _getExprType(FunName(_,_,t))          { return t; }
AST _getExprType(FunOf(p,leaf(i)))		  { return _getExprType(paramsOf(instantiateName(p))[toInt(i)]); }
AST _getExprType(ExternalFun(_,_,_,_,_,t))  { return t; }
AST _getExprType(Builtin(n,_))            { return _getExprType(n); }
AST _getExprType(Equal(_, a, b))          { return SpecialType(Predicate()); }
AST _getExprType(Struct(t, _))            { return t; }
AST _getExprType(DotOp(e, n, t))          { return t; }
AST _getExprType(BlockExpr(ss, t))          { return t; }
AST _getExprType(Literal(l))              { return SpecialType(Literal()); }
AST _getExprType(Param(_,_,t))            { return t; }
AST _getExprType(AnonParam(_,t))          { return t; }
AST _getExprType(IfThenElseExpr(_,t,_))   { return _getExprType(t); }
AST _getExprType(Guard(e, _))             { return _getExprType(e); }	
AST _getExprType(x:Instance(n, b))        { return _getExprType(instantiateName(x)); }	
default AST _getExprType(AST e)           { throw "getExprType(): not implemented for <consOf(e)>"; }


public bool xor(bool a, bool b) { return (a && !b) || (!a && b); }

public AST rename(AST tree, map[AST,AST] renaming) {
	return top-down-break visit(tree) {
		case n:Name(_): if(n in renaming) insert renaming[n]; else fail;
		case n:TypeName(IdName(_,_)): if(n in renaming) insert renaming[n];
		case n:FunName(IdName(_,_),_,_): if(n in renaming) insert renaming[n];
		case n:ProcName(IdName(_,_),_): if(n in renaming) insert renaming[n];
		case n:TypeName(_): if(n in renaming) insert renaming[n]; else fail;
		case n:FunName(_,_,_): if(n in renaming) insert renaming[n]; else fail;
		case n:ProcName(_,_): if(n in renaming) insert renaming[n]; else fail;
		case b:Binding(x, y): {
			z = rename(y, renaming);
			if(y != z)
				insert Binding(x, z);
//			if(y in renaming)
//				insert renaming[z];
		}
//		case Concrete(_): ;
		case i:Instance(x, y): ;
		case x:ExternalType(lang, m, onPart, n): {
			onPart = rename(onPart, renaming);
			insert preserveAnnos(ExternalType(lang, m, onPart, n), x);
		}
		case x:ExternalFun(lang, m, onPart, n, ps, t): {
			ps = rename(ps, renaming);
			t = rename(t, renaming);
			onPart = rename(onPart, renaming);
			insert preserveAnnos(ExternalFun(lang, m, onPart, n, ps, t), x);
		}
		case x:ExternalProc(lang, m, onPart, n, ps): {
			ps = rename(ps, renaming);
			onPart = rename(onPart, renaming);
			insert preserveAnnos(ExternalProc(lang, m, onPart, n, ps), x);
		}
	};
}

public AST LetMulti(list[AST] decls, AST body) {
	r = _LetMulti(decls, body);
	assert Let(_,seq(_)) := r || Block(seq(_)) := r;
	assert !(Let(seq(_),_) := r);
	return r;
	
}

private AST _LetMulti([AST decl, *decls], AST body) {
	if(decls != [])
		return Let(decl, seq([LetMulti(decls, body)]));
	else
		return Let(decl, body);
}

private AST _LetMulti([], AST body) {
	return Block(body);
}

@doc{The contents (as a string) of a file bundled with the compiler.}
@reflect{Throwing exceptions}
@javaClass{org.magnolialang.magnolia.lib.MagnoliaCompileLib}
public java str getBundledResource(str name);
