/**************************************************************************
 * Copyright (c) 2014 Anya Helene Bagge
 * Copyright (c) 2014 University of Bergen
 * 
 * All rights reserved. This program and the accompanying materials
 * are made available under the terms of the Eclipse Public License v1.0
 * which accompanies this distribution, and is available at
 * http://www.eclipse.org/legal/epl-v10.html
 * 
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version. See http://www.gnu.org/licenses/
 * 
 * 
 * See the file COPYRIGHT for more information.
 * 
 * Contributors:
 * * Anya Helene Bagge
 * 
 *************************************************************************/
module org::magnolialang::magnolia::lib::Modules
import IO;
import org::magnolialang::magnolia::AbstractSyntax;
import org::magnolialang::magnolia::MagnoliaFacts;
import org::magnolialang::magnolia::typechecker::toplevel::TopLevelOps;
import org::magnolialang::magnolia::lib::MagnoliaCompileLib;
import org::magnolialang::magnolia::lib::Names;
import org::magnolialang::magnolia::lib::Environment;
import org::nuthatchery::pica::terms::Terms;
import org::magnolialang::compiler::Util;
import Set;
import Node;
import List;
import String;
import util::Benchmark;
import util::Math;

data DefKind = RequireKind() | ProvideKind() | ImplKind() | UndefinedKind();

data Module = Module(AST name, map[AST, DefKind] kinds, map[AST,AST] dtors, map[AST,AST] bodies, map[AST,set[AST]] attrs);


public Module Module(AST name) {
	return Module(name, (), (), (), ());
}

public DefKind defKind(Module m, AST qName) {
	return m.kinds[qName] ? UndefinedKind();
}

public bool hasApi(Module m, AST qName) {
	return defKind(m, qName) in {RequireKind(), ProvideKind()};
}

public bool hasDef(Module m, AST qName) {
	return qName in m.kinds;
}

public bool isImplemented(Module m, AST qName) {
	return defKind(m, qName) == ImplKind();
}

public bool isRequired(Module m, AST qName) {
	return defKind(m, qName) == RequireKind();
}

public bool isProvided(Module m, AST qName) {
	return defKind(m, qName) == RequireKind();
}

public bool isInstance(Module m, AST qName) {
	return isInstance(getBody(m, qName));
}

public AST getDeclarator(Module m, AST qName) {
	return m.dtors[qName];
}

public AST getBody(Module m, AST qName) {
	return m.bodies[qName];
}

public Module setBody(Module m, AST qName, AST body) {
	if(!hasDef(m, qName)) {
		throw InternalError("Modules::setBody called on undefined name", qName);
	}
	assert !(Define(_,_,_) := body);
	m.bodies[qName] = body;
	
	return m;
}

public set[AST] getAttrs(Module m, AST qName) {
	return m.attrs[qName];
}

public Module setAttrs(Module m, AST qName, set[AST] attrs) {
	if(!hasDef(m, qName)) {
		throw InternalError("Modules::setAttrs called on undefined name", qName);
	}
	
	m.attrs[qName] = attrs;
	
	return m;
}

public AST getDefine(Module m, AST qName) {
	body = m.bodies[qName];
	attrs = m.attrs[qName];
	
	return Define(m.dtors[qName], seq(toList(attrs)), body);
}

public DefInfo getDefInfo(Module m, AST qName) {
	info = definfo();
	info.attrs = toList(m.attrs[qName]);
	// info.parent = env.scName;
	// info.locInfo = LocInfo(n@\loc);
	info.decl = m.dtors[qName];
	info.kind = kindOf(qName, "unknown");
	info.qName = qName;
	info.defName = basename(qName);
	info.body = m.bodies[qName];
	
	return info;
}

public Module addApi(Module m, d:Define(dtor, seq(attrs), body)) {
	if(hasDef(m, qName)) {
		throw InternalError("adding API definition which already exists", qName);
	}
	return replaceApi(m, defNameOf(dtor), dtor, body, toSet(attrs));
}

public Module replaceApi(Module m, d:Define(dtor, seq(attrs), body)) {
	return replaceApi(m, defNameOf(dtor), dtor, body, toSet(attrs));
}

public Module replaceApi(Module m, AST qName, AST dtor, AST body, set[AST] attrs) {
	if(EmptyBody() := body) {
		//if(Require() notin attrs) {
		//	throw InternalError("adding API requirement should have require attribute", qName);
		//}
		return replaceInModule(m, qName, RequireKind(), dtor, body, attrs);
	}
	else {
		if(Require() in attrs) {
			throw InternalError("adding API definition should not have require attribute", qName);
		}
		return replaceInModule(m, qName, ProvideKind(), dtor, body, attrs);
	}
}

public Module addImpl(Module m, d:Define(dtor, seq(attrs), body)) {
	qName = defNameOf(dtor);
	if(hasDef(m, qName)) {
		throw InternalError("Internal compiler error: adding concrete definition which already exists", qName);
	}
	return replaceInModule(m, qName, ImplKind(), dtor, body, toSet(attrs));
}

public Module replaceImpl(Module m, d:Define(dtor, seq(attrs), body)) {
	return replaceInModule(m, defNameOf(dtor), ImplKind(), dtor, body, toSet(attrs));
}

private Module replaceInModule(Module m, AST qName, DefKind kind, AST dtor, AST body, set[AST] attrs) {
	assert !(Define(_,_,_) := body);
	m.kinds[qName] = kind;
	m.bodies[qName] = body;
	m.attrs[qName] = attrs;
	m.dtors[qName] = dtor;
	return m;
}

public set[AST] requiresOf(Module m) {
	return {n | n <- m.kinds, m.kinds[n] == RequireKind()};
}

public set[AST] providesOf(Module m) {
	return {n | n <- m.kinds, m.kinds[n] == ProvideKind()};
}

public set[AST] implementsOf(Module m) {
	return {n | n <- m.kinds, m.kinds[n] == ImplKind()};
}




public list[AST] getParamList(Module m, AST qName) {
	return paramsOf(getDeclarator(m, qName)) ? [];
}

public AST getReturnType(Module m, AST qName) {
	return returnOf(getDeclarator(m, qName)) ? Undefined();
}

public void printModule(Module m) {
	termPrintln("Module ", m.name, " contains: ");
	termPrintln("  Requires API:");
	for(qn <- requiresOf(m)) {
		termPrintln("    ", qn);
	}
	termPrintln("  Provides API:");
	for(qn <- providesOf(m)) {
		termPrintln("    ", qn, " -\> ", m.api[qn]);
	}
	termPrintln("  Definitions:");
	for(qn <- implementsOf(m)) {
		termPrintln("    ", qn);
	}
}