/**************************************************************************
 * Copyright (c) 2011-2014 Anya Helene Bagge
 * Copyright (c) 2011-2014 University of Bergen
 * 
 * All rights reserved. This program and the accompanying materials
 * are made available under the terms of the Eclipse Public License v1.0
 * which accompanies this distribution, and is available at
 * http://www.eclipse.org/legal/epl-v10.html
 * 
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version. See http://www.gnu.org/licenses/
 * 
 * 
 * See the file COPYRIGHT for more information.
 * 
 * Contributors:
 * * Anya Helene Bagge
 * 
 *************************************************************************/
module org::magnolialang::magnolia::lib::Names

import org::nuthatchery::pica::terms::Terms;
import org::magnolialang::magnolia::AbstractSyntax;
import org::magnolialang::compiler::Util;
import org::magnolialang::magnolia::MagnoliaFacts;
import org::magnolialang::magnolia::lib::MagnoliaCompileLib;
import org::nuthatchery::pica::memo::Memo;
import IO;
import String;
import List;


@doc{Return true iff argument is a name.}
public default bool isName(AST n) { return false; }
public bool isName(Name(leaf(_))) { return true; }
public bool isName(TypeName(n)) { return isName(n); }
public bool isName(AlertName(n)) { return isName(n); }
public bool isName(FunName(n,ps,rt)) { return isName(n); }
public bool isName(ProcName(n,ps)) { return isName(n); }
public bool isName(VarName(n,t)) { return isName(n); }
public bool isName(TopName(n,k)) { return isName(n); }
public bool isName(IdName(n,i)) { return true; }
public bool isName(SpecialType(_)) { return true; }
public bool isName(Unknown()) { return true; }
public bool isName(QName(seq(ns))) { return !any(n <- ns, isQualifiedName(n)); }

@doc{Return true iff argument is a unique id name.}
public default bool isId(AST n) { return false; }
public bool isId(Name(leaf(_))) { return false; }
public bool isId(TypeName(n)) { return isId(n); }
public bool isId(AlertName(n)) { return isId(n); }
public bool isId(FunName(n,ps,rt)) { return isId(n); }
public bool isId(ProcName(n,ps)) { return isId(n); }
public bool isId(VarName(n,t)) { return isId(n); }
public bool isId(TopName(n,k)) { return isId(n); }
public bool isId(IdName(n,i)) { return true; }
public bool isId(SpecialType(_)) { return true; }
public bool isId(Unknown()) { return true; }
public bool isId(QName(seq(ns))) { return isId(ns[size(ns)-1]); }


@doc{Generate a unique ID string}
@javaClass{org.magnolialang.magnolia.lib.MagnoliaCompileLib}
private java str makeId();

public AST makeId(AST n, str s) {
	return transformName(n, AST(y) {
		if(Name(x) := y) 
			return preserveAnnos(IdName(x, leaf(s)), x);
		else if(IdName(x,leaf(z)) := y) 
			return preserveAnnos(IdName(x, leaf("<s>z<z>")), x);
	});
}

public AST makeId(AST n) {
	return makeId(n, makeId());
}


public AST unIdName(AST n) {
	if(IdName(x, _) := unqualify(basename(n)))
		return Name(x);
	else
		return n;
}

@doc{Return true iff argument is a qualified name.}
public default bool isQualifiedName(AST n) {
	if(!isName(n))
		throw InternalError("isQualifiedName(): not a name <yieldTerm(n)>", n);
	return false;
}
public bool isQualifiedName(Name(leaf(_))) { return false; }
public bool isQualifiedName(IdName(_,_)) { return false; }
public bool isQualifiedName(TypeName(n)) { return isQualifiedName(n); }
public bool isQualifiedName(AlertName(n)) { return isQualifiedName(n); }
public bool isQualifiedName(FunName(n,ps,rt)) { return isQualifiedName(n); }
public bool isQualifiedName(ProcName(n,ps)) { return isQualifiedName(n); }
public bool isQualifiedName(VarName(n,t)) { return isQualifiedName(n); }
public bool isQualifiedName(TopName(n,_)) { return isQualifiedName(n); }
public bool isQualifiedName(QName(seq(ns))) { return !any(n <- ns, isQualifiedName(n)); }


@doc{Qualify 'name' with the given scope name.}
public AST qualify(AST scopeName, AST name) {
	if(!isName(name))
		throw InternalError("qualify(): not a name <yieldTerm(name)>", name);
	
	if(TopName(n, _) := scopeName)
		scopeName = n;
		
	if(Unknown() := name || SpecialType(_) := name)
		return name;
	else if(QName(seq(ns)) := scopeName) {
		if(QName(seq(ns2)) := name)
			return preserveAnnos(QName(seq(ns + ns2)), name);
		else {
			return preserveAnnos(QName(seq(ns + name)), name);
		}
	}
	else if(!isQualifiedName(scopeName)) {
		if(QName(seq(ns2)) := name)
			throw InternalError("qualify(): unqualified scope, but qualified name; did you switch parameters? qualify(<yieldTerm(scopeName)>, <yieldTerm(name)>)", <scopeName, name>);
		return preserveAnnos(QName(seq([mn, name])), name);
	}
	else
		throw InternalError("qualify(): Ill-formed module name <yieldTerm(scopeName)>", scopeName);
}


@doc{Return 'name' with qualifiactions (if any) removed.}
public default AST unqualify(AST name) {
	throw InternalError("unqualify(): Unknown constructor <consIdOf(name)>", name);
}
public AST unqualify(n:Name(_))            { return n; }
public AST unqualify(n:IdName(_,_))        { return n; }
public AST unqualify(x:TypeName(n))        { return x[arg0 = unqualify(n)]; }
public AST unqualify(x:AlertName(n))       { return x[arg0 = unqualify(n)]; }
public AST unqualify(x:FunName(n, ps, t))  { return x[arg0 = unqualify(n)]; }
public AST unqualify(x:ProcName(n, ps))    { return x[arg0 = unqualify(n)]; }
public AST unqualify(x:VarName(n, t))      { return x[arg0 = unqualify(n)]; }
public AST unqualify(x:TopName(n,_))       { return x[arg0 = unqualify(n)]; }
public AST unqualify(QName(seq(ns)))       { return ns[size(ns)-1]; }
public AST unqualify(n:Unknown())          { return n; }
public AST unqualify(n:SpecialType(_))     { return n; }


@doc{Return 'name' with the base name transformed.}
public default AST transformName(AST name, AST(AST) trans) {
	throw InternalError("transformName(): Unknown constructor <consIdOf(name)>", name);
}
public AST transformName(x:Name(n), AST(AST) trans)            { return trans(x); }
public AST transformName(x:IdName(n,_), AST(AST) trans)        { return trans(x); }
public AST transformName(x:TypeName(n), AST(AST) trans)        { return x[arg0 = transformName(n, trans)]; }
public AST transformName(x:AlertName(n), AST(AST) trans)       { return x[arg0 = transformName(n, trans)]; }
public AST transformName(x:FunName(n, ps, t), AST(AST) trans)  { return x[arg0 = transformName(n, trans)]; }
public AST transformName(x:ProcName(n, ps), AST(AST) trans)    { return x[arg0 = transformName(n, trans)]; }
public AST transformName(x:VarName(n, t), AST(AST) trans)      { return x[arg0 = transformName(n, trans)]; }
public AST transformName(x:TopName(n,_), AST(AST) trans)       { return x[arg0 = transformName(n, trans)]; }
public AST transformName(QName(seq([*ns, n])), AST(AST) trans) { return QName(seq([*ns, transformName(n, trans)])); }
public AST transformName(n:Unknown(), AST(AST) trans)          { return n; }
public AST transformName(n:SpecialType(_), AST(AST) trans)     { return n; }

@doc{Return 'name' with the base name transformed.}
public default AST transformName(AST name, str(str) trans) {
	throw InternalError("transformName(): Unknown constructor <consIdOf(name)>", name);
}
public AST transformName(n:leaf(s), str(str) trans)            { return n[strVal = trans(s)]; }
public AST transformName(x:Name(n), str(str) trans)            { return x[arg0 = transformName(n, trans)]; }
public AST transformName(x:IdName(n,_), str(str) trans)          { return x[arg0 = transformName(n, trans)]; }
public AST transformName(x:TypeName(n), str(str) trans)        { return x[arg0 = transformName(n, trans)]; }
public AST transformName(x:AlertName(n), str(str) trans)       { return x[arg0 = transformName(n, trans)]; }
public AST transformName(x:FunName(n, ps, t), str(str) trans)  { return x[arg0 = transformName(n, trans)]; }
public AST transformName(x:ProcName(n, ps), str(str) trans)    { return x[arg0 = transformName(n, trans)]; }
public AST transformName(x:VarName(n, t), str(str) trans)      { return x[arg0 = transformName(n, trans)]; }
public AST transformName(x:TopName(n,_), str(str) trans)       { return x[arg0 = transformName(n, trans)]; }
public AST transformName(QName(seq([*ns, n])), str(str) trans) { return QName(seq([*ns, transformName(n, trans)])); }
public AST transformName(n:Unknown(), str(str) trans)          { return n; }
public AST transformName(n:SpecialType(_), str(str) trans)     { return n; }

@doc{Return 'name' with a qualifiaction prefix removed.}
public default AST unprefix(AST name, AST pfx) {
	throw InternalError("unprefix(): Unknown constructor <consIdOf(name)>", name);
}
public AST unprefix(n:Name(_), AST pfx)            { return n; }
public AST unprefix(n:IdName(_,_), AST pfx)          { return n; }
public AST unprefix(x:TypeName(n), AST pfx)        { return x[arg0 = unprefix(n,pfx)]; }
public AST unprefix(x:AlertName(n), AST pfx)       { return x[arg0 = unprefix(n,pfx)]; }
public AST unprefix(x:FunName(n, ps, t), AST pfx)  { return x[arg0 = unprefix(n,pfx)][arg1 = unprefixParams(ps,pfx)][arg2 = unprefix(t,pfx)]; }
public AST unprefix(x:ProcName(n, ps), AST pfx)    { return x[arg0 = unprefix(n,pfx)][arg1 = unprefixParams(ps,pfx)]; }
public AST unprefix(x:VarName(n, t), AST pfx)      { return x[arg0 = unprefix(n,pfx)][arg1 = unprefix(t,pfx)]; }
public AST unprefix(x:TopName(n,_), AST pfx)       { return x[arg0 = unprefix(n,pfx)]; }
public AST unprefix(x:AnonParam(m, t), AST pfx)    { return x[arg1 = unprefix(t,pfx)]; }
public AST unprefix(x:QName(seq(ns)), QName(seq(pfx))) {
	if([pfx, sfx*] := ns)
		return QName(seq(sfx));
	else
		return x;
}
public AST unprefix(n:Unknown(), AST pfx)          { return n; }
public AST unprefix(n:SpecialType(_), AST pfx)          { return n; }

public AST unprefixParams(seq(ps), AST pfx) {
	return seq([unprefix(p, pfx) | p <- ps]);
}


@doc{Return the qualifier of 'name'}
public AST qualifierOf(AST name) {
	assert isName(name);
	
	if(QName(seq([*ns, n])) := basename(name))
		return QName(seq(ns));
	else
		return QName(seq([]));
}


@doc{Return name as a simple name, without kind decoration.
E.g., basename(function x.y.z.foo(...) : t) = foo}
public AST basename(AST name) {
	if(!isName(name))
		throw InternalError("basename(): not a name <yieldTerm(name)>", name);
	
	switch(name) {
		case TypeName(n):
			return n;
		case FunName(n, _, _):
			return n;
		case ProcName(n, _):
			return n;
		case VarName(n, _):
			return n;
		case TopName(n,_):
			return n;
		case AlertName(n):
			return n;
		default:
			return name;
	}
}


@doc{Return true if n1 and n2 are names of the same kind
  (e.g, functions, procedures, ...).}
public bool sameKindName(AST n1, AST n2) {
	if(!isName(n1))
		throw InternalError("sameKindName(): not a name <yieldTerm(n1)>", n1);
	if(!isName(n2))
		throw InternalError("sameKindName(): not a name <yieldTerm(n2)>", n2);
	
	switch(<unqualify(n1),unqualify(n2)>) {
		case <TypeName(_), TypeName(_)>:
			return true;
		case <FunName(_,_,_), FunName(_,_,_)>:
			return true;
		case <ProcName(_,_), ProcName(_,_)>:
			return true;
		case <VarName(_,_), VarName(_,_)>:
			return true;
		case <TopName(_,_), TopName(_,_)>:
			return true;
		case <AlertName(_), AlertName(_)>:
			return true;
		case <Name(_), Name(_)>:
			return true;
		case <IdName(_,_), IdName(_,_)>:
			return true;
		case <Name(_), _>:
			return true;
	}
	return false;
}

@doc{Return the name of the given construct, formatted for display / debug.}
public default str dispNameOf(AST tree, bool short) {
			throw InternalError("dispNameOf(): Unknown constructor <consIdOf(tree)>", tree);
}
public str dispNameOf(DefInfo info, bool short) {
	if(info.kind == "axiom" && FunClause(n,ps,t) := info.decl)
		return dispNameOf(FunClause(n,ps,SpecialType(Axiom())), short);
	else
		return dispNameOf(info.decl, short);
		
}
public str dispNameOf(def:Define(FunClause(n,ps,t),_,_), bool short) {
	if(hasModifier(def, Axiom()))
		return dispNameOf(FunClause(n,ps,SpecialType(Axiom())), short);
	else
		return dispNameOf(FunClause(n,ps,t), short);
}
public str dispNameOf(Define(d,_,_), bool short)        { return dispNameOf(d, short); }
public str dispNameOf(TypeClause(Unknown()), bool short)  { return "?"; }
public str dispNameOf(TypeClause(n), bool short)          { return "type <dispNameOf(basename(n), short)>"; }
public str dispNameOf(AlertClause(n, _), bool short)         { return "alert <dispNameOf(basename(n), short)>"; }
public str dispNameOf(LiteralClause(n, as), bool short)   { return "literal type <dispNameOf(basename(n), short)>"; }
public str dispNameOf(FunClause(n, ParamList(seq(ps)), t), bool short) {
	return dispOpNameOf(n, ps, t, funParamToStr, short);
}
public str dispNameOf(ProcClause(n, ParamList(seq(ps))), bool short) {
	return dispOpNameOf(n, ps, Nop(), procParamToStr, short);
}
public str dispNameOf(Name(leaf(n)), bool short)          { return n; }
public str dispNameOf(IdName(leaf(n), leaf(i)), bool short)  { return "<n>«<i>»"; }
public str dispNameOf(TypeName(n), bool short)            { return "type <dispNameOf(n, short)>"; }
public str dispNameOf(AlertName(n), bool short)           { return "alert <dispNameOf(n, short)>"; }
public str dispNameOf(FunName(n, seq(ps), t), bool short) {
	return dispOpNameOf(n, ps, t, funParamToStr, short);
}
public str dispNameOf(ProcName(n, seq(ps)), bool short)   {
	return dispOpNameOf(n, ps, Nop(), procParamToStr, short);
}
public str dispNameOf(VarName(Unknown(), t), bool short)  { return "var _ : <nameToStr(t, short)>"; }
public str dispNameOf(VarName(n, t), bool short)          { return "var <nameToStr(n, short)> : <nameToStr(t, short)>"; }
public str dispNameOf(VarDecl(m, n, t, e), bool short)	    { return "var <nameToStr(n, short)> : <nameToStr(t, short)>"; }
public str dispNameOf(Param(m, n, t), bool short)	    { return "<modeToStr(m)> <nameToStr(n, short)> : <nameToStr(t, short)>"; }
public str dispNameOf(TopName(n,leaf(k)), bool short)     { return "<k> <nameToStr(n, short)>"; }
public str dispNameOf(AnonParam(m, t), bool short)        { return "var _ : <nameToStr(t, short)>"; }
public str dispNameOf(QName(seq(ns)), bool short)         { if(short) return nameToStr(ns[size(ns)-1]); else return strJoin([nameToStr(n, short) | n <- ns], "."); }
public str dispNameOf(Unknown(), bool short)              { return "?"; }
public str dispNameOf(SpecialType(n), bool short)         { return toLowerCase(consOf(n)); }
public str dispNameOf(AST tree) {	return dispNameOf(tree, false); }
public str dispNameOf(ConceptClause(n), bool short)       { return "concept <dispNameOf(basename(n), short)>"; }
public str dispNameOf(ImplClause(n), bool short)          { return "implementation <dispNameOf(basename(n), short)>"; }
public str dispNameOf(ProgramClause(n), bool short)       { return "program <dispNameOf(basename(n), short)>"; }
public str dispNameOf(RenClause(n), bool short)           { return "renaming <dispNameOf(basename(n), short)>"; }
public str dispNameOf(SatClause(n), bool short)           { return "satisfaction <dispNameOf(basename(n), short)>"; }
public str shortDispNameOf(tree) {	return dispNameOf(tree, true); }
private str dispOpNameOf(AST n, list[AST] ps, AST t, str(AST,bool) p2s, bool short) {
	str typePart = "";
	str kind = "";
	switch(t) {
		case SpecialType(Predicate()):
			kind = "predicate";
		case SpecialType(Axiom()):
			kind = "axiom";
		case SpecialType(Test()):
			kind = "test";
		case SpecialType(Renaming()):
			kind = "renaming";
		case SpecialType(Decl()):
			kind = "decl";
		case SpecialType(DeclList()):
			kind = "decllist";
		case SpecialType(Literal()):
			kind = "literal";
		case Nop():
			kind = "procedure";
		default: {
			kind = "function";
			typePart = " : <nameToStr(t, short)>";
		}
	}	 
	return "<kind> <nameToStr(n, short)>(<strJoin([p2s(x, short) | x <- ps], ", ")>)<typePart>";
}

@doc{Return the name of the given construct, formatted for display / debug.}
public default str nameToStr(AST tree, bool short) {
			throw InternalError("nameToStr(): Unknown constructor <consIdOf(tree)>", tree);
}
// public str nameToStr(ProcClause(n, _))       { return nameToStr(n); }
public str nameToStr(Name(leaf(n)), bool short)          { return n; }
public str nameToStr(IdName(leaf(n), leaf(i)), bool short)  { return "<n>«<i>»"; }
public str nameToStr(Instance(n,seq(bs)), bool short)  		 { return "<nameToStr(n, short)>[<strJoin(["<shortNameToStr(x)> =\> <shortNameToStr(y)>" | Binding(x, y) <- bs], ", ")>]"; }
public str nameToStr(Lang(leaf(n)), bool short)          { return n; }
public str nameToStr(TypeName(Unknown()), bool short)    { return "?"; }
public str nameToStr(TypeName(n), bool short)            { return nameToStr(n, short); }
public str nameToStr(AlertName(n), bool short)            { return nameToStr(n, short); }
public str nameToStr(Type(n), bool short)    	         { return nameToStr(n, short); }
public str nameToStr(FunName(n, _, _), bool short)       { return nameToStr(n, short); }
public str nameToStr(ProcName(n, _), bool short)         { return nameToStr(n, short); }
public str nameToStr(VarName(Unknown(), t), bool short)  { return "_"; }
public str nameToStr(VarName(n, t), bool short)          { return nameToStr(n, short); }
public str nameToStr(TopName(n,_), bool short)           { return nameToStr(n, short); }
public str nameToStr(AnonParam(m, t), bool short)        { return "_"; }
public str nameToStr(Param(m,n,t), bool short)           { return nameToStr(n, short); }
public str nameToStr(QName(seq(ns)), bool short)         { if(short) return nameToStr(ns[size(ns)-1]); else return strJoin([nameToStr(n, short) | n <- ns], "."); }
public str nameToStr(Unknown(), bool short)              { return "?"; }
public str nameToStr(SpecialType(Axiom()), bool short)         { return "\\axiom"; }
public str nameToStr(SpecialType(Predicate()), bool short)     { return "\\predicate"; }
public str nameToStr(SpecialType(Test()), bool short)     { return "\\test"; }
public str nameToStr(SpecialType(Renaming()), bool short)     { return "\\renaming"; }
public str nameToStr(SpecialType(Literal()), bool short)     { return "\\literal"; }
public str nameToStr(SpecialType(Decl()), bool short)     { return "\\decl"; }
public str nameToStr(SpecialType(DeclList()), bool short)     { return "\\decllist"; }
public str nameToStr(SpecialType(Expr(t)), bool short)     { return "\\expr[<nameToStr(t, short)>]"; }
public str nameToStr(AST tree) {	return nameToStr(tree, false); }
public str shortNameToStr(tree) {	return nameToStr(tree, true); }

private default str procParamToStr(AST tree, bool short) { throw InternalError("funParamToStr(): Unknown constructor <consIdOf(tree)>", tree); }
private str procParamToStr(AnonParam(m, t), bool short) { return "<modeToStr(m)> :<nameToStr(t, short)>"; }
private str procParamToStr(Param(m, n, t), bool short)  { return "<modeToStr(m)> <nameToStr(n, short)> : <nameToStr(t, short)>"; }
private default str funParamToStr(AST tree, bool short) { throw InternalError("funParamToStr(): Unknown constructor <consIdOf(tree)>", tree); }
private str funParamToStr(AnonParam(m, t), bool short) { return ":<nameToStr(t, short)>"; }
private str funParamToStr(Param(m, n, t), bool short)  { return "<nameToStr(n, short)> : <nameToStr(t, short)>"; }

public str modeToStr(Obs()) { return "obs"; }
public str modeToStr(Upd()) { return "upd"; }
public str modeToStr(Out()) { return "out"; }
public str modeToStr(Exp()) { return "exp"; }
public str modeToStr(Giv()) { return "giv"; }
public str modeToStr(Del()) { return "del"; }
public str modeToStr(Nrm()) { return "nrm"; }


public AST strToName(str name) {
	if(/\./ := name) {
		return QName(seq([Name(leaf(n)) | n <- strSplit(name, ".")]));
	}
	else {
		return Name(leaf(name));
	}
}

@doc{Return the name of the definition/declaration.}
public default AST defNameOf(AST tree) {
	throw InternalError("defNameOf(): Unknown constructor <consIdOf(tree)>", tree);
}
public AST defNameOf(Define(d,_,_))		{ return defNameOf(d); }
public AST defNameOf(TypeClause(n))			{ return n; }
public AST defNameOf(LiteralClause(n,_))	{ return n; }
public AST defNameOf(FunClause(n, _, _))	{ return n; }
public AST defNameOf(PredClauseSugar(n, _))		{ return n; }
public AST defNameOf(ProcClause(n, _))		{ return n; }
public AST defNameOf(AxiomClauseSugar(n, _))		{ return n; }
public AST defNameOf(AlertClause(n, _))		{ return n; }
public AST defNameOf(n:Name(leaf(_)))		{ return n; }
public AST defNameOf(n:IdName(leaf(_), leaf(_)))		{ return n; }
public AST defNameOf(n:TypeName(_))			{ return n; }
public AST defNameOf(n:Instance(idName,_))		{ return idName; }
public AST defNameOf(n:FunName(_, _, _))	{ return n; }
public AST defNameOf(n:ProcName(_, _))		{ return n; }
public AST defNameOf(n:VarName(_, _))		{ return n; }
public AST defNameOf(n:TopName(_,_))	    { return n; }
public AST defNameOf(n:AlertName(_))	    { return n; }
public AST defNameOf(AnonParam(m, t))		{ return Unknown(); }
public AST defNameOf(n:QName(_))			{ return n; }
public AST defNameOf(n:Unknown())			{ return n; }
public AST defNameOf(n:SpecialType(_))   	{ return n; }
public AST defNameOf(Var(n))                { return n; }
public AST defNameOf(Type(n))                { return n; }

public AST renameDef(d:Define(dcl,_,_), n)		{ return d[arg0=renameDef(dcl, n)]; }
public AST renameDef(dcl:TypeClause(_), n)		{ return dcl[arg0=n]; }
public AST renameDef(dcl:FunClause(_,_,_), n)	{ return dcl[arg0=n]; }
public AST renameDef(dcl:ProcClause(_,_), n)	{ return dcl[arg0=n]; }
public AST renameDef(dcl:AlertClause(_,_), n)	{ return dcl[arg0=n]; }
   
public default list[AST] paramsOf(AST name) {
	throw InternalError("paramsOf(): name doesn\'t have parameters \'<yieldTerm(name)>\'", name);
}

public list[AST] paramsOf(TypeName(n)) {
	return [];
}

public list[AST] paramsOf(SpecialType(n)) {
	return [];
}

public list[AST] paramsOf(AlertName(n)) {
	return [];
}

public list[AST] paramsOf(FunName(n,seq(ps),rt)) {
	return ps;
}
public list[AST] paramsOf(ProcName(n,seq(ps))) {
	return ps;
}

public default AST returnTypeOf(AST name) {
	throw InternalError("returnTypeOf(): name doesn\'t have a return type <yieldTerm(name)>", name);
}

public AST returnTypeOf(FunName(n,seq(ps),rt)) {
	return rt;
}

@doc{Return the name of the given construct, formatted for display / debug.}
public default AST nameToDcl(AST tree) {
			throw InternalError("nameToDcl(): Unknown constructor <consIdOf(tree)>", tree);
}
// public str nameToDcl(ProcClause(n, _))       { return nameToDcl(n); }
public AST nameToDcl(x:TypeName(n))          { return TypeClause(x); }
public AST nameToDcl(x:AlertName(n))         { return AlertClause(x); }
public AST nameToDcl(x:FunName(n, ps, rt))   { return FunClause(x, ParamList(ps), rt); }
public AST nameToDcl(x:ProcName(n, ps))      { return ProcClause(x, ParamList(ps)); }
public AST nameToDcl(x:VarName(n, t))        { return VarClause(x, t); }
public AST nameToDcl(x:SpecialType(_))       { return TypeClause(x); }
