/**************************************************************************
 * Copyright (c) 2014 Anya Helene Bagge
 * Copyright (c) 2014 University of Bergen
 * 
 * All rights reserved. This program and the accompanying materials
 * are made available under the terms of the Eclipse Public License v1.0
 * which accompanies this distribution, and is available at
 * http://www.eclipse.org/legal/epl-v10.html
 * 
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version. See http://www.gnu.org/licenses/
 * 
 * 
 * See the file COPYRIGHT for more information.
 * 
 * Contributors:
 * * Anya Helene Bagge
 * 
 *************************************************************************/
package org.magnolialang.magnolia.resources;

import java.util.List;

import org.rascalmpl.value.IConstructor;

public interface IDefinition extends IMagnoliaCodeUnit {
	public static final String ATTR_DOC = "Doc";


	/**
	 * Guaranteed not to contain spaces, and to be equal to the id-string of any
	 * other definition with the same overload signature. May contain
	 * parentheses, commas and operator symbols.
	 * 
	 * @return A string uniquely identifying this definition within its scope
	 */
	public String toIdString();


	/**
	 * @param name
	 *            Name of the attribute, e.g., Foo/2
	 * @return Value of the attribute, e.g., Foo(..., ...)
	 * @throws IllegalArgumentException
	 *             if the attribute does not exist
	 */
	IConstructor getAttribute(String name);


	/**
	 * @return Base name, as a term (e.g., qualified but
	 *         without
	 *         FunName(...))
	 */
	IConstructor getBaseName();


	/**
	 * @return Body of this definition
	 */
	IConstructor getBody();


	/**
	 * @return Definition clause of this definition, e.g.
	 *         FunClause(name,paramList,returnType)
	 */
	IConstructor getDefClause();


	/**
	 * @return Number of parameters (zero if def has no parameters)
	 */
	int getNumParams();


	/**
	 * @param i
	 * @return Mode of parameter i, e.g., Obs()
	 * @throws IndexOutOfBoundsException
	 *             if no such parameter exists
	 * @throws UnsupportedOperationException
	 *             if definition does not have parameters
	 */
	IConstructor getParamMode(int i);


	/**
	 * @param i
	 * @return Name of parameter i, e.g. Name(leaf("foo"))
	 * @throws IndexOutOfBoundsException
	 *             if no such parameter exists
	 * @throws UnsupportedOperationException
	 *             if definition does not have parameters
	 */
	IConstructor getParamName(int i);


	/**
	 * @return Parameter list (unmodifiable)
	 * @throws UnsupportedOperationException
	 *             if definition does not have parameters
	 */
	List<IConstructor> getParams();


	/**
	 * @param i
	 * @return Type of parameter i
	 * @throws IndexOutOfBoundsException
	 *             if no such parameter exists
	 * @throws UnsupportedOperationException
	 *             if definition does not have parameters
	 */
	IConstructor getParamType(int i);


	/**
	 * @return Unqualified basename of the definition (as a term)
	 */
	IConstructor getUnqualifiedName();


	/**
	 * Check if a definition has an attribute.
	 * 
	 * Attribute name should be of the form 'Name/N' where N is the number of
	 * parameters in the attribute constructor.
	 * 
	 * @param name
	 *            Attribute name
	 * @return True if definition has the given attribute
	 */
	boolean hasAttribute(String name);


	/**
	 * Check if a definition has a modifier.
	 * 
	 * A modifier is a nullary attribute. The name is given without arity.
	 * 
	 * @param name
	 *            Modifier name
	 * @return True if definition has the given modifier
	 * @see {@link #hasAttribute(String)}
	 */
	boolean hasModifier(String name);


	/**
	 * @return True if the definition has parameters (even though the parameter
	 *         list may be empty)
	 */
	boolean hasParams();
}
