/**************************************************************************
 * Copyright (c) 2012-2014 Anya Helene Bagge
 * Copyright (c) 2012-2014 University of Bergen
 *
 * All rights reserved. This program and the accompanying materials
 * are made available under the terms of the Eclipse Public License v1.0
 * which accompanies this distribution, and is available at
 * http://www.eclipse.org/legal/epl-v10.html
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version. See http://www.gnu.org/licenses/
 *
 *
 * See the file COPYRIGHT for more information.
 *
 * Contributors:
 * * Anya Helene Bagge
 *
 *************************************************************************/
package org.magnolialang.magnolia.resources;

import java.io.IOException;
import java.io.InputStream;
import java.io.OutputStream;
import java.net.URI;
import java.net.URISyntaxException;
import java.util.ArrayList;
import java.util.Collection;
import java.util.HashSet;
import java.util.List;
import java.util.Set;
import java.util.concurrent.Callable;
import java.util.concurrent.ExecutionException;
import java.util.concurrent.ExecutorService;
import java.util.concurrent.Future;
import java.util.concurrent.atomic.AtomicReference;

import org.rascalmpl.value.IConstructor;
import org.rascalmpl.value.IList;
import org.rascalmpl.value.ISet;
import org.rascalmpl.value.ISourceLocation;
import org.rascalmpl.value.ITuple;
import org.rascalmpl.value.IValue;
import org.eclipse.jdt.annotation.Nullable;
import org.magnolialang.infra.Infra;
import org.magnolialang.magnolia.Magnolia;
import org.magnolialang.magnolia.resources.internal.AlgoDefinition;
import org.magnolialang.magnolia.resources.internal.DefinitionData;
import org.nuthatchery.pica.errors.CancelledException;
import org.nuthatchery.pica.errors.ImplementationError;
import org.nuthatchery.pica.errors.Severity;
import org.nuthatchery.pica.rascal.RascalUtil;
import org.nuthatchery.pica.resources.IFact;
import org.nuthatchery.pica.resources.ILanguage;
import org.nuthatchery.pica.resources.ISerializer;
import org.nuthatchery.pica.resources.IXRefInfo;
import org.nuthatchery.pica.resources.handles.IFileHandle;
import org.nuthatchery.pica.resources.internal.AbstractManagedResource;
import org.nuthatchery.pica.resources.internal.facts.GenericFact;
import org.nuthatchery.pica.resources.managed.IManagedCodeUnit;
import org.nuthatchery.pica.resources.managed.IManagedResource;
import org.nuthatchery.pica.resources.regions.ICodeRegion;
import org.nuthatchery.pica.resources.regions.IOffsetLength;
import org.nuthatchery.pica.resources.regions.OffsetLength;
import org.nuthatchery.pica.resources.storage.IStorage;
import org.nuthatchery.pica.tasks.ITaskMonitor;
import org.nuthatchery.pica.terms.TermAdapter;
import org.nuthatchery.pica.terms.TermFactory;
import org.nuthatchery.pica.terms.TermLocator;
import org.nuthatchery.pica.util.ISignature;
import org.nuthatchery.pica.util.Pair;
import org.nuthatchery.pica.util.Quadruple;
import org.nuthatchery.pica.util.Signature;
import org.nuthatchery.pica.util.Triple;
import org.nuthatchery.pica.util.depgraph.ITopologicalWorkQueue;
import org.nuthatchery.pica.util.depgraph.IWritableDepGraph;
import org.nuthatchery.pica.util.depgraph.UnsyncedDepGraph;
import org.rascalmpl.interpreter.control_exceptions.Throw;
import org.rascalmpl.uri.URIUtil;

public class MagnoliaModule extends AbstractMagnoliaManagedResource implements IMagnoliaModule {
	protected final static String MARK_SOURCE = MagnoliaModule.class.getName();
	protected final static String MARK_SOURCE_TYPECHECK = MARK_SOURCE + "#typecheck";
	protected final static String MARK_SOURCE_TOP_DEPENDS = MARK_SOURCE + "#topDepends";
	protected final static String MARK_SOURCE_CHECK_DEPS = MARK_SOURCE + "#checkDepends";

	private final IConstructor defInfo;
	private final DefinitionData defData;
	private final IFact<Quadruple<IConstructor, Collection<MagnoliaModule>, ISet, ISet>> nameresFact;
	private final IFact<Triple<IConstructor, Collection<IDefinition>, ISet>> typecheckFact;
	private final ISignature signature;

	/**
	 * Held while doing full typechecking (e.g., from getTypechecked()).
	 */
	private final Object typecheckLock = new Object();


	public MagnoliaModule(IFileHandle res, DefinitionData defData) {
		super(defData.getURI(), res);
		this.defData = defData;
		this.defInfo = defData.getDefInfo(); // .set("resource", this);
		this.nameresFact = new GenericFact<>("nameres");
		this.typecheckFact = new GenericFact<>("typecheck");
		this.signature = new Signature(System.identityHashCode(this));
	}


	@Override
	public boolean containsOffset(int offset) {
		return defData.containsOffset(offset);
	}


	/* (non-Javadoc)
	 * @see java.lang.Object#equals(java.lang.Object)
	 */
	@Override
	public boolean equals(@Nullable Object obj) {
		if(this == obj) {
			return true;
		}
		if(obj == null) {
			return false;
		}
		if(getClass() != obj.getClass()) {
			return false;
		}
		MagnoliaModule other = (MagnoliaModule) obj;
		if(!defData.equals(other.defData)) {
			return false;
		}
		return true;
	}


	/**
	 * The desugared but otherwise unprocessed AST.
	 *
	 * @see org.nuthatchery.pica.resources.managed.IManagedCodeUnit#getAST(org.rascalmpl.debug.ITaskMonitor)
	 */
	@Override
	public IConstructor getAST(ITaskMonitor tm) {
		return defData.getAST();
	}


	@Override
	public IConstructor getAttribute(String name) {
		return defData.getAttribute(name);
	}


	@Override
	public IConstructor getBaseName() {
		return defData.getBaseName();
	}


	@Override
	public IConstructor getBody() {
		return defData.getBody();
	}


	@Override
	public Collection<IDefinition> getChildren(ITaskMonitor tm) {
		return typecheckFact(tm).getSecond();
	}


	@Override
	public IConstructor getDefClause() {
		return defData.getDefClause();
	}


	@Override
	public IConstructor getDefInfo(ITaskMonitor tm) {
		return defInfo;
	}


	@Override
	public Collection<MagnoliaModule> getDepends(ITaskMonitor tm) {
		return nameresFact(tm).getSecond();
	}


	@Override
	public ISet getExternals(ITaskMonitor tm) {
		return nameresFact(tm).getThird();
	}


	@Override
	public String getFragmentName() {
		throw new UnsupportedOperationException();
	}


	@Override
	public IConstructor getId() {
		return defData.getId();
	}


	@Override
	public String getKind(ITaskMonitor tm) {
		return defData.getKind();
	}


	@Override
	public ILanguage getLanguage() {
		return getParent().getLanguage();
	}


	@Override
	public long getLength() {
		return defData.getLength();
	}


	@Override
	public String getName() {
		return defData.getFullName();
	}


	@Override
	public ISourceLocation getNameLoc() {
		return defData.getNameLoc();
	}


	@Override
	public int getNumParams() {
		return 0;
	}


	@Override
	public long getOffset() throws UnsupportedOperationException {
		return defData.getOffset();
	}


	@Override
	public IConstructor getParamMode(int i) {
		throw new UnsupportedOperationException();
	}


	@Override
	public IConstructor getParamName(int i) {
		throw new UnsupportedOperationException();
	}


	@Override
	public List<IConstructor> getParams() {
		throw new UnsupportedOperationException();
	}


	@Override
	public IConstructor getParamType(int i) {
		throw new UnsupportedOperationException();
	}


	@Override
	public IMagnoliaPackage getParent() {
		return (IMagnoliaPackage) defData.getParent();
	}


	@Override
	public ISignature getSourceSignature(ITaskMonitor tm) {
		return signature;
	}


	@Override
	public ISourceLocation getSrcLoc() {
		return defData.getSrcLoc();
	}


	@Override
	@Nullable
	public IStorage getStorage() {
		return null;
	}


	@Override
	public Collection<MagnoliaModule> getTransitiveDepends(ITaskMonitor tm) {
		Set<MagnoliaModule> depends = new HashSet<MagnoliaModule>();
		List<MagnoliaModule> todo = new ArrayList<MagnoliaModule>(getDepends(tm));
		while(!todo.isEmpty()) {
			MagnoliaModule mod = todo.remove(0);
			depends.add(mod);
			for(IManagedCodeUnit m : mod.getDepends(tm)) {
				if(m != this && !depends.contains(m)) {
					todo.add((MagnoliaModule) m);
				}
			}
		}
		return depends;
	}


	@Override
	public IConstructor getTypechecked(final ITaskMonitor tm) {
		return typecheckFact(tm).getFirst();
	}


	@Override
	public IConstructor getUnqualifiedName() {
		return defData.getUnqualifiedName();
	}


	@Override
	public Collection<Pair<IOffsetLength, Object>> getXRefs(ITaskMonitor tm) {
		ArrayList<Pair<IOffsetLength, Object>> result = new ArrayList<>();
		ISet xrefs = nameresFact(tm).getFourth();
		for(IValue v : xrefs) {
			ITuple t = (ITuple) v;
			ISourceLocation loc = (ISourceLocation) t.get(0);
			result.add(new Pair<>(RascalUtil.toCodeRegion(loc), t.get(1)));
		}

		xrefs = typecheckFact(tm).getThird();
		for(IValue v : xrefs) {
			ITuple t = (ITuple) v;
			ISourceLocation loc = (ISourceLocation) t.get(0);

			result.add(new Pair<>(RascalUtil.toCodeRegion(loc), t.get(1)));
		}
		return result;
	}


	@Override
	@Nullable
	public IXRefInfo getXRefs(IOffsetLength ol, ITaskMonitor tm) {
		// Check if it's our own name
		ISourceLocation loc;
		try {
			loc = RascalUtil.toSourceLocation(URIUtil.changeFragment(uri, null), ol);
			IConstructor name = (IConstructor) defInfo.get("defName");
			IList bindings = null;
			IConstructor found = TermLocator.locateName(name, loc);
			if(found != null) {
				return XRefInfo.makeFromResource(this);
			}

			// Perhaps it's an attribute?
			// ...

			// Or maybe it's in our pre-flattened body
			found = TermLocator.locateName(nameresFact(tm).getFirst(), loc);
			IConstructor env = getParent().getPackageEnv(tm);
			if(found != null) {
				try {
					IConstructor info = (IConstructor) Magnolia.getInstance().getCompiler().call(tm, "lookupDef", found, env);
					if(info.get("resource") instanceof IManagedCodeUnit) {
						return XRefInfo.makeFromResource((IMagnoliaCodeUnit) info.get("resource"));
					}

					return XRefInfo.makeFromDefInfo(info, bindings);
				}
				catch(Throw e) {
				}
			}

			// Finally, check if it is in our typechecked body
			IConstructor tc = typecheckFact(tm).getFirst();
			found = TermLocator.locateName(tc, loc);

			if(found != null) {
				System.err.println("getXRefs Found: " + TermAdapter.yieldTerm(found, false));
				env = (IConstructor) Magnolia.getInstance().getCompiler().call(tm, "loadDefs", tc, env);
				if(found.getName().equals("Instance")) {
					bindings = (IList) ((IConstructor) found.get(1)).get(0);
					found = (IConstructor) found.get(0);
				}
				try {
					IConstructor info = (IConstructor) Magnolia.getInstance().getCompiler().call(tm, "lookupDef", found, env);
					return XRefInfo.makeFromDefInfo(info, bindings);
				}
				catch(Throw e) {
					if(found.asAnnotatable().hasAnnotation("hoverInfo")) {
						return XRefInfo.makeFromHoverInfo((IConstructor) found.asAnnotatable().getAnnotation("hoverInfo"));
					}
					System.err.println("Not found in environment: " + TermAdapter.yieldTerm(found, false));
					return XRefInfo.makeFromName(found);
				}

				// need to look it up in our own body
			}
		}
		catch(URISyntaxException e1) {
			// TODO Auto-generated catch block
			e1.printStackTrace();
		}
		return null;
	}


	@Override
	public boolean hasAttribute(String name) {
		return defData.hasAttribute(name);
	}


	/* (non-Javadoc)
	 * @see java.lang.Object#hashCode()
	 */
	@Override
	public int hashCode() {
		final int prime = 31;
		int result = 1;
		result = prime * result + ((defData == null) ? 0 : defData.hashCode());
		return result;
	}


	@Override
	public boolean hasIncompleteDepends(ITaskMonitor tm) {
		return false;
	}


	@Override
	public boolean hasModifier(String name) {
		return defData.hasModifier(name);
	}


	@Override
	public boolean hasParams() {
		return false;
	}


	@Override
	public boolean isCodeUnit() {
		return true;
	}


	@Override
	public boolean isFragment() {
		return true;
	}


	@Override
	public boolean isRoot() {
		return false;
	}


	@Override
	public void onDependencyChanged() {
		nameresFact.dispose();
		typecheckFact.dispose();
	}


	@Override
	public void onResourceChanged() {
		throw new UnsupportedOperationException();
	}


	@Override
	public String toIdString() {
		return defData.toIdString();
	}


	@Override
	public String toString() {
		return defData.toString();
	}


	/**
	 * Typecheck this module.
	 *
	 * @param env
	 *            An environment in which all dependencies of this module has
	 *            been flattened and typechecked
	 * @param tm
	 *            A progress monitor
	 * @return The flattened and typechecked AST of the module
	 */
	@Override
	public Triple<IConstructor, Collection<IDefinition>, ISet> typecheck(IConstructor env, ITaskMonitor tm) {
		synchronized(typecheckFact) {
			Triple<IConstructor, Collection<IDefinition>, ISet> fact = typecheckFact.getValue(getSourceSignature(tm));
			if(fact != null) {
				return fact;
			}
			IConstructor info = defInfo.set("body", nameresFact(tm).getFirst());
			IConstructor result = (IConstructor) Magnolia.getInstance().getCompiler().call(tm, "flatten", info, env);
			getParent().clearMarks(MARK_SOURCE_TYPECHECK, this);
			getParent().addMarks(result, MARK_SOURCE_TYPECHECK, this);
			IConstructor typechecked;
			ISet typecheckedXRefs;
			Collection<IDefinition> children = new ArrayList<>();

			if(result.has("val")) {
				typechecked = (IConstructor) result.get("val");
				typecheckedXRefs = (ISet) result.get("xrefs");

				if(TermAdapter.isCons(typechecked, "DeclBody", 1)) {
					IList defs = (IList) TermAdapter.getArg(typechecked, 0).get("args");

					for(IValue d : defs) {
						IConstructor def = (IConstructor) d;
						assert def != null;
						if(TermAdapter.isCons(def, "Define", 3)) {
							DefinitionData data = new DefinitionData(def, this);
							children.add(new AlgoDefinition((IFileHandle) resource, data));
							//System.err.println(data);
						}
					}
				}
			}
			else {
				System.err.println("Typecheck failed: " + getName());
				System.err.println("  result: " + result);
				typechecked = TermFactory.cons("DeclBody", TermFactory.seq());
				typecheckedXRefs = TermFactory.vf.set();
			}

			fact = new Triple<>(typechecked, children, typecheckedXRefs);

			typecheckFact.setValue(fact, getSourceSignature(tm));
			return fact;
		}
	}


	/**
	 * @param tm
	 * @return (Resolved AST, dependencies, externals, nameresolvedxrefs)
	 */
	private Quadruple<IConstructor, Collection<MagnoliaModule>, ISet, ISet> nameresFact(ITaskMonitor tm) {
		synchronized(nameresFact) {
			Quadruple<IConstructor, Collection<MagnoliaModule>, ISet, ISet> fact = nameresFact.getValue(getSourceSignature(tm));
			if(fact != null) {
				return fact;
			}

			Collection<MagnoliaModule> depends = new HashSet<MagnoliaModule>();
			IConstructor env = getParent().getPackageEnv(tm);

			IConstructor result = (IConstructor) Magnolia.getInstance().getCompiler().call(tm, "getTopDepends", defInfo.get("body"), getId(), env);
			getParent().clearMarks(MARK_SOURCE_TOP_DEPENDS, this);
			getParent().addMarks(result, MARK_SOURCE_TOP_DEPENDS, this);

			ITuple val = (ITuple) result.get("val");
			ISet deps = (ISet) val.get(0);
			IConstructor nameResolved = (IConstructor) val.get(1);
			for(IValue v : deps) {
				IConstructor qn = Magnolia.basename((IConstructor) v);
				IConstructor pkg = Magnolia.qualifierOf(qn);

				IManagedCodeUnit pkgUnit = getParent().getManager().findCodeUnit(getLanguage(), pkg);
				if(pkgUnit instanceof IMagnoliaPackage) {
					MagnoliaModule m = (MagnoliaModule) ((IMagnoliaPackage) pkgUnit).getChild(v, tm);
					depends.add(m);
				}
				else {
					Infra.get().logMessage("Not a MagnoliaModule: " + v);
				}
			}
			ISet externals = (ISet) val.get(2);
			ISet nameResolvedXRefs = (ISet) result.get("xrefs");
			fact = new Quadruple<>(nameResolved, depends, externals, nameResolvedXRefs);

			nameresFact.setValue(fact, getSourceSignature(tm));
			return fact;
		}
	}


	private Triple<IConstructor, Collection<IDefinition>, ISet> typecheckFact(final ITaskMonitor tm) {
		synchronized(typecheckLock) {
			synchronized(typecheckFact) {
				Triple<IConstructor, Collection<IDefinition>, ISet> fact = typecheckFact.getValue(getSourceSignature(tm));
				if(fact != null) {
					return fact;
				}
			}

			// build a dependency graph of modules
			Collection<MagnoliaModule> directDeps = getDepends(tm);
			Set<MagnoliaModule> allDeps = new HashSet<>();

			// check for dependency cycles
			// (we do this in an extra step so we can give better error message)
			for(MagnoliaModule m : directDeps) {
				Collection<MagnoliaModule> transitiveDepends = m.getTransitiveDepends(tm);
				if(!transitiveDepends.contains(this) && m != this) {
					allDeps.add(m);
					allDeps.addAll(transitiveDepends);
				}
				else {// avoid cycles
					getParent().getManager().addMark("Module " + getName() + " depends on " + m.getName() + " which has a circular dependency on " + getName(), RascalUtil.toCodeRegion(getNameLoc()), Severity.ERROR, MARK_SOURCE_CHECK_DEPS, getURI());
				}
			}

			// build dependency graph
			IWritableDepGraph<MagnoliaModule> depGraph = new UnsyncedDepGraph<MagnoliaModule>();
			depGraph.add(this);

			for(MagnoliaModule m : allDeps) {
				depGraph.add(m);
				depGraph.add(this, m);

				for(MagnoliaModule d : m.getDepends(tm)) {
					depGraph.add(m, d);
				}
			}

			final AtomicReference<IConstructor> envHolder = new AtomicReference<>(getParent().getPackageEnv(tm));

			final ITopologicalWorkQueue<MagnoliaModule> queue = depGraph.topologicalWorkQueue();
			MagnoliaModule m = queue.waitForNext();
			ExecutorService threadPool = this.getParent().getManager().getResourceManager().getThreadPool();
			List<Future<Triple<IConstructor, Collection<IDefinition>, ISet>>> tasks = new ArrayList<>();
			int resultPos = -1;
			while(m != null) {
				//System.err.println("Flattening " + m.getName());
				final MagnoliaModule module = m;
				if(m == this) {
					resultPos = tasks.size();// position in list of result for this module
				}
				tasks.add(threadPool.submit(new Callable<Triple<IConstructor, Collection<IDefinition>, ISet>>() {
					@Override
					public Triple<IConstructor, Collection<IDefinition>, ISet> call() {
						try {
							IConstructor info = module.getDefInfo(tm);
							Triple<IConstructor, Collection<IDefinition>, ISet> result = module.typecheck(envHolder.get(), tm);
							IConstructor body = result.getFirst();
							if(body != null) {
								info = info.set("body", body);

								// need to update the environment with the typechecked definition
								// since envHolder can be changed from other threads, we do compareAndSet to make sure
								// that we retry in case the environment was changed while we were attempting to update
								IConstructor env, newEnv;
								do {
									env = envHolder.get();
									newEnv = (IConstructor) Magnolia.getInstance().getCompiler().call(tm, "replaceDef", module.getId(), info, env);
								}
								while(!envHolder.compareAndSet(env, newEnv));
							}
							return result;
						}
						finally {
							queue.done(module);
						}
					}
				}));
				m = queue.waitForNext();// will wait when not all dependencies have been processed
			}

			for(Future<Triple<IConstructor, Collection<IDefinition>, ISet>> task : tasks) {// wait for all tasks to finish
				try {
					task.get();
				}
				catch(InterruptedException e) {
					throw new CancelledException("Interrupted while waiting for task to finish", e);
				}
				catch(ExecutionException e) {
					e.printStackTrace();
					Throwable cause = e.getCause();
					if(cause instanceof RuntimeException) {
						throw (RuntimeException) cause;
					}
					else {
						throw new ImplementationError("Something went wrong...", e);
					}
				}
			}
			if(resultPos != -1) {
				Triple<IConstructor, Collection<IDefinition>, ISet> result;
				try {
					result = tasks.get(resultPos).get();
				}
				catch(InterruptedException e) {
					throw new CancelledException("While typechecking: ", e);
				}
				catch(ExecutionException e) {
					Throwable cause = e.getCause();
					if(cause instanceof RuntimeException) {
						throw (RuntimeException) cause;
					}
					else {
						throw new ImplementationError("Exception caught in typechecking thread", e);
					}
				}
				if(result != null) {
					return result;
				}
				else {
					throw new ImplementationError("While typechecking: got null result from subtask");
				}
			}
			else {
				throw new ImplementationError("While typechecking: subtask not completed");
			}
		}
	}


	@Override
	public ISignature getFullSignature(ITaskMonitor tm) {
		// TODO Auto-generated method stub
		return null;
	}


	@Override
	@Nullable
	public IManagedResource getChild(Object childId, ITaskMonitor rm) {
		throw new UnsupportedOperationException();
	}


	@Override
	public ICodeRegion<URI> getNameRegion() {
		return RascalUtil.toCodeRegion(getNameLoc());
	}


	@Override
	public ICodeRegion<URI> getSrcRegion() {
		return RascalUtil.toCodeRegion(getSrcLoc());
	}
}

class ModuleSerializer implements ISerializer<IMagnoliaModule> {

	public ModuleSerializer() {

	}


	@Override
	public IMagnoliaModule read(InputStream stream) throws IOException {
		throw new UnsupportedOperationException();
	}


	@Override
	public void write(IMagnoliaModule data, OutputStream stream) throws IOException {
		// throw new UnsupportedOperationException();
	}

}
