/**************************************************************************
 * Copyright (c) 2012-2014 Anya Helene Bagge
 * Copyright (c) 2012-2014 Tero Hasu
 * Copyright (c) 2012-2014 University of Bergen
 *
 * All rights reserved. This program and the accompanying materials
 * are made available under the terms of the Eclipse Public License v1.0
 * which accompanies this distribution, and is available at
 * http://www.eclipse.org/legal/epl-v10.html
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version. See http://www.gnu.org/licenses/
 *
 *
 * See the file COPYRIGHT for more information.
 *
 * Contributors:
 * * Anya Helene Bagge
 * * Tero Hasu
 *
 *************************************************************************/
package org.magnolialang.magnolia.resources;

import static org.nuthatchery.pica.errors.Severity.ERROR;
import static org.nuthatchery.pica.terms.TermFactory.vf;

import java.io.IOException;
import java.io.InputStream;
import java.net.URI;
import java.security.MessageDigest;
import java.util.ArrayList;
import java.util.Collection;
import java.util.Collections;
import java.util.HashSet;
import java.util.List;
import java.util.Set;
import java.util.concurrent.atomic.AtomicInteger;

import org.rascalmpl.value.IAnnotatable;
import org.rascalmpl.value.IConstructor;
import org.rascalmpl.value.IListWriter;
import org.rascalmpl.value.ISet;
import org.rascalmpl.value.ISetWriter;
import org.rascalmpl.value.ISourceLocation;
import org.rascalmpl.value.IString;
import org.rascalmpl.value.ITuple;
import org.rascalmpl.value.IValue;
import org.rascalmpl.value.type.Type;
import org.rascalmpl.value.visitors.IValueVisitor;
import org.eclipse.jdt.annotation.Nullable;
import org.magnolialang.magnolia.Magnolia;
import org.magnolialang.magnolia.MagnoliaFacts;
import org.magnolialang.magnolia.frontend.MagnoliaLoader;
import org.magnolialang.magnolia.resources.internal.DefinitionData;
import org.nuthatchery.pica.errors.ErrorMarkers;
import org.nuthatchery.pica.errors.Severity;
import org.nuthatchery.pica.rascal.ICompiler;
import org.nuthatchery.pica.rascal.RascalUtil;
import org.nuthatchery.pica.rascal.ValuesFactFactory;
import org.nuthatchery.pica.resources.IFact;
import org.nuthatchery.pica.resources.ILanguage;
import org.nuthatchery.pica.resources.IProjectManager;
import org.nuthatchery.pica.resources.IResourceRegion;
import org.nuthatchery.pica.resources.IXRefInfo;
import org.nuthatchery.pica.resources.ResourceRegion;
import org.nuthatchery.pica.resources.handles.IFileHandle;
import org.nuthatchery.pica.resources.internal.AbstractManagedResource;
import org.nuthatchery.pica.resources.internal.facts.GenericFact;
import org.nuthatchery.pica.resources.managed.IManagedCodeUnit;
import org.nuthatchery.pica.resources.managed.IManagedContainer;
import org.nuthatchery.pica.resources.managed.IManagedResource;
import org.nuthatchery.pica.resources.marks.IMark;
import org.nuthatchery.pica.resources.marks.MarkBuilder;
import org.nuthatchery.pica.resources.regions.ICodeRegion;
import org.nuthatchery.pica.resources.regions.IOffsetLength;
import org.nuthatchery.pica.resources.storage.IStorage;
import org.nuthatchery.pica.tasks.ITaskMonitor;
import org.nuthatchery.pica.tasks.NullTaskMonitor;
import org.nuthatchery.pica.terms.TermAdapter;
import org.nuthatchery.pica.terms.TermFactory;
import org.nuthatchery.pica.terms.TermImploder;
import org.nuthatchery.pica.terms.TermLocator;
import org.nuthatchery.pica.util.Hash;
import org.nuthatchery.pica.util.ISignature;
import org.nuthatchery.pica.util.NullablePair;
import org.nuthatchery.pica.util.NullnessHelper;
import org.nuthatchery.pica.util.Pair;
import org.nuthatchery.pica.util.Signature;
import org.nuthatchery.pica.util.Triple;
import org.rascalmpl.interpreter.NullRascalMonitor;
import org.rascalmpl.parser.gtd.exception.ParseError;
import org.rascalmpl.values.uptr.ITree;

import edu.umd.cs.findbugs.annotations.SuppressFBWarnings;

public class MagnoliaPackage extends AbstractMagnoliaManagedResource implements IMagnoliaPackage {
	protected final static String MARK_SOURCE = MagnoliaPackage.class.getName();
	protected final static String MARK_SOURCE_READING_FILE = MARK_SOURCE + "#loadError";
	protected final static String MARK_SOURCE_PARSING = MARK_SOURCE + "#syntaxError";
	protected final static String MARK_SOURCE_GETPACKAGEENV = MARK_SOURCE + "#getPackageEnv";
	protected final static String MARK_SOURCE_RESOLVE_IMPORTS = MARK_SOURCE + "#resolveImports";

	protected final IConstructor id;
	protected final Magnolia lang = Magnolia.getInstance();
	protected final ICompiler compiler = lang.getCompiler();
	private final String name;

	/* Note on synchronization: Must synchronize on each fact before attempting to read its value. Note
	 * the prescribed order if multiple locks are needed. Be careful with calling methods on other
	 * objects while a lock is held. */
	/* Synchronization policy: All private methods should handle synchronization; no public method should synchronize. */

	/**
	 * Avoid accessing the fact fields directly; use the corresponding private
	 * method, which will deal with recomputation if necessary.
	 */
	/**
	 * No other fact methods are safe to call when locked on treeFact.
	 */
	private final IFact<IConstructor> treeFact;
	/**
	 * A pair of a DefInfo and a set of xrefs from the package header.
	 *
	 * Safe facts when synced on childrenFact: {@link #treeFact(ITaskMonitor)}
	 * .
	 */
	private final IFact<Pair<IConstructor, ISet>> pkgInfoFact;

	/**
	 * A triple of a) the packages we depend on b) xrefs from dependency
	 * resolution c) boolean indicating whether all dependencies were
	 * found.
	 *
	 * Safe facts when synced on loaderFact:
	 * {@link #pkgInfoFact(ITaskMonitor)}, {@link #treeFact(ITaskMonitor)}.
	 */
	private final IFact<Triple<Set<IMagnoliaPackage>, ISet, Boolean>> loaderFact;

	/**
	 * Safe facts when synced on childrenFact:
	 * {@link #pkgInfoFact(ITaskMonitor)}, {@link #treeFact(ITaskMonitor)}.
	 */
	private final IFact<Pair<Collection<IMagnoliaModule>, ISet>> childrenFact;
	/**
	 * Safe to call when synced on packageEnvFact:
	 * {@link #loaderFact(ITaskMonitor)},
	 * {@link #childrenFact(ITaskMonitor)}, etc.
	 */
	private final IFact<IConstructor> packageEnvFact;

	/**
	 * Safe to call when synced on packageEnvFact:
	 * {@link #packageEnvFact(ITaskMonitor)}, etc.
	 */
	private final IFact<IConstructor> typecheckedFact;
	@Nullable
	private volatile ISignature sourceSignature = null;

	@Nullable
	private volatile ISignature fullSignature = null;

	@Nullable
	private volatile ISignature depSignature = null;

	private final Object fullSignatureLock = new Object();
	private final Object depSignatureLock = new Object();
	private final IStorage storage;
	private final IProjectManager manager;
	private final AtomicInteger revision = new AtomicInteger(0);


	public MagnoliaPackage(IProjectManager manager, IFileHandle file, IStorage storage, IConstructor id, ILanguage lang) {
		super(file.getURI(), file);
		this.manager = manager;
		this.id = id;
		this.name = NullnessHelper.assertNonNull(lang.getNameString(id).intern());
		this.storage = storage;
		ValuesFactFactory factory = ValuesFactFactory.getFactory(storage);
		treeFact = factory.makeIConstructorFact("ast", TermFactory.Type_AST);
		childrenFact = new GenericFact<>("children");
		pkgInfoFact = new GenericFact<>("pkgInfo");// use a non-serializable fact for this
		packageEnvFact = factory.makeIConstructorFact("packageEnv");
		typecheckedFact = factory.makeIConstructorFact("typechecked", TermFactory.Type_AST);
		loaderFact = new GenericFact<>("loader");
	}


	@Override
	public void addMarks(IConstructor result, String source, IManagedCodeUnit context) {
		addMarks((ISet) result.get("errors"), source, context);
		addMarks((ISet) result.get("warnings"), source, context);
	}


	/** xxx Should this be a part of IMagnoliaPackage? */
	public void clearFacts() {
		treeFact.setValue(null, null);
		childrenFact.setValue(null, null);
		pkgInfoFact.setValue(null, null);
		packageEnvFact.setValue(null, null);
		typecheckedFact.setValue(null, null);
		loaderFact.setValue(null, null);
	}


	@Override
	public void clearMarks(String source, IManagedCodeUnit context) {
		manager.clearMarks(source, context.getURI());
	}


	@Override
	public boolean containsOffset(int offset) {
		try {
			return offset >= 0 && offset < getLength();
		}
		catch(IOException e) {
			e.printStackTrace();
			return false;
		}
	}


	@Override
	public IConstructor getAST(ITaskMonitor tm) {
		return treeFact(tm);
	}


	@Override
	@Nullable
	public IManagedCodeUnit getChild(Object childId, ITaskMonitor tm) {
		if(childId instanceof IConstructor) {
			Collection<IMagnoliaModule> cs = getChildren(tm);
			for(IMagnoliaModule c : cs) {
				if(c.getId().isEqual((IValue) childId)) {
					return c;
				}
			}
			return null;
		}
		else {
			throw new IllegalArgumentException();
		}
	}


	@Override
	public Collection<IMagnoliaModule> getChildren(ITaskMonitor tm) {
		return childrenFact(tm).getFirst();
	}


	@Override
	public IConstructor getDefInfo(ITaskMonitor tm) {
		return pkgInfoFact(tm).getFirst();
	}


	@Override
	public Collection<? extends IMagnoliaPackage> getDepends(ITaskMonitor tm) {
		return loaderFact(tm).getFirst();
	}


	/**
	 * Get a byte array containing a hash that identifies the current source
	 * code version of this package and all its dependencies.
	 *
	 * The return value must not be modified.
	 *
	 * @param tm
	 *            A monitor
	 * @return A hash of the package and its dependencies
	 */
	@Override
	public ISignature getFullSignature(ITaskMonitor tm) {
		// TODO: also include allDepsOk
		synchronized(fullSignatureLock) {
			ISignature sig = fullSignature;
			if(sig != null) {
				return sig;
			}
			MessageDigest md = Hash.getMessageDigest();
			getSourceSignature(tm).digest(md);
			Collection<? extends IManagedCodeUnit> depends = getTransitiveDepends(tm);
			for(IManagedCodeUnit d : depends) {
				d.getSourceSignature(tm).digest(md);
			}

			Triple<Set<IMagnoliaPackage>, ISet, Boolean> fact = loaderFact(tm);

			if(!fact.getThird()) {
				int r = revision.get();
				md.update((byte) (r & 0xff));
				md.update((byte) ((r >>> 8) & 0xff));
				md.update((byte) ((r >>> 16) & 0xff));
				md.update((byte) ((r >>> 24) & 0xff));
			}
			fullSignature = sig = new Signature(md.digest());
			return sig;
		}
	}


	@Override
	public IConstructor getId() {
		return id;
	}


	@Override
	public String getKind(ITaskMonitor tm) {
		return MagnoliaFacts.KIND_PACKAGE;
	}


	@Override
	public ILanguage getLanguage() {
		return lang;
	}


	@Override
	public IProjectManager getManager() {
		return manager;
	}


	@Override
	public String getName() {
		return name;
	}


	@Override
	public ISourceLocation getNameLoc() {
		return TermAdapter.locOf(id);
	}


	@Override
	public long getOffset() {
		return 0L;
	}


	/**
	 * Create an environment for this package, suitable for name resolution.
	 *
	 * @param tm
	 * @return The environment
	 */
	@Override
	public IConstructor getPackageEnv(ITaskMonitor tm) {
		return packageEnvFact(tm);
	}


	/** Delegates to managedFile. */
	@Override
	@Nullable
	public IManagedContainer getParent() {
		return null;
	}


	/**
	 * Get a byte array containing a hash that identifies the current source
	 * code of this package.
	 *
	 * The return value must not be modified.
	 *
	 * @param tm
	 *            A monitor
	 * @return A hash of the source code
	 */
	@Override
	public ISignature getSourceSignature(ITaskMonitor tm) {
		// TODO: synchronize
		ISignature sig = sourceSignature;
		if(sig != null) {
			return sig;
		}
		loadFile(tm);
		sig = sourceSignature;
		assert sig != null;
		return sig;
	}


	@Override
	public ISourceLocation getSrcLoc() {
		return TermFactory.vf.sourceLocation(uri);
	}


	@Override
	public IStorage getStorage() {
		return storage;
	}


	@Override
	@SuppressFBWarnings("UCF_USELESS_CONTROL_FLOW")
	public Collection<? extends IManagedCodeUnit> getTransitiveDepends(ITaskMonitor tm) {
		Set<IManagedCodeUnit> depends = new HashSet<IManagedCodeUnit>();
		List<IManagedCodeUnit> todo = new ArrayList<IManagedCodeUnit>(getDepends(tm));
		while(!todo.isEmpty()) {
			IManagedCodeUnit pkg = todo.remove(0);
			depends.add(pkg);
			for(IManagedCodeUnit p : pkg.getDepends(tm)) {
				if(p != this && !depends.contains(p)) {
					todo.add(p);
				}
				else if(p == this) {
					;// dependency cycle
				}
			}
		}
		return depends;
	}


	@Override
	public IConstructor getTypechecked(ITaskMonitor tm) {
		return typecheckedFact(tm);
	}


	@Override
	public Collection<Pair<IOffsetLength, Object>> getXRefs(ITaskMonitor tm) {
		ArrayList<Pair<IOffsetLength, Object>> result = new ArrayList<>();

		ISet xrefs = pkgInfoFact(tm).getSecond();
		xrefs = xrefs.union(loaderFact(tm).getSecond());

		for(IValue v : xrefs) {
			ITuple t = (ITuple) v;
			ISourceLocation loc = (ISourceLocation) t.get(0);

			result.add(new Pair<>(RascalUtil.toCodeRegion(loc), t.get(1)));
		}

		Collection<? extends IManagedCodeUnit> children = getChildren(tm);
		xrefs = childrenFact(tm).getSecond();
		if(xrefs != null && children != null) {
			for(IValue v : xrefs) {
				ITuple t = (ITuple) v;
				ISourceLocation loc = (ISourceLocation) t.get(0);

				result.add(new Pair<>(RascalUtil.toCodeRegion(loc), t.get(1)));
			}

			for(IManagedCodeUnit u : children) {
				result.addAll(u.getXRefs(tm));
			}
		}

		return result;
	}


	@Override
	@Nullable
	public IXRefInfo getXRefs(IOffsetLength reg, ITaskMonitor tm) {
		ISourceLocation loc = RascalUtil.toSourceLocation(getURI(), reg);

		for(IManagedCodeUnit child : getChildren(tm)) {
			if(TermLocator.isInside(loc, ((IMagnoliaModule) child).getSrcLoc())) {
				return child.getXRefs(reg, tm);
			}
		}
		return null;
	}


	@Override
	public boolean hasIncompleteDepends(ITaskMonitor tm) {
		Triple<Set<IMagnoliaPackage>, ISet, Boolean> fact = loaderFact(tm);
		return !fact.getThird();
	}


	@Override
	public boolean isCodeUnit() {
		return true;
	}


	@Override
	public boolean isFragment() {
		return true;
	}


	@Override
	public void onDependencyChanged() {
		fullSignature = null;
		depSignature = null;
		revision.incrementAndGet();
		Collection<IMagnoliaModule> children = childrenFact(null).getFirst();
		for(IMagnoliaModule mod : children) {
			mod.onDependencyChanged();
		}

		System.err.println("MagnoliaPackage: DEP CHANGED: " + this);
	}


	@Override
	public void onResourceChanged() {
		sourceSignature = null;
		fullSignature = null;
		depSignature = null;
		revision.incrementAndGet();
		System.err.println("MagnoliaPackage: CHANGED: " + this);
	}


	@Override
	public String toString() {
		return "Package(" + getName() + ")";
	}


	private Pair<Collection<IMagnoliaModule>, ISet> childrenFact(ITaskMonitor tm) {
		synchronized(childrenFact) {
			Pair<Collection<IMagnoliaModule>, ISet> fact = childrenFact.getValue(getSourceSignature(tm));
			if(fact != null) {
				return fact;
			}

			Pair<List<DefinitionData>, ISet> pkgContents = MagnoliaLoader.getPkgContents(this, manager, tm);

			Collection<IMagnoliaModule> children = new ArrayList<IMagnoliaModule>();
			for(DefinitionData childData : pkgContents.getFirst()) {
				assert childData != null;
				children.add(new MagnoliaModule((IFileHandle) resource, childData));
			}

			children = Collections.unmodifiableCollection(children);
			Pair<Collection<IMagnoliaModule>, ISet> result = new Pair<>(children, pkgContents.getSecond());

			childrenFact.setValue(result, getSourceSignature(tm));
			return result;
		}
	}


	/**
	 * Get a signature that identifies the current source
	 * code version of this package and its direct dependencies.
	 *
	 * Returns null if dependencies are unavailable.
	 *
	 * @param tm
	 *            A monitor
	 * @param depends
	 *            A collection of dependencies
	 * @param allDepsOk
	 *            True if dependency resolution was successful; otherwise, the
	 *            computed signature will include the current revision number
	 *            (which will be updated whenever dependencies change)
	 * @return A hash of the package and its dependencies
	 */
	private ISignature getDepSignature(ITaskMonitor tm, Collection<? extends IManagedCodeUnit> depends, boolean allDepsOk) {
		// TODO: needs sync
		synchronized(depSignatureLock) {
			ISignature sig = depSignature;
			if(sig != null) {
				return sig;
			}
			MessageDigest md = Hash.getMessageDigest();
			getSourceSignature(tm).digest(md);
			for(IManagedCodeUnit d : depends) {
				d.getSourceSignature(tm).digest(md);
			}
			if(!allDepsOk) {
				int r = revision.get();
				md.update((byte) (r & 0xff));
				md.update((byte) ((r >>> 8) & 0xff));
				md.update((byte) ((r >>> 16) & 0xff));
				md.update((byte) ((r >>> 24) & 0xff));
			}
			depSignature = sig = new Signature(md.digest());
			return sig;
		}
	}


	/**
	 * @param tm
	 * @return A triple of a) the packages we depend on b) xrefs from dependency
	 *         resolution c) boolean indicating whether all dependencies were
	 *         found
	 */
	private Triple<Set<IMagnoliaPackage>, ISet, Boolean> loaderFact(ITaskMonitor tm) {
		synchronized(loaderFact) {
			// The loader fact depends on the dependencies of this packages, which are stored in the fact.
			// To resolve the cycle, we obtain the fact without checking the signature, then compute the
			// signature using the dependencies stored there (if any).
			//
			// This will trigger recomputation if:
			//   * Dependencies have not yet been computed
			//   * The this package or the packages we depend on have changed
			//   * Files have been added, and not all dependencies were found on the last pass
			NullablePair<Triple<Set<IMagnoliaPackage>, ISet, Boolean>, ISignature> factPair = loaderFact.getValue();

			Triple<Set<IMagnoliaPackage>, ISet, Boolean> first = factPair.getFirst();
			if(first != null && factPair.getSecond() != null) {
				ISignature sig = getDepSignature(tm, first.getFirst(), first.getThird());

				if(factPair.getSecond().equals(sig)) {
					return first;
				}
			}

			IConstructor defInfo = getDefInfo(tm);
			Set<IMagnoliaPackage> depends = new HashSet<>();
			ISetWriter xrefs = MagnoliaFacts.xrefsWriter();
			manager.clearMarks(MARK_SOURCE_RESOLVE_IMPORTS, getURI());
			boolean allDepsOk = true;
			for(IValue i : (ISet) defInfo.get("depends")) {
				IConstructor imp = (IConstructor) i;
				IManagedCodeUnit pkg = manager.findCodeUnit(lang, imp);
				if(pkg == null) {
					pkg = manager.findCodeUnit(lang, Magnolia.qualify(Magnolia.qualifierOf(id), imp));
				}
				if(pkg instanceof IMagnoliaPackage) {
					depends.add((IMagnoliaPackage) pkg);
					if(imp.asAnnotatable().hasAnnotation("loc")) {
						xrefs.insert(TermFactory.vf.tuple(imp.asAnnotatable().getAnnotation("loc"), TermFactory.vf.constructor(MagnoliaFacts.Cons_Ref, (IValue) pkg.getId())));
					}
				}
				else {
					manager.addMark("Import '" + lang.getNameString(imp) + "' not found", RascalUtil.toCodeRegion((ISourceLocation) compiler.call(tm, "locOf", imp)), Severity.ERROR, MARK_SOURCE_RESOLVE_IMPORTS, getURI());
					allDepsOk = false;
				}
			}
			Triple<Set<IMagnoliaPackage>, ISet, Boolean> fact = new Triple<>(depends, xrefs.done(), allDepsOk);
			loaderFact.setValue(fact, getDepSignature(tm, depends, allDepsOk));
			return fact;
		}
	}


	private IFileHandle getFile() {
		return (IFileHandle) resource;
	}


	private @Nullable char[] loadFile(ITaskMonitor tm) {
		// load the file
		char[] cs = null;
		try {
			manager.clearMarks(MARK_SOURCE_READING_FILE, getURI());
			cs = getFile().getContentsCharArray(tm);
			sourceSignature = Hash.hashChars(cs);
/*			tree = treeFact.getValue(sourceSignature); // maybe we still have it cached?
			if(tree != null)
				return tree;
 */ }
		catch(IOException e) {
			sourceSignature = Hash.hashChars(new char[0]);
			manager.addMark(e.getMessage(), RascalUtil.toCodeRegion(vf.sourceLocation(getURI())), Severity.ERROR, MARK_SOURCE_READING_FILE, getURI());
		}
		return cs;
	}


	private IConstructor packageEnvFact(ITaskMonitor tm) {
		synchronized(packageEnvFact) {
			IConstructor packageEnv = packageEnvFact.getValue(getFullSignature(tm));
			if(packageEnv != null) {
				return packageEnv;
			}
			IConstructor defInfo = pkgInfoFact(tm).getFirst();
			ISetWriter inScopeDefs = vf.setWriter();
			for(IManagedCodeUnit pkg : getDepends(tm)) {
				if(pkg instanceof IMagnoliaPackage) {
					for(IMagnoliaCodeUnit mod : ((IMagnoliaPackage) pkg).getChildren(tm)) {
						inScopeDefs.insert(mod.getDefInfo(tm));
					}
				}
			}
			for(IMagnoliaCodeUnit mod : childrenFact(tm).getFirst()) {
				inScopeDefs.insert(mod.getDefInfo(tm));
			}

			IConstructor result = (IConstructor) compiler.call(tm, "getPackageEnv", defInfo, inScopeDefs.done());
			manager.clearMarks(MARK_SOURCE_GETPACKAGEENV, getURI());
			addMarks(result, MARK_SOURCE_GETPACKAGEENV, this);
			packageEnv = (IConstructor) result.get("val");
			assert packageEnv != null;
			packageEnvFact.setValue(packageEnv, fullSignature);
			return packageEnv;
		}
	}


	private Pair<IConstructor, ISet> pkgInfoFact(ITaskMonitor tm) {
		synchronized(pkgInfoFact) {
			ISignature sig = getSourceSignature(tm);
			Pair<IConstructor, ISet> pkgInfo = pkgInfoFact.getValue(sig);
			if(pkgInfo != null) {
				return pkgInfo;
			}

			Pair<IConstructor, ISet> result = MagnoliaLoader.getPkgInfo(this, manager, tm);
			pkgInfo = new Pair<>(result.getFirst()/*.set("resource", this)*/, result.getSecond());
			pkgInfoFact.setValue(pkgInfo, getSourceSignature(tm));
			return pkgInfo;
		}
	}


	private IConstructor treeFact(ITaskMonitor tm) {
		synchronized(treeFact) {
			IConstructor tree = null;
			ISignature sig = sourceSignature;
			if(sig != null) {
				tree = treeFact.getValue(sig);
			}
			if(tree != null) {
				return tree;
			}

			char[] cs = loadFile(tm);

			// parse it
			try {
				if(cs != null) {
					manager.clearMarks(MARK_SOURCE_PARSING, getURI());
					ITree pt = Magnolia.getInstance().getParser().parseModule(getURI(), cs);
					// tree = TermImploder.implodeTree(pt);
					IConstructor implodeTree = TermImploder.implodeTree(pt);
					tree = MagnoliaLoader.desugar(tm, implodeTree);
				}
			}
			catch(ParseError e) {
				ISourceLocation file = vf.sourceLocation(e.getLocation());
				ISourceLocation location = vf.sourceLocation(file, e.getOffset(), e.getLength(), e.getBeginLine(), e.getEndLine(), e.getBeginColumn(), e.getEndColumn());
				IMark mark = new MarkBuilder().message(e).region(RascalUtil.toCodeRegion(location)).severity(ERROR).source(MARK_SOURCE_PARSING).context(getURI()).done();
				manager.addMark(mark);
			}
			catch(Exception e) {
				e.printStackTrace();
				manager.addMark(e.getMessage(), RascalUtil.toCodeRegion(vf.sourceLocation(getURI())), ERROR, MARK_SOURCE_PARSING, getURI());

			}

			if(tree == null) {
				tree = TermFactory.cons("MagnoliaTree", TermFactory.cons("PackageHead", id, TermFactory.seq()), TermFactory.seq());
			}

			assert tree != null;
			assert sourceSignature != null;

			treeFact.setValue(tree, sourceSignature);

			return tree;
		}
	}


	private IConstructor typecheckedFact(ITaskMonitor tm) {
		synchronized(typecheckedFact) {
			IConstructor typechecked = typecheckedFact.getValue(getFullSignature(tm));
			if(typechecked != null) {
				return typechecked;
			}
			IListWriter list = TermFactory.vf.listWriter();
			for(IMagnoliaCodeUnit u : getChildren(tm)) {
				IConstructor tc = u.getTypechecked(tm);
				IConstructor info = u.getDefInfo(tm);
				info = info.set("body", tc);
				list.append(compiler.call(tm, "defInfoToAST", info));
			}
			IConstructor info = getDefInfo(tm);
			info = info.set("body", TermFactory.seq(list.done()));
			typechecked = (IConstructor) compiler.call(tm, "defInfoToAST", info);
			typecheckedFact.setValue(typechecked, getFullSignature(tm));
			return typechecked;
		}
	}


	protected void addMarks(ISet marks, String source, IManagedCodeUnit context) {
		for(IValue mark : marks) {
			IConstructor m = (IConstructor) mark;
			String msg = ((IString) m.get("message")).getValue();
			assert msg != null;
			Severity severity = RascalUtil.getSeverity((IConstructor) m.get("severity"));
			ISourceLocation loc = (ISourceLocation) m.get("location");
			assert loc != null;
			if(!loc.getURI().equals(context.getResource().getURI())) {
				msg = msg + " (while processing " + context.getName() + ")";
				// System.err.println("addMarks: context != loc: " + context.getContainingFile().getURI() + " != " + loc.getURI() + ": " + msg);
			}
			manager.addMark(msg, RascalUtil.toCodeRegion(loc), severity, source, context.getURI());
		}
	}


	protected IResourceRegion getTitleRegion() {
		return new ResourceRegion(this, 0, 1);
	}


	@Override
	public long getLength() throws UnsupportedOperationException, IOException {
		return getFile().getLength(new NullTaskMonitor());
	}


	@Override
	public boolean isRoot() {
		return false;
	}


	@Override
	public ICodeRegion<URI> getNameRegion() {
		return RascalUtil.toCodeRegion(getNameLoc());
	}


	@Override
	public ICodeRegion<URI> getSrcRegion() {
		return RascalUtil.toCodeRegion(getSrcLoc());
	}

}
