/**************************************************************************
 * Copyright (c) 2014 Anya Helene Bagge
 * Copyright (c) 2014 University of Bergen
 *
 * All rights reserved. This program and the accompanying materials
 * are made available under the terms of the Eclipse Public License v1.0
 * which accompanies this distribution, and is available at
 * http://www.eclipse.org/legal/epl-v10.html
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version. See http://www.gnu.org/licenses/
 *
 *
 * See the file COPYRIGHT for more information.
 *
 * Contributors:
 * * Anya Helene Bagge
 *
 *************************************************************************/
package org.magnolialang.magnolia.resources.internal;

import static org.nuthatchery.pica.terms.TermFactory.vf;

import java.net.URI;
import java.net.URISyntaxException;
import java.util.ArrayList;
import java.util.Collections;
import java.util.HashMap;
import java.util.HashSet;
import java.util.List;
import java.util.Map;
import java.util.Random;
import java.util.Set;

import org.rascalmpl.value.IConstructor;
import org.rascalmpl.value.IList;
import org.rascalmpl.value.ISourceLocation;
import org.rascalmpl.value.IValue;
import org.eclipse.jdt.annotation.Nullable;
import org.magnolialang.magnolia.Magnolia;
import org.magnolialang.magnolia.MagnoliaFacts;
import org.magnolialang.magnolia.resources.IMagnoliaCodeUnit;
import org.nuthatchery.pica.errors.ImplementationError;
import org.nuthatchery.pica.terms.TermAdapter;
import org.nuthatchery.pica.terms.TermFactory;
import org.nuthatchery.pica.util.NullnessHelper;
import org.rascalmpl.uri.URIUtil;

/**
 * Helper class to store and decode definition ASTs
 *
 */
public class DefinitionData {
	public static final String KIND_TYPE = "type";
	public static final String KIND_FUNCTION = "function";
	public static final String KIND_PROCEDURE = "procedure";
	public static final String KIND_AXIOM = "axiom";
	public static final String KIND_TEST = "test";
	public static final String KIND_CONCEPT = "concept";
	public static final String KIND_IMPLEMENTATION = "implementation";
	public static final String KIND_PROGRAM = "program";
	public static final String KIND_RENAMING = "renaming";
	public static final String KIND_SATISFACTION = "satisfaction";
	public static final int OPTION_QUALIFY = 1;
	public static final int OPTION_NONE = 0;

	private final DefinitionFlags flags = new DefinitionFlags();
	private final IConstructor body;
	private final IConstructor clause;
	private final IList attrs;
	private final IConstructor id;
	private final String name;
	private final String kind;
	private final boolean hasParams;
	private final boolean isAlgo; // should maybe be false for all TopDefines (incl ops)?
	private final List<IConstructor> params;
	private final ISourceLocation srcLoc;
	private final ISourceLocation nameLoc;
	private final IConstructor def;
	private final IConstructor baseName;
	private final Map<String, IConstructor> attrMap = new HashMap<>();
	private final Set<String> modifiers = new HashSet<>();
	private final IMagnoliaCodeUnit parent;
	private final URI uri;
	private final IConstructor unqName;


	public DefinitionData(IConstructor def, IMagnoliaCodeUnit parent) {
		this(def, parent, OPTION_NONE);
	}


	/**
	 * @param def
	 *            A definition AST, e.g., Define(clause, attrs, body)
	 * @param parent
	 * @param option
	 */
	public DefinitionData(IConstructor def, IMagnoliaCodeUnit parent, int option) {
		if(!(TermAdapter.isCons(def, "Define", 3) || TermAdapter.isCons(def, "TopDefine", 3))) {
			throw new IllegalArgumentException("Not a definition");
		}

		this.def = def;
		this.clause = TermAdapter.getArg(def, 0);
		this.attrs = NullnessHelper.assertNonNull((IList) TermAdapter.getArg(def, 1).get("args"));
		this.body = TermAdapter.getArg(def, 2);
		this.parent = parent;

		for(IValue a : attrs) {
			IConstructor attr = (IConstructor) a;
			assert attr != null;
			String attrName = TermAdapter.getName(attr);
			int arity = TermAdapter.arity(attr);
			if(arity == 0) {
				modifiers.add(attrName);
			}
			attrMap.put(attrName + "/" + arity, attr);
		}

		if(TermAdapter.isCons(clause, "TypeClause", 1)) {
			kind = KIND_TYPE;
			hasParams = false;
			isAlgo = true;
		}
		else if(TermAdapter.isCons(clause, "FunClause", 3)) {
			if(hasModifier("Axiom")) {
				kind = KIND_AXIOM;
			}
			else {
				kind = KIND_FUNCTION;
			}
			hasParams = true;
			isAlgo = true;
		}
		else if(TermAdapter.isCons(clause, "ProcClause", 2)) {
			if(hasModifier("Test")) {
				kind = KIND_TEST;
			}
			else {
				kind = KIND_PROCEDURE;
			}
			hasParams = true;
			isAlgo = true;
		}
		else if(TermAdapter.isCons(clause, "ConceptClause", 1)) {
			kind = KIND_CONCEPT;
			hasParams = false;
			isAlgo = false;
		}
		else if(TermAdapter.isCons(clause, "ImplClause", 1)) {
			kind = KIND_IMPLEMENTATION;
			hasParams = false;
			isAlgo = false;
		}
		else if(TermAdapter.isCons(clause, "ProgramClause", 1)) {
			kind = KIND_PROGRAM;
			hasParams = false;
			isAlgo = false;
		}
		else if(TermAdapter.isCons(clause, "RenClause", 1)) {
			kind = KIND_RENAMING;
			hasParams = false;
			isAlgo = false;
		}
		else if(TermAdapter.isCons(clause, "SatClause", 1)) {
			kind = KIND_SATISFACTION;
			hasParams = false;
			isAlgo = false;
		}
		else {
			throw new IllegalArgumentException();
		}

		if((option & OPTION_QUALIFY) != 0) {
			this.baseName = Magnolia.qualify(parent.getId(), TermAdapter.getArg(clause, 0));
			if(!isAlgo) {
				this.id = TermAdapter.preserveAnnos(TermFactory.cons("TopName", baseName, TermFactory.leaf(kind)), baseName);
			}
			else {
				this.id = baseName;
			}
		}
		else {
			this.id = TermAdapter.getArg(clause, 0);
			this.baseName = Magnolia.basename(id);
		}
		this.unqName = Magnolia.unqualify(baseName);
		this.name = Magnolia.yieldName(baseName);
		this.srcLoc = TermAdapter.locOf(def);
		this.nameLoc = TermAdapter.locOf(id);

		if(hasParams) {
			IConstructor paramList = TermAdapter.getArg(clause, 1);
			paramList = TermAdapter.getArg(paramList, 0);
			params = new ArrayList<>(TermAdapter.arity(paramList));
			for(IConstructor c : TermAdapter.getChildren(paramList)) {
				params.add(c);
			}
		}
		else {
			params = NullnessHelper.assertNonNull(Collections.EMPTY_LIST);
		}

		StringBuilder fragment = new StringBuilder();
		String origFragment = parent.getURI().getFragment();
		if(origFragment != null) {
			fragment.append(origFragment);
			fragment.append(".");
		}
		fragment.append(toIdString());
		URI uri;
		try {
			uri = URIUtil.changeFragment(parent.getURI(), fragment.toString());
		}
		catch(URISyntaxException e) {
			e.printStackTrace();
			try {
				uri = URIUtil.changeFragment(parent.getURI(), String.valueOf(new Random().nextInt()));
			}
			catch(URISyntaxException e1) {
				throw new ImplementationError("Failed to create URI", e);
			}

		}
		assert uri != null;
		this.uri = uri;
	}


	public boolean containsOffset(int offset) {
		return offset >= getOffset() && offset < getOffset() + getLength();
	}


	/* (non-Javadoc)
	 * @see java.lang.Object#equals(java.lang.Object)
	 */
	@Override
	public boolean equals(@Nullable Object obj) {
		if(this == obj) {
			return true;
		}
		if(obj == null) {
			return false;
		}
		if(getClass() != obj.getClass()) {
			return false;
		}
		DefinitionData other = (DefinitionData) obj;
		if(!clause.equals(other.clause)) {
			return false;
		}
		return true;
	}


	/**
	 * @return Original raw AST of the definition
	 */
	public IConstructor getAST() {
		return def;
	}


	/**
	 * @param name
	 *            Name of the attribute, e.g., Foo/2
	 * @return Value of the attribute, e.g., Foo(..., ...)
	 * @throws IllegalArgumentException
	 *             if the attribute does not exist
	 */
	public IConstructor getAttribute(String name) {
		IConstructor attr = attrMap.get(name);
		if(attr == null) {
			throw new IllegalArgumentException();
		}
		else {
			return attr;
		}
	}


	/**
	 * @return Basename of the definition, as a term (e.g., qualified but
	 *         without
	 *         FunName(...))
	 */
	public IConstructor getBaseName() {
		return baseName;
	}


	/**
	 * @return Body of this definition
	 */
	public IConstructor getBody() {
		return body;
	}


	/**
	 * @return Definition clause of this definition, e.g.
	 *         FunClause(name,paramList,returnType)
	 */
	public IConstructor getDefClause() {
		return clause;
	}


	/**
	 * @return This definition as a DefInfo constructor
	 */
	public IConstructor getDefInfo() {
		IConstructor info = MagnoliaFacts.defInfo(vf);
		info = info.set("srcLoc", srcLoc);

		info = info.set("defName", unqName);
		info = info.set("kind", vf.string(kind));

		info = info.set("attrs", attrs);
		info = info.set("decl", clause);
		info = info.set("body", body);
		info = info.set("locInfo", vf.constructor(MagnoliaFacts.Cons_LocInfo1, nameLoc));

		info = info.set("parent", parent.getId());

		info = info.set("qName", id);

		assert info != null;
		return info;
	}


	/**
	 * @return Name of the definition
	 */
	public String getFullName() {
		return name;
	}


	/**
	 * @return Fully-qualified name of the definition, as a term, e.g.
	 *         TypeName(Name(leaf("T")))
	 */
	public IConstructor getId() {
		return id;
	}


	/**
	 * @return One of {@link #KIND_TYPE}, {@link #KIND_FUNCTION},
	 *         {@link #KIND_PROCEDURE}
	 */
	public String getKind() {
		return kind;
	}


	public int getLength() {
		if(srcLoc.hasOffsetLength()) {
			return srcLoc.getLength();
		}
		else {
			throw new UnsupportedOperationException();
		}
	}


	/**
	 * @return Location of declaration's name
	 */
	public ISourceLocation getNameLoc() {
		return nameLoc;
	}


	/**
	 * @return Number of parameters
	 */
	public int getNumParams() {
		return params.size();
	}


	public int getOffset() {
		if(srcLoc.hasOffsetLength()) {
			return srcLoc.getOffset();
		}
		else {
			throw new UnsupportedOperationException();
		}
	}


	/**
	 * @param i
	 * @return Mode of parameter i, e.g., Obs()
	 * @throws IndexOutOfBoundsException
	 *             if no such parameter exists
	 * @throws UnsupportedOperationException
	 *             if definition does not have parameters
	 */
	public IConstructor getParamMode(int i) {
		if(hasParams) {
			return TermAdapter.getArg(NullnessHelper.assertNonNull(params.get(i)), 0);
		}
		else {
			throw new UnsupportedOperationException();
		}
	}


	/**
	 * @param i
	 * @return Name of parameter i, e.g. Name(leaf("foo"))
	 * @throws IndexOutOfBoundsException
	 *             if no such parameter exists
	 * @throws UnsupportedOperationException
	 *             if definition does not have parameters
	 */
	public IConstructor getParamName(int i) {
		if(hasParams) {
			return TermAdapter.getArg(NullnessHelper.assertNonNull(params.get(i)), 1);
		}
		else {
			throw new UnsupportedOperationException();
		}
	}


	/**
	 * @return Parameter list (unmodifiable)
	 * @throws UnsupportedOperationException
	 *             if definition does not have parameters
	 */
	public List<IConstructor> getParams() {
		if(hasParams) {
			return params;
		}
		else {
			throw new UnsupportedOperationException();
		}
	}


	/**
	 * @param i
	 * @return Type of parameter i
	 * @throws IndexOutOfBoundsException
	 *             if no such parameter exists
	 * @throws UnsupportedOperationException
	 *             if definition does not have parameters
	 */
	public IConstructor getParamType(int i) {
		if(hasParams) {
			return TermAdapter.getArg(NullnessHelper.assertNonNull(params.get(i)), 2);
		}
		else {
			throw new UnsupportedOperationException();
		}
	}


	public IMagnoliaCodeUnit getParent() {
		return parent;
	}


	/**
	 * @return Source location, including head and body.
	 */
	public ISourceLocation getSrcLoc() {
		return srcLoc;
	}


	/**
	 * @return Unqualified basename of the definition (as a term)
	 */
	public IConstructor getUnqualifiedName() {
		return unqName;
	}


	public URI getURI() {
		return uri;
	}


	/**
	 * Check if a definition has an attribute.
	 *
	 * Attribute name should be of the form 'Name/N' where N is the number of
	 * parameters in the attribute constructor.
	 *
	 * @param name
	 *            Attribute name
	 * @return True if definition has the given attribute
	 */
	public boolean hasAttribute(String name) {
		return attrMap.containsKey(name);
	}


	/* (non-Javadoc)
	 * @see java.lang.Object#hashCode()
	 */
	@Override
	public int hashCode() {
		final int prime = 31;
		int result = 1;
		result = prime * result + ((clause == null) ? 0 : clause.hashCode());
		return result;
	}


	/**
	 * @param name
	 *            Modifier name
	 * @return True if definition has the given modifier
	 */
	public boolean hasModifier(String name) {
		return modifiers.contains(name);
	}


	/**
	 * @return True if the definition has parameters (even though the parameter
	 *         list may be empty)
	 */
	public boolean hasParams() {
		return hasParams;
	}


	public boolean isAlgo() {
		return isAlgo;
	}


	public boolean isTop() {
		return !isAlgo;
	}


	/**
	 * Guaranteed not to contain spaces, and to be equal to the id-string of any
	 * other definition with the same overload signature. May contain
	 * parentheses, commas and operator symbols.
	 *
	 * @return A string uniquely identifying this definition within its scope
	 */
	public String toIdString() {
		StringBuilder buf = new StringBuilder();

		buf.append(kind);
		buf.append("_");
		buf.append(name);
		if(hasParams) {
			buf.append("(");
			String sep = "";
			for(IConstructor p : params) {
				assert p != null;
				buf.append(sep);
				buf.append(Magnolia.yieldName(TermAdapter.getArg(p, 2)));
				sep = ",";
			}
			buf.append(")");
		}

		return NullnessHelper.assertNonNull(buf.toString());
	}


	@Override
	public String toString() {
		StringBuilder buf = new StringBuilder();

		buf.append(kind);
		buf.append(" ");
		buf.append(name);
		if(hasParams) {
			buf.append("(");
			String sep = "";
			for(IConstructor p : params) {
				assert p != null;
				buf.append(sep);
				buf.append(TermAdapter.getArg(p, 0).getName().toLowerCase());
				buf.append(" ");
				buf.append(Magnolia.yieldName(TermAdapter.getArg(p, 1)));
				buf.append(" : ");
				buf.append(Magnolia.yieldName(TermAdapter.getArg(p, 2)));
				sep = ", ";
			}
			buf.append(")");
		}

		return NullnessHelper.assertNonNull(buf.toString());
	}

}
