/**************************************************************************
 * Copyright (c) 2011-2014 Anya Helene Bagge
 * Copyright (c) 2011-2014 Tero Hasu
 * Copyright (c) 2011-2014 University of Bergen
 * 
 * All rights reserved. This program and the accompanying materials
 * are made available under the terms of the Eclipse Public License v1.0
 * which accompanies this distribution, and is available at
 * http://www.eclipse.org/legal/epl-v10.html
 * 
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version. See http://www.gnu.org/licenses/
 * 
 * 
 * See the file COPYRIGHT for more information.
 * 
 * Contributors:
 * * Anya Helene Bagge
 * * Tero Hasu
 * 
 *************************************************************************/
module org::magnolialang::magnolia::typechecker::Flattener
import IO;
import org::magnolialang::magnolia::AbstractSyntax;
import org::magnolialang::magnolia::MagnoliaFacts;
import org::magnolialang::magnolia::typechecker::toplevel::TopLevelOps;
import org::magnolialang::magnolia::typechecker::toplevel::Filters;
import org::magnolialang::magnolia::typechecker::toplevel::TopPrimitives;
import org::magnolialang::magnolia::lib::MagnoliaCompileLib;
import org::magnolialang::magnolia::lib::Renaming;
import org::magnolialang::magnolia::typechecker::TypeChecker;
import org::magnolialang::magnolia::lib::Environment;
import org::magnolialang::magnolia::analysis::Permissions;
import org::magnolialang::magnolia::typechecker::algo::Expressions;
import org::magnolialang::magnolia::typechecker::algo::Primitives;
import org::magnolialang::magnolia::typechecker::algo::Statements;
import org::magnolialang::magnolia::typechecker::algo::Types;
import org::magnolialang::magnolia::typechecker::algo::Merge;
import org::magnolialang::magnolia::evaluator::Values;
import org::magnolialang::magnolia::lib::Names;
import org::nuthatchery::pica::terms::Terms;
import org::magnolialang::compiler::Util;

import Set;
import Node;
import List;
import util::Benchmark;
import Exception;

public Result[tuple[set[AST],AST,set[AST]]] getTopDepends(AST tree, AST this, Env env) {
	set[AST] result = {};
	set[AST] exts = {};
	tree = visit(tree) {
		case ModuleName(name): {
			candidates = lookupName(name, ExprKind(), env);
			if({} := candidates) {
				env = errorMark(env, "Undefined name <nameToStr(name)>", locOf(name));
			}
			else if({qn} := candidates) {
				if(qn != this) {
					result += lookupDef(qn, env).qName;
					env = addTokenRef(name, qn, env);
					insert qn;
				}
				else {
					env = errorMark(env, "Circular dependency on <nameToStr(name)>", locOf(name));
					insert preserveAnnos(DeclList(seq([])), name);
				}
			}
			else {
				env = errorMark(env, "Ambiguous name <nameToStr(name)>", locOf(name));
				candidates = candidates - {this};
				for(c <- candidates) {
					env = addTokenRef(name, c, env);
				}
				qn = getOneFrom(candidates);
				result += lookupDef(qn, env).resource;
				env = addTokenRef(name, qn, env);
				insert qn;
			}
		}
		case External(lang, name, _): {
			exts += External(lang, name);
		}
	}
	// termPrintln("getTopDepends: Contains externals: ", exts); 

	return makeResult(<result,tree,exts>, env);
}

public Result[Env] getPackageEnv(DefInfo info, set[DefInfo] inScopeDefs) {
//	termPrintln("Make env for ", info.qName);
//	termPrintln("Imported stuff:");
//	for(i <- inScopeDefs)
//		termPrintln("  ", i.qName);
	body = info.body;
	//termPrintln("Body: ", body);
	env = newEnv(info.qName, info.parent, locOf(body));
	env = addPrimitives(env);
	env = addTopScope(inScopeDefs, env);
	return makeResult(env, env);
}

public Result[AST] flatten(DefInfo info, Env env) {
	//	termPrintln("Will flatten ", info.qName);
	body = info.body;
	try {
		env = enterScope(info.qName, body@\loc, env);
	}
	catch NoSuchAnnotation(_): {
		termPrintln("ERROR: body lack location annotation (in ", info.qName, "): ", body);
		env = enterScope(info.qName, locOf(body), env);
	}
	if(isOperationKind(info.kind)) {
		<body, env> = typeCheckTopExpr(info, body, env);
	}
	else {
		// env = addTokenRef(info.defName, Ref(info.qName), env);
		list[AST] expectations = [];
		
		while(Expect(el:leaf(err), e) := body) {
			env = expectError(err, 1, env);
			expectations += el;
			body = e;
		}

		<body, env> = flattenTopExpr(preserveAnnos(Close(body), body), env);

		
		if(DeclBody(seq(ds)) := body) {
			switch(info.kind) {
				case "concept": {
					ds = for(d <- ds) {
						//if(hasModifier(d, Require())) {
						//	env = errorMarkInDecl(env, "Dropping illegal attribute \'require\' from \'<shortDispNameOf(d)>\' in \'<shortDispNameOf(info.qName)>\'", Error(), d);
						//}
						if(hasModifier(d, Extend())) {
							env = errorMarkInDecl(env, "Dropping illegal attribute \'extend\' from \'<shortDispNameOf(d)>\' in \'<shortDispNameOf(info.qName)>\'", Error(), d);
						}
						if(!hasModifier(d, Axiom()) && hasBody(d)) {
							if(!hasModifier(d, Derived())) {
								env = errorMarkInDecl(env, "Dropping body from definition \'<shortDispNameOf(d)>\' in \'<shortDispNameOf(info.qName)>\' (concepts should only contain declarations)", Error(), d);
							}
							d.arg2 = EmptyBody();
						}
						
						d = changeModifiers(d, {Assume(), Require(), Use(), Extend()}, {});
						append d;
					} 
					body = preserveAnnos(DeclBody(seq(ds)), body);
				}
				case "program": {
					<body, env> = analysePermissions(body, env); 
					ds = for(d <- ds) {
						if(hasModifier(d, Require())) {
							env = errorMarkInDecl(env, "Requirement \'<shortDispNameOf(d)>\' not satisfied in \'<shortDispNameOf(info.qName)>\' (use implementation instead?)", Error(), d);
						}
						else if(hasModifier(d, Assume())) {
							if(hasBody(d))
								d = changeModifiers(d, {Assume()}, {});
							;// env = errorMarkInDecl(env, "Assuming that \'<shortDispNameOf(d)>\' will be provided in \'<shortDispNameOf(info.qName)>\'", Info(), d);
						}
						else if(!hasBody(d)) {
							env = errorMarkInDecl(env, "Declaration \'<shortDispNameOf(d)>\' has no definition in \'<shortDispNameOf(info.qName)>\'", Error(), d);
						}
						append d;
					}
					body = preserveAnnos(DeclBody(seq(ds)), body);
				}
				case "satisfaction": {
					<body, env> = analysePermissions(body, env); 
					ds = for(d <- ds) {
						if(hasModifier(d, Require())) {
							env = errorMarkInDecl(env, "Requirement \'<shortDispNameOf(d)>\' not satisfied in \'<shortDispNameOf(info.qName)>\'", Error(), d);
						}
						else if(hasModifier(d, Assume())) {
							if(hasBody(d))
								d = changeModifiers(d, {Assume()}, {});
							;// env = errorMarkInDecl(env, "Assuming that \'<shortDispNameOf(d)>\' will be provided in \'<shortDispNameOf(info.qName)>\'", Info(), d);
						}
						else if(!hasBody(d) && !hasModifier(d, External()) && !hasModifier(d, Derived())) {
							env = errorMarkInDecl(env, "Declaration \'<shortDispNameOf(d)>\' has no definition in \'<shortDispNameOf(info.qName)>\'", Error(), d);
						}
						append d;
					}
					body = preserveAnnos(DeclBody(seq(ds)), body);
				}
			}
		}
		for(el:leaf(err) <- expectations) {
			env = checkExpected(err, locOf(el), env);
		}
	}
	
	if(info.kind == "renaming" && RenameList(seq(rns)) := body) {
		env = checkRenaming(rns, locOf(info.defName), env);
	}
	env = exitScope(env);

	//termPrintln("flatten: flattened: ", info.qName);
	//for(<qn,_> <- env.defs)
	//	termPrintln("flatten: def: ", qn);
		
	return makeResult(body, env);
}

public tuple[AST,Env] flattenTopExpr(AST body, Env env) {
	AST result = body;
	location = locOf(body);

	try {	
		switch(body) {
			case DeclBody(_): {
				result = body; // already flattened
			}
			case Close(b): {
				<b, env> = flattenTopExpr(b, env);
				if(isFlat(b)) {
					<b, env> = check(b, false, env);
					<result, env> = makeOverloadUnique(b, joinDef, env);
					<result, env> = checkModifiers(result, env);
				}
				else {
					result = b;
				}
			}
			case DeclList(seq(ds)): { // AlgDecl::DeclBody
				<result, env> = flattenDeclList(ds, env);
			}
			case Decl(mods, dc, attrs, bd): {// AlgDecl::DeclNoSemi
				<result, env> = flattenDeclList([preserveAnnos(Define(dc, attrs, bd), body)], env);
			}
			case ModuleName(n):
				;
			case Expect(el:leaf(err), e): {
				env = expectError(err, 1, env);
				<result, env> = flattenTopExpr(preserveAnnos(Close(e), e), env);
				env = checkExpected(err, locOf(el), env);
			}
			case TopName(_,_): {
				try	{
					DefInfo info = lookupDef(body, env);
					// <result, env> = flattenTopExpr(info.body, env);  // trackOrigin(getFact(tr, #Flattened, info.qName).val, body);
					result = info.body;
					if(isFlat(result)) {
						result = addOrigin(result, shortNameToStr(body), location);
					}
					env = addTokenRef(body, Ref(info.qName), env);
				}
				catch NoSuchKey(_): {
					env = errorMark(env, "INTERNAL COMPILER ERROR: Error looking up resolved name <nameToStr(body)>",  location);
					result = DeclBody(seq([]));
				}
			}
//				catch str e: env = errorMark(env, e, location);
//				catch: env = errorMark(env, "Error resolving name <nameToStr(body)>", location);
			case Models(a, b): { // satisfaction expr
				<result, env> = applyOp(preserveAnnos(Name(leaf("models")), body), [a, DeclBody(seq([])), b], location, env);
			}
			case WithModels(a, w, b): {// satisfaction expr
				<result, env> = applyOp(preserveAnnos(Name(leaf("models")), body), [a, w, b], location, env);
			}
			case Renamed(a, r): {
				<a, env> = flattenTopExpr(Close(a), env);
				<rs, env> = flattenRenaming(RenameList(r), env);
				env = checkRenaming(rs, locOf(r), env);
				if(isFlat(a)) {
					<result, env> = simpleRename(env, a, rs);
					// TODO: check that renaming is valid
					//env = pushErrorLoc(locOf(r), "Error in renaming: ", "While applying renaming: ", env);
					//<result, env> = makeOverloadUnique(result, joinDef, env);
					//env = popErrorLoc(env);
				}
				else {
					result = DeclBody(seq([]));
				}
			}
			case Filtered(a, op, b): {
				<result, env> = applyFilter(op, AllFilter(), a, b, location, env);
			}
			case r:RenameApp(_, _): {
				<renamings, env> = flattenRenaming(r, env);
				result = RenameList(seq(renamings));
			}
			case r:RenameCompose(_, _): {
				<renamings, env> = flattenRenaming(r, env);
				result = RenameList(seq(renamings));
			}
			case r:RenameList(_): {
				<renamings, env> = flattenRenaming(r, env);
				result = RenameList(seq(renamings));
			}
			case SignatureOf(a): {
				<a, env> = flattenTopExpr(a, env);
				if(isFlat(a)) {
					result = DeclBody(seq(signatureOf(a)));
				}
			}
			case RequiresOf(a): {
				<a, env> = flattenTopExpr(a, env);
				if(isFlat(a)) {
					result = DeclBody(seq(onOf(a)));
				}
			}
			case DefinesOf(a): {
				<a, env> = flattenTopExpr(a, env);
				if(isFlat(a)) {
					result = DeclBody(seq(definesOf(a)));
				}
			}
			case External(_, _, _): {
				<result, env> = flattenDeclList([body], env);
			}
			case TopApp(op, seq(as)): {
				<result, env> = applyOp(op, as, location, env);
			}
			case TopMap(fun, a): {
				<a, env> = flattenTopExpr(a, env);
				if(DeclBody(seq(ds)) := a) {
					<mappedDs, env> = applyMap(fun, ds, env);
					result = DeclBody(seq(mappedDs));
				}
			}
			default: {
				env = errorMark(env, "MagnoliaFlattener::flattenTopExpr: Unknown top expr: <consOf(body)>", Warning(), locOf(body));
				termPrintln("Body: ", body);
				result = DeclBody(seq([]));
			}
		}
	}
	catch e:InternalError(msg, l): {
		env = errorMark(env, "While processing <consOf(body)>: <msg>", locOf(l));
		//throw e;
	}
	
	result = preserveAnnos(result, body);
	if((DeclBody(seq(ds)) := result) && (result@from)?) {
		result.arg0.args = [d[@from=(result@from)] | d <- ds];
	}
	return <result, env>;
}


public tuple[list[AST], Env] applyMap(Modifier(m), list[AST] decls, Env env) {
	list[AST] result;
	
	result = for(decl <- decls) {
		if(Define(_,_,_) := decl) {
			decl = setModifier(decl, m);
		}
		append decl;
	}
	
	return <result, env>;
}

public default tuple[list[AST], Env] applyMap(AST fun, list[AST] decls, Env env) {
	list[AST] result;
	
	if(Name(leaf(name)) := fun && name in modifierOps)
		result = for(decl <- decls) {
			if(Define(_,_,_) := decl) {
				decl = setModifier(decl, modifierOps[name]);
			}
			append decl;
		}
	else {	
		result = for(decl <- decls) {
			<decl, env> = applyDefOp(fun, decl, locOf(decl), env);
			append decl;
		}
	}
	return <result, env>;
}


public tuple[AST, Env] flattenDeclList(list[AST] decls, Env env) {
	list[AST] result = [];

    // This function updates 'result' and 'env'
	// decl - surrounding declaration
	// body - expression to flatten
	// f    - function applied to flattened expression
	void flattenAndAdd(AST decl, AST body, tuple[AST,Env](AST,Env) f) {				
		<body, env> = flattenTopExpr(body, env);
		if(DeclBody(seq(_)) := body) {
			<body, env> = f(body, env);
			result += fullOf(body);
		}
		else {
			if(!(Name(_) := body || QName(_) := body))
				println("MagnoliaFlattener::flattenDeclList:  no expansion found: <trunc(decl)>");
			result += decl;
		}
	}
	
	for(decl <- decls) {
		switch(decl) {
			case External(lang, name, ext): {
				if(!isValidLanguage(lang, env)) {
					env = errorMark(env, "Unknown language \'<nameToStr(lang)>\'", locOf(lang));
				}
				flattenAndAdd(decl, preserveAnnos(Close(ext), ext), tuple[AST,Env](AST d, Env env) { return makeExternal(lang, name, d, env); });
			}
			case Define(_,_,_): {
				result += decl;
			}
			case Nop(): {
				;
			}
			default: {
				flattenAndAdd(decl, decl, tuple[AST,Env](AST d, Env env) { return <d, env>; });
			}
		}
	}
	return <DeclBody(seq(result)), env>;
}

public Env checkRenaming(list[AST] renamings, loc renameLoc, Env env) {
	set[AST] lefts = {};
	for(Rename(a, b) <- renamings) {
		if(a in lefts) {
			env = errorMark(env, "Ambiguous left-hand side \'<nameToStr(a)>\' in renaming", Warning(), renameLoc);
		}
		else {
			lefts += {a};
		}
	}
	return env;
}

tuple[list[AST], Env] flattenRenaming(AST rename, Env env) {
	switch(rename) {
		case RenameList(seq(decls)): {
			list[AST] renamings = [];
			for(r <- decls) {
				if(Rename(_,_) := r)
					renamings += r;
				else {
					<r, env> = flattenTopExpr(r, env);
					if(RenameList(seq(rs)) := r) {
						renamings += rs;
					}
					else {
						env = errorMark(env, "Not a valid renaming", locOf(rename));
					}
				}
			}
			return <renamings, env>;
		}
		case RenameCompose(left, rs:seq(_)): {
			<as, env> = flattenRenaming(left, env);
			<bs, env> = flattenRenaming(RenameList(rs), env);
			list[AST] result = bs;
			map[AST,AST] renaming = ();
			for(Rename(from, to) <- bs) {
				renaming[from] = to;
			}
			
			for(a <- as) {
				if(Rename(from, to) := a && to in renaming) {
					result += preserveAnnos(Rename(from, renaming[to]), a);
				}
				else {
					result += a;
				}
			}
			return <result, env>;
		}
		case RenameApp(n, args): {
			switch(n) {
				case Name(leaf("invert")): {
					if(seq([_]) := args) {
						<as, env> = flattenRenaming(RenameList(args), env);
						list[AST] result = [];
						for(r:Rename(from, to) <- as) {
							result += preserveAnnos(Rename(to, from), args);
						} 
						
						return <result, env>;
					}
					else {
						env = errorMark(env, "Wrong number of arguments for \'<nameToStr(n)>\': expected 1", locOf(rename));
					}
				}
				default: {
					env = errorMark(env, "Unknown renaming function \'<nameToStr(n)>\'", locOf(n));
					return <[], env>;
				}
			}
		}
		default: {
			env = errorMark(env, "MagnoliaFlattener::flattenRenaming: Unknown renaming: <consOf(rename)>", locOf(rename));
			return <[], env>;
		}
	}
}						


public tuple[AST,Env] typeCheckTopExpr(DefInfo info, AST body, Env env) {
	switch(info.decl) {
		case FunClause(n,ParamList(seq(ps)),t): {
			<psChecked, env> = checkParams(ps, env);
			<body, bodyt, env> = checkExprBody(body, env);
			if(bodyt != Unknown()) {
				if(bodyt != t)
					env = errorMark(env, "Function return type \'<dispNameOf(t, env)>\' does not match function body type \'<dispNameOf(bodyt, env)>\'", locOf(n));
			}
		}
		case ProcClause(n,ParamList(seq(ps))): {
			<psChecked, env> = checkParams(ps, env);
			<body, env> = checkStatBody(body, env);
		}
		default:
			termPrintln("Don\'t know what to do with ", info.decl);
	}
	
	return <body, env>;
}

@doc{
Find all in-scope toplevel defs, and add them to the environment; for use in typechecking top-level operations. 
}
public Env addTopScope(set[DefInfo] inScopeDefs, Env env) {
	for(DefInfo info <- inScopeDefs) {
		kind = ExprKind();
		qName = info.qName;
		name = info.defName;
		info.body = Nop();
		if(isOperationKind(info.kind)) {
			switch(info.decl) {
				case FunClause(n,ParamList(seq(ps)),t): {
					<psChecked, env> = checkParams(ps, env);
					<t, _, env> = checkType(t, env);
					qName = FunName(qualify(info.parent, n), seq([anonParam(p) | p <- psChecked]), t);
					kind = ExprKind();
				}
				case ProcClause(n,ParamList(seq(ps))): {
					<psChecked, env> = checkParams(ps, env);
					qName = ProcName(qualify(info.parent, n), seq([anonParam(p) | p <- psChecked]));
					kind = StatKind();
				}
				default:
					termPrintln("Don\'t know what to do with ", info.decl);
			}
		}
		//termPrintln("addTopScope: adding name ", name, " =\> ", qName, " (", kind, ")");
		env = addName(name, qName, kind, env);
		env = addDef(qName, info, env);
	}

	return env;
}


tuple[AST,Env] checkModifiers(DeclBody(seq(ds)), Env env) {
	list[AST] result = [];
	for(d <- ds) {
		if(Define(_,_,body) := d) {
			if(hasModifier(d, Protect())) {
				orgLoc = locOf(getDeclarator(d));
				<via, actualLoc> = getLatestOrigin(d, orgLoc);
				trace = "";
				if(orgLoc != actualLoc)
					trace += "from <locInfo(orgLoc, actualLoc)>";
				if(via != "")
					trace += ", via <via>"; 
				env = errorMark(env, "Unused protect \'<dispNameOf(d, env)>\' <trace == "" ? "" : "(<trace>)">", Warning(), actualLoc);
			}
			else if(hasModifier(d, Derived()) && hasModifier(d, Require())) {
				; // drop these declarations
			}
			else if(hasModifier(d, Assume()) && body != EmptyBody()) {
				result += changeModifiers(d, {Assume()}, {}); // remove assume if decl has a body
			}
			else {
				result += d;
			}
		}
		else {
			result += d;
		}
	}
	
	return <DeclBody(seq(result)), env>;
}

default tuple[AST,Env] checkModifiers(AST ast, Env env) {
	return <ast, env>;
}
