/**************************************************************************
 * Copyright (c) 2011-2014 Anya Helene Bagge
 * Copyright (c) 2011-2014 University of Bergen
 * 
 * All rights reserved. This program and the accompanying materials
 * are made available under the terms of the Eclipse Public License v1.0
 * which accompanies this distribution, and is available at
 * http://www.eclipse.org/legal/epl-v10.html
 * 
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version. See http://www.gnu.org/licenses/
 * 
 * 
 * See the file COPYRIGHT for more information.
 * 
 * Contributors:
 * * Anya Helene Bagge
 * 
 *************************************************************************/
module org::magnolialang::magnolia::typechecker::TypeChecker
import IO;
import org::magnolialang::magnolia::AbstractSyntax;
import org::magnolialang::magnolia::MagnoliaFacts;
import org::magnolialang::magnolia::typechecker::toplevel::TopLevelOps;
import org::magnolialang::magnolia::lib::MagnoliaCompileLib;
import org::magnolialang::magnolia::lib::Names;
import org::magnolialang::magnolia::lib::Modules;
import org::magnolialang::magnolia::lib::Environment;
import org::magnolialang::magnolia::typechecker::algo::Expressions;
import org::magnolialang::magnolia::typechecker::algo::Attributes;
import org::magnolialang::magnolia::typechecker::algo::Statements;
import org::magnolialang::magnolia::typechecker::algo::DerivedOps;
import org::magnolialang::magnolia::typechecker::algo::Types;
import org::magnolialang::magnolia::typechecker::algo::Primitives;
import org::magnolialang::magnolia::typechecker::algo::Merge;
import org::magnolialang::magnolia::typechecker::algo::Alerts;
import org::nuthatchery::pica::terms::Terms;
import org::magnolialang::compiler::Util;
import org::nuthatchery::pica::memo::Memo;
import Set;
import Node;
import List;
import String;
import util::Benchmark;
import util::Math;


public Env loadDefs(AST defs, Env env) {
	if(DeclBody(seq(ds)) := defs)
		return loadDefs(defs, false, env);
	else
		return env;
}

public Env loadDefs(DeclBody(seq(ds)), bool sloppy, Env env) {
	// Predeclare type names
	for(d:Define(def, attrs, body) <- ds) {
		env = loadDef(d, sloppy, env);
	}
	return env;
}

public Env loadDef(d:Define(def, attrs, body), bool sloppy, Env env) {
	DefInfo info = addAttrs(definfo(), attrs);
	info.body = body;
	info.parent = env.scName;
	info.decl = def;
	info.qName = defNameOf(def);
	info.defName = unqualify(basename(info.qName));
	if((def@from)?)
		info.locInfo = LocInfo(info.defName@\loc, d@from);
	else
		info.locInfo = LocInfo(info.defName@\loc);
	env = replaceDef(info.qName, info, env);
	return env;
}

public tuple[bool,Env] loadNames(DeclBody(seq(ds)), bool sloppy, Env env) {
	// Predeclare type names
	for(d:Define(def, attrs, body) <- ds) {
		env = loadName(def, sloppy, env);
	}
	return <false, env>;
}

public Env loadName(AST def, bool sloppy, Env env) {
	qName = defNameOf(def);
	n = unqualify(basename(qName)); 
	if(DEBUG_TYPECHECK)
		termPrintln("CHECK: adding def ", n, " =\> ", qName);
	switch(def) {
		case TypeClause(_):
			env = addName(n, qName, TypeKind(), true, env);
		case FunClause(_,_,_):
			env = addName(n, qName, ExprKind(), true, env);
		case ProcClause(_,_):
			env = addName(n, qName, StatKind(), true, env);
		default:
			if(!sloppy)
				termPrintln("Don\'t know what to do with ", def);
	}
	return env;
}
public default tuple[bool,Env] loadNames(AST body, bool sloppy, Env env) {
	return <true, env>;
}
@doc{
The algorithm is as follows:

 * First, we examine all the type declarations, and add their names to the environment
 
 * Then, we examine all the type definitions, and add the field declarations to the environment.
   Any conflicting field declarations are reported, and errors are recovered by picking an arbitrary
   field declaration (which should be marked as "in error", to reduce cascading error messages).
   
 * Next, we examine all operation declarations, check the types used, and add the declarations
   to the environment. Conflicting declarations (overload-equal but not decl-equal – e.g., different
   return types or parameter modes) are reported,  and errors are recovered by picking an arbitrary
   declaration (which should be marked as "in error", to reduce cascading error messages).
   
 * We now have a complete environment, and are ready to typecheck expressions and statements.
   All operation definitions (and type definitions with initialisers) are checked, and added to
   the result, even though some of them may be in conflict.
   
 * Finally, we have a list of typechecked declarations and definitions, which we then check for conflicting
   definitions, and try to resolve the conflicts.
}
public tuple[AST,Env] check_impl(db:DeclBody(seq(ds)), bool sloppy, Env env) {
	env = enterNewScope(env);
	env = addPrimitives(env);
	Module m = Module(Unknown());
	list[AST] result = [];

	<doneImplDs, doneApiDs, typeDs, otherDs> = splitAndSortDecls(ds);

	// first, load stuff that has already been processed
	// we do the implementations first, so they are available if we need to merge stuff from the api
	for(d <- doneImplDs) {
		if(DEBUG_INSTANCES)
 			termPrintln("Concrete def: ", d);
 		env = loadName(getDeclarator(d), sloppy, env);
 		env = loadDef(d, sloppy, env);
	 	<m, env> = mergeImpl(m, d, env);
 	}
 	for(d <- doneApiDs) {
 		if(DEBUG_INSTANCES)
 			termPrintln("Instance def: ", d);
 		env = loadName(getDeclarator(d), sloppy, env);
 		env = loadDef(d, sloppy, env);
 		<m, env> = mergeApi(m, d, env);
	}
	
	
	// Predeclare type names
	// 1) First, we examine all the type declarations, and add their names to the environment
	for(d <- typeDs) {
		<ds1, env> = qualifyDeclName(d, env);
		for(d1 <- ds1) {
			<m, env> = mergeApi(m, d1, env);
		}
	}
	
		
	
	// 2) Then, we examine all the type definitions, and add the field declarations to the environment.
	for(qName <- providesOf(m), !isInstance(m, qName)) {
		if(isTypeDecl(qName) || isAlertDecl(qName))
			<m, env> = checkDecl(m, qName, env);
	}
	
	// Precheck operation defs
	// 3) Next, we examine all operation declarations, check the types used, ...
	for(d <- otherDs) {
		<ds1, env> = qualifyDeclName(d, env);
		for(d1 <- ds1) {
			//termPrintln("Adding op ", d1);
			<m, env> = mergeApi(m, d1, env);
		}
	}
	
	// 4) We now have a complete environment, and are ready to typecheck expressions and statements.
	for(qName <- providesOf(m), !isInstance(m, qName)) {
		if(!(isTypeDecl(qName) || isAlertDecl(qName)))
			<m, env> = checkDecl(m, qName, env);
	}


	typeParams = {qName | qName <- requiresOf(m), isTypeDecl(qName)};
	moduleParams = requiresOf(m);
	//termPrintln("moduleParams: ", moduleParams);
	map[AST,AST] names = ();
	map[AST,AST] instances = ();
	list[tuple[AST,list[AST]]] impls = [];
	rel[AST,AST] useGraph = {};
	map[AST,tuple[AST,AST,set[AST],AST]] todo = ();
	
	set[AST] findUsedNames(value tree) {
		set[AST] result = {};
		top-down-break visit(tree) {
			case x:FunName(_,ps,rt): result += {x} + findUsedNames(rt) + findUsedNames(ps);
			case x:ProcName(_,ps): result += x + findUsedNames(ps);
			case x:TypeName(_): result += x;
		//	case x:ExternalType(_,_,extParams,_): result += findUsedNames(extParams);
		//	case x:ExternalFun (_,_,extParams,_,_,_): result += findUsedNames(extParams);
		//	case x:ExternalProc(_,_,extParams,_,_): result += findUsedNames(extParams);
			case x:Instance(xIdName, seq(xBindings)): {
				result += ({} | it+findUsedNames(y) | Binding(_, y) <- xBindings);
			}
		}
		return result;
	}
	void findParamsAndInstances(AST qName)
	{
		body = getBody(m, qName);
		if(Instance(idName, seq(bindings)) := body) {
			if(DEBUG_INSTANCES)
				termPrintln("Processing: ", qName, " =\> ", body);
			instances[qName] = body;
			useGraph += {qName} * findUsedNames(body);
		}
		else if(isInstance(body)) {
			throw "Whoops! Is instance but didn\'t match pattern: <consOf(body)>";
		}
		else if(body != EmptyBody()) {
			dtor = getDeclarator(m, qName); 
			attrs = getAttrs(m, qName);
			idName = makeId(rename(qName, instances));
			dtor = renameDef(dtor, idName);
			//implParams = isTypeDecl(qName) ? typeParams : moduleParams;
			set[AST] usedNames = findUsedNames(<dtor, attrs, body>);
			useGraph += {idName} * usedNames;
			if(DEBUG_INSTANCES)
				termPrintln("Processing: ", qName, " =\> ", idName);
			//termPrintln("  moduleParams: ", implParams);
			//termPrintln("  usedNames: ", usedNames);
			//implParams = implParams & usedNames;
			//termPrintln("  implParams: ", implParams);
			//instance = Instance(idName, seq([Binding(x, x) | x <- implParams]));
			//instances[qName] = instance;
			names[qName] = idName;
			todo[idName] = <qName, dtor, attrs, body>;
			//termPrintln("Concrete implementation: ", qName, " =\> ", idName);
			//termPrintln(" body: ", d.arg2);
			//termPrintln("    d: ", d);
		}
	}

	void makeImplsAndInstanceList(AST idName) {
		<qName, dtor, attrs, body> = todo[idName]; 
		if(DEBUG_INSTANCES)
			termPrintln("useGraph2: ", idName);
		useGraph0 = useGraph<0>;
		if(DEBUG_INSTANCES) {
			for(x <- useGraph[idName]) {
				termPrintln("     uses: ", x, " (<x notin useGraph0>)");
			} 
		}
		implParams = [x | x <- useGraph[idName], x notin useGraph0];
		if(DEBUG_INSTANCES)
			termPrintln("implParams2: ", implParams);
		instance = Instance(idName, seq([Binding(x,  x) | x <- implParams]));
		//termPrintln("  instance: ", instance);
		instances[qName] = instance;
		m = setBody(m, qName, instance);
		m = setAttrs(m, qName, (attrs - {External()}) + {Api()});
		d = Define(dtor, seq([Impl()] + toList(attrs)), body);
		impls += <d, implParams>;
	}

	for(qName <- [x | x <- providesOf(m), isTypeDecl(x)]) {
		findParamsAndInstances(qName);
	} 
	useGraph = {<x, names[y] ? y> | <x,y> <- useGraph}+;
	for(idName <- todo) {
		makeImplsAndInstanceList(idName);
	}
	// Make sure all instances are fully expanded
	for(i <- instances) {
		instances[i] = lockInstances(instances[i], instances);
	}
	
	
	todo = ();
	for(qName <- [x | x <- providesOf(m), !isTypeDecl(x)]) {
		findParamsAndInstances(qName);
	} 
	useGraph = {<x, names[y] ? y> | <x,y> <- useGraph}+;
	for(idName <- todo) {
		makeImplsAndInstanceList(idName);
	}
	// Make sure all instances are fully expanded
	for(i <- instances) {
		instances[i] = lockInstances(instances[i], instances);
	}

	
	if(DEBUG_INSTANCES) {
		for(i <- instances) {
			termPrintln("Renamed Instance: ", i);
			termPrintln("      =\> ", instances[i]);
		}
	}
	
	for(<d:Define(dtor, seq(attrs), body), implParams> <- impls) {
		//termPrintln("Processing: ", dtor);
		//termPrintln("implParams3: ", implParams);
		map[AST,AST] renaming = ();
		for(k <- instances) {
			if(k in implParams) {
				//implParams -= k;
				renaming[k] = instances[k];
			//	termPrintln("Locking parameter: ", k);
			//	termPrintln("                =\> ", instances[k]);
			}
			else {
				renaming[k] = instances[k];
			//	termPrintln("Locking free name: ", k);
			//	termPrintln("                =\> ", instances[k]);
			}
		}
		for(k <- implParams) {
			//termPrintln("Keeping parameter: ", k);
			renaming[k] = k;
		}

		d = lockInstances(d, renaming);
					
		env = checkImpl(d, implParams, env);
		m = addImpl(m, d);
	}

	result = [getDefine(m, x) | x <- requiresOf(m) + providesOf(m) + implementsOf(m)];
	env = exitDefScope(env);
	return <DeclBody(seq(result)), env>;	
}

public tuple[AST,Env](AST,bool,Env) check = memo(check_impl);

public AST lockInstances(AST d, map[AST,AST] renaming) {

	AST inst(AST n, AST old) {
		if(Instance(_,bs) := n) {
			if(DEBUG_INSTANCES)
				termPrintln("lockInstances: ", old, " =\> ", n);
			r = n[arg1=lockInstances(bs, renaming)];
			if(DEBUG_INSTANCES)
				termPrintln("               ", old, " =\> ", r);
			return r;
		 }
		else
			return n;
	}
		d = top-down-break visit(d) {
			case n:TypeName(_): if(n in renaming) insert inst(renaming[n], n);
			case n:FunName(_,_,_): if(n in renaming) insert inst(renaming[n], n);
			case n:ProcName(_,_): if(n in renaming) insert inst(renaming[n], n);
		case x:Binding(a, b): insert preserveAnnos(Binding(a, lockInstances(b, renaming)), x);
		case x:ExternalType(lang, m, onPart, n): {
			onPart = lockInstances(onPart, renaming);
			insert preserveAnnos(ExternalType(lang, m, onPart, n), x);
		}
		case x:ExternalFun(lang, m, onPart, n, ps, t): {
			ps = lockInstances(ps, renaming);
			t = lockInstances(t, renaming);
			onPart = lockInstances(onPart, renaming);
			insert preserveAnnos(ExternalFun(lang, m, onPart, n, ps, t), x);
		}
		case x:ExternalProc(lang, m, onPart, n, ps): {
			ps = lockInstances(ps, renaming);
			onPart = lockInstances(onPart, renaming);
			insert preserveAnnos(ExternalProc(lang, m, onPart, n, ps), x);
		}
		};
		return d;
}


public tuple[list[AST], Env] qualifyDeclName(d:Define(def, attrs, body), Env env) {
	list[AST] result = [];
	
	if(hasModifier(d, Require()) && body != EmptyBody()) {
		env = errorMarkInDecl(env, "Dropping illegal attribute \'require\' from \'<shortDispNameOf(d)>\'", Warning(), d);
		d = changeModifiers(d, {Require()}, {});
	}
	
	if(TypeClause(n) := def) {
		n = basename(n);
		qName = preserveAnnos(TypeName(n), n);
		if(DEBUG_TYPECHECK)
			termPrintln("CHECK: adding type ", qName, " from ", locOf(qName));

		if(!(d@from)?)
			env = addTokenRef(n, qName, env);
		d.arg0 = preserveAnnos(TypeClause(qName), def);
		env = addName(n, qName, TypeKind(), false, env);
		
		<derived, env> = deriveOps(n, qName, TypeKind(), d, env);
			
		for(<n1, qName1, kind1, d1> <- derived) { // all decls coming from this one, including itself
			env = addName(n1, qName1, kind1, env);
			if(!hasDef(qName1, env))
				env = replaceDef(qName1, definfo(), env); // note that qName1 has been defined, don't care about definition
			result += d1;
		}
		
	}
	else if(AlertClause(n, supers) := def) {
		n = basename(n);
		qName = preserveAnnos(AlertName(n), n);
		if(DEBUG_TYPECHECK)
			termPrintln("CHECK: adding alert ", qName, " from ", locOf(qName));

		if(!(d@from)?)
			env = addTokenRef(n, qName, env);
		d.arg0 = preserveAnnos(AlertClause(qName, supers), def);
		env = addName(n, qName, AlertKind(), false, env);

		<derived, env> = deriveOps(n, qName, AlertKind(), d, env);
			
		for(<n1, qName1, kind1, d1> <- derived) { // all decls coming from this one, including itself
			env = addName(n1, qName1, kind1, env);
			if(!hasDef(qName1, env))
				env = replaceDef(qName1, definfo(), env); // note that qName1 has been defined, don't care about definition
			result += d1;
		}
	}
	else if(LiteralClause(n, as) := def) {
	;
/*		n = basename(n);
		qName = preserveAnnos(TypeName(n), n);
		env = addName(n, qName, TypeKind(), false, env);
		d.arg0 = preserveAnnos(TypeClause(qName), def);
		d = addModifier(d, Require());
		<derived, attr, env> = deriveLiteralOps(def, qName, env);
		for(x <- derived)
			termPrintln("adding derived: ", x);
		opDeclsTodo += derived;
		if(attr != Nop()) {
			d.arg1.args += [attr];
		}
		termPrintln("adding: ", n, d);
		result = [d];
		*/
	}
	else if(FunClause(n,ParamList(seq(ps)),t) := def || ProcClause(n,ParamList(seq(ps))) := def) {
		AST qName;
		NameKind kind;
		n = basename(n);
		env = enterScope(env);
		if((d@from)?)
			env.info = SourceInfo();
		<psChecked, env> = checkParams(ps, env);
		if(isFunctionDecl(d)) {
			<t, _, env> = checkType(t, env);
			qName = preserveAnnos(FunName(n, seq([anonParam(p) | p <- psChecked]), t), n);
			d.arg0 = preserveAnnos(FunClause(qName,ParamList(seq(psChecked)), t), def);
			kind = ExprKind();
		}
		else {
			qName = preserveAnnos(ProcName(n, seq([anonParam(p) | p <- psChecked])), n);
			d.arg0 = preserveAnnos(ProcClause(qName,ParamList(seq(psChecked))), def);
			kind = StatKind();
		}

		if(DEBUG_TYPECHECK)
			termPrintln("CHECK: adding op ", qName, " from ", locOf(qName));
		
		<derived, env> = deriveOps(n, qName, kind, d, env);
		env = exitScope(env);
			
		for(<n1, qName1, kind1, d1> <- derived) { // all decls coming from this one, including itself
			env = addName(n1, qName1, kind1, env);
			if(!hasDef(qName1, env))
				env = replaceDef(qName1, definfo(), env); // note that qName1 has been defined, don't care about definition
			result += d1;
		}
	}
	return <result, env>;
}

public tuple[Module, Env] checkDecl(Module m, AST qName, Env env) {
	dcl = nameToDcl(qName);
	attrs = getAttrs(m, qName);
	body = getBody(m, qName);
	
	env = enterScope(env);
	//if((d@from)?) {
	//	env.info = SourceInfo();
	//}
	<_, env> = checkParams(getParamList(m, qName), env);  // load params (if any) into environment
	// TODO: <attrs, env> = checkAttrs(attrs, d, env);
	
	
	if(isTypeDecl(qName)) {
		<body, env> = checkTypeFieldDecls(body, env);  // TODO: this should be done separately, and before
		<body, env> = checkTypeBody(body, env);
		m = setBody(m, qName, body);
		env = replaceDef(qName, getDefInfo(m, qName), env);
	}
	else if(isAlertDecl(qName)) {
		throw "Not really implemented yet...";
		<supers, env> = checkAlertSupers(supers, env);
		<body, env> = checkAlertBody(body, env);
		m = setBody(m, qName, body);
		env = replaceDef(qName, getDefInfo(m, qName), env);
	}
	else if(isFunctionDecl(qName)) {
		t = getReturnType(m, qName);
		env = enterScope(env);
		env = replaceDef(preserveAnnos(returnTypeName, t), makeDefInfo(preserveAnnos(returnTypeName, t), env)[body=t], env);
		<body, bodyt, env> = checkExprBody(body, env);
		if(Body(b) := body) {
			<b, bodyt, env> = checkLiteral(b, bodyt, t, env);
			body.arg0 = b;
		}
		env = exitScope(env);
		if(bodyt != Unknown()) {
			if(bodyt != t) {
				termPrintln("TypeChecker: bodyt=", bodyt, ", t=", t, ", body=", body);
				env = errorMark(env, "Function return type \'<dispNameOf(t, env)>\' does not match function body type \'<dispNameOf(bodyt, env)>\'", locOf(qName));
			}
		}
		m = setBody(m, qName, body);
		env = replaceDef(qName, getDefInfo(m, qName), env);
	}
	else if(isProcedureDecl(qName)) {
		<body, env> = checkStatBody(body, env);
		m = setBody(m, qName, body);
		env = replaceDef(qName, getDefInfo(m, qName), env);
	}
	else {
		throw InternalError("Typechecker::checkDecl don\'t know what to do with this kind of decl", qName);
	}
	
	env = exitScope(env);
	
	return <m, env>;
}



tuple[Module, Env] mergeImpl(Module m, d:Define(dtor, seq(attrs), body), Env env) {
	qName = defNameOf(dtor);
	
	if(hasDef(m, qName)) {
		<d, env> = joinDef(d, getDefine(m, qName), env);
	}
	
	m = replaceImpl(m, d);
	
	return <m, env>;
}

tuple[Module, Env] mergeApi(Module m, d:Define(dtor, seq(attrs), body), Env env) {
	qName = defNameOf(dtor);
	
	if(hasDef(m, qName)) {
		<d, env> = joinDef(d, getDefine(m, qName), env);
	}
	
	m = replaceApi(m, d);
	
	return <m, env>;
}

default tuple[AST,Env] check_impl(AST ast, bool sloppy, Env env) {
	if(!sloppy) {
		println("CHECK: no rule for <consOf(ast)>");
		env = errorMark(env, "Internal Compiler Error: StaticSemantics::check(): no check rule for <consOf(ast)>", Info(), locOf(ast));
	}
	return <ast, env>;
}


DefInfo addAttrs(DefInfo info, AST attrs) {
	list[AST] result = info.attrs;
	if(seq(as) := attrs)
		result += as;
	info.attrs = result;
	return info;
}


public AST anonParam(p:Param(m, n, t)) {
	return preserveAnnos(AnonParam(m, t), p);
}

public default AST anonParam(AST p) {
	termPrintln("TypeChecker::anonParam: unknown AST: ", p);
	return p;
}



private tuple[list[AST],list[AST],list[AST],list[AST]] splitAndSortDecls(list[AST] ds){
	list[AST] doneImplDs = [];
	list[AST] doneApiDs = [];
	list[AST] typeDs1 = [];
	list[AST] typeDs2 = [];
	list[AST] otherDs1 = [];
	list[AST] otherDs2 = [];

	for(d <- ds) {
		if(isConcrete(d)) {
			doneImplDs += d;
		}
		else if(isInstance(d)) {
			doneApiDs += d;
		}
		else if(hasModifier(d, Require())) {
			if(isTypeDecl(d))
				typeDs1 += d;
			else
				otherDs1 += d;
		}
		else {
			if(isTypeDecl(d))
				typeDs1 += d;
			else
				otherDs1 += d;
		}
	}
	
	return <doneImplDs, doneApiDs, typeDs1 + typeDs2, otherDs1 + otherDs2>;
}

private Env checkImpl(Define(dtor, seq(attrs), body), list[AST] args, Env env) {
	//termPrintln("Checking implementation: ", dtor);
	void checkName(AST name, list[AST] args, loc l) {
		//termPrintln("checkName: ", name);
		if(!(isId(name) || isInstance(name) || name in args)) {
			env = errorMark(env, "INTERNAL ERROR (broken invariant): name <yieldTerm(name)> should be bound or id", Internal(), l);			
			termPrintln("*** INTERNAL ERROR: name should be bound or id: ", name);
			termPrintln("        In");
			termPrintln("           * dtor: ", dtor);
			termPrintln("           * args: ", args);
			termPrintln("           * body: ", body);
		}
	}
	void checkRec(AST tree, list[AST] args) {
		top-down-break visit(tree) {
		case x:Fun(n:FunName(_,_,_)): checkName(n, args, locOf(x));
		case x:Proc(n:ProcName(_,_)): checkName(n, args, locOf(x));
		case x:ExternalType(_, _, ExternalParams(seq(ns1), seq(ns2)), _): for(n <- ns1+ns2) checkName(n, args, locOf(x));
		case x:ExternalFun(_, _, ExternalParams(seq(ns1), seq(ns2)), _, _, _): for(n <- ns1+ns2) checkName(n, args, locOf(x));
		case x:ExternalProc(_, _, ExternalParams(seq(ns1), seq(ns2)), _, _): for(n <- ns1+ns2) checkName(n, args, locOf(x));
		case x:Binding(a,b): if(a != b) checkName(b, args, locOf(x));	
		case x:Instance(n, seq(bs)): checkRec(n, args + [b | Binding(b,_) <- bs]);
		}
	}
	
	checkRec(body, args);
	
	return env;
}