/**************************************************************************
 * Copyright (c) 2012-2014 Anya Helene Bagge
 * Copyright (c) 2012-2014 University of Bergen
 * 
 * All rights reserved. This program and the accompanying materials
 * are made available under the terms of the Eclipse Public License v1.0
 * which accompanies this distribution, and is available at
 * http://www.eclipse.org/legal/epl-v10.html
 * 
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version. See http://www.gnu.org/licenses/
 * 
 * 
 * See the file COPYRIGHT for more information.
 * 
 * Contributors:
 * * Anya Helene Bagge
 * 
 *************************************************************************/
module org::magnolialang::magnolia::typechecker::algo::Attributes
import IO;
import org::magnolialang::magnolia::AbstractSyntax;
import org::magnolialang::magnolia::MagnoliaFacts;
import org::magnolialang::magnolia::typechecker::toplevel::TopLevelOps;
import org::magnolialang::magnolia::lib::MagnoliaCompileLib;
import org::magnolialang::magnolia::lib::Names;
import org::magnolialang::magnolia::lib::Environment;
import org::magnolialang::magnolia::lib::Renaming;
import org::magnolialang::magnolia::typechecker::algo::DerivedOps;
import org::magnolialang::magnolia::typechecker::algo::Expressions;
import org::magnolialang::magnolia::typechecker::algo::Types;
import org::magnolialang::magnolia::typechecker::algo::Alerts;
import org::magnolialang::magnolia::typechecker::algo::Overload;
import org::nuthatchery::pica::terms::Terms;
import org::magnolialang::compiler::Util;
import Set;
import Node;
import List;
import String;

set[AST] modifiers = {Api(), Impl(), ImplyMod(), Assume(), Broken(), Predicate(), Axiom(), Derived(), Extend(), NoBuild(), Protect(), Require(), Default(), External(), Use()};

public tuple[AST, Env] checkAttrs(seq(attrs), AST decl, Env env) {
	<attrs1, env> = checkAttrs(attrs, decl, env);
	return <seq(attrs1), env>;
}

public tuple[list[AST], Env] checkAttrs(list[AST] attrs, AST decl, Env env) {
	attrs = for(a <- attrs) {
		<a, env> = checkAttr(a, decl, env);
		append a;
	}
	
	return <attrs, env>;
}

tuple[AST, Env] checkAttr(attr:Guard(expr), AST decl, Env env) {
	<expr, t, env> = checkExpr(expr, env);
	
	if(t != SpecialType(Predicate()) && t != Unknown())
		env = errorMark(env, "Guard should be a predicate, not \'<dispNameOf(t, env)>\'", locOf(attr));
	
	return <attr[arg0=expr], env>;
}

tuple[AST, Env] checkAttr(attr:Alert(alert, spec), AST decl, Env env) {
	<alert, env> = checkAlert(alert, env);
	switch(spec) {
		case AlertPre(expr): {
			<expr, t, env> = checkExpr(expr, env);		
			if(t != SpecialType(Predicate()) && t != Unknown())
				env = errorMark(env, "Alert expression should be a predicate, not \'<dispNameOf(t, env)>\'", locOf(attr));
			spec.arg0 = expr;
		}
		case AlertPreNeg(expr): {
			<expr, t, env> = checkExpr(expr, env);		
			if(t != SpecialType(Predicate()) && t != Unknown())
				env = errorMark(env, "Alert expression should be a predicate, not \'<dispNameOf(t, env)>\'", locOf(attr));
			spec.arg0 = expr;
		}
		case AlertPost(expr): {
			<expr, t, env> = checkExpr(expr, env);		
			if(t != SpecialType(Predicate()) && t != Unknown())
				env = errorMark(env, "Alert expression should be a predicate, not \'<dispNameOf(t, env)>\'", locOf(attr));
			spec.arg0 = expr;
		}
		case AlertPostNeg(expr): {
			<expr, t, env> = checkExpr(expr, env);		
			if(t != SpecialType(Predicate()) && t != Unknown())
				env = errorMark(env, "Alert expression should be a predicate, not \'<dispNameOf(t, env)>\'", locOf(attr));
			spec.arg0 = expr;
		}
		case AlertThrows(_): {
			env = errorMark(env, "Checking throws not implemented yet", Warning(locOf(sepc)));
		}
		default: {
			env = errorMark(env, "Unknown alert spec \'<consOf(spec)>\'", locOf(spec));
		}
	}
	
	return <preserveAnnos(Alert(alert, spec), attr), env>;
}

tuple[AST, Env] checkAttr(attr:Origins(os), AST decl, Env env) {
	return <attr, env>;
}

tuple[AST, Env] checkAttr(attr:Concrete(_), AST decl, Env env) {
	return <attr, env>;
}

tuple[AST, Env] checkAttr(attr:Protect(), AST decl, Env env) {
	if(getDefBody(decl) != EmptyBody()) {
		env = errorMark(env, "Dropping body from definition \'<shortDispNameOf(decl)>\' (protect declarations should not have bodies)", Error(), locOf(decl));
	}
	return <attr, env>;
}

tuple[AST, Env] checkAttr(attr:IntegerLiteralAttr(typ, zerof, onef, plus, mul, neg), AST decl, Env env) {
	for(qName <- [typ, zerof, onef, plus, mul, neg]) {
		if(qName != Nop() && !hasDef(qName, env)) {
			termPrintln("literal attr: missing decl: ", qName);
			env = errorMark(env, "Internal Compiler Error: checkAttr: \'<dispNameOf(qName)>\' should be declared in environment", locOf(qName));
		}
	}
	return <attr, env>;
}

default tuple[AST, Env] checkAttr(AST attr, AST decl, Env env) {
	if(attr in modifiers)
		return <attr, env>;
	location = locOf(attr);
	if(location != unknownLoc) { // attr without loc is likely compiler-internal
		name = consOf(attr);
		env = errorMark(env, "Unknown attribute \'<name>\'", locOf(attr));
	}
	return <attr, env>;
}

public tuple[AST,Env] mergeAttrs(decl1:Define(dtor1, seq(attrs1), body1), decl2:Define(dtor2, seq(attrs2), _), Env env) {
	<ren, env> = makeParamRenaming(dtor2, dtor1, env);
	<attrs2, env> = simpleRenameNoWarn(env, attrs2, ren);
	
	<decl1.arg1.args, env> = mergeAttrs(attrs1, attrs2, env);
	return <decl1, env>;
}

public tuple[list[AST],Env] mergeAttrs(list[AST] attrs1, list[AST] attrs2, Env env) {
	list[AST] attrs = attrs1;

	for(attr <- attrs2) {
		switch(attr) {
			case Origins(_):
				;
			case IntegerLiteralAttr(typeName, zeroName, oneName, plusName, mulName, negName): {
				ok = true;
				for(a:IntegerLiteralAttr(_, _, _, _, _, _) <- attrs) {
					if(a != attr) {
						env = errorMark(env, "Incompatible integer literal specification for \'<dispNameOf(typeName)>\'", locOf(attr));
						ok = false;
					}
				}
				if(ok) {
					attrs += [attr];
				}
			}
			case Guard(expr): {
				if(attr notin attrs)
					attrs += [attr];
			}
			case Alert(_,_): {
				if(attr notin attrs)
					attrs += [attr];
			}
			case Concrete(_): {
				if(attr notin attrs)
					attrs += [attr];
			}
			default: {
				if(attr notin modifiers)
					env = errorMark(env, "Don\'t know how to merge attribute \'<consOf(attr)>\'", locOf(attr));
			}
		}
	}

	return <attrs, env>;
}

public tuple[AST,Env] mergeProtect(decl1:Define(dtor1, seq(attrs1), body1), decl2:Define(dtor2, seq(attrs2), _), Env env) {
	<ren, env> = makeParamRenaming(dtor2, dtor1, env);
	<attrs2, env> = simpleRenameNoWarn(env, attrs2, ren);

	if(isOperation(decl1)) {
		<decl1.arg1.args, env> = mergeProtect(attrs1, attrs2, decl2,  env);
	}
	return <decl1, env>;
}


public tuple[list[AST],Env] mergeProtect(list[AST] attrs1, list[AST] attrs2, AST decl2, Env env) {
	list[AST] attrs = attrs1;

	bool ok = false;
	
	for(attr <- attrs2) {
		switch(attr) {
			case Guard(expr): {
				ok = true;
				if(attr notin attrs)
					attrs += [attr];
			}
			case Alert(_,_): {
				ok = true;
				if(attr notin attrs)
					attrs += [attr];
			}
		}
	}
	
	//if(!ok) {
	//	env = errorMark(env, "Protect declaration with no guard", locOf(decl2));
	//}
	return <attrs, env>;
}


public AST getAttr(str name, Define(_, seq(attrs), _)) {
	return getAttr(name, attrs);
}

public AST getAttr(str name, list[AST] attrs) {
	for(attr <- attrs) {
		if(consOf(attr) == name)
			return attr;
	}
	return Nop();
}
