/**************************************************************************
 * Copyright (c) 2012-2014 Anya Helene Bagge
 * Copyright (c) 2012-2014 University of Bergen
 * 
 * All rights reserved. This program and the accompanying materials
 * are made available under the terms of the Eclipse Public License v1.0
 * which accompanies this distribution, and is available at
 * http://www.eclipse.org/legal/epl-v10.html
 * 
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version. See http://www.gnu.org/licenses/
 * 
 * 
 * See the file COPYRIGHT for more information.
 * 
 * Contributors:
 * * Anya Helene Bagge
 * 
 *************************************************************************/
module org::magnolialang::magnolia::typechecker::algo::Primitives
import IO;
import org::magnolialang::magnolia::AbstractSyntax;
import org::magnolialang::magnolia::MagnoliaFacts;
import org::magnolialang::magnolia::lib::MagnoliaCompileLib;
import org::magnolialang::magnolia::lib::Names;
import org::magnolialang::magnolia::lib::Environment;
import org::nuthatchery::pica::terms::Terms;
import org::magnolialang::compiler::Util;
import Set;
import Node;
import List;
import String;
import util::Benchmark;
import util::Math;


public Env addPrimitives(Env env) {
	for(t <- [Axiom(), Predicate()]) {
		env = addName(SpecialType(t), SpecialType(t), TypeKind(), env);
		info = definfo();
		info.locInfo = LocInfo();
		info.decl = TypeClause(SpecialType(t));
		info.defName = SpecialType(t);
		info.qName = SpecialType(t);
		info.parent = QName(seq([]));
		info.attrs = [Builtin()];
		env = addDef(SpecialType(t), info, env);
	}
	env = addPredicate("TRUE", [], "\<p class=\"dochead\"\>Logical true  – \<i\>basic predicate.\</i\>\</p\>\<p\>Always true.\</p\>", env);
	env = addPredicate("FALSE", [], "\<p class=\"dochead\"\>Logical false  – \<i\>basic predicate.\</i\>\</p\>\<p\>Always false.\</p\>", env);
	env = addPredicate("!_", ["e"], "\<p class=\"dochead\"\>Logical not (negation) – \<i\>predicate connective.\</i\>\</p\>
                                     \<p\>
                                       \<code\>!p\</code\> is true if and only if \<code\>p\</code\> is false.
                                     \</p\>
                                     ", env);
	env = addPredicate("_=\>_", ["e1", "e2"], "\<p class=\"dochead\"\>Logical implication – \<i\>predicate connective.\</i\>\</p\>
	                                           \<p\>
	                                             \<code\>p =\> q\</code\> is true if and only if 
	                                             \<ul\>
	                                               \<li\>both \<code\>p\</code\> and \<code\>q\</code\> are true, or\</li\>
	                                               \<li\>\<code\>p\</code\> is false.\</li\>
	                                             \</ul\>
	                                           \</p\>
	                                           \<p\>This operator is short-circuiting, and will only evaluate \<code\>q\<code\> if \<code\>p\</code\> is true.\</p\>
	                                          ", env);
	env = addPredicate("_\<=\>_", ["e1", "e2"], "\<p class=\"dochead\"\>Logical if-and-only-if (biconditional) – \<i\>predicate connective.\</i\>\</p\>
	                                             \<p\>
	                                               \<code\>p &lt;=\> q\</code\> is true if and only if 
	                                               \<ul\>
	                                                 \<li\>both \<code\>p\</code\> and \<code\>q\</code\> are true, or\</li\>
	                                                 \<li\>both \<code\>p\</code\> and \<code\>q\</code\> are false.\</li\>
	                                               \</ul\>
	                                             \</p\>
	                                            ", env);
	env = addPredicate("_&&_", ["e1", "e2"], "\<p class=\"dochead\"\>Logical and (conjunction) – \<i\>predicate connective.\</i\>\</p\>
	                                          \<p\>
	                                            \<code\>p &amp;&amp; q\</code\> is true if and only if 
	                                            \<ul\>
	                                              \<li\>both \<code\>p\</code\> and \<code\>q\</code\> are true.\</li\>
	                                            \</ul\>
	                                          \</p\>
	                                          \<p\>This operator is short-circuiting, and will only evaluate \<code\>q\<code\> if \<code\>p\</code\> is true.\</p\>
	                                          ", env);
	env = addPredicate("_||_", ["e1", "e2"], "\<p class=\"dochead\"\>Logical inclusive or (disjunction) – \<i\>predicate connective.\</i\>\</p\>
	                                          \<p\>
	                                            \<code\>p || q\</code\> is true if and only if 
	                                            \<ul\>
	                                              \<li\>one or both of \<code\>p\</code\> and \<code\>q\</code\> are true.\</li\>
	                                            \</ul\>
	                                          \</p\>
	                                          \<p\>This operator is short-circuiting, and will only evaluate \<code\>q\<code\> if \<code\>p\</code\> is false.\</p\>
	                                         ", env);
	env = addSpecialOp("_-\>_", [<"e1", TypeName(Name(leaf("T")))>, <"e2", TypeName(Name(leaf("T")))>], SpecialType(Predicate()),
	                               "\<p class=\"dochead\"\>Left-to-right substitutability – \<i\>assertion operator.\</i\>\</p\>
                                    \<p\>
                                      \<code\>e1 -\> e2\</code\> implies that
                                      \<ul\>
                                        \<li\>e2 can be substituted for e1 in any expression, without affecting the value of that expression.\</li\>
                                      \</ul\>
                                    \</p\>
                                    \<p\>As a predicate, \<code\>e1 -\> e2\</code\> is true iff the value of \<code\>e1\</code\> is equivalent to the value of \<code\>e2\</code\>.\</p\>
	                               ", env);
	env = addSpecialOp("_==_", [<"e1", TypeName(Name(leaf("T")))>, <"e2", TypeName(Name(leaf("T")))>], SpecialType(Predicate()),
	                               "\<p class=\"dochead\"\>Substitutability – \<i\>assertion operator.\</i\>\</p\>
                                    \<p\>
                                      \<code\>e1 &lt;-\> e2\</code\> implies that
                                      \<ul\>
                                        \<li\>e1 can be substituted for e2, or vice-versa, in any expression, without affecting the value of that expression.\</li\>
                                      \</ul\>
                                    \</p\>
                                    \<p\>As a predicate, \<code\>e1 == e2\</code\> is true iff the value of \<code\>e1\</code\> is equivalent to the value of \<code\>e2\</code\>.\</p\>
	                               ", env);
	env = addSpecialOp("_!=_", [<"e1", TypeName(Name(leaf("T")))>, <"e2", TypeName(Name(leaf("T")))>], SpecialType(Predicate()),
	                               "\<p class=\"dochead\"\>Substitutability – \<i\>assertion operator.\</i\>\</p\>
                                    \<p\>
                                      \<code\>e1 &lt;-\> e2\</code\> implies that
                                      \<ul\>
                                        \<li\>e1 can be substituted for e2, or vice-versa, in any expression, without affecting the value of that expression.\</li\>
                                      \</ul\>
                                    \</p\>
                                    \<p\>As a predicate, \<code\>e1 &lt;-\> e2\</code\> is true iff the value of \<code\>e1\</code\> is equivalent to the value of \<code\>e2\</code\>.\</p\>
	                               ", env);
	return env;
}

public AST predicateName(str strName, int n) {
	return FunName(QName(seq([Name(leaf(strName))])), seq([AnonParam(Obs(), SpecialType(Predicate())) | _ <- [0 .. n]]), SpecialType(Predicate()));
}

public AST builtin(str name, list[str] paramTypes) {
	return Builtin(predicateName(name, size(paramTypes)), leaf("<name>(<strJoin(paramTypes, ",")>)"));
}

public AST binaryPredicate(str name) {
	return builtin(name, ["\\predicate", "\\predicate"]);
}

public AST unaryPredicate(str name) {
	return builtin(name, ["\\predicate", "\\predicate"]);
}

public Env addPredicate(str strName, list[str] params, str doc, Env env) {
	return addSpecialOp(strName, [<p, SpecialType(Predicate())> | p <- params], SpecialType(Predicate()), doc, env);
}

public Env addSpecialOp(str strName, list[tuple[str,AST]] params, AST rt, str doc, Env env) {
	name = Name(leaf(strName));
	funName = FunName(QName(seq([name])), seq([AnonParam(Obs(), t) | <_,t> <- params]), rt);
	info = definfo();
	info.locInfo = LocInfo();
	info.decl = FunClause(funName, ParamList(seq([Param(Obs(), Name(leaf(p)), t) | <p,t> <- params])), rt);
	info.defName = name;
	info.qName = name;
	info.parent = QName(seq([]));
	info.attrs = [Builtin()];
	info.doc = doc;
	info.body = Builtin(funName, leaf("<strName>(<strJoin([nameToStr(y) | <x,y> <- params], ",")>)"));
	env = addName(name, funName, ExprKind(), env);
	env = addDef(funName, info, env);
	return env;
}

public Env addSpecialType(str strName, str doc, Env env) {
	name = Name(leaf(strName));
	typeName = TypeName(QName(seq([name])));
	info = definfo();
	info.locInfo = LocInfo();
	info.decl = TypeClause(typeName);
	info.defName = name;
	info.qName = typeName;
	info.parent = QName(seq([]));
	info.attrs = [Builtin()];
	info.doc = doc;
	info.body = Builtin(typeName, leaf(strName));
	env = addName(name, typeName, TypeKind(), env);
	env = addDef(typeName, info, env);
	return env;
}
