/**************************************************************************
 * Copyright (c) 2012-2014 Anya Helene Bagge
 * Copyright (c) 2012-2014 Tero Hasu
 * Copyright (c) 2012-2014 University of Bergen
 * 
 * All rights reserved. This program and the accompanying materials
 * are made available under the terms of the Eclipse Public License v1.0
 * which accompanies this distribution, and is available at
 * http://www.eclipse.org/legal/epl-v10.html
 * 
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version. See http://www.gnu.org/licenses/
 * vardecl
 * 
 * See the file COPYRIGHT for more information.
 * 
 * Contributors:
 * * Anya Helene Bagge
 * * Tero Hasu
 * 
 *************************************************************************/
module org::magnolialang::magnolia::typechecker::algo::Statements
import IO;
import org::magnolialang::magnolia::AbstractSyntax;
import org::magnolialang::magnolia::MagnoliaFacts;
import org::magnolialang::magnolia::typechecker::toplevel::TopLevelOps;
import org::magnolialang::magnolia::lib::MagnoliaCompileLib;
import org::magnolialang::magnolia::lib::Names;
import org::magnolialang::magnolia::lib::Environment;
import org::magnolialang::magnolia::typechecker::algo::Expressions;
import org::magnolialang::magnolia::typechecker::algo::DerivedOps;
import org::magnolialang::magnolia::typechecker::algo::Types;
import org::magnolialang::magnolia::typechecker::algo::Overload;
import org::nuthatchery::pica::terms::Terms;
import org::magnolialang::compiler::Util;
import org::nuthatchery::pica::memo::Memo;
import Set;
import Node;
import List;
import String;


public tuple[AST,Env] checkStatBody_impl(b:Body(body), Env env) {
	if(Block(_) := body)
		<body, env> = checkStat(body, env);
	else
		<body, env> = checkStat(preserveAnnos(Block(seq([body])), body), env);
	return <b[arg0=body], env>;
}
public tuple[AST,Env] checkStatBody_impl(b:EmptyBody(), Env env) {	return <b, env>; }

public tuple[AST,Env] checkStatBody_impl(b:Instance(_,_), Env env) {	return <b, env>; }

public tuple[AST,Env](AST,Env) checkStatBody = memo(checkStatBody_impl);

public tuple[AST,Env] checkStat(AST stat, Env env) {
	// termPrintln("checkStat: ", stat);
	<s, env> = chkStat(stat, env);
	return <preserveAnnos(s, stat), env>;
}

tuple[AST,Env] chkStat(seq(ss), Env env) {
	ss = for(s <- ss) {
		<s, env> = checkStat(s, env);
		append s;
	}
	return <seq(ss), env>; 
}

tuple[AST,Env] chkStat(b:Block(body), Env env) {
	<s, env> = checkStat(body, env);
	return <Block(s), env>;
}

// Statement:   ;
tuple[AST,Env] chkStat(stat:Nop(), Env env) {
	return <stat, env>;
}

tuple[AST,Env] chkStat(stat:Expect(el:leaf(err), s), Env env) {
	env = expectError(err, 1, env);
	<s, env> = checkStat(s, env);
	env = checkExpected(err, locOf(el), env);
	return <s, env>;
}

// Statement:   if e then s1 else s2
tuple[AST,Env] chkStat(stat:If(e, s1, s2), Env env) {
	<e, et, env> = checkExpr(e, env);
	if(!isAssignable(SpecialType(Predicate()), et, env))
		env = errorMark(env, "If condition should be a predicate (type was <dispNameOf(et, env)>)", locOf(e));
	<s1, env> = checkStat(s1, env);
	<s2, env> = checkStat(s2, env);
	
	return <If(e, s1, s2), env>;
}

// Statement:   while e do ss end
tuple[AST,Env] chkStat(stat:While(e, ss), Env env) {
	<e, et, env> = checkExpr(e, env);
	if(!isAssignable(SpecialType(Predicate()), et, env))
		env = errorMark(env, "While condition should be a predicate (type was <dispNameOf(et, env)>)", locOf(e));
	<ss, env> = checkStat(ss, env);
	
	return <While(e, ss), env>;
}

// Statement:   for v in e do ss end
tuple[AST,Env] chkStat(stat:For(v, e, ss), Env env) {
	<e, et, env> = checkExpr(e, env);
	env = addName(v, et, env);
	<ss, env> = checkStat(ss, env);
	
	return <For(v, e, ss), env>;
}

// Statement:   guard ss by e end
tuple[AST,Env] chkStat(stat:Guard(ss, e), Env env) {
	<e, et, env> = checkExpr(e, env);
	if(!isAssignable(SpecialType(Predicate()), et, env))
		env = errorMark(env, "Guard condition should be a predicate (type was <dispNameOf(et, env)>)", locOf(e));
	<ss, env> = checkStat(ss, env);
	
	return <Guard(ss, e), env>;
}

// Statement:   call proc(args)
tuple[AST,Env] chkStat(stat:Call(proc, seq(args)), Env env) {
	list[AST] argTypes = [];
	list[AST] argValues = [];
	for(a <- args) {
		<av, at, env> = checkExpr(a, env);
		argTypes += at;
		argValues += av;
	}
	
	if(Proc(Instance(_,_)) := proc) {
		// already checked
		return <Call(proc, seq(argValues)), env>;
	}
	else if(Proc(name) := proc) {
		//		termPrintln("chkStat: ", name);
		
		ps = lookupName(name, StatKind(), env);
		if(ps == {}) {
			env = errorMark(env, "Undefined procedure \'<dispNameOf(name, env)>\'", locOf(name));
			return <Call(fakeProc(name, argTypes), seq(argValues)), env>;
		}
		else {
			ops = overload(ps, argTypes, env);
			if(ops == {}) {
				if(size(ps) == 1) {
					if(DEBUG_OVERLOAD) {
						termPrintln("Argument mismatch: formals: ", getOneFrom(ps).arg1.args);
						termPrintln("                   actuals: ", argTypes);
					}
					env = addTokenRef(name, Ref(getOneFrom(ps)), env);
					env = errorMark(env, "Call to \'<dispNameOf(name, env)>(<dispArgList(argTypes, env)>)\' doesn\'t match parameters \'<dispArgList(getOneFrom(ps).arg1.args, env)>\'", locOf(name));
				}
				else {
					for(p <- ps)
						env = addTokenRef(name, Ref(p), env);
					env = errorMark(env, "No matching overloaded procedure for \'<dispNameOf(name, env)>(<dispArgList(argTypes, env)>)\' (<size(ps)> candidates)", locOf(name));
				}
				return <Call(fakeProc(name, argTypes), seq(argValues)), env>;
			}
			else if({p:ProcName(pn,seq(pps))} := ops) {
				if(DEBUG_OVERLOAD) {
					termPrintln("Resolved <dispNameOf(name)>(<dispArgList(argTypes, env)>)");
					termPrintln("   =\> ", p);
				}

				<argValues, env> = checkLiteralArgs(argValues, pps, env);
				env = addTokenRef(name, Ref(p), env);
				return <Call(Proc(p), seq(argValues)), env>;
			}
			else {
				env = errorMark(env, "Ambiguous name \'<dispNameOf(name, env)>\'", locOf(name));
				for(p <- ops)
					env = addTokenRef(name, Ref(p), env);
				return <Call(fakeProc(name, argTypes), seq(argValues)), env>;
			}
		}
	}
	else if(ExternalProc(lang, ns, onPart, name, ParamList(seq(ps))) := proc) {
		env = enterScope(env);
		<ps, env> = checkParams(ps, env);
		env = exitScope(env);
		if(DEBUG_OVERLOAD) {
			termPrintln("StaticSemantics: External procedure: ", lang, ns, name);
			termPrintln("  Param list: ", ps);
		}
		// TODO: add tokenRef to C++ file
		return <Call(ExternalProc(lang, ns, onPart, name, ParamList(seq(ps))), seq(argValues)), env>;
	}
	else {
		env = errorMark(env, "Internal Compiler Error: StaticSemantics::chkStat(): unknown procedure kind <yieldTerm(proc)>", locOf(proc));
		return <Call(proc, seq(argValues)), env>;
	}
}

AST fakeProc(AST name, list[AST] argTypes) {
	return Proc(ProcName(name, seq([AnonParam(Obs(), t) | t <- argTypes])));
}

// Statement:   yield e;
tuple[AST,Env] chkStat(stat:Yield(e), Env env) {
	<e, et, env> = checkExpr(e, env);
	// TODO: check that et is assignable to yield type
	return <Yield(e), env>;
}

// Statement:   break;
tuple[AST,Env] chkStat(stat:Break(), Env env) {
	return <stat, env>;
}

// Statement:   return;
tuple[AST,Env] chkStat(stat:Return(), Env env) {
	return <Return(), env>;
}

// Statement:   return e;
tuple[AST,Env] chkStat(stat:Return(e), Env env) {
	<e, et, env> = checkExpr(e, env);
	// TODO: check that et is assignable to return type
	if(hasDef(returnTypeName, env)) {
		rt = lookupDef(returnTypeName, env).body;
		if(!isAssignable(rt, et, env) && et != Unknown() && rt != Unknown()) {
			env = errorMark(env, "Type \'<dispNameOf(et, env)>\' not returnable as \'<dispNameOf(rt, env)>\'", locOf(stat));
		}
		return <Return(e)[@\type=et], env>;
	}
	else {
		env = errorMark(env, "Attempting to returning value of type \'<dispNameOf(et, env)>\' from operation with no return value", locOf(stat));
		// TODO: check if there are any output args of the same type, and suggest assigning to them
		return <Return(), env>;
	}
	
}

// Statement:   v = e;
tuple[AST,Env] chkStat(stat:Assign(v, e), Env env) {
	if(DotOp(de,dn,dt) := v) {
		if(dt == Unknown())
			return checkStat(Call(Proc(Name(leaf("_set_<nameToStr(dn)>"))), seq([de, e])), env);
		else
			return <stat, env>;
	}
	else {	 
		<vc, vt, env> = checkExpr(v, env);
		<ec, et, env> = checkExpr(e, env);
	
		<ec, et, env> = checkLiteral(ec, et, vt, env);

		if(!isAssignable(vt, et, env) && et != Unknown()) {
			if(DEBUG_TYPECHECK)
				termPrintln("Expression type: ", et);
			env = errorMark(env, "Type \'<dispNameOf(et, env)>\' not assignable to \'<dispNameOf(vt, env)>\'", locOf(stat));
		}
	
		if(Var(n) := vc) {
			return <Assign(vc, ec), env>;
		}
		// TODO: check if vc is an indexing expression
		else {
			env = errorMark(env, "Expression \'<unparse(v)>\' cannot be updated", locOf(v));
			return <Assign(vc, ec), env>;
		}
	}
}

// Statement:   let bindings in stats
tuple[AST,Env] chkStat(stat:Let(decl, seq(stats)), Env env) {
	env = enterScope(env);
	if(VarDecl(m, n, t, e) := decl) {
		<_, t, env> = checkType(t, env);
		<ec, et, env> = checkExpr(e, env);
		<ec, et, env> = checkLiteral(ec, et, t, env);

		if(t == Unknown()) {
			t = et;
		}
		else if(!isAssignable(t, et, env) && et != Unknown()) {
			if(DEBUG_TYPECHECK)
				termPrintln("Expression type: ", et);
			env = errorMark(env, "Type \'<dispNameOf(et, env)>\' not bindable to \'<dispNameOf(t, env)>\'", locOf(decl));
			t = et;
		}
		
		
		qName = preserveAnnos(VarName(n,t), n);
		info = makeDefInfo(qName, env);
		info.defName = n;
		info.srcLoc = locOf(decl);
		info.decl = preserveAnnos(VarDecl(m, n, t, ec), decl);
		info.body = e;
		env = addName(n, qName[@hoverInfo = info], ExprKind(), env);
		decl = preserveAnnos(VarDecl(m, n[@hoverInfo = info], t, ec), decl);
	}
	else
		termPrintln("Wrong vardecl (chkStat): ", decl);
	
	<seq(checkedStats), env> = chkStat(seq(stats), env);
	env = exitScope(env);
	return <Let(decl, seq(checkedStats)), env>;
}

// Statement:   assert e clauses;
tuple[AST,Env] chkStat(stat:Assert(e, seq(clauses)), Env env) {
	// TODO: check that e is a predicate
	<e, et, env> = checkExpr(e, env);

	if(!isAssignable(SpecialType(Predicate()), et, env) && et != Unknown())
		env = errorMark(env, "Assert condition should be an axiom or predicate (type was <dispNameOf(et, env)>)", locOf(e));
	
	clauses = [checkAssertClause(c, env) | c <- clauses];
	
	
	return <Assert(e, seq(clauses)), env>;
}

// Statement:   assume e clauses;
tuple[AST,Env] chkStat(stat:Assume(e, seq(clauses)), Env env) {
	// TODO: check that e is a predicate
	<e, et, env> = checkExpr(e, env);

	if(!isAssignable(SpecialType(Predicate()), et, env) && et != Unknown())
		env = errorMark(env, "Assume condition should be an axiom or predicate (type was <dispNameOf(et, env)>)", locOf(e));
	
	clauses = [checkAssertClause(c, env) | c <- clauses];
	
	
	return <Assume(e, seq(clauses)), env>;
}

	// TODO: make these internal to the previous fun when possible
	// TODO: check that the By clause is an axiom call
	// TODO: get the updated env value
	AST checkAssertClause(By(e), Env env) { return By(checkExpr(e, env)[0]); }
	
	AST checkAssertClause(clause:QED(), Env env) { return clause; }
	
	AST checkAssertClause(clause:Src(_), Env env) { return clause; }

	AST checkAssertClause(clause:Doc(_), Env env) { return clause; }
	
	default AST checkAssertClause(AST tree, Env env) {
		env = errorMark(env, "Internal Compiler Error: StaticSemantics::chkStat(): no checkAssertClause rule for <consOf(tree)>", locOf(tree));
		return tree;
	}

default tuple[AST,Env] chkStat(AST tree, Env env) {
	println("CHECK: no rule for <tree>");
	env = errorMark(env, "Internal Compiler Error: StaticSemantics::chkStat(): no checkStat rule for <consOf(tree)>", locOf(tree));
	return <tree, env>;
}

str dispArgList(list[AST] args, Env env) {
	str dname(AnonParam(_, t)) { return dname(t); }
	str dname(AST n) { return nameToStr(unprefix(defNameOf(n), env.modName)); }
	return strJoin([dname(a) | a <- args], ", ");
}
