/**************************************************************************
 * Copyright (c) 2011-2013 Anya Helene Bagge
 * Copyright (c) 2011-2013 University of Bergen
 * 
 * All rights reserved. This program and the accompanying materials
 * are made available under the terms of the Eclipse Public License v1.0
 * which accompanies this distribution, and is available at
 * http://www.eclipse.org/legal/epl-v10.html
 * 
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version. See http://www.gnu.org/licenses/
 * 
 * 
 * See the file COPYRIGHT for more information.
 * 
 * Contributors:
 * * Anya Helene Bagge
 * 
 *************************************************************************/
module org::magnolialang::magnolia::typechecker::algo::Types
import IO;
import org::magnolialang::magnolia::AbstractSyntax;
import org::magnolialang::magnolia::MagnoliaFacts;
import org::magnolialang::magnolia::lib::MagnoliaCompileLib;
import org::magnolialang::magnolia::lib::Environment;
import org::magnolialang::magnolia::typechecker::algo::Expressions;
import org::magnolialang::magnolia::typechecker::algo::Attributes;
import org::magnolialang::magnolia::lib::Names;
import org::nuthatchery::pica::terms::Terms;
import Set;
import Node;
import List;

public AST returnTypeName = Name(leaf("return type"));

@doc{True if t2 is assignable to t1.}
public bool isAssignable(AST t1, AST t2, Env env) {
	if(t1 == t2)
		return true;
	else if(SpecialType(Literal()) := t2) {
		if(hasDef(t1, env)) {
			info = lookupDef(t1, env);
			if(IntegerLiteralAttr(_,_,_,_,_,_) := getAttr("IntegerLiteralAttr", info.attrs))
				return true;
			else
				return false;
		}
		else {
			return false;
		}
	}
	else if(SpecialType(Predicate()) := t1  && SpecialType(_) := t2)
		return true;
	else if(t1 == Unknown())
		return true;
	else
		return false;
}

@doc{Least upper bound on types.
  Find the narrowest type that can hold values of types t1 and t2.
} 
public AST lub(AST t1, AST t2) {
	if(t1 == t2)
		return t1;
	else if(SpecialType(_) := t1 && SpecialType(_) := t2)
		return SpecialType(Predicate());
	else
		return Unknown();
}

public AST lub(set[AST] types) {
	<result, types> = takeOneFrom(types);
	while(!isEmpty(types)) {
		<typ, types> = takeOneFrom(types);
		result = lub(result, typ);
	}
	return result;
}


@doc{Second element may be Unknown() if type isn't declared.
Otherwise the first two elements are equal.}
public default tuple[AST,AST,Env] checkType(AST typ, Env env) {
	env = errorMark(env, "Internal Compiler Error: Types: no checkType rule for <consOf(typ)>", locOf(typ));
	return <typ, typ, env>;
}

public tuple[AST,AST,Env] checkType(typ:Unknown(), Env env) {
	return <typ, typ, env>;
}

public tuple[AST,AST,Env] checkType(Type(typ), Env env) {
	return chkType(typ, env);
}

public tuple[AST,AST,Env] checkType(typ:TypeName(_), Env env) {
	return <typ, typ, env>;
}

public tuple[AST,AST,Env] checkType(typ:Instance(t,_), Env env) {
	<_, _, env> = checkType(t, env);
	return <typ, typ, env>;
}

public tuple[AST,AST,Env] checkType(typ:SpecialType(_), Env env) {
	return chkType(typ, env);
}

public tuple[AST,AST,Env] checkType(typ:Expect(el:leaf(err), t), Env env) {
	env = expectError(err, 1, env);
	<t, u, env> = checkType(t, env);
	env = checkExpected(err, locOf(el), env);
	return <y, u, env>;
}

@doc{Accept argument as a valid type without checking}
public default tuple[AST,Env] makeType(AST typ, Env env) {
	// env = errorMark(env, "Internal Compiler Error: Types: no makeType rule for <consOf(typ)>", locOf(typ));
	return <typ, env>;
}

public tuple[AST,Env] makeType(Type(typ), Env env) {
	return <TypeName(typ), env>;
}


private tuple[AST,AST,Env] chkType(AST typ, Env env) {
	ts = lookupName(typ, TypeKind(), env);
	if(ts == {}) {
		env = errorMark(env, "Undefined type \'<dispNameOf(typ, env)>\'", locOf(typ));
		if(TypeName(_) := typ)
			return <typ, Unknown(), env>;
		else
			return <preserveAnnos(TypeName(typ), typ), Unknown(), env>;
	}
	else if({t} := ts) {
		try {
			if(Body(dt:TypeName(_)) := (lookupDef(t, env).body ? EmptyBody())) {
				env = addTokenRef(typ, dt, env);
				return <preserveAnnos(t, typ), preserveAnnos(dt, typ), env>;
			}
		}
		catch NoSuchKey(_): {
			env = errorMark(env, "Internal Compiler Error: StaticSemantics: resolved name has no definition: <dispNameOf(t, env)>", locOf(typ));
		}
		env = addTokenRef(typ, t, env);
		return <preserveAnnos(t, typ), preserveAnnos(t, typ), env>;
	}
	else {
		env = errorMark(env, "Ambiguous name \'<dispNameOf(typ, env)>\'", locOf(typ));
		t = preserveAnnos(getOneFrom(ts), typ);
		for(tt <- ts)
			env = addTokenRef(typ, tt, env);
		if(TypeName(_) := t)
			return <t, t, env>;
		else
			return <preserveAnnos(TypeName(t), typ), t, env>;
	}
} 

@doc{Check the field declarations of a struct type body. Field initialisers (if present) are *not* checked.}
public tuple[AST, Env] checkTypeFieldDecls(b:Body(tb), Env env) {
	<tb, env> = chkTypeFieldDecls(tb, false, env);
	return <preserveAnnos(Body(tb), b), env>;
}

public tuple[AST, Env] checkTypeFieldDecls(b:EmptyBody(), Env env) {
	return <b, env>;
}

public default tuple[AST, Env] checkTypeFieldDecls(AST body, Env env) {
	env = errorMark(env, "Internal Compiler Error: StaticSemantics::checkTypeFieldDecls(): no check rule for <consOf(body)>", locOf(body));
	return <body, env>;
}

public tuple[AST, Env] checkTypeBody(b:Body(tb), Env env) {
	<tb, env> = chkTypeFieldDecls(tb, true, env);
	return <preserveAnnos(Body(tb), b), env>;
}

public tuple[AST, Env] checkTypeBody(b:EmptyBody(), Env env) {
	return <b, env>;
}

public default tuple[AST, Env] checkTypeBody(AST body, Env env) {
	env = errorMark(env, "Internal Compiler Error: StaticSemantics::checkTypeBody(): no check rule for <consOf(body)>", locOf(body));
	return <body, env>;
}

tuple[AST, Env] chkTypeFieldDecls(s:Struct(seq(ds)), bool checkInits, Env env) {
	ds = for(d <- ds) {
		<d, env> = chkTypeField(d, checkInits, env);
		append d;
	};
	return <preserveAnnos(Struct(seq(ds)), s), env>;
}

tuple[AST, Env] chkTypeFieldDecls(t:TypeName(_), bool checkInits, Env env) {
	return <t, env>; // it's already checked
}

tuple[AST, Env] chkTypeFieldDecls(n:Type(_), bool checkInits, Env env) {
	<t, _, env> = checkType(n, env);
	return <t, env>; // will now refer to a type name
}

tuple[AST, Env] chkTypeFieldDecls(t:ExternalType(lang, modl, _, name), bool checkInits, Env env) {
	return <t, env>;
}

tuple[AST,Env] chkTypeFieldDecls(typ:Expect(el:leaf(err), t), bool checkInits, Env env) {
	env = expectError(err, 1, env);
	<typ, env> = chkTypeFieldDecls(t, checkInits, env);
	env = checkExpected(err, locOf(el), env);
	return <typ, env>;
}

default tuple[AST, Env] chkTypeFieldDecls(AST body, bool checkInits, Env env) {
	env = errorMark(env, "Internal Compiler Error: StaticSemantics::chkTypeFieldDecls(): no check rule for <consOf(body)>", locOf(body));
	return <body, env>;
}

tuple[AST, Env] chkTypeField(f:TypeField(d:VarDecl(m, n, t, e)), bool checkInits, Env env) {
	if(Unknown() := t) {
		env = errorMark(env, "Field must have a declared type", locOf(d));
	}
	else {
		<t, _, env> = checkType(t, env);
	}
	if(m != Upd()) {
		env = errorMark(env, "Internal Compiler Error: StaticSemantics::chkTypeField(): mode should be upd", locOf(f));
	}
	if(checkInits) {
		<e, et, env> = checkExpr(e, env);
		if(!isAssignable(t, et, env) && et != Unknown() && t != Unknown()) {
			env = errorMark(env, "Type \'<dispNameOf(et, env)>\' not assignable to \'<dispNameOf(t, env)>\'", locOf(d));
		}
	}
	return <preserveAnnos(TypeField(preserveAnnos(VarDecl(m, n, t, e), d)), f), env>;
}

default tuple[AST, Env] chkTypeField(AST d, bool checkInits, Env env) {
	env = errorMark(env, "Internal Compiler Error: StaticSemantics::chkTypeField(): no check rule for <consOf(d)>", locOf(d));
	return <d, env>;
}

public tuple[AST, Env] mergeTypeDecls(AST decl1, AST decl2, Env env) {
	
	<body, env> = mergeTypeFieldDecls(decl1, decl2, defNameOf(decl1), env);
	if(body == getDefBody(decl1)) {
		return mergeAttrs(decl1, decl2, env);
	} 
	else if(body == getDefBody(decl2)) {
		return mergeAttrs(decl2, decl1, env);
	}
	else {
		decl1.arg2 = body;
		return mergeAttrs(decl1, decl2, env);
	}
}

public tuple[AST, Env] mergeTypeFieldDecls(AST decl1, AST decl2, AST typeName, Env env) {
	body1 = getDefBody(decl1);
	body2 = getDefBody(decl2);
	if(body1 == body2) {
		return <body2, env>;
	}
	else if(<Body(Struct(seq(ds1))), Body(Struct(seq(ds2)))> := <body1, body2>) {
		map[AST, tuple[AST,AST]] names = ();
		list[AST] result = [];

		for(d2:TypeField(VarDecl(m2, n2, t2, e2)) <- ds2) {
			names[n2] = <t2, d2>;
			result += d2;
		}
		for(d1:TypeField(VarDecl(m1, n1, t1, e1)) <- ds1) {
			if(n1 in names) {
				<t2, d2> = names[n1];
				if(t1 != t2) {
					<via1, loc1> = getLatestOrigin(decl1, locOf(d1));
					<via2, loc2> = getLatestOrigin(decl2, locOf(d2));
					otherDef1 = "(other definition <via2 == "" ? "" : "at <locInfo(locOf(d2), loc1)>, via <via2> ">at <locInfo(loc2, loc1)>)";
					otherDef2 = "(other definition <via1 == "" ? "" : "at <locInfo(locOf(d1), loc2)>, via <via1> ">at <locInfo(loc1, loc2)>)";

					env = errorMark(env, "Conflicting type for field \'<nameToStr(n1)>\' of type \'<nameToStr(typeName)>\': expected \'<nameToStr(t2)>\', was \'<nameToStr(t1)>\' <otherDef1>", loc1);
					env = errorMark(env, "Conflicting type for field \'<nameToStr(n1)>\' of type \'<nameToStr(typeName)>\': expected \'<nameToStr(t1)>\', was \'<nameToStr(t2)>\' <otherDef2>", loc2);
				}	
			}
			else {
				result += d1;
			}
		}
		termPrintln("mergeTypeFieldDecls: merged to: ", result);
		return <preserveAnnos(Body(Struct(seq(result))), body1), env>;
	}
	else if(<EmptyBody(), _> := <body1, body2>) {
		return <body2, env>;
	}
	else if(<_, EmptyBody()> := <body1, body2>) {
		return <body1, env>;
	}
	else {
		<via1, loc1> = getLatestOrigin(decl1, locOf(decl1));
		<via2, loc2> = getLatestOrigin(decl2, locOf(decl2));
		otherDef1 = "(other definition <via2 == "" ? "" : "via <via2> ">at <locInfo(loc2, loc1)>)"; 
		otherDef2 = "(other definition <via1 == "" ? "" : "via <via1> ">at <locInfo(loc1, loc1)>)"; 

		env = errorMark(env, "Conflicting definitions of type \'<nameToStr(typeName)>\' <otherDef1>", loc1);
		env = errorMark(env, "Conflicting definitions of type \'<nameToStr(typeName)>\' <otherDef2>", loc2);
		return <body2, env>;
	}
}
