/**************************************************************************
 * Copyright (c) 2012-2013 Anya Helene Bagge
 * Copyright (c) 2012-2013 University of Bergen
 * 
 * All rights reserved. This program and the accompanying materials
 * are made available under the terms of the Eclipse Public License v1.0
 * which accompanies this distribution, and is available at
 * http://www.eclipse.org/legal/epl-v10.html
 * 
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version. See http://www.gnu.org/licenses/
 * 
 * 
 * See the file COPYRIGHT for more information.
 * 
 * Contributors:
 * * Anya Helene Bagge
 * 
 *************************************************************************/
module org::magnolialang::magnolia::typechecker::toplevel::Filters

import org::magnolialang::magnolia::typechecker::Flattener;
import org::magnolialang::magnolia::AbstractSyntax;
import org::nuthatchery::pica::terms::Terms;
import org::magnolialang::magnolia::lib::MagnoliaCompileLib;
import org::magnolialang::magnolia::lib::Names;
import org::magnolialang::magnolia::typechecker::TypeChecker;
import org::magnolialang::magnolia::lib::Environment;
import org::magnolialang::magnolia::lib::CxxAPI;
import org::magnolialang::magnolia::MagnoliaFacts;
import org::magnolialang::magnolia::typechecker::toplevel::TopLevelOps;
import org::magnolialang::magnolia::evaluator::Values;
import IO;
import Set;
import List;
import Node;

data Filter = AllFilter() | OnFilter() | DefinesFilter();

private str retainFilterDoc = "\<p\>The \'\<code\>retain\</code\>\' filter will keep all declarations on the
left-hand side that are also on the right-hand side,[unimplemented: \<i\>and\</i\> any other declarations that are used
by the retained declarations.]\</p\>
\<p\>The variant \'\<code\>on retain\</code\>\' and \'\<code\>declared retain\</code\>\' will apply the filter only
to the on-part or the declared-part (respectively) of the left-hand side, all other declarations will be
kept.\</p\>";

private str removeFilterDoc = "\<p\>The \'\<code\>remove\</code\>\' filter will remove all declarations on the
left-hand side that are also on the right-hand side,[unimplemented: \<i\>and\</i\> any other declarations that are used
by the removed declarations.]\</p\>
\<p\>The variant \'\<code\>on retain\</code\>\' and \'\<code\>declared retain\</code\>\' will apply the filter only
to the on-part or the declared-part (respectively) of the left-hand side, all other declarations will be
kept.\</p\>";
public tuple[AST, Env] applyFilter(AST op, Filter fKind, AST lhs, AST rhs, loc location,  Env env) {
	<lhs, env> = flattenTopExpr(lhs, env);
	env = enterScope(env);
	<_, env> = loadNames(lhs, false, env);
	<rhs, env> = flattenTopExpr(rhs, env);
	env = exitScope(env);
	if(!isFlat(lhs) || !isFlat(rhs)) {
		switch(fKind) {
			case AllFilter(): return <Filtered(lhs, op, rhs), env>;
			case OnFilter(): return <OnFilter(lhs, op, rhs), env>;
			case DefinesFilter(): return <DeclaredFilter(lhs, op, rhs), env>;
		};
	}
	if(fKind == OnFilter())
		env = errorMark(env, "\'on\' filter not supported yet, will filter everything instead", Warning(), locOf(op));
	else if(fKind == DefinesFilter())
		env = errorMark(env, "\'declared\' filter not supported yet, will filter everything instead", Warning(), locOf(op));
	
	try {
		set[AST] filterDecls = {canonical(d) | d <- fullOf(rhs)};
		if(leaf("retain") := op) {
			<lhs.arg0.args, leftOver> = retain(lhs.arg0.args, filterDecls);
			for(d <- leftOver)
				env = errorMark(env, "Retained declaration not found in left-hand side", Warning(), locOf(d));
			env = addTokenRefForOp(op, lhs, retainFilterDoc, env);
		}
		else if(leaf("remove") := op) {
			<lhs.arg0.args, leftOver> = remove(lhs.arg0.args, filterDecls);
			for(d <- leftOver)
				env = errorMark(env, "Removed declaration not found in left-hand side", Warning(), locOf(d));
			env = addTokenRefForOp(op, lhs, removeFilterDoc, env);
		}
		return <lhs, env>;
	}
	catch InternalError(msg, ast): {
		env = errorMark(env, msg, Internal(), locOf(ast));
	}
	return <Filtered(lhs, op, rhs), env>;
}


private tuple[list[AST], set[AST]] retain(list[AST] decls, set[AST] toRetain) {
	list[AST] result = [];
	for(d <- decls) {
		if(canonical(d) in toRetain) {
			toRetain -= canonical(d);
			result += d;
		}
	}
	return <result, toRetain>;
}

private tuple[list[AST], set[AST]] remove(list[AST] decls, set[AST] toRemove) {
	list[AST] result = [];
	for(d <- decls) {
		if(canonical(d) notin toRemove)
			result += d;
		else
			toRemove -= canonical(d);
			
	}
	return <result, toRemove>;
}
