/**************************************************************************
 * Copyright (c) 2012-2014 Anya Helene Bagge
 * Copyright (c) 2012-2014 University of Bergen
 * 
 * All rights reserved. This program and the accompanying materials
 * are made available under the terms of the Eclipse Public License v1.0
 * which accompanies this distribution, and is available at
 * http://www.eclipse.org/legal/epl-v10.html
 * 
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version. See http://www.gnu.org/licenses/
 * 
 * 
 * See the file COPYRIGHT for more information.
 * 
 * Contributors:
 * * Anya Helene Bagge
 * 
 *************************************************************************/
module org::magnolialang::magnolia::typechecker::toplevel::TopTestRunner
import IO;
import org::magnolialang::magnolia::AbstractSyntax;
import org::magnolialang::magnolia::MagnoliaFacts;
import org::magnolialang::magnolia::typechecker::Flattener;
import org::magnolialang::magnolia::typechecker::toplevel::TopLevelOps;
import org::magnolialang::magnolia::typechecker::toplevel::TopPrimitives;
import org::magnolialang::magnolia::lib::MagnoliaCompileLib;
import org::magnolialang::magnolia::lib::Names;
import org::magnolialang::magnolia::lib::Renaming;
import org::magnolialang::magnolia::typechecker::TypeChecker;
import org::magnolialang::magnolia::lib::Environment;
import org::magnolialang::magnolia::typechecker::algo::Primitives;
import org::magnolialang::magnolia::evaluator::Values;
import org::nuthatchery::pica::terms::Terms;
import org::magnolialang::compiler::Util;
import Set;
import Node;
import String;
import List;
import Exception;
import util::Benchmark;

data Exception
	= ReturnValue(Value val, Env env)
	| AssertionFailed(Env env)
	| EvaluationError(Env env)
	;
	
public bool topTestRun(Transaction tr, type[TestReport] k, tuple[list[type[&T]],value] key) {
	name = getCombiKeyName(key);
	info = getFact(tr, #DefInfo, name);
	if(info.kind == KIND_PACKAGE) {
		set[ErrorMark] errors = {};
		for(n <- getFact(tr, #Contents, name).val) {
			nInfo = getFact(tr, #DefInfo, n);
			if(nInfo.kind == "test") {
				fact = getFact(tr, #TestReport, makeCombiKey(key, n));
				errors += fact.errors;
			}
		}
		result = errors == {} ? Success() : Failed("test failed");
		setFact(tr, #TestReport, key, TestReport(result, errors));
	}
	else {
		set[ErrorMark] errors = {};
		arg = getFact(tr, getCombiKeyHead(key), getCombiKeyTail(key));

		time = getMilliTime();
		if(arg has val) {
			result = Success();
			if(info.kind == "test" && ProcClause(_,ParamList(seq([]))) := info.decl) {
				tree = getVal(arg);
				errors = getErrors(arg);
				if(errors == {}) {
					errors = testTopLevel(tr, info, tree);
				}
				if(errors != {}) {
					result = Failed("test failed");
				}
			}
			setFact(tr, #TestReport, key, TestReport(result, errors));
			
		}
		else {
			println("TESTRUNNER ERROR: <arg>");
		}
		println("Test run: <getMilliTime()-time> ms (for <nameToStr(astValue(key[1]))>)");
		
	}
	return true;
}

set[ErrorMark] testTopLevel(Transaction tr, DefInfo info, AST tree) {
	println("Testing <nameToStr(info.qName)>");
	body = tree;
	termPrintln("  ", tree);
	env = newEnv(tr, info.qName, info.parent, body@\loc);
	env = addPrimitives(env);
	env = addTopPrimitives(env);
	scope = getFact(tr, #Scope, info.qName);
	env = addTopScope(scope, env, true);
	// env = addTokenRef(info.defName, Ref(info.qName), env);
	ok = true;
	env = enterScope(env);
	try {
		<result, env> = evaluateBody(body, env);
	}
	catch EvaluationError(e):
		env = e;
	catch AssertionFailed(e): {
		env = e;
		ok = false;
	}
	env = exitScope(env);
	
	errors = env.errors;
	if(errors == {})
		setMark(tr, "OK", Info(), locOf(info.decl));
	else if(!ok)
		setMark(tr, "FAIL", Error(), locOf(info.decl));
	else
		setMark(tr, "ERROR", Error(), locOf(info.decl));
	
	return env.errors;
}

tuple[Value, Env] evaluateBody(Body(tree:Block(_)), Env env) {
	result = VoidValue();
	try {
		env = evaluateStat(tree, env);
	}
	catch ReturnValue(v, e):
		<result, env> = <v, e>;
		 
	return <result, env>;	
}

tuple[Value, Env] evaluateBody(Body(tree), Env env) {
	result = VoidValue();
	try {
		<result, env> = evaluateExpr(tree, env);
	}
	catch ReturnValue(v, e):
		<result, env> = <v, e>;
		 
	return <result, env>;	
}

Env evaluateStat(Block(seq(ss)), Env env) {
	for(s <- ss) {
		env = evaluateStat(s, env);
	}
	
	return env;
}

Env evaluateStat(Return(e), Env env) {
	<res, env> = evaluateExpr(e, env);
	throw ReturnValue(res, env);
}

Env evaluateStat(Nop(), Env env) {
	return env;
}

tuple[Value,Env] evaluateExpr(expr:BlockExpr(seq(ss), rt), Env env) {
	result = VoidValue();
	try {
		for(s <- ss) {
			env = evaluateStat(s, env);
		}
	}
	catch ReturnValue(v, e):
		<result, env> = <v, e>;
	
	return <result, env>;
}

tuple[Value,Env] evaluateExpr(expr:Var(n), Env env) {
	// termPrintln("evaluatExpr:Var: ", Var(n));
	if(TopName(_,_) := n)
		return <AstValue(lookupDef(n, env).body), env>;
	else
		return <getVar(n, env), env>;
}

tuple[Value,Env] evaluateExpr(expr:Equal(op, a, b), Env env) {
	if(Fun(name) := op) {
		bool result = false;		
		bool mode =  nameToStr(name) == "_!=_";
		DefInfo info;
		if(isQualifiedName(name)) {
			try {
				info = lookupDef(name, env);
				return evaluateExpr(Apply(op, seq([a, b])), env);
			}
			catch NoSuchKey(_): {
				;
			}
		}

		<av, env> = evaluateExpr(a, env);
		<bv, env> = evaluateExpr(b, env);
//		termPrintln("Equal av: ", av);
//		termPrintln("Equal bv: ", bv);
			
		if(<AstValue(def1:Define(_,_,_)), AstValue(def2:Define(_,_,_))> := <av, bv>) {
			return <BoolValue(canonical(def1) == canonical(def2)), env>;
		}
		else if(<AstValue(DeclBody(seq(defs1))), AstValue(DeclBody(seq(defs2)))> := <av, bv>) {
			if(size(defs1) != size(defs2))
				result = false;
			else
				result = size(sigDiff(defs1, defs2)) == 0;
		}
		else if(<AstValue(RenameList(seq(defs1))), AstValue(RenameList(seq(defs2)))> := <av, bv>) {
			if(size(defs1) != size(defs2))
				result = false;
			else
				result = size(sigDiff(defs1, defs2)) == 0;
		}
		else if(av == bv) {
			result = true;
		}
		else if(<AstValue(ast1), AstValue(ast2)> := <av, bv>) {
			env = errorMark(env, "No equals defined for <consOf(ast1)> == <consOf(ast1)>", locOf(a));
			result = false;
		}
		else if(<BoolValue(b1), BoolValue(b2)> := <av, bv>) {
			result = b1 == b2;
		}		
		else if(<IntValue(v1), IntValue(v2)> := <av, bv>) {
			result = v1 == v2;
		}		
		else if(<RealValue(v1), RealValue(v2)> := <av, bv>) {
			result = v1 == v2;
		}		
		else if(<StrValue(v1), BoolValue(v2)> := <av, bv>) {
			result = v1 == v2;
		}
		else {
			env = errorMark(env, "No equals defined for <nameToStr(typeOf(av))> == <nameToStr(typeOf(bv))>", locOf(a));
			result = false;
		}
		return <BoolValue(xor(result, mode)), env>;
	}
}
tuple[Value,Env] evaluateExpr(expr:Apply(func, seq(as)), Env env) {
	
	if(Fun(name) := func) {
		DefInfo info;
		if(isQualifiedName(name)) {
			try {
				info = lookupDef(name, env);
			}
			catch NoSuchKey(_): {
				termPrintln("TopTestRunner::evaluateExpr::Apply: function not found: ", name);
				env = errorMark(env, "No function definition found for \'<dispNameOf(name, env)>(...)\'", locOf(name));
				throw EvaluationError(env);
			}
		}
		else {
			fs = lookupName(name, ExprKind(), env);
			if({f} := fs) {
				info = lookupDef(f, env);
			}
			else {
				termPrintln("evaluateExpr::Apply: ", name);
				env = errorMark(env, "No function found for \'<dispNameOf(name, env)>(...)\'", locOf(name));
				throw EvaluationError(env);
			}
			
		}

		if(FunClause(_, ParamList(seq(ps)), _) := info.decl) {
			list[Value] avs;
			avs = for(a <- as) {
				<av, env> = evaluateExpr(a, env);
				append av;
			};
			
			termPrintln("Applying function: ", info.qName);
			env = enterScope(env);
			for(<av,p:Param(m,n,t)> <- zip(avs, ps)) {
				termPrintln("  arg: ", p, " = ", av);
				env = addVar(VarName(n, t), env);
				env = setVar(VarName(n, t), av, env);
			}
			// termPrintln("  body: ", info.body);
			<res, env> = evaluateBody(info.body, env);
			//termPrintln("Bindings: ", env.bindings);
			//termPrintln("Stack: ", head(env.stack, env.stackTop));
			env = exitScope(env);
			//termPrintln("After Bindings: ", env.bindings);
			//termPrintln("After Stack: ", head(env.stack, env.stackTop));
			//termPrintln("Evaluation result: ", res);
			return <res, env>;
		}
		else {
			termPrintln("info.decl: ", info.decl);
			env = errorMark(env, "Internal Error: not a function <consOf(info.decl)>", locOf(name));
			throw EvaluationError(env);
		}
	}
	else if(ExternalFun(lang, ns, onPart, name, ParamList(seq(ps0)), rt) := func) {
		env = enterScope(env);
		<ps, env> = checkParams(ps0, env);
		<rt, _, env> = checkType(rt, env);
		env = exitScope(env);
		if(DEBUG_OVERLOAD) {
			termPrintln("StaticSemantics: External function: ", lang, ns, name);
			termPrintln("  Param list: ", ps);
			termPrintln("  Return type: ", rt);
		}
		// TODO: add xref to C++ file
		return <Apply(ExternalFun(lang, ns, onPart, name, ParamList(seq(ps)), rt), seq(argValues)), env>;
	}
	else if(Builtin(n, s) := func) {
		switch(func) {
			case Builtin(_,leaf("TRUE()")): {
				acceptArgs([], as, expr, env);
				return <BoolValue(true), env>;
			}
			case Builtin(_,leaf("FALSE()")): {
				acceptArgs([], as, expr, env);
				return <BoolValue(false), env>;
			}
			case Builtin(_,leaf("!_(\\predicate)")): {
				<vs, env> = acceptArgs([boolType], as, expr, env);
				if([BoolValue(a)] := vs)
					return <BoolValue(!a), env>;
			}
			case Builtin(_,leaf("_&&_(\\predicate,\\predicate)")): {
				<vs, env> = acceptArgs([boolType, boolType], as, expr, env);
				if([BoolValue(a), BoolValue(b)] := vs)
					return <BoolValue(a && b), env>;
			}
			case Builtin(_,leaf("_||_(\\predicate,\\predicate)")): {
				<vs, env> = acceptArgs([boolType, boolType], as, expr, env);
				if([BoolValue(a), BoolValue(b)] := vs)
					return <BoolValue(a || b), env>;
			}
			case Builtin(_,leaf("_\<=\>_(\\predicate,\\predicate)")): {
				<vs, env> = acceptArgs([boolType, boolType], as, expr, env);
				if([BoolValue(a), BoolValue(b)] := vs)
					return <BoolValue(a == b), env>;
			}
			case Builtin(_,leaf("_==_(\\predicate,\\predicate)")): {
				<vs, env> = acceptArgs([boolType, boolType], as, expr, env);
				if([BoolValue(a), BoolValue(b)] := vs)
					return <BoolValue(a == b), env>;
			}
			case Builtin(_,leaf("_!=_(\\predicate,\\predicate)")): {
				<vs, env> = acceptArgs([boolType, boolType], as, expr, env);
				if([BoolValue(a), BoolValue(b)] := vs)
					return <BoolValue(a != b), env>;
			}
			case Builtin(_,leaf("_=\>_(\\predicate,\\predicate)")): {
				<vs, env> = acceptArgs([boolType, boolType], as, expr, env);
				if([BoolValue(a), BoolValue(b)] := vs)
					return <BoolValue((a && b) || !a), env>;
			}
			case Builtin(_,leaf("_in_(\\decl,\\decllist)")): {
				<vs, env> = acceptArgs([declType, decllistType], as, expr, env);
				return <BoolValue(isIn(vs[0].a, vs[1].a)), env>;
			}
			case Builtin(_,leaf("_[_](\\decl,\\renaming)")): {
				<vs, env> = acceptArgs([declType, renamingType], as, expr, env);
				if([AstValue(body), AstValue(RenameList(seq(rns)))] := vs) {
					<result, env> = simpleRename(env, body, rns);
					return <AstValue(result), env>;
				}
				env = errorMark(env, "Bad renaming \'<vs[1]>\'", locOf(as[1]));
				throw EvaluationError(env);
			}
			case Builtin(_,leaf("_[_](\\decllist,\\renaming)")): {
				<vs, env> = acceptArgs([decllistType, renamingType], as, expr, env);
				if([AstValue(body), AstValue(RenameList(seq(rns)))] := vs) {
					<result, env> = simpleRename(env, body, rns);
					return <AstValue(result), env>;
				}
				env = errorMark(env, "Bad renaming \'<vs[1]>\'", locOf(as[1]));
				throw EvaluationError(env);
			}
			case Builtin(_,leaf(b)): {
				env = errorMark(env, "Unknown builtin \'<b>\'", locOf(name));
				throw EvaluationError(env);
			}
			default: {
				env = errorMark(env, "Unknown builtin \'<dispNameOf(name, env)>(...)\'", locOf(name));
				throw EvaluationError(env);
			}
		}
	}
	else if(FunOf(p:ProcName(_,seq(ps)), leaf(i)) := func) {
		env = addTokenRef(func, Ref(p), env);
		if(AnonParam(_,rt) := ps[toInt(i)])
			return <Apply(func, seq(argValues)), env>;
		else {
			env = errorMark(env, "Internal Compiler Error: StaticSemantics::chkExpr(): unknown parameter <ps[toInt(i)]>", locOf(func));
			throw EvaluationError(env);
		}
	}
	else {
		env = errorMark(env, "Internal Compiler Error: StaticSemantics::chkExpr(): unknown function kind <yieldTerm(func)>", locOf(func));
		throw EvaluationError(env);
	}
}

Env evaluateStat(stat:Call(proc, seq(as)), Env env) {
	if(Proc(name) := proc) {
		DefInfo info;
		if(isQualifiedName(name)) {
			try {
				info = lookupDef(name, env);
			}
			catch NoSuchKey(_): {
				termPrintln("TopTestRunner::evaluateStat::Call: procedure not found: ", name);
				env = errorMark(env, "No procedure definition found for \'<dispNameOf(name, env)>(...)\'", locOf(name));
				throw EvaluationError(env);
			}
		}
		else {
			ps = lookupName(name, StatKind(), env);
			if({p} := ps) {
				info = lookupDef(p, env);
			}
			else {
				termPrintln("evaluateStat::Call: ", name);
				env = errorMark(env, "No procedure found for \'<dispNameOf(name, env)>(...)\'", locOf(name));
				throw EvaluationError(env);
			}
			
		}

		if(ProcClause(_, ParamList(seq(ps))) := info.decl) {
			list[Value] avs;
			avs = for(a <- as) {
				<av, env> = evaluateExpr(a, env);
				append av;
			};
			
			termPrintln("Calling procedure: ", info.qName);
			env = enterScope(env);
			for(<av,p:Param(m,n,t)> <- zip(avs, ps)) {
				termPrintln("  arg: ", p, " = ", av);
				env = addVar(VarName(n, t), env);
				env = setVar(VarName(n, t), av, env);
			}
			// termPrintln("  body: ", info.body);
			<_, env> = evaluateBody(info.body, env);
			//termPrintln("Bindings: ", env.bindings);
			//termPrintln("Stack: ", head(env.stack, env.stackTop));
			env = exitScope(env);
			//termPrintln("After Bindings: ", env.bindings);
			//termPrintln("After Stack: ", head(env.stack, env.stackTop));
			return env;
		}
		else {
			termPrintln("info.decl: ", info.decl);
			env = errorMark(env, "Internal Error: not a procedure <consOf(info.decl)>", locOf(name));
			throw EvaluationError(env);
		}
	}
	else {
		env = errorMark(env, "Internal Compiler Error: TopTestRunner::evaluateStat(): unknown procedure kind <yieldTerm(proc)>", locOf(proc));
		throw EvaluationError(env);
	}
}

tuple[Value,Env] evaluateExpr(expr:DotOp(e, n, _), Env env) {
	termPrintln("evaluatExpr:DotOp: ", e, ", ", n);
	<vs, env> = acceptArgs([decllistType], [e], expr, env);
	if([AstValue(DeclBody(seq(decls)))] := vs) {
		for(def:Define(dcl,_,_) <- decls) {
			termPrintln("evaluateExpr:DotOp: def: ", defNameOf(dcl));
			if(basename(defNameOf(dcl)) == n) {
				return <AstValue(def), env>;
			} 
		}
		env = errorMark(env, "No such declaration: <nameToStr(n)>", locOf(n));
		throw EvaluationError(env);
	}
	else {
		env = errorMark(env, "Internal Compiler Error: evaluateExpr::DotOp(): top-level definition not flattened (<vs[0]>)", locOf(e));
		throw EvaluationError(env);
	}
}

tuple[Value,Env] evaluateExpr(expr:RenameList(_), Env env) {
	return <AstValue(expr), env>;
}

tuple[Value,Env] evaluateExpr(expr:Rename(_,_), Env env) {
	return <AstValue(RenameList(expr)), env>;
}
// Statement:   assert e clauses;
Env evaluateStat(stat:Assert(e, seq(clauses)), Env env) {
	// TODO: check that e is a predicate
	res = BoolValue(true);
	try {
		<res, env> = evaluateExpr(e, env);
	}
	catch AssertionFailed(_): {
		env = errorMark(env, "Assertion failed", locOf(stat));
		throw AssertionFailed(env);
	}

	if(res == BoolValue(false)) {
		env = errorMark(env, "Assertion failed", locOf(stat));
		throw AssertionFailed(env);
	}
	else if(res != BoolValue(true)) {
		env = errorMark(env, "Assert condition not TRUE or FALSE (was <res>)", locOf(e));
		throw EvaluationError(env);
	}
	return env;
}

// Statement:   assert e clauses;
Env evaluateStat(stat:Assume(e, seq(clauses)), Env env) {
	// TODO: check that e is a predicate
	res = BoolValue(true);
	try {
		<res, env> = evaluateExpr(e, env);
	}
	catch AssertionFailed(_): {
		env = errorMark(env, "Assertion failed", locOf(stat));
		throw AssertionFailed(env);
	}

	if(res == BoolValue(false)) {
		env = errorMark(env, "Assertion failed", locOf(stat));
		throw AssertionFailed(env);
	}
	else if(res != BoolValue(true)) {
		env = errorMark(env, "Assert condition not TRUE or FALSE (was <res>)", locOf(e));
		throw EvaluationError(env);
	}
	return env;
}

default tuple[Value, Env] evaluateExpr(AST tree, Env env) {
	println("EVALUATE: no rule for <tree>");
	env = errorMark(env, "Internal Compiler Error: TopTestRunner::evaluateExpr: no evaluate rule for <consOf(tree)>", locOf(tree));
	throw EvaluationError(env);
}

default Env evaluateStat(AST tree, Env env) {
	println("EVALUATE: no rule for <tree>");
	env = errorMark(env, "Internal Compiler Error: TopTestRunner::evaluateStat: no evaluate rule for <consOf(tree)>", locOf(tree));
	throw EvaluationError(env);
}

default tuple[Value, Env] evaluateBody(AST tree, Env env) {
	println("EVALUATE: no rule for <tree>");
	env = errorMark(env, "Internal Compiler Error: TopTestRunner::evaluateBody: no evaluate rule for <consOf(tree)>", locOf(tree));
	throw EvaluationError(env);
}

tuple[list[Value], Env] acceptArgs(list[AST] types, list[AST] args, AST expr, Env env) {
	list[Value] avs;
	if(size(types) != size(args)) {
		env = errorMark(env, "Internal Compiler Error: TopTestRunner::acceptArg: wrong number of arguments, expected <size(types)>, got <size(args)>", locOf(expr));
		throw EvaluationError(env);
	}
	
	avs = for(<t, a> <- zip(types, args)) {
		<av, env> = evaluateExpr(a, env);
		if(!(typeOf(av) == t)) {
			env = errorMark(env, "Internal Compiler Error: TopTestRunner::acceptArg: expected <nameToStr(t)>, got <nameToStr(typeOf(av))>", locOf(a));
			throw EvaluationError(env);
		}
		append av;
	};

	return <avs, env>;
}