(ns org.magnolialang.testutil.bankers-deque
  (:require [clojure.math.numeric-tower :as math]))

(comment

  Implements a banker's deque, as per Okasaki. Uses Clojure's lazy
  sequences as the stream type. The invariant maintained for [lenf f
  lenr r] is (and (<= lenf (+ (* *dq-factor* lenr) 1)) (<= lenr (+ (*
  *dq-factor* lenf) 1))), for some constant '*dq-factor*', in order to
  maintain balance. 'f' means "front" and 'r' means "rear". The
  algorithm supports every deque operation in O(1) amortized time.

  We define a deque type that can directly be used as both a Clojure
  record type and a (opaque) Rascal value.

  (use 'org.magnolialang.testutil.bankers-deque :reload-all)
  (reval "identity = &T (&T x) { return x; };")
  (rcall "identity" db-empty)

  )

(def ^:dynamic *dq-factor* 7)

;; xxx Should implement 'lub' and 'isSubTypeof'?
;; At least Type::typeOf does not work on this.
;; Note that in Clojure, 'identical?' compares by identity.
(def ^:private dq-type
  (proxy [org.rascalmpl.value.type.ExternalType] []))

(defrecord Deque [lenf f lenr r]
  org.rascalmpl.value.IExternalValue
  (getType [this] dq-type)
  ;;(accept [this v] (.visitExternal v this))
  ;;(isEqual [this v] (.equals this v))
  (accept [this v] nil)
  (isEqual [this v] false)
  )

;; An empty double-ended queue.
(def dq-empty
  (Deque. 0 () 0 ()))

(def ^:private dq-class (class dq-empty))

(defn dq-is? [q]
  (instance? dq-class q))

(defn- div2 [n]
  (let [x (math/round (/ n 2))]
    [x (- n x)]))

;; Constructs a queue with the specified content, balancing first if
;; necessary to meet the invariant.
(defn- check [lenf f lenr r]
  (cond
   (> lenf (+ (* *dq-factor* lenr) 1))
   (let [[lenf' lenr'] (div2 (+ lenf lenr))
         [f' f-rest] (split-at lenf' f) ;; gives seqs, not lists
         r' (concat r (reverse f-rest))]
     (Deque. lenf' f' lenr' r'))

   (> lenr (+ (* *dq-factor* lenf) 1))
   (let [[lenf' lenr'] (div2 (+ lenf lenr))
         [r' r-rest] (split-at lenr' r)
         f' (concat f (reverse r-rest))]
     (Deque. lenf' f' lenr' r'))

   :else
   (Deque. lenf f lenr r)))

;; Whether empty.
(defn dq-empty? [q]
  (and (= (:lenf q) 0) (= (:lenr q) 0)))

;; Push element to front.
(defn dq-push-f [q e]
  (check (+ (:lenf q) 1) (conj (:f q) e) (:lenr q) (:r q)))

;; Push element to rear.
(defn dq-push-r [q e]
  (check (:lenf q) (:f q) (+ (:lenr q) 1) (conj (:r q) e)))

;; Returns the first element from the front, or nil if none.
(defn dq-peek-f [q]
  (let [lenf (:lenf q)
        lenr (:lenr q)]
    (cond
     (and (= lenf 0) (= lenr 0)) nil
     ;; Note that the invariant guarantees that both 'f' and 'r' are
     ;; non-empty if there are at least two elements in the queue. It
     ;; would perhaps be clearer to write (last r) instead of (first r)
     ;; here, but it's the same either way since we know the length of
     ;; 'r' must be exactly 1 here.
     (= lenf 0) (first (:r q))
     :else (first (:f q)))))

;; Returns the first element from the rear, or nil if none.
(defn dq-peek-r [q]
  (let [lenf (:lenf q)
        lenr (:lenr q)]
    (cond
     (and (= lenf 0) (= lenr 0)) nil
     (= lenr 0) (first (:f q))
     :else (first (:r q)))))

;; Drops the first element from the front. Does nothing if there isn't
;; one.
(defn dq-tail-f [q]
  (let [lenf (:lenf q)
        lenr (:lenr q)]
    (cond
     (and (= lenf 0) (= lenr 0)) q
     (= lenr 0) dq-empty
     :else (check (- lenf 1) (rest (:f q)) lenr (:r q)))))

;; Drops the first element from the rear. Does nothing if there isn't
;; one.
(defn dq-tail-r [q]
  (let [lenf (:lenf q)
        lenr (:lenr q)]
    (cond
     (and (= lenf 0) (= lenr 0)) q
     (= lenf 0) dq-empty
     :else (check lenf (:f q) (- lenr 1) (rest (:r q))))))

;; Return first element from the front, plus all but the first element
;; from the front. Same semantics as for [(dq-peek-f q) (dq-tail-f
;; q)], but more efficient.
(defn dq-pop-f [q]
  (let [lenf (:lenf q)
        lenr (:lenr q)]
    (cond
     (and (= lenf 0) (= lenr 0)) [nil q]
     (= lenf 0) [(first (:r q)) dq-empty]
     :else (let [f (:f q)]
             [(first f)
              (check (- lenf 1) (rest f) lenr (:r q))]))))

;; Appends another deque 'qa' to the front.
(defn dq-prepend [q qa]
  (let [lenf (:lenf q)
        lenr (:lenr q)
        qa-lenf (:lenf qa)
        qa-lenr (:lenr qa)]
    (cond
     (and (= lenf 0) (= lenr 0)) qa
     (and (= qa-lenf 0) (= qa-lenr 0)) q
     :else (check (+ qa-lenf qa-lenr lenf)
                  (concat (:f qa) (reverse (:r qa)) (:f q))
                  lenr (:r q)))))

;; Appends another deque 'qa' to the rear.
(defn dq-append [q qa]
  (let [lenf (:lenf q)
        lenr (:lenr q)
        qa-lenf (:lenf qa)
        qa-lenr (:lenr qa)]
    (cond
     (and (= lenf 0) (= lenr 0)) qa
     (and (= qa-lenf 0) (= qa-lenr 0)) q
     :else (check lenf (:f q)
                  (+ lenr qa-lenf qa-lenr)
                  (concat (:r qa) (reverse (:f qa)) (:r q))))))

;;; 
;;; An implementation of the Rascal interface.
;;; 

(gen-class
 :name org.magnolialang.testutil.BankersDequeRsc
 :init init
 :constructors {[org.rascalmpl.value.IValueFactory] []}
 :state vf
 :methods
 [
  [tseqEmpty [] org.rascalmpl.value.IValue]
  [tseqIs [org.rascalmpl.value.IValue] org.rascalmpl.value.IBool]
  [tseqIsEmpty [org.rascalmpl.value.IValue] org.rascalmpl.value.IBool]
  [tseqPushF [org.rascalmpl.value.IValue org.rascalmpl.value.IValue] org.rascalmpl.value.IValue]
  [tseqPushR [org.rascalmpl.value.IValue org.rascalmpl.value.IValue] org.rascalmpl.value.IValue]
  [tseqPeekF [org.rascalmpl.value.IValue] org.rascalmpl.value.IValue]
  [tseqPeekR [org.rascalmpl.value.IValue] org.rascalmpl.value.IValue]
  [tseqTailF [org.rascalmpl.value.IValue] org.rascalmpl.value.IValue]
  [tseqTailR [org.rascalmpl.value.IValue] org.rascalmpl.value.IValue]
  [tseqPrepend [org.rascalmpl.value.IValue org.rascalmpl.value.IValue] org.rascalmpl.value.IValue]
  [tseqAppend [org.rascalmpl.value.IValue org.rascalmpl.value.IValue] org.rascalmpl.value.IValue]
  [tseqMake [org.rascalmpl.value.IList] org.rascalmpl.value.IValue]
  [tseqPopF [org.rascalmpl.value.IValue] org.rascalmpl.value.ITuple]
  ]
 )

(defn -init [vf]
  [[] vf])

(defn -tseqEmpty [this]
  dq-empty)

(defn -tseqIs [this q]
  (.bool (.vf this) (dq-is? q)))

(defn -tseqIsEmpty [this q]
  (.bool (.vf this) (dq-empty? q)))

(defn -tseqPushF [this q e]
  (dq-push-f q e))

(defn -tseqPushR [this q e]
  (dq-push-r q e))

(defn -tseqPeekF [this q]
  (dq-peek-f q))

(defn -tseqPeekR [this q]
  (dq-peek-r q))

(defn -tseqTailF [this q]
  (dq-tail-f q))

(defn -tseqTailR [this q]
  (dq-tail-r q))

(defn -tseqPrepend [this q qa]
  (dq-prepend q qa))

(defn -tseqAppend [this q qa]
  (dq-append q qa))

(defn -tseqMake [this lst]
  (reduce
   dq-push-r
   dq-empty
   (iterator-seq (.iterator lst))))

;; Collection<IValue> -> IValue[]
(defn- value-array [coll]
  (into-array org.rascalmpl.value.IValue coll))

(defn -tseqPopF [this q]
  (let [r (dq-pop-f q)]
    (.tuple (.vf this) (value-array r))))
