;; E.g. (use 'org.magnolialang.testutil.core :reload-all)
(ns org.magnolialang.testutil.core
  (:use
   clojure.java.javadoc
   )
  (:import
   (org.magnolialang.infra ConsoleInfra)
   (org.magnolialang.magnolia MagnoliaFacts)
   (org.magnolialang.magnolia.resources ConsoleMagnoliaPackage)
   (org.magnolialang.testutil RascalEval RascalExt TestUtil)
   ))

;;; --------------------------------------------------
;;; testing
;;; --------------------------------------------------

(defn hello-str []
  (str "This is " org.magnolialang.magnolia.Magnolia/MAGNOLIA "."))

(defn say-hello []
  (println (hello-str))) 

;;; --------------------------------------------------
;;; Clojure
;;; --------------------------------------------------

(defmacro thunk [& body]
  `(fn [] ~@body))

;;; --------------------------------------------------
;;; javadoc
;;; --------------------------------------------------

(defn init-javadoc []
  (add-local-javadoc
   (.getCanonicalPath
    (java.io.File. (System/getProperty "java.home") "../docs/api/"))))

;;; --------------------------------------------------
;;; infra
;;; --------------------------------------------------

(defn set-infra []
  (org.magnolialang.infra.ConsoleInfra/setInfra))

(defn get-infra []
  (org.magnolialang.infra.ConsoleInfra/getInfra))

(defn path-env-var-to-list [n]
  (java.util.ArrayList.
   (map
    #(-> % (java.io.File.) (.getAbsolutePath))
    (let [ev (System/getenv n)]
      (if ev (seq (.split #":" ev)) [])))))

(def MAGNOLIA_RASCAL_MODULE_PATH
  (path-env-var-to-list "MAGNOLIA_RASCAL_MODULE_PATH"))

(def MAGNOLIA_MODULE_PATH
  (path-env-var-to-list "MAGNOLIA_MODULE_PATH"))

(org.magnolialang.infra.ConsoleInfra/setRascalSearchPath
 MAGNOLIA_RASCAL_MODULE_PATH)
(org.magnolialang.infra.ConsoleInfra/setMagnoliaSearchPath
 MAGNOLIA_MODULE_PATH)
(set-infra)

(def infra (get-infra))

(def tu-state)

;;; --------------------------------------------------
;;; plain Rascal
;;; --------------------------------------------------

(defn get-ri []
  (RascalEval.))

;; The Rascal evaluator instance to use. By replacing this with a
;; fresh copy you can "reload" Rascal modules. Clojure modules can
;; also be reloaded, naturally. However, unless you're running in an
;; controlled environment (such as Eclipse) with custom class loaders,
;; reloading Java classes is not possible, as Java itself does not
;; allow it.
(def rsc-evaluator (get-ri))

(defn result-to-string [r]
  (RascalEval/resultToString r))

;; Evaluate Rascal expression in an Evaluator.
;; E.g. (reval "registerCompiler")
(defn reval [s]
  (.getValue (.eval rsc-evaluator s)))

(defn reval-str [s]
  (result-to-string (.eval rsc-evaluator s)))

;; E.g. (reval-more "import org::magnolialang::magnolia::resources::Resources; getFact; 5;")
(defn reval-more [s]
  (.getValue (.evalMore rsc-evaluator s)))

(defn reval-more-str [s]
  (result-to-string (.evalMore rsc-evaluator s)))

(defn rapply [s as]
  (.call rsc-evaluator s as))

(defn rcall [s & as]
  (rapply s as))

(defn rcar [pair] ;; Pair<T1, T2> -> T1
  (.getFirst pair))

(defn rcdr [pair] ;; Pair<T1, T2> -> T2
  (.getSecond pair))

(defn rpair->vector [pair]
  (vector (rcar pair) (rcdr pair)))

;; Has a nice 'toString' method that shows all the imported modules.
(defn get-heap [] ;; -> GlobalEnvironment
  (.__getHeap (.getEvaluator rsc-evaluator)))

(defn get-root-scope [] ;; -> ModuleEnvironment
  (.__getRootScope (.getEvaluator rsc-evaluator)))

;; Has a 'toString' method that (recursively) shows name, imports, and
;; extends of the module. Inherits from 'Environment'.
;;
;; E.g.
;; (get-module "List")
;; (.getName (get-module "List")) => "List"
;; (.getExtends (get-module "Prelude"))
;; (.getImports (get-module "List"))
;; (.getImportsTransitive (get-module "List"))
;; (.getStore (get-module "List")) -> TypeStore
;; (.getFunctions (get-module "List")), including 'private' ones
;; (.getTests (get-module "List"))
;; (.getVariables (get-module "List"))
;; (map rpair->vector (seq (.getFunctions (.getModule (get-heap) "List"))))
(defn get-module [n] ;; string -> ModuleEnvironment
  (.getModule (get-heap) n))

(defn show-public-functions [n]
  (let [mod (get-module n)
        xs (seq (.getFunctions mod))]
    (doseq [x xs]
      (let [name (rcar x)
            ols (seq (rcdr x))
            p-ols (for [x ols :when (.isPublic x)] x)]
        (when-not (empty? p-ols)
          (prn [name p-ols]))))))

;; This produces VoidResult, which
;; org.rascalmpl.interpreter.result.Result.toString dislikes
;; (unless using a patched version of VoidResult).
;; E.g. (result-to-string (make-rvoid))
(defn make-void-result []
  (.evalMore rsc-evaluator ";"))

;; E.g. (rimport "org::magnolialang::tasks::Manager")
(defn rimport [s]
  (.doImport rsc-evaluator s))

;; For reloading Rascal modules individually.
;; E.g. (rreload "org::nuthatchery::pgf::rascal::Token")
(defn rreload [& s]
  (let [evaluator (.getEvaluator rsc-evaluator)
        errorLoc (java.net.URI/create "clojure:///")
        mods (java.util.HashSet. s)
        rm (.getMonitor infra)]
    (.reloadModules evaluator rm mods errorLoc)))

;; Whether all the specified Rascal modules have been imported.
(defn rimported? [& ss]
  (let [evaluator (.getEvaluator rsc-evaluator)
        heap (.getHeap evaluator)]
    (every? (fn [name]
              (.existsModule heap name)) ss)))

;; Imports all specified modules not already imported. Returns a
;; sequence of the imported modules.
(defn ensure-rimport [& ss]
  (let [evaluator (.getEvaluator rsc-evaluator)
        heap (.getHeap evaluator)
        lst (for [s ss :when (not (.existsModule heap s))] s)]
    (doseq [s lst] (rimport s))))

(defn rvalue? [o]
  (RascalExt/isIValue o))

(defn rstring [s]
  (.string org.magnolialang.terms.TermFactory/vf s))

(defn rbool [b]
  (.bool org.magnolialang.terms.TermFactory/vf b))

(defn rinteger [i]
  ;; has other overloads
  (.integer org.magnolialang.terms.TermFactory/vf i))

(defn to-value-array [coll]
  (RascalExt/toValueArray coll))

(defn rlist [& lst]
  ;; invoking function taking IValue...,
  ;; which is sugar for IValue[]
  (.list org.magnolialang.terms.TermFactory/vf
         (to-value-array lst)))

(defn rset [& lst]
  ;; invoking function taking IValue...,
  ;; which is sugar for IValue[]
  (.set org.magnolialang.terms.TermFactory/vf
        ;; takes any Collection
        (to-value-array lst)))

(defn ctor-get-field [v s]
  (RascalExt/ctorGetField v s))

(defn ctor-get-val [v]
  (ctor-get-field v "val"))

;; E.g. (spit "/tmp/foo" (yield-term (ctor-get-val (get-parse-tree "basic.Basic"))))
(defn yield-term
  ([v a?] (org.magnolialang.terms.TermAdapter/yieldTerm v a?))
  ([v] (yield-term v false)))

(defn get-parser [s]
  (.getParser infra s))

;; (rimport "util::tasks::Manager")
(defn start-transaction []
  (rcall "startTransaction"))

(defn end-transaction [tr]
  (rcall "endTransaction" tr))

(defn abandon-transaction [tr]
  (rcall "abandonTransaction" tr))

;; E.g. (in-transaction tr (prn tr))
(defmacro in-transaction
  ([tr & act]
     `(let [~tr (start-transaction)]
        (try ~@act (finally (end-transaction ~tr))))))

;;; --------------------------------------------------
;;; Rascal with Magnolia codebase Rascal modules
;;; --------------------------------------------------

(defn ensure-common []
  (ensure-rimport
   "IO"
   "org::magnolialang::testutil::RascalExt"
   "org::magnolialang::testutil::RascalPp"
   "org::magnolialang::magnolia::lib::Pretty"))

;; Gives us an Evaluator rather than a pool.
;; Has compiler search paths and such set up.
(defn make-evaluator []
  (.makeEvaluator (.getEvaluatorFactory infra)))

(defn get-ric []
  (let [evaluator (make-evaluator)
        ri (RascalEval. evaluator)]
    ri))

;; This sets up an evaluator that has the Magnolia compiler modules
;; importable, and some extra Rascal utilities ready for use.
(defn init-evaluator []
  (def rsc-evaluator (get-ric))
  (ensure-common))

;; (rimport "org::magnolialang::testutil::RascalPp")
(defn rprint
  ([v]
     (println (.getValue (rcall "prettyRascalValue" v))))
  ([v annos?]
     (println (.getValue (rcall "prettyRascalValue" v
                                (rcall "StyleA" (rbool annos?)))))))

(defn rprint-ca [v]
  (println (.getValue (rcall "prettyRascalValue" v
                             (rcall "StyleC" (rbool true))))))

(defn rtostring [v]
  (.getValue (rcall "javaToString" v)))

;; (rimport "org::magnolialang::testutil::TestUtil")
;; Note that (keep-annos ast) removes all annotations.
(defn keep-annos [ast & annos]
  (rcall "keepAnnotationsRec" ast (apply rset (map rstring annos))))

;; (rimport "org::magnolialang::testutil::TestUtil")
(defn del-annos [ast & annos]
  (rcall "delAnnotationsRec" ast (apply rset (map rstring annos))))

;; (rimport "org::magnolialang::magnolia::lib::Pretty")
;; E.g. (print-ast-as-mg (get-imploded-tree "basic.Basic"))
(defn print-ast-as-mg [ast]
  (println (.getValue (rcall "prettyMagnolia" ast))))

;;; --------------------------------------------------
;;; Magnolia compiler pipeline access
;;; --------------------------------------------------

;; (.getId (magnolia-language)) => "Magnolia"
;; (.getExtensions (magnolia-language)) => [".mg"]
;; (.getModuleName (.getParser (magnolia-language))) => "org::magnolialang::syntax::Magnolia"
(defn magnolia-language []
  (org.magnolialang.magnolia.Magnolia/getInstance))

(defn get-magnolia-parser []
  (get-parser (.getModuleName (.getParser (magnolia-language)))))

;; This is time consuming for the first time, but must be done to use
;; the Magnolia toolchain proper.
(defn get-compiler []
  (.getCompiler (magnolia-language)))

;; This may not be needed anymore, but at least an empty declaration
;; should be retained for backward compatibility.
(defn init-compiler []
  (init-evaluator)
  (def tu-state (new TestUtil))
  ;; (.addFile tu-state
  ;;           "org.magnolialang.magnolia.templates.MagicMain"
  ;;           (java.io.File.
  ;;            "src/org/magnolialang/magnolia/templates/MagicMain.cxx"))
  ;; (org.magnolialang.resources.LanguageRegistry/registerLanguage (magnolia-language))
  ;; (let [parser (.getParser (magnolia-language))]
  ;;   (.initialize parser))
  ;;(rimport "org::magnolialang::testutil::TestUtil")
  )

(defn refresh-compiler []
  (.refresh (get-compiler)))

(defn get-pkg-file [mod-name]
  (let [lang (magnolia-language)
        mod-id (.getNameAST lang mod-name)
        file (or (ConsoleInfra/findModuleFile mod-id lang)
                 (throw (java.io.FileNotFoundException.
                         (str "module " mod-name))))]
    file))

(defn call-compiler [f & as]
  (TestUtil/callCompiler f (to-value-array as)))

(defn get-pkg [mod-name]
  (let [lang (magnolia-language)
        mod-id (.getNameAST lang mod-name)
        file (or (ConsoleInfra/findModuleFile mod-id lang)
                 (throw (java.io.FileNotFoundException.
                         (str "module " mod-name))))
        pkg (ConsoleMagnoliaPackage/makeFileBackedTransientPackage file mod-id)]
    pkg))

(defn get-pkg-info [mod-name]
  (let [rm (.getMonitor infra)
        pkg (get-pkg mod-name)]
    (org.magnolialang.magnolia.frontend.MagnoliaLoader/getPkgInfo pkg rm)))

(defn get-name-ast [s]
  (.getNameAST (magnolia-language) s))

(defn get-fact-uri [s]
  (.getURI (get-pkg s)))

;; Original source text.
(defn get-fact-contents-string [s]
  (.getContentsString (get-pkg s)))

(defn get-fact-imploded-tree [s]
  (.getAST (get-pkg s) (.getMonitor infra)))

(defn get-fact-source-signature [s]
  (.getSourceSignature (get-pkg s) (.getMonitor infra)))

(defn get-fact-full-signature [s]
  (.getFullSignature (get-pkg s) (.getMonitor infra)))

;;(defn load-def-info [s] ;; normally private
;;  (.loadDefInfo (get-pkg s) (.getMonitor infra)))

;; 'toString' on the result appears to cause a stack overflow,
;; presumably due to the result referring to itself.
(defn get-fact-def-info [s]
  (.getDefInfo (get-pkg s) (.getMonitor infra)))

(defn get-fact-depends [s]
  (.getDepends (get-pkg s) (.getMonitor infra)))

(defn get-fact-transitive-depends [s]
  (.getTransitiveDepends (get-pkg s) (.getMonitor infra)))

(defn get-fact-desugared-tree [s]
  (call-compiler "desugarTree" (get-fact-imploded-tree s)))

;; E.g.
;; (print-ast-as-mg (ensugar-tree (get-fact-desugared-tree "basic.Basic")))
(defn ensugar-tree [ast]
  (call-compiler "ensugarTree" ast))

;; E.g.
;; (print-ast-as-mg (get-fact-type-checked "basic.Basic"))
;; (spit "/tmp/foo" (with-out-str (print-ast-as-mg (get-fact-type-checked "basic.Basic"))))
(defn get-fact-type-checked [s]
  (.getTypechecked (get-pkg s) (.getMonitor infra)))

(defn get-fact-children [s]
  (.getChildren (get-pkg s) (.getMonitor infra)))

;; E.g. (doseq [x (get-contents-seq "basic.Basic")] (prn x))
(defn get-contents-seq [s] ;; string -> seq
  (iterator-seq (.iterator (get-fact-children s))))

;; E.g. (get-cxx-code "basic.Basic")
(defn get-cxx-code [s]
  (call-compiler "cxxCode" (get-fact-def-info s)))

(defn view-flattened-module [s]
  (print-ast-as-mg (get-fact-type-checked s)))

(defn trigger-file-changed [s]
  (.triggerFileChanged (get-pkg s) (.getMonitor infra)))
