(ns org.magnolialang.testutil.pdb-values
  (:use
   org.magnolialang.testutil.core
   ;;[clojure.set :only [union]]
   clojure.walk
   ))

;; Has other overloads, but typically java.lang.Long (Clojure) -> long
;; (Java) -> IInteger (Rascal).
(defn c-r-integer [i]
  (.integer org.magnolialang.terms.TermFactory/vf i))

;; Has other overloads, but typically java.lang.Double (Clojure) ->
;; double (Java) -> IReal (Rascal).
(defn c-r-real [d]
  (.real org.magnolialang.terms.TermFactory/vf d))

;; java.lang.String (Clojure) -> IString (Rascal).
(defn c-r-string [s]
  (.string org.magnolialang.terms.TermFactory/vf s))

;; Constructs a ISourceLocation (Rascal).
;; Has no direct Clojure counterpart.
;; First argument URI or String.
(defn r-loc
  ([path offset length beginLine endLine beginCol endCol]
     (.uri org.magnolialang.terms.TermFactory/vf
           path offset length beginLine endLine beginCol endCol))
  ([path]
     (.uri org.magnolialang.terms.TermFactory/vf path)))

;; Collection<IValue> -> IValue[]
(defn- value-array [coll]
  (into-array org.rascalmpl.value.IValue coll))

;; Collection<Type> -> Type[]
(defn- type-array [coll]
  (into-array org.rascalmpl.value.type.Type coll))

;; Constructs an ITuple (Rascal).
;; Has no Clojure counterpart.
(defn r-tuple
  ([] ;; nullary tuple
     (.tuple org.magnolialang.terms.TermFactory/vf))
  ([& lst]
     (.tuple org.magnolialang.terms.TermFactory/vf
             (value-array lst))))

;; Constructs an INode (Rascal).
;; Has no Clojure counterpart.
(defn r-node
  ([s] ;; nullary generic tree node
     (.node org.magnolialang.terms.TermFactory/vf s))
  ([s & lst]
     (.node org.magnolialang.terms.TermFactory/vf s
            (value-array lst))))

(defn- error [& msgs]
  (throw (Exception. (apply str msgs))))

(defn tuple-type [ts]
  (.tupleType org.magnolialang.terms.TermFactory/tf (type-array ts)))

(defn- c-j-map [m]
  (reduce-kv
   (fn [w k v]
     (.put w k v) w)
   (java.util.HashMap.)
   m))

;; GlobalEnvironment unfortunately provides no way to ask for all the
;; loaded modules (apart from parsing toString output), so we require
;; a fully qualified name to find the constructor.
;;
;; E.g.
;; (rimport "org::magnolialang::testutil::TestInliner")
;; (r-constructor "org::magnolialang::testutil::TestInliner" "NoAenv")
;;
;; (rimport "org::magnolialang::syntax::MagnoliaAST")
;; (def nop (r-constructor "org::magnolialang::syntax::MagnoliaAST" "Nop"))
;; (r-constructor "org::magnolialang::syntax::MagnoliaAST" "Body" nop)
(defn r-constructor
  ([mod-name s]
     (let [m (get-module mod-name)
           ts (.getStore m)
           cs (.lookupConstructors ts s)
           n-cs (for [t (seq cs)
                      :when (empty? (.getFieldTypes t))]
                  t)
           n (count n-cs)]
       (when (= n 0) (error "no nullary ctor for " s))
       (when (> n 1) (error "ambiguous nullary ctor " s))
       (.constructor org.magnolialang.terms.TermFactory/vf (first n-cs))))
  ([mod-name s & args]
     (let [m (get-module mod-name)
           ts (.getStore m)
           ats (map #(.getType %) args)
           args-t (tuple-type ats)
           ct (.lookupFirstConstructor ts s args-t)]
       (.constructor org.magnolialang.terms.TermFactory/vf
                     ct (value-array args)))))

;; E.g.
;; (rimport "org::magnolialang::testutil::TestInliner")
;; (rimport "org::magnolialang::syntax::MagnoliaAST")
;; (r-constructor-annos "org::magnolialang::syntax::MagnoliaAST" "Nop" {"inlId" (c-r-integer 5)})
(defn r-constructor-annos [mod-name s an & args]
  (let [m (get-module mod-name)
        ts (.getStore m)
        ats (map #(.getType %) args)
        args-t (tuple-type ats)
        ct (.lookupFirstConstructor ts s args-t)
        an (c-j-map an)]
    (.constructor org.magnolialang.terms.TermFactory/vf
                  ct an (value-array args))))

;; E.g. (c-r-set #{(c-r-integer 5)})
(defn c-r-set [xs]
  (.set org.magnolialang.terms.TermFactory/vf
        (value-array (seq xs))))

;; E.g. (r-set (c-r-integer 5))
(defn r-set [& args]
  (.set org.magnolialang.terms.TermFactory/vf
        (value-array args)))

(defn c-r-list [xs]
  (.list org.magnolialang.terms.TermFactory/vf
         (value-array xs)))

(defn r-list [& args]
  (.list org.magnolialang.terms.TermFactory/vf
        (value-array args)))

;; E.g. (r-empty-list (.getType (c-r-integer 5)))
(defn r-empty-list [t]
  (.list org.magnolialang.terms.TermFactory/vf t))

;; E.g. (let [t (.getType (c-r-integer 5))] (r-empty-relation t t t))
(defn r-empty-relation [& ts]
  (.relation org.magnolialang.terms.TermFactory/vf (tuple-type ts)))

;; Arguments must be tuples.
;; E.g. (let [i c-r-integer t r-tuple] (r-relation (t (i 1) (i 2)) (t (i 3) (i 4))))
(defn r-relation [& args]
  (.relation org.magnolialang.terms.TermFactory/vf
             (value-array args)))

(defn r-empty-map [kt vt]
  (.map org.magnolialang.terms.TermFactory/vf kt vt))

(defn- write-map [w m]
  (doseq [[k v] (seq m)]
    (.put w k v))
  (.done w))

;; E.g. (let [i c-r-integer s c-r-string kt (.getType (i 0)) vt (.getType (s ""))] (r-typed-map kt vt {(i 1) (s "2") (i 3) (s "4")}))
(defn r-typed-map [kt vt m]
  (write-map
   (.mapWriter org.magnolialang.terms.TermFactory/vf kt vt) m))

;; E.g. (let [i c-r-integer s c-r-string] (c-r-map {(i 1) (s "2") (i 3) (s "4")}))
(defn c-r-map [m]
  (write-map
   (.mapWriter org.magnolialang.terms.TermFactory/vf) m))

(defn c-r-bool [v]
  (.bool org.magnolialang.terms.TermFactory/vf v))
  
;; xxx gives incorrect result if doesn't fit into a long.
(defn r-c-integer [v]
  (.longValue v))

;; xxx may lose accuracy if doesn't fit into a double
(defn r-c-real [v]
  (.doubleValue v))

(defn r-c-string [v]
  (.getValue v))

(defn r-c-bool [v]
  (.getValue v))

(defn r-c-set [v]
  (into #{} (iterator-seq (.iterator v))))

;; note that IRelation is-a ISet
(defn relation-to-set [v]
  (into #{} (iterator-seq (.iterator v))))

;; E.g.
;; (r-c-list (c-r-list (list (c-r-integer 1))))
(defn r-c-list [v]
  (reverse
   ;; 'conj' each to beginning of list
   (into () (iterator-seq (.iterator v)))))

;; No Clojure counterpart for the rather common
;; org.rascalmpl.value.ITuple. This is perhaps the closest
;; equivalent, should we want to convert to some native type.
(defn tuple-to-vector [v]
  (into [] (iterator-seq (.iterator v))))

;; No Rascal counterpart for the common clojure.lang.IPersistentVector
;; either, but this may sometimes be useful.
(defn vector-to-tuple [v]
  (apply r-tuple v))

(defn r-c-map [v]
  (reduce
   (fn [m e]
     (let [k (.getKey e)
           v (.getValue e)]
       (assoc m k v)))
   {}
   (iterator-seq (.entryIterator v))))

(defmulti c-r-shallow class)

(defmethod c-r-shallow java.lang.Long [v] (c-r-integer v))
(defmethod c-r-shallow java.lang.Double [v] (c-r-real v))
(defmethod c-r-shallow java.lang.String [v] (c-r-string v))
(defmethod c-r-shallow java.lang.Boolean [v] (c-r-bool v))
(defmethod c-r-shallow clojure.lang.ISeq [v] (c-r-list v))
(defmethod c-r-shallow clojure.lang.IPersistentMap [v] (c-r-map v))
(defmethod c-r-shallow clojure.lang.IPersistentSet [v] (c-r-set v))
(defmethod c-r-shallow clojure.lang.IPersistentVector [v] (vector-to-tuple v))

;; No counterpart in Rascal, so leave as is. This will probably cause
;; problems as the result then is not necessarily an IValue.
(defmethod c-r-shallow :default [v] v)

;; E.g.
;; (r-c-shallow (c-r-shallow true))
;; (r-c-shallow (c-r-shallow 1))
;; (r-c-shallow (c-r-shallow 1.5))
;; (r-c-shallow (c-r-shallow "foo"))
;; (r-c-shallow (c-r-shallow ()))
;; (r-c-shallow (c-r-shallow #{}))
;; (r-c-shallow (c-r-shallow {}))
(defmulti r-c-shallow class)

(defmethod r-c-shallow org.rascalmpl.value.IInteger [v] (r-c-integer v))
(defmethod r-c-shallow org.rascalmpl.value.IReal [v] (r-c-real v))
(defmethod r-c-shallow org.rascalmpl.value.IBool [v] (r-c-bool v))
(defmethod r-c-shallow org.rascalmpl.value.IString [v] (r-c-string v))
(defmethod r-c-shallow org.rascalmpl.value.IList [v] (r-c-list v))
(defmethod r-c-shallow org.rascalmpl.value.IMap [v] (r-c-map v))
(defmethod r-c-shallow org.rascalmpl.value.ISet [v] (r-c-set v))
(defmethod r-c-shallow org.rascalmpl.value.ITuple [v] (tuple-to-vector v))

;; Not having a counterpart should be okay. Can retain Rascal value in Clojure.
(defmethod r-c-shallow :default [v] v)

;; A version of 'walk' (of clojure.walk) that supports some Rascal
;; data types also.
;;
;; Normal 'postwalk' walks java.util.Map$Entry objects too and they
;; seem to also be clojure.lang.PersistentVector objects, which we try
;; to turn into tuples. Here we do not expose map entries at all, but
;; just keys and values thereof.
(defn walk-cr [inner outer form]
  (cond
   (list? form) (outer (apply list (map inner form)))
   (map? form) (outer (into {} (for [[k v] (seq form)] [(inner k) (inner v)])))
   (seq? form) (outer (doall (map inner form)))
   (coll? form) (outer (into (empty form) (map inner form)))
   :else (outer form)))

;; Converts from Clojure to Rascal, recursively, bottom-up.
;; This may very well fail due to typing restrictions.
(defn c-r [v]
  (walk-cr c-r c-r-shallow v))

;; Converts from Rascal to Clojure, recursively, top-down.
;; E.g. (r-c (c-r-shallow {(c-r-integer 1) (c-r-string "2")}))
;;   => {1 "2"}
(defn r-c [v]
  (walk-cr r-c identity (r-c-shallow v)))

(comment
  (use 'clojure.walk)
  (use 'org.magnolialang.testutil.pdb-values :reload-all)
  (map c-r [1 "foo" 4.5 false {1 "2" 3 "4"}])
  (c-r [1 "foo" 4.5 false {1 "2" 3 "4"}])
  (c-r (r-c (apply r-relation (map c-r [[1 2 "a"] [4 5 "b"]]))))
  )
