;; E.g. (use 'org.magnolialang.testutil.rascal-info :reload-all)
(ns org.magnolialang.testutil.rascal-info
  (:use
   org.magnolialang.testutil.core
   [clojure.set :only [union]]))

;; assoc with multi-map
(defn- assoc-multi [m k v]
  (assoc m k (conj (get m k #{}) v)))

;; merge of multi-maps
;;
;; E.g.
;; (merge-multi {"foo" #{1 2 3} "bar" #{3 4}} {"foo" #{2 3 4} "baz" #{1}} {"foo" #{6}})
;; => {"baz" #{1}, "bar" #{3 4}, "foo" #{1 2 3 4 6}}
(defn merge-multi [& ms]
  (reduce
   (fn [r e]
     (reduce-kv
      (fn [m k v]
        (assoc m k (union v (get m k #{}))))
      r
      e))
   {}
   ms))

;; requires (rimport "Prelude") first
(defn get-prelude-modules [] ;; -> set of string
  (set (.getExtends (get-module "Prelude"))))

(defn import-modules [lst]
  (doseq [x lst] (rimport x)))

(defn import-prelude-modules []
  (rimport "Prelude")
  (import-modules (get-prelude-modules)))

;; Including RascalFunction, JavaMethod, and ConstructorFunction.
;; E.g.
;; (doseq [x (get-public-functions "String")] (prn x))
;; (first ((first (get-public-functions "String")) 1))
(defn get-public-functions [n]
  (let [mod (get-module n)
        xs (seq (.getFunctions mod))]
    (for [x xs
          :let [name (rcar x)
                ols (seq (rcdr x))
                p-ols (for [x ols
                            :when (.isPublic x)]
                        x)]
          :when (not (empty? p-ols))]
      [name p-ols])))

(defn- fn-to-str [mn f]
  (str
   ;;(if (.isDefault f) "default " "")
   "function " (.getName f) " : "
   (.toString (.getFunctionType f)) " in " mn))

;; E.g.
;; (rimport "String")
;; (get-public-functions-mm "String")
(defn get-public-functions-mm [mn]
  (into {} (for [[n f-lst] (get-public-functions mn)]
             (let [f-set (reduce
                          (fn [r x] (conj r (fn-to-str mn x)))
                          #{} f-lst)]
               [n f-set]))))

;; Gets information about all public variables declared in a module.
;;
;; E.g.
;; (rimport "SystemAPI")
;; (get-public-variables-mm "SystemAPI")
(defn get-public-variables-mm [n]
  (let [mod (get-module n)
        mn (.getName mod)
        ;; get HashMap, could (into {} jhm) to turn into a Clojure map
        jhm (.getVariables mod)]
    (into {}
          (for [e (seq jhm) ;; get HashMap.Entry sequence
                :let [v (.getValue e)
                      public? (.isPublic v)]
                :when public?]
            (let [n (.getKey e)
                  ;; using .toString on 'value' would give value as
                  ;; well, which might be useful for constants
                  t (.toString (.getType v))
                  s (str "variable " n " : " t " in " mn)]
              [n #{s}])))))

;; Gets information about all annotations declared in a module.
;; Returns a map with anno name as key, and a set of anno descriptions
;; as value.
;;
;; E.g.
;; (rimport "ParseTree")
;; (get-annotations-mm "ParseTree")
(defn get-annotations-mm [n]
  (let [mod (get-module n)
        mn (.getName mod)
        store (.getStore mod) ;; -> TypeStore
        jhm (.getAnnotations store)] ;; -> HashMap
    (reduce
     (fn [r e]
       (let [k (.getKey e) ;; -> Type
             n (.toString k) ;; ADT name
             v (.getValue e)] ;; -> Map<String, Type>
         (reduce
          (fn [r e]
            (let [k (.getKey e) ;; -> String
                  v (.getValue e) ;; -> Type
                  t (.toString v)
                  s (str "annotation " k " : " t " for " n " in " mn)]
              ;;(prn [k (class k) s])
              (assoc-multi r k s)))
          r
          (seq v))))
     {}
     (seq jhm))))

;; We do not include 'alias' declared names, as their use appears
;; unreliable in Rascal.
;;
;; E.g.
;; (rimport "ParseTree")
;; (.getStore (get-module "ParseTree")) -> TypeStore
;; (.getAbstractDataTypes (.getStore (get-module "ParseTree")))
;; (.getAliases (.getStore (get-module "ParseTree")))
(defn get-abstract-data-types [mn]
  (let [mod (get-module mn)
        store (.getStore mod) ;; -> TypeStore
        jcoll (.getAbstractDataTypes store)] ;; -> Collection
    (into #{}
          (map
           (fn [type]
             [type (.lookupAlternatives store type)])
           jcoll))))

(defn- ctor-to-str [ctor]
  ;; (.getName ctor) -> String
  ;; (.getFieldTypes ctor) -> TupleType
  (let [n (.getName ctor)
        ts (iterator-seq (.iterator (.getFieldTypes ctor)))] ;; of Type
    (str n "("
         (clojure.string/join ", " (map #(.toString %) ts))
         ")")))

(defn- adt-to-str [mn adt ctors]
  (str "data " (.toString adt) " in " mn
       (if (empty? ctors) ""
           (str " = "
                (clojure.string/join
                 " | " (map ctor-to-str ctors))))))

(defn get-types-mm [mn]
  (into {}
        (map
         (fn [[adt ctors]]
           (let [n (.getName adt)
                 s (adt-to-str mn adt ctors)]
             [n #{s}]))
         (get-abstract-data-types mn))))

(defn get-module-info [mn]
  (merge-multi
   (get-public-functions-mm mn)
   (get-public-variables-mm mn)
   (get-annotations-mm mn)
   (get-types-mm mn)))

(defn get-prelude-info []
  (apply merge-multi (map get-module-info (get-prelude-modules))))

(defn- quote-s [n]
  (str "\"" n "\""))

(defn- fmt-elisp [n xs]
  (str "    ("
       (quote-s n)
       " "
       (quote-s
        (clojure.string/join "; OR\\n" xs))
       ")\n"))

(defn get-prelude-info-elisp []
  (str "(defvar rascal-prelude-candidates '(\n"
       (apply str
              (map
               (fn [[n ss]]
                 (fmt-elisp n (sort ss)))
               (sort (seq (get-prelude-info)))))
       ") \"public names in Rascal Prelude\")\n\n"
       "(provide 'rascal-mode-prelude)\n"
       ))

;; lein run -m org.magnolialang.testutil.rascal-info/generate-prelude-info-elisp /tmp/foo.el
(defn generate-prelude-info-elisp [out-file]
  (init-compiler)
  (import-prelude-modules)
  (spit out-file (get-prelude-info-elisp))
  (println "wrote" out-file))

