/**************************************************************************
 * Copyright (c) 2014 Anya Helene Bagge
 * Copyright (c) 2014 University of Bergen
 * 
 * All rights reserved. This program and the accompanying materials
 * are made available under the terms of the Eclipse Public License v1.0
 * which accompanies this distribution, and is available at
 * http://www.eclipse.org/legal/epl-v10.html
 * 
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version. See http://www.gnu.org/licenses/
 * 
 * 
 * See the file COPYRIGHT for more information.
 * 
 * Contributors:
 * * Anya Helene Bagge
 * 
 *************************************************************************/
package org.magnolialang.util;

import java.util.Arrays;

/**
 * This class contains utility methods for manipulating Java stack traces.
 */
public class StackTraceUtil {

	private StackTraceUtil() {

	}


	/**
	 * Patch in a new entry in the stack trace of an exception.
	 * 
	 * The entry is inserted right after (on top of) the calling method's trace
	 * entry.
	 * 
	 * The bottom of the exception and the calling method's stacks need to be
	 * equal up to the point of the calling method – this should normally be the
	 * case if the exception has been caught in the calling method.
	 * 
	 * @param exception
	 *            An exception
	 * @param className
	 *            Class name, for the new stack entry
	 * @param methodName
	 *            Method name, for the new stack entry
	 * @param fileName
	 *            File name, for the new stack entry
	 * @param lineNumber
	 *            Line number, for the new stack entry
	 * @return A modified exception (unmodified if patching failed for some
	 *         reason)
	 */
	public static <E extends Throwable> E patch(E exception, String className, String methodName, String fileName, int lineNumber) {
		StackTraceElement[] stackTrace = exception.getStackTrace();
		StackTraceElement[] myTrace = Thread.currentThread().getStackTrace(); // includes entry for current method
		int insertLoc = stackTrace.length - (myTrace.length - 2); //  and getStackTrace(), so -2
		if(insertLoc > 0) {
			StackTraceElement[] newTrace = Arrays.copyOf(stackTrace, stackTrace.length + 1);
			newTrace[insertLoc++] = new StackTraceElement(className, methodName, fileName, lineNumber);
			while(insertLoc < newTrace.length) {
				newTrace[insertLoc] = stackTrace[insertLoc - 1];
				insertLoc++;
			}
			exception.setStackTrace(newTrace);
			exception.printStackTrace();
		}
		return exception;
	}
}
